/* $Id: hwpc.c,v 1.13 2009/03/25 16:29:58 rkufrin Exp $ */

/****************************************************************************
 *
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *          Copyright(C) 2004-2009, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * hwpc.c: native method definitions for PerfSuite Java API.
 *
 * Author: Rick Kufrin
 * Date:   Jan 23, 2004
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <jni.h>
#include <pshwpc.h>

#define EXCEPTION      "java/lang/Exception"
#define INITERROR      "java/lang/ExceptionInInitializerError"
#define INDEXEXCEPTION "java/lang/ArrayIndexOutOfBoundsException"

static void
error(JNIEnv *env, int ret, char *exceptionclass)
{
    jclass Exception;

    Exception = (*env)->FindClass(env, exceptionclass);
    (*env)->ThrowNew(env, Exception, ps_strerror(ret));
}

JNIEXPORT void JNICALL
Java_org_perfsuite_hwpc_PS_1hwpc_init(JNIEnv *env, jclass class)
{
    int ret;

    ret = ps_hwpc_init();

    if ( ret != PS_SUCCESS ) {
        error(env, ret, INITERROR);
    }
}

JNIEXPORT jint JNICALL
Java_org_perfsuite_hwpc_PS_1hwpc_numevents(JNIEnv *env, jclass class)
{
    int ret, numevents;

    ret = ps_hwpc_numevents(&numevents);

    if ( ret != PS_SUCCESS ) {
        error(env, ret, EXCEPTION);
    }

    return numevents;
}

JNIEXPORT void JNICALL 
Java_org_perfsuite_hwpc_PS_1hwpc_eventnames(JNIEnv *env, jclass class,
                                            jobjectArray names)
{
    int ret, numevents, i;
    char **cnames = NULL;

    ret = ps_hwpc_eventnames(&cnames);

    if ( ret != PS_SUCCESS ) {
        error(env, ret, EXCEPTION);
    }

    ret = ps_hwpc_numevents(&numevents);

    if ( ret != PS_SUCCESS ) {
        error(env, ret, EXCEPTION);
    }

    for (i = 0; i < numevents; i++) {
       (*env)->SetObjectArrayElement(env, names, i,
                                     (*env)->NewStringUTF(env, cnames[i]));
       free(cnames[i]);
    }
    free(cnames);
}


JNIEXPORT void JNICALL
Java_org_perfsuite_hwpc_PS_1hwpc_start(JNIEnv *env, jclass class)
{
    int ret;

    ret = ps_hwpc_start();

    if ( ret != PS_SUCCESS ) {
        error(env, ret, EXCEPTION);
    }
}

JNIEXPORT void JNICALL
Java_org_perfsuite_hwpc_PS_1hwpc_readevents(JNIEnv *env, jclass class,
                                            jlongArray jarray)
{
    int numevents, ret;
    jsize jlength = (*env)->GetArrayLength(env, jarray);
    jlong *body;
    jclass Exception;

    ret = ps_hwpc_numevents(&numevents);
    if ( ret != PS_SUCCESS ) {
        error(env, ret, EXCEPTION);
    }

    if ( jlength < numevents ) {
        Exception = (*env)->FindClass(env, INDEXEXCEPTION);
        (*env)->ThrowNew(env, Exception, "Not enough elements to store data");

    }

    body = (*env)->GetLongArrayElements(env, jarray, 0);

    ret = ps_hwpc_read(body);
    if ( ret != PS_SUCCESS ) {
        error(env, ret, EXCEPTION);
    }

    (*env)->ReleaseLongArrayElements(env, jarray, body, 0);
}

JNIEXPORT void JNICALL
Java_org_perfsuite_hwpc_PS_1hwpc_suspend(JNIEnv *env, jclass class)
{
    int ret;

    ret = ps_hwpc_suspend();

    if ( ret != PS_SUCCESS ) {
        error(env, ret, EXCEPTION);
    }
}

JNIEXPORT void JNICALL
Java_org_perfsuite_hwpc_PS_1hwpc_stop(JNIEnv *env, jclass class,
                                      jstring filename)
{
    int ret;
    const char *filenameUTF;

    filenameUTF = (*env)->GetStringUTFChars(env, filename, NULL);

    ret = ps_hwpc_stop(filenameUTF);
    if ( ret != PS_SUCCESS ) {
        (*env)->ReleaseStringUTFChars(env, filename, filenameUTF);
        error(env, ret, EXCEPTION);
    }

    (*env)->ReleaseStringUTFChars(env, filename, filenameUTF);
}

JNIEXPORT void JNICALL
Java_org_perfsuite_hwpc_PS_1hwpc_shutdown(JNIEnv *env, jclass class)
{
    int ret;

    ret = ps_hwpc_shutdown();

    if ( ret != PS_SUCCESS ) {
        error(env, ret, EXCEPTION);
    }
}
