/* $Id: psjprof.c,v 1.6 2012/10/17 14:55:53 ruiliu Exp $ */

/****************************************************************************
 *
 *                      University of Illinois/NCSA
 *                          Open Source License
 * 
 *          Copyright(C) 2011-2012, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                             Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * psjprof: Java profiling implementation using JVMTI and hardware
 *          performance counters.
 *
 * Author: Rick Kufrin, Rui Liu
 * Date:   2011-05
 *
 * This file is part of PerfSuite.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <semaphore.h>
#include <fcntl.h>   /* To make the pipe non-blocking. */
#define _GNU_SOURCE
#define __USE_GNU    /* For hsearch_r, hcreate_r, hdestroy_r. */
#include <search.h>  /* For hsearch, hcreate, hdestroy. */
/* For pthread_self required by PAPI_thread_init, and for the mutexes. */
#include <pthread.h>

#include <jni.h>
#include <jvmti.h>

#include <papi.h>
/* "psdebug.h" provides ps_debug....  It also includes <perfsuite.h>,
 * which contains the defs of BEGIN_C_DECLS, PARAMS, etc.
 */
#include <psdebug.h>
#include "write-xml.h"
#include "splay.h"


/* Usage:
 *     java -agentlib:psjprof=events=<event_name>@<event_threshold>[+<event_name>@<event_threshold>...][,domain={user | kernel | all}][,max_num_threads=<maximum_number_of_threads>][,pipe={blocking | non-blocking}][,ignored_thread_names=name1[+name2...]]
 *
 * Example:
 *     java -agentlib:psjprof=events=PAPI_TOT_CYC@1000000000+PAPI_L2_TCA@1000000,domain=user,max_num_threads=10000,pipe=blocking,ignored_thread_names=thread-1
 *
 * Default values:
 *     events:               no default value;
 *     domain:               "user";
 *     max_num_threads:      100;
 *     pipe:                 "non-blocking";
 *     ignored_thread_names: no default value.
 */


#define GET_STACK_DEPTH 5

static int g_max_num_threads = 100;
static jthread *g_threadList = NULL;
/* Used to signal that a thread is finished, so there is
 * no need to get its trace any more.
 */
static sig_atomic_t *   g_threadFinished; /* Has g_max_num_threads entries. */
static pthread_mutex_t * g_eventSetMutex; /* Has g_max_num_threads entries. */
static jvmtiEnv * g_jvmti = NULL;

/* Used to communicate the event set and the event between the
 * PAPI overflow handler and the helper_loop function.
 * Initialized in vm_init, closed in vm_death.
 */
static int g_pipefd[2];
typedef struct {
    int eventset;
    long long overflow_vector;
} PSJP_PIPE_DATA;
static const int PIPE_DATA_SIZE = sizeof (PSJP_PIPE_DATA);
enum {
    PSJP_PIPE_NON_BLOCKING,
    PSJP_PIPE_BLOCKING
};
static int g_pipe_behavior = PSJP_PIPE_NON_BLOCKING;

/* Used to synchronize the threads running the vm_death and helper_loop
 * functions regarding when to call PAPI_shutdown().
 */
static sem_t g_sem;

/* Profile multiple events.
 * The data type of PAPI_overflow()'s threshold is "int",
 * so we can not use "long" or "long long".
 *
 * Possible values of "type" are "preset" and "native".
 */
static unsigned int g_numEvents;  /* The number of *valid* PAPI events. */
typedef struct {
    const char *  name;
    int           code;
    const char *  type;
    int           threshold;
} PSJP_EVT_VALID;
static PSJP_EVT_VALID *g_events;
static const char * PAPI_EVENT_TYPE_PRESET = "preset";
static const char * PAPI_EVENT_TYPE_NATIVE = "native";

/* Input event configuration. */
typedef struct {
    const char *  event_name;
    int           threshold;
} PSJP_EVT_CONF;
/* Number of event configurations in the command line options,
 * assigned as numSeparators+1.
 */
static PSJP_EVT_CONF * g_eventConfs;
static int g_numEventConfs;
/* Event domain: PAPI_DOM_ALL, USER, KERNEL.  Default to "all". */
static int g_eventDomain = PAPI_DOM_USER;
static char * g_eventDomainStr = "user";

/* For debugging. */
# define PS_DEBUGENV "PSJP_DEBUG"
static ps_debuglevel_t g_debug_level = PS_DEBUG_UNINITIALIZED;
#define debug(x,y...)  if (g_debug_level > x) { printf(y); }
#define debugputs(x,y) if (g_debug_level > x) { puts(y); }


/* "sample_trees" is of type "SplayTree **",
 * i.e., an array of "SplayTree *" (g_numEvents of them).
 * to support profiling on multiple events.
 */
typedef struct {
    int                 eventset;
    SplayTree **        sample_trees;
    unsigned long long  starttime;
} hwpc_papi_tls_t;

typedef struct {
    const char * name;
    const char * classname;
    const char * sourceFileName;
    jvmtiLineNumberEntry *lineNumberTable;    
    int count;
} METHOD_DATA;

/* For writing out the XML file. */
static psjp_globals_t globals;

/* Threads to ignore. */
const int  g_numIgnoredThreadNames = 10;
const char ** g_ignoredThreadNames;

BEGIN_C_DECLS

extern SplayTreeNode * splay_insert PARAMS((MethodLocPair key,
                                            SplayTree * root));
extern int splay_treewalk           PARAMS((SplayTree* root,
                                            SplayTreeNode*** sarray,
                                            unsigned int *p_numNodes,
                                            int *p_totalCount));

extern int _ps_cpuid PARAMS((ps_cpuid_t *cpuid));

static inline void error_assert      PARAMS((const char *condition,
                                             const char *file, int line));
static inline void print_jvmti_error PARAMS((jvmtiEnv *jvmti,
                                             jvmtiError errnum,
                                             const char *str));
static inline void check_jvmti_error PARAMS((jvmtiEnv *jvmti,
                                             jvmtiError errnum,
                                             const char *str));

JNIEXPORT jint JNICALL Agent_OnLoad PARAMS((JavaVM *vm, char *options,
                                            void *reserved));
static void JNICALL vm_init         PARAMS((jvmtiEnv *jvmti, JNIEnv *env,
                                            jthread thread));
static void JNICALL vm_death        PARAMS((jvmtiEnv *jvmti, JNIEnv *env));
static void JNICALL thread_start    PARAMS((jvmtiEnv *jvmti, JNIEnv *env,
                                            jthread thread));
static void JNICALL thread_end      PARAMS((jvmtiEnv *jvmti, JNIEnv *env,
                                            jthread thread));
static void JNICALL method_entry    PARAMS((jvmtiEnv *jvmti, JNIEnv* env,
                                            jthread thread,
                                            jmethodID methodID));
static void JNICALL helper_loop     PARAMS((jvmtiEnv *jvmti, JNIEnv *env,
                                            void *p));

static int  get_tok       PARAMS((char **src, char *buf, int buflen, int sep));
static void option_error  PARAMS((char *description));
static void parse_options PARAMS((char *command_line_options));
static void jvmtiDeallocate    PARAMS((void *ptr));
static void version_check      PARAMS((jint cver, jint rver));
static void createHelperThread PARAMS((JNIEnv *env, jvmtiEnv *jvmti,
                                       const char *name,
                                       jvmtiStartFunction func));
static void callVoidMethod     PARAMS((JNIEnv *env, jobject object,
                                       jmethodID method, jboolean arg));
static inline int shouldIgnoreThisThread PARAMS((jvmtiThreadInfo * info));

static int startPAPI PARAMS((void));
static void stopPAPI PARAMS((int EventSet));
static void handler  PARAMS((int EventSet, void *address,
                             long long overflow_vector, void *context));

static void jvmtiGetMethodInfo  PARAMS((jmethodID methodID,
                                        METHOD_DATA * methodData,
                                        int called_from_method_entry));
static jint getLineNumber       PARAMS((jmethodID method, jlocation location,
                                        jvmtiLineNumberEntry *ppentries,
                                        int count));
static inline jint map_loc2line PARAMS((jlocation location,
                                        jvmtiLineNumberEntry *table,
                                        jint count));

END_C_DECLS

static inline void
error_assert(const char *condition, const char *file, int line)
{
    printf ("ASSERTION FAILURE: %s [%s:%d]\n", condition, file, line);
    exit (1);
}

#define PSJP_ASSERT(cond)                                              \
    (((int)(cond))?(void)0:error_assert(#cond, __FILE__, __LINE__))

/* Macro to get JNI and JVMTI function pointers. */
#define   JNI_FUNC_PTR(env,f) (*((*(env))->f))
#define JVMTI_FUNC_PTR(env,f) (*((*(env))->f))

/* PAPI definitions. */
#define PAPI_ERROR_MSG_RETURN(retval,msg) { fprintf(stderr, "Error code %d (%s), %s, file %s, line %d.\n", retval,PAPI_strerror(retval),msg,__FILE__,__LINE__);  exit(retval); }

/* To make it easy to macro out all pthread_mutex codes. */
#ifdef NO_MUTEX
# define MUTEX_INIT(x,y)
# define MUTEX_LOCK(x)
# define MUTEX_UNLOCK(x)
# define MUTEX_DESTROY(x)
#else
# define MUTEX_INIT(x,y) pthread_mutex_init(x,y)

# define MUTEX_LOCK(x) {                                                \
        int mutex_retval = pthread_mutex_lock (x);                      \
        if (mutex_retval != 0) {                                        \
            debug (PS_DEBUG_WARNING,                                    \
                   "pthread_mutex_lock, error code: %d, line: %d.\n",   \
                   mutex_retval, __LINE__);                             \
        }                                                               \
    }

# define MUTEX_UNLOCK(x) {                                              \
        int mutex_retval = pthread_mutex_unlock (x);                    \
        if (mutex_retval != 0) {                                        \
            debug (PS_DEBUG_WARNING,                                    \
                   "pthread_mutex_unlock, error code: %d, line: %d.\n", \
                   mutex_retval, __LINE__);                             \
        }                                                               \
    }
# define MUTEX_DESTROY(x) pthread_mutex_destroy(x)
#endif

/* Every JVMTI interface returns an error code, which should be checked
 *   to avoid any cascading errors down the line.
 *   The interface GetErrorName() returns the actual enumeration constant
 *   name, making the error messages much easier to understand.
 */
static inline void
print_jvmti_error(jvmtiEnv *jvmti, jvmtiError errnum, const char *str)
{
    if ( errnum != JVMTI_ERROR_NONE ) {
	char       *errnum_str = NULL;
	(*jvmti)->GetErrorName(jvmti, errnum, &errnum_str);
        fprintf (stderr, "JVMTI ERROR: %s: code %d(%s).\n",
                 (str==NULL?"":str), errnum, 
                 (errnum_str==NULL?"Unknown":errnum_str));
    }
    return;
}
/* The "check_jvmti_error" function EXITS after printing out the error msg. */
static inline void
check_jvmti_error(jvmtiEnv *jvmti, jvmtiError errnum, const char *str)
{
    if ( errnum != JVMTI_ERROR_NONE ) {
        print_jvmti_error (jvmti, errnum, str);
        exit (1);
    }
    return;
}


/* Agent_OnLoad() is called first.
 * - Obtain the JVMTI environment;
 * - Obtain the necessary capabilities;
 * - Set the event callbacks, and enable the event notification;
 * - Parse the command line options.
 *
 * Only after Agent_OnLoad returns, does the Java VM continue to run.
 */
JNIEXPORT jint JNICALL
Agent_OnLoad(JavaVM *vm, char *options, void *reserved)
{
    jint                rc;
    jvmtiEventCallbacks callbacks;
    jvmtiEnv           *jvmti;
    jvmtiError          error;
    jvmtiCapabilities   needed_capabilities;

    /* Get the debug level from the environment. */
    if ( g_debug_level == PS_DEBUG_UNINITIALIZED ) {
        char * envval = getenv(PS_DEBUGENV);
        if ( envval != NULL ) {
            g_debug_level = atoi(envval);
        }
        else {
            g_debug_level = PS_DEBUG_OFF;
        }
    }

    debugputs (PS_DEBUG_VERBOSE, "Entering Agent_OnLoad...");

    /* Get the JVMTI environment. */
    rc = (*vm)->GetEnv(vm, (void **)&jvmti, JVMTI_VERSION);
    if (rc != JNI_OK) {
	fprintf (stderr,
                 "ERROR: Unable to create jvmtiEnv, GetEnv failed, error=%d\n",
                 rc);
	return -1;
    }
    g_jvmti = jvmti;

    /* Add the capabilities.
     * Fill in only the required ones.
     */
    (void)memset(&needed_capabilities,0,sizeof(needed_capabilities));
    needed_capabilities.can_get_line_numbers     = 1;
    needed_capabilities.can_get_source_file_name = 1;
    needed_capabilities.can_generate_method_entry_events = 1;
    needed_capabilities.can_access_local_variables = 1;
    error = JVMTI_FUNC_PTR(jvmti,AddCapabilities)
        (jvmti, &needed_capabilities);
    check_jvmti_error(jvmti, error, "get the necessary JVMTI capabilities");

    /* Set the callbacks and enable the event notifications. */
    memset(&callbacks, 0, sizeof(callbacks));
    callbacks.VMInit                  = &vm_init;
    callbacks.VMDeath                 = &vm_death;
    callbacks.ThreadStart             = &thread_start;
    callbacks.ThreadEnd               = &thread_end;
    callbacks.MethodEntry             = &method_entry;
    error = (*jvmti)->SetEventCallbacks(jvmti, &callbacks, sizeof(callbacks));
    check_jvmti_error(jvmti, error, "set event callbacks");
    error = (*jvmti)->SetEventNotificationMode(jvmti, JVMTI_ENABLE, 
                                             JVMTI_EVENT_VM_INIT, NULL);
    check_jvmti_error(jvmti, error, "set event notify, VMInit");
    error = (*jvmti)->SetEventNotificationMode(jvmti, JVMTI_ENABLE, 
                                             JVMTI_EVENT_VM_DEATH, NULL);
    check_jvmti_error(jvmti, error, "set event notify, VMDeath");
    error = (*jvmti)->SetEventNotificationMode(jvmti, JVMTI_ENABLE, 
                                             JVMTI_EVENT_THREAD_START, NULL);
    check_jvmti_error(jvmti, error, "set event notify, ThreadStart");
    error = (*jvmti)->SetEventNotificationMode(jvmti, JVMTI_ENABLE, 
                                             JVMTI_EVENT_THREAD_END, NULL);
    check_jvmti_error(jvmti, error, "set event notify, ThreadEnd");
    error = (*jvmti)->SetEventNotificationMode(jvmti, JVMTI_ENABLE, 
                                             JVMTI_EVENT_METHOD_ENTRY, NULL);
    check_jvmti_error(jvmti, error, "set event notify, MethodEntry");


    parse_options (options);

    return 0;
}


/* Callback for JVMTI_EVENT_VM_INIT */
static void JNICALL
vm_init (jvmtiEnv *jvmti, JNIEnv *env, jthread thread)
{
    jvmtiError error;
    jint       runtime_version;
    int        retval;
    char * eventStr;
    int num_events = 0;
    int PAPI_event;
    int i;
    char version[PAPI_MAX_STR_LEN];

    debugputs (PS_DEBUG_VERBOSE, "Entering vm_init()...");

    /* The exact JVMTI version doesn't have to match, however this
     *  code demonstrates how you can check that the JVMTI version seen 
     *  in the jvmti.h include file matches that being supplied at runtime
     *  by the VM. 
     */
    error = (*jvmti)->GetVersionNumber(jvmti, &runtime_version);
    check_jvmti_error(jvmti, error, "get JVMTI version number");
    version_check(JVMTI_VERSION, runtime_version);

    /* Allocate memory for the thread list and threadFinished list. */
    g_threadList = (jthread *) calloc (g_max_num_threads, sizeof (jthread));
    g_threadFinished = (sig_atomic_t *)
        calloc (g_max_num_threads, sizeof (sig_atomic_t));
    g_eventSetMutex = (pthread_mutex_t *)
        calloc (g_max_num_threads, sizeof (pthread_mutex_t));

    /* PAPI library init, and thread init. */
    if ((retval = PAPI_library_init (PAPI_VER_CURRENT)) != PAPI_VER_CURRENT) {
        PAPI_ERROR_MSG_RETURN(retval, "PAPI_library_init error");
    }
    if (PAPI_thread_init (pthread_self) != PAPI_OK) {
        PAPI_ERROR_MSG_RETURN(retval, "PAPI_thread_init error");
    }
    /* Get domain from the command line option.
     * Could consider later using libpshwpc to parse and set it.
     */
    debug (PS_DEBUG_INFO, "Setting PAPI event domain to %d.\n", g_eventDomain);
    if (PAPI_set_domain (g_eventDomain) != PAPI_OK) {
        PAPI_ERROR_MSG_RETURN(retval, "PAPI_set_domain error");
    }

    /* Initialize the semaphore. */
    if (sem_init(&g_sem, 0, 0) == -1) {
        perror ("sem_init failed");
        exit (1);
    }

    /* Here we can be sure that g_numEventConfs > 0, otherwise it
     * should have already exited.  See parse_options(...).
     * Assert it again for safety.
     */
    if (g_numEventConfs <= 0) {
        puts ("No profile event is given. Please specify the profile event(s). Examples:\n    java -agentlib:psjprof=events=PAPI_TOT_CYC@1000000000 MyClass\n    java -agentlib:psjprof=events=PAPI_TOT_CYC@1000000000+PAPI_L2_TCA@1000000 MyClass");
        exit (1);
    }
    g_events =
        (PSJP_EVT_VALID *) calloc (g_numEventConfs, sizeof (PSJP_EVT_VALID));

    /* Add the input event configuration. */
    for (i = 0; i < g_numEventConfs; i++) {
        eventStr = (char *) g_eventConfs[i].event_name;
        if ( (retval = PAPI_event_name_to_code (eventStr, &PAPI_event))
             != PAPI_OK ) {
            fprintf (stderr,
                     "Error in PAPI_event_name_to_code for event '%s': %s.  The given event is not a valid event.\n",
                     eventStr, PAPI_strerror (retval));
            exit (retval);
        }
        if ((retval=PAPI_query_event (PAPI_event)) != PAPI_OK) {
            fprintf (stderr,
                     "Error in PAPI_query_event for event '%s': %s.  This usually means that the event is a valid event, but unavailable on this machine.\n",
                     eventStr, PAPI_strerror (retval));
            exit (retval);
        }

        g_events[num_events].name      = g_eventConfs[i].event_name;
        g_events[num_events].code      = PAPI_event;
        g_events[num_events].threshold = g_eventConfs[i].threshold;
        if (PAPI_event | PAPI_PRESET_MASK) {
            g_events[num_events].type = PAPI_EVENT_TYPE_PRESET;
        } else {
            g_events[num_events].type = PAPI_EVENT_TYPE_NATIVE;
        }

        num_events++;
    }

    g_numEvents = num_events;
    debug (PS_DEBUG_INFO, "In vm_init, g_numEvents = %u.\n", g_numEvents);

    /* Initialize the pipe. */
    if (pipe (g_pipefd) == -1) {
        perror ("pipe");
        exit (1);
    }

    if (PSJP_PIPE_NON_BLOCKING == g_pipe_behavior) {
        /* Make the write end of the pipe non-blocking. */
        if (fcntl (g_pipefd[1], F_SETFL, O_NONBLOCK) == -1) {
            perror ("fcntl");
            exit (1);
        }
        debugputs (PS_DEBUG_INFO, "Set the pipe to non-blocking.");
    } else {
        debugputs (PS_DEBUG_INFO, "Set the pipe to blocking.");
    }

    /* Get the cpuid data and mhz, needed by the XML output function. */
    retval = _ps_cpuid(&globals.cpuid);
    if ( retval != PS_SUCCESS ) {
        fprintf (stderr, "ps_cpuid failed: code %d (%s).\n",
                 retval, ps_strerror (retval));
        exit(1);
    }
    retval = ps_cpuspeed(&globals.mhz);
    if ( retval != PS_SUCCESS ) {
        fprintf (stderr, "ps_cpuspeed failed: code %d (%s).\n",
                 retval, ps_strerror (retval));
        exit(1);
    }
    /* Get the PAPI version (for PAPI version >= 3.0.8.2).
     * Since 2005-04-19, papi.h has PAPI_VERSION_INCREMENT.
     * Next released PAPI version was 3.2.1 on 2005-11-23.
     */
    retval = PAPI_get_opt(PAPI_LIB_VERSION, NULL);
    sprintf(version, "%d.%d.%d.%d",
            PAPI_VERSION_MAJOR(retval),
            PAPI_VERSION_MINOR(retval),
            PAPI_VERSION_REVISION(retval),
            PAPI_VERSION_INCREMENT(retval));
    globals.papi_version = strdup(&version[0]);
    /* Get "pid".  It'll be the same for all threads. */
    globals.pid = getpid();
    /* Initialize the main class name to "java", and args to "". */
    globals.main_class_name = "java";
    globals.main_method_args = "";
    /* Assign the event domain. */
    globals.event_domain = g_eventDomainStr;

    /* Create a helper (daemon) thread. */
    createHelperThread(env, jvmti, "PerfSuite Java profiling helper thread",
                      &helper_loop);

    return;
}


/* Check the compatibility of the compile-time and run-time JVMTI versions. */
static void
version_check(jint cver, jint rver)
{
    jint cmajor, cminor;
    jint rmajor, rminor;

    cmajor = (cver & JVMTI_VERSION_MASK_MAJOR) >> JVMTI_VERSION_SHIFT_MAJOR;
    cminor = (cver & JVMTI_VERSION_MASK_MINOR) >> JVMTI_VERSION_SHIFT_MINOR;
    rmajor = (rver & JVMTI_VERSION_MASK_MAJOR) >> JVMTI_VERSION_SHIFT_MAJOR;
    rminor = (rver & JVMTI_VERSION_MASK_MINOR) >> JVMTI_VERSION_SHIFT_MINOR;

    if ( (cmajor > rmajor) || (cmajor == rmajor && cminor > rminor) ) {
	fputs
            ("ERROR: Compile Time JVMTI and Run Time JVMTI are incompatible\n",
             stderr);
        exit (1);
    }
    return;
}

static void
createHelperThread(JNIEnv *env, jvmtiEnv *jvmti,
                   const char *name, jvmtiStartFunction func)
{
    jvmtiError          error;

    jclass          threadClass;
    jmethodID       threadConstructor;
    jmethodID       threadSetDaemon;
    jthread         thread;
    jstring         nameString;
    jthreadGroup    systemThreadGroup;
    jthreadGroup *  groups;
    jint            groupCount;
        
    PSJP_ASSERT(name!=NULL);
    PSJP_ASSERT(func!=NULL);

    thread                  = NULL;
    systemThreadGroup       = NULL;
    groups                  = NULL;
    threadClass             = (*env)->FindClass (env, "java/lang/Thread");
    PSJP_ASSERT(threadClass!=NULL);
    threadConstructor       =
        (*env)->GetMethodID(env, threadClass, "<init>", 
                            "(Ljava/lang/ThreadGroup;Ljava/lang/String;)V");
    threadSetDaemon         =
        (*env)->GetMethodID(env, threadClass, "setDaemon", "(Z)V");
        
    error = JVMTI_FUNC_PTR(jvmti,GetTopThreadGroups)
        (jvmti, &groupCount, &groups);
    if ( error == JVMTI_ERROR_NONE ) {
        if ( groupCount > 0 ) {
            systemThreadGroup = groups[0];
        }
        jvmtiDeallocate(groups);
            
        nameString  = JNI_FUNC_PTR(env,NewStringUTF)(env, name);
        PSJP_ASSERT(nameString!=NULL);
        thread      = JNI_FUNC_PTR(env,NewObject)
            (env, threadClass, threadConstructor, 
             systemThreadGroup, nameString);
        PSJP_ASSERT(thread!=NULL);
        callVoidMethod(env, thread, threadSetDaemon, JNI_TRUE);
           
        error = JVMTI_FUNC_PTR(jvmti,RunAgentThread)
            (jvmti, thread, func, NULL, JVMTI_THREAD_MAX_PRIORITY);
        
    }
    
    check_jvmti_error(jvmti, error, "create helper thread");
    return;
}

static void
callVoidMethod(JNIEnv *env, jobject object, jmethodID method, jboolean arg)
{
    PSJP_ASSERT(env!=NULL);
    PSJP_ASSERT(object!=NULL);
    PSJP_ASSERT(method!=NULL);
    JNI_FUNC_PTR(env,CallVoidMethod)(env, object, method, arg);
    return;
}

static void JNICALL
helper_loop (jvmtiEnv *jvmti, JNIEnv *env, void *p)
{
    jvmtiError error;
    SplayTree ** sampleTrees;
    hwpc_papi_tls_t * tls;
    MethodLocPair key1;

    jthread thisThread;
    jint depth = GET_STACK_DEPTH;
    jvmtiFrameInfo frames[GET_STACK_DEPTH];
    int fcount;
    int fi;

    int EventSet = 0;
    long long overflow_vector;
    PSJP_PIPE_DATA pipedata;
    int *Events;
    int number, i, retval;
    int eventIndex;
    int hasError = 0;

    Events = (int *) malloc(g_numEvents*sizeof(int));
    if (NULL == Events) {
        puts ("malloc error for Events in helper_loop.");
        exit (1);
    }

    while ( 1 ) {
        if (read (g_pipefd[0], &pipedata, PIPE_DATA_SIZE) == -1) {
            /* The following line is commented out after development.
             * This block is usually entered at JVM destroy time.
             * After the pipe file descriptor is closed in VM_death(),
             * read() returns EBADF (Bad file descriptor).
             *
             * perror ("read");
             */
            break;
        }

        EventSet = pipedata.eventset;
        overflow_vector = pipedata.overflow_vector;

        /* Validate EventSet.
         * If the value is invalid, indicating
         * that this sample has an issue, ignore it.
         */
        if ( (EventSet < 0) || (EventSet >= g_max_num_threads) ) {
            debug (PS_DEBUG_WARNING, "Error: EventSet (%d) is negative or exceeds the limit %d.\n", EventSet, g_max_num_threads-1);
            continue;
        }

        MUTEX_LOCK (&g_eventSetMutex[EventSet]);

        /* If this thread is finished, ignore this sample. */
        if (g_threadFinished[EventSet]) {
            debug (PS_DEBUG_INFO, "Info: the thread for the event set %d is finished, skip sampling it.\n", EventSet);
            MUTEX_UNLOCK (&g_eventSetMutex[EventSet]);
            continue;
        }

        /* Events[number-1] won't be out of bound, according to the
         * PAPI man page for the function.  On input, "number" indicates
         * the size of the array "Events"; on output, it is set to
         * the number of indexes stored in the array.  In the case
         * where the array is too short to store all the indexes,
         * "number" is set to the size of the array, so
         * Events[number-1] won't be out of bound.
         */
        number = g_numEvents;
        retval = PAPI_get_overflow_event_index
            (EventSet, overflow_vector, Events, &number);
        if (retval == PAPI_OK) {
            for(i=0; i<number; i++) {
                eventIndex = Events[i];
                if (eventIndex < 0 || (unsigned int) eventIndex >= g_numEvents) {
                    debug (PS_DEBUG_WARNING, "Error: eventIndex (%d) is negative or exceeds the limit %d.\n", eventIndex, g_numEvents-1);
                    /* Ignore this entire sample if the data is invalid. */
                    hasError = 1;
                    break;
                }
            }
        } else {
            debug (PS_DEBUG_WARNING,
                   "PAPI_get_overflow_event_index returned %d.\n", retval);
            hasError = 1;
            /* Ignore the sample if there is an error
             * getting the event index.
             *
             * One scenario observed:
             * the thread has finished, so it returns -11 (PAPI_ENOEVST)
             * -- the event set does not exist.
             *
             * After commenting out PAPI_destroy_eventset call,
             * PAPI_get_overflow_event_index won't return -11,
             * but still could return -1 (PAPI_EINVAL),
             * indicating the eventset is empty.
             * This will occur only after the thread has finished.
             */
        }

        if (hasError) {
            MUTEX_UNLOCK (&g_eventSetMutex[EventSet]);
            continue;
        }


        thisThread = g_threadList[EventSet];

        /* Get the stack trace. */
        error = JVMTI_FUNC_PTR(jvmti,GetStackTrace)
            (jvmti, thisThread, 0, depth, frames, &fcount);
        if ( JVMTI_ERROR_NONE == error ) {
            if (fcount > 0) {
                error = JVMTI_FUNC_PTR(jvmti, GetThreadLocalStorage)
                    (jvmti, thisThread, (void **) &tls);
                check_jvmti_error(jvmti, error,
                                  "GetThreadLocalStorage in helper_loop");
                sampleTrees = tls->sample_trees;
                /* The following loop can be simplified, but left here
                 * for possible future work of getting the call path,
                 * instead of just the last method being called.
                 */
                for (fi = 0; fi < fcount; fi++) {
                    /* Get the methodID and location. */
                    key1.methodID = (void *) frames[fi].method;
                    key1.location = frames[fi].location;
                    for (i = 0; i < number; i++) {
                        eventIndex = Events[i];
                        splay_insert (key1, sampleTrees[eventIndex]);
                    }
                    /* Only use the top frame for now. */
                    break;
                }
            } else {
                debugputs (PS_DEBUG_VERBOSE, "Got no frame in GetStackTrace.");
            }
        } else {
            print_jvmti_error (jvmti, error, "get thread stack trace");
            MUTEX_UNLOCK (&g_eventSetMutex[EventSet]);
            /* Ignore the sample if there is an error
             * getting the thread's stack traces.
             */
            continue;
        }
        MUTEX_UNLOCK (&g_eventSetMutex[EventSet]);
    }

    free (Events);

    if (sem_post(&g_sem) == -1) {
        perror ("sem_post");
    }

    return;
}

static inline int
shouldIgnoreThisThread (jvmtiThreadInfo * info)
{
    int i;
    const char * iname;
    if ( (info->is_daemon) ||
         (NULL == info->name) ||
         (strcmp (info->name, "Attach API teardown") == 0) ||
         (strstr(info->name, "DestroyJavaVM")) ) {
        /* Sun Java 1.5, 1.6: "DestroyJavaVM";
         * IBM Java 1.6: "DestroyJavaVM helper thread".
         * IBM Java 1.6: "Attach API teardown".
         */
        return 1;
    }
    for (i = 0; i < g_numIgnoredThreadNames; i++) {
        iname = g_ignoredThreadNames[i];
        if (NULL == iname) {
            break;
        }
        if (0 == strcmp (info->name, iname)) {
            return 1;
        }
    }
    return 0;
}

static void JNICALL 
thread_start (jvmtiEnv *jvmti, JNIEnv *env, jthread thread)
{
    jvmtiError        error;
    jvmtiThreadInfo   threadInfo;
    jthread           thisThread;
    hwpc_papi_tls_t * tls = NULL;
    SplayTree **      sampleTrees = NULL;
    int               PAPI_EventSet;
    unsigned int      i;

    (void)memset((void*)&threadInfo, 0, sizeof(jvmtiThreadInfo));
    error = JVMTI_FUNC_PTR(jvmti,GetThreadInfo)
        (jvmti, thread, &threadInfo);

    if (shouldIgnoreThisThread (&threadInfo)) {
        return;
    }

    debug (PS_DEBUG_INFO, "in thread_start, past daemon check, thread name: '%s'\n", threadInfo.name);

    sampleTrees = (SplayTree **) malloc (g_numEvents * sizeof (SplayTree *));
    for (i = 0; i < g_numEvents; i++) {
        sampleTrees[i] = (SplayTree *) calloc (1, sizeof (SplayTree));
    }
    tls = (hwpc_papi_tls_t*) calloc (1, sizeof (hwpc_papi_tls_t));

    PAPI_EventSet = startPAPI ();

    tls->sample_trees = sampleTrees;
    tls->eventset = PAPI_EventSet;
    debug (PS_DEBUG_INFO, "PAPI_EventSet = %d.\n", PAPI_EventSet);

    /* Set tls for the thread. */
    thisThread = (*env)->NewGlobalRef (env, thread);
    error = JVMTI_FUNC_PTR(jvmti,SetThreadLocalStorage)
        (jvmti, thisThread, tls);
    check_jvmti_error(jvmti, error, "SetThreadLocalStorage in thread_start");

    /* Add the thread to the thread list. */
    g_threadList[PAPI_EventSet] = thisThread;

    /* startPAPI validated the value of the created event set,
     * so no need to check here.
     */
    MUTEX_INIT (&g_eventSetMutex[PAPI_EventSet], NULL);

    ps_rtc(&tls->starttime);

    debugputs (PS_DEBUG_VERBOSE, " exiting thread_start...");

    return;
}

static int
startPAPI (void)
{

    int retval;
    unsigned int i;
    int eventset = PAPI_NULL;

    /* event set must be null before calling PAPI_create_eventset. */
    if ((retval=PAPI_create_eventset (&eventset)) != PAPI_OK) {
        PAPI_ERROR_MSG_RETURN(retval, "PAPI_create_eventset");
    }

    /* Check whether the returned event set is within the valid range,
     * and exit if not.
     */
    if ( (eventset < 0) || (eventset >= g_max_num_threads) ) {
        PAPI_ERROR_MSG_RETURN(retval, "PAPI_create_eventset created an eventset whose value is out of range");
    }

    /* Add the event, and register the overflow with the threshold. */ 
    for (i = 0; i < g_numEvents; i++) {
        if ((retval=PAPI_add_event (eventset, g_events[i].code)) != PAPI_OK) {
            PAPI_ERROR_MSG_RETURN(retval, "PAPI_add_event");
        }
        if ((retval = PAPI_overflow (eventset, g_events[i].code,
                                     g_events[i].threshold, 0,
                                     handler)) != PAPI_OK) {
            PAPI_ERROR_MSG_RETURN(retval, "PAPI_overflow");
        }
    }

    /* Start PAPI. */
    if ( (retval=PAPI_start (eventset)) != PAPI_OK) {
        PAPI_ERROR_MSG_RETURN(retval, "PAPI_start");
    }

    return eventset;
}

/* The handler called by PAPI_overflow. */
static void
handler (int EventSet, void *address, long long overflow_vector, void *context)
{
    PSJP_PIPE_DATA pipedata;
    ssize_t status;
    pipedata.eventset = EventSet;
    pipedata.overflow_vector = overflow_vector;
    status = write (g_pipefd[1], &pipedata, PIPE_DATA_SIZE);
    /* Ignore the return value of write().
     * Here we used to check the return value against -1 for error, and
     * call perror() to print if so.  Never saw an error in developement.
     * perror() is not async-signal-safe, so removed it in production code.
     */

    return;
}

static void JNICALL 
thread_end (jvmtiEnv *jvmti, JNIEnv *env, jthread thread)
{
    SplayTreeNode ** sarray = NULL;
    unsigned int numNodes = 0;
    int status = 0;
    unsigned int i;
    jvmtiError error;
    jvmtiThreadInfo threadInfo;
    hwpc_papi_tls_t * tls;
    SplayTree *  sampleTree;
    SplayTree ** sampleTrees;
    int PAPI_EventSet;
    struct hsearch_data method_table;
    int lineno;
    unsigned int maxTreeNumNodes = 0;
    unsigned int event_idx;

    /* For writing out XML files. */
    FILE * fout = NULL;
    int close = 1;
    /* A non-0 non--1 initial value to check whether it changed as expected. */
    int tid = -2;
    ps_time_t usertime, systime;
    double usersecs, systemsecs;
    unsigned long long currenttime;
    char filename[256], hostname[256], *firstdot;
    int totalCount;   /* The number of all samples in a tree. */
    psjp_thread_event_profile_t * profs;
    char * base_class_name = "java";


    ps_rtc(&currenttime);

    debugputs (PS_DEBUG_VERBOSE, "entering thread_end...");

    (void)memset((void*)&threadInfo, 0, sizeof(jvmtiThreadInfo));
    error = JVMTI_FUNC_PTR(jvmti,GetThreadInfo)
        (jvmti, thread, &threadInfo);
    if (shouldIgnoreThisThread (&threadInfo)) {
        return;
    }

    /* Determine the CPU usage of this process/(thread?). */
    if ( ps_cpuusage(0, &usertime, &systime) != PS_SUCCESS ) {
        debugputs (PS_DEBUG_WARNING, "Error in retrieving process time");
    }
    debug (PS_DEBUG_VERBOSE,
           "CPU time secs/msecs (user: %ld/%ld, sys: %ld/%ld)\n",
           usertime.secs, usertime.msecs, systime.secs, 
           systime.msecs);
    usersecs = usertime.secs + (usertime.msecs*1.e-6);
    systemsecs = systime.secs + (systime.msecs*1.e-6);

    error = JVMTI_FUNC_PTR (jvmti, GetThreadLocalStorage)
        (jvmti, thread, (void **)&tls);
    check_jvmti_error(jvmti, error, "GetThreadLocalStorage in thread_end");
    sampleTrees = tls->sample_trees;
    PAPI_EventSet = tls->eventset;

    /* Validate the value of the event set. */
    if ( (PAPI_EventSet < 0) || (PAPI_EventSet >= g_max_num_threads) ) {
        debug (PS_DEBUG_WARNING, "Error: EventSet (%d) in the thread's local storage is negative or exceeds the limit %d.\n", PAPI_EventSet, g_max_num_threads-1);
        /* Silent return.  To see the error, one can open the debug output
         * by setting PSJP_DEBUG to 2 or above.
         */
        return;
    }

    MUTEX_LOCK (&g_eventSetMutex[PAPI_EventSet]);
    g_threadFinished[PAPI_EventSet] = 1;
    MUTEX_UNLOCK (&g_eventSetMutex[PAPI_EventSet]);

    tid = PAPI_EventSet;
    /* Use PAPI_EventSet's value as the tid, before calling stopPAPI().
     * After stopPAPI(), PAPI_EventSet's value becomes -1.
     */
    stopPAPI (PAPI_EventSet);

    for (i = 0; i < g_numEvents; i++) {
        if (maxTreeNumNodes < sampleTrees[i]->numNodes) {
            maxTreeNumNodes = sampleTrees[i]->numNodes;
        }
    }
    debug (PS_DEBUG_INFO,
           "maxTreeNumNodes = %u, g_numEvents = %u.\n",
           maxTreeNumNodes, g_numEvents);

    /* If no sample was collected, still need to write this information
     * in the XML files, since this information itself could be useful.
     */

    /* Use a cache for method -> (name, linenumbertable) lookup.
     * Create a hash table to be used as the cache.
     * 
     * Do this only if at least one sample was collected.
     */
    memset (&method_table, 0, sizeof (method_table));
    if (maxTreeNumNodes > 0) {
        if (hcreate_r (maxTreeNumNodes, &method_table) == 0) {
            fputs ("hcreate_r failed.\n", stderr);
            exit (1);
        }
    }

    /* Use "calloc" instead of "malloc", so that when not populated,
     * the value of number of total samples will be 0.
     */
    profs = (psjp_thread_event_profile_t *)
        calloc (g_numEvents, sizeof (psjp_thread_event_profile_t));
    for (event_idx = 0; event_idx < g_numEvents; event_idx++) {

        psjp_loc_t *locs;

        /* Fill in the data, to be written out later:
         *     event_name, event_type ("preset", "native"), event_threshold.
         */
        profs[event_idx].event_name      = g_events[event_idx].name;
        profs[event_idx].event_type      = g_events[event_idx].type;
        profs[event_idx].event_threshold = g_events[event_idx].threshold;

        sampleTree = sampleTrees[event_idx];

        /* No sample was collected for this event in this thread.
         * Skip tree walk and line number resolution.
         */
        if (0 == sampleTree->numNodes) {
            continue;
        }

        status = splay_treewalk (sampleTree, &sarray, &numNodes, &totalCount);
        if (status != PS_SUCCESS) {
            debug (PS_DEBUG_WARNING,
                   "error: splay_treewalk() returned %d, '%s'.\n",
                   status, ps_strerror (status));
            continue;
        }

        /* "2 *" below is because the array duplicated the data. */
        debug (PS_DEBUG_INFO,
               "Total profiling memory buffer space required (bytes) = %zu.\n",
               2 * numNodes * sizeof (SplayTreeNode));

        locs = (psjp_loc_t *)malloc(numNodes * sizeof(psjp_loc_t));

        for (i = 0; i < numNodes; i++) {
            jmethodID method = sarray[i]->key.methodID;
            jlocation location = sarray[i]->key.location;

            ENTRY e, *ep;
            METHOD_DATA * methodData =
                (METHOD_DATA *) malloc (sizeof(METHOD_DATA));
            /* 32 bit: 0x + 8 chars, 64 bit: 0x + 16 chars,
             * so needs 18 + 1 (for ending null) = 19 bytes,
             * allocate 24 bytes for better alignment.
             */
            const int MAX_LEN = 24;
            char * methodStr = (char *) malloc (MAX_LEN);
            snprintf (methodStr, MAX_LEN, "%p", method);

            e.key = methodStr;
            status = hsearch_r (e, FIND, &ep, &method_table);
            if (ep == NULL) {
                jvmtiGetMethodInfo (method, methodData, 0);
                e.data = methodData;
                status = hsearch_r (e, ENTER, &ep, &method_table);
                if (status == 0) {
                    fputs ("hsearch_r failed.\n", stderr);
                }
            }

            lineno = getLineNumber (method, location,
                                    ((METHOD_DATA *) ep->data)->lineNumberTable,
                                    ((METHOD_DATA *) ep->data)->count);

            locs[i].module   = ((METHOD_DATA *) ep->data)->classname;
            locs[i].file     = ((METHOD_DATA *) ep->data)->sourceFileName;
            locs[i].function = ((METHOD_DATA *) ep->data)->name;
            locs[i].line = lineno;
            locs[i].num_samples = sarray[i]->count;

        }

        /* Fill in the data, to be written out later. */
        profs[event_idx].num_nodes = numNodes;
        profs[event_idx].locs = locs;
        profs[event_idx].num_total_samples = totalCount;
    }


    /* Write the XML file.
     * Done changing hardcoded "filename" ("java") and "exebase" ("java").
     * _ps_hwpc_select_stream calls generate_fd(), which calls
     * get_values(), which then allocates memory and get myid.  This is
     * unnecessary overhead for Java profiling XML output, and the "myid"
     * value is incorrect for use in this case.  So implemented it again.
     * The downside is that only one type of file name is used, namely,
     *    <command>.<tid>.<pid>.<hostname>.xml.
     */
    if ( gethostname(hostname, 256) == 0 ) {
        firstdot = strchr(hostname, '.');
        if ( firstdot != NULL ) {
            *firstdot = '\0';
        }
    } else {
        strcpy(hostname, "nohost");
    }

    /* base_class_name: initialized to "java". */
    if (globals.main_class_name != NULL) {
        char * last_dot = strrchr (globals.main_class_name, '.');
        if ( (last_dot != NULL) && (*(last_dot+1) != 0)) {
            base_class_name = last_dot+1;
        } else {
            base_class_name = globals.main_class_name;
        }
    }

    /* Use the last segment of the fully-qualified class name to generate
     * the output file name.  Could consider using the thread's name too.
     */
    sprintf(filename, "%s.%d.%d.%s.%s",
            base_class_name, tid, globals.pid, hostname, PS_SUFFIX_XML);
    fout = fopen (filename, "w");
    if (NULL == fout) {
        debug (PS_DEBUG_WARNING, "thread_end: error opening output stream for file named '%s'", filename);
        return;
    }

    _ps_hwpc_xml_write_java (fout, "java profiling",
                             &globals,
                             tid, currenttime - tls->starttime,
                             usersecs, systemsecs,
                             g_numEvents, profs,
                             close, threadInfo.name);

    return;
}

static void
stopPAPI (int EventSet)
{
    /* Assume that it is safe if multiple threads call PAPI_stop
     * at the same time.
     *
     * The values read by PAPI_stop are discarded.
     * First tried to use one global variable for it to avoid multiple malloc,
     * but potentially multiple threads writing into the same area could
     * cause problem, so did not do this level of optimization yet.
    */
    int retval = 0;
    long long * values;

    values = (long long *) malloc (g_numEvents*sizeof(long long));
    if (NULL == values) {
        puts ("malloc error for values in stopPAPI.");
        exit (1);
    }

    /* Stop the counters and read the values (which are discarded). */
    if ( (retval=PAPI_stop (EventSet, values)) != PAPI_OK) {
        PAPI_ERROR_MSG_RETURN(retval, "PAPI_stop");
    }

    free (values);

    /* PAPI API doc says that
     * PAPI_cleanup_eventset() removes all events from a PAPI event set
     * and turns off profiling and overflow for all events in the eventset.
     * This can not be called if the EventSet is not stopped.
     * PAPI_destroy_eventset() deallocates the memory associated with
     * an empty PAPI event set.
     *
     * so it seems that we do not need to explicitly turn overflow off.
     */
    /* clear the overflow status */
    /*
      if ((retval = PAPI_overflow (EventSet, PAPI_event, 0, 0,
      handler)) != PAPI_OK)
      ERROR_RETURN(retval);
    */

    /************************************************************************
     * PAPI_cleanup_eventset can only be used after the counter has been    *
     * stopped then it remove all events in the eventset                    *
     ************************************************************************/
    if ( (retval=PAPI_cleanup_eventset (EventSet)) != PAPI_OK) {
        PAPI_ERROR_MSG_RETURN(retval, "PAPI_cleanup_eventset");
    }

    /* We should not destroy the event set.  If the event set is destroyed,
     * PAPI will recycle the value of the event set.  This behavior
     * intervenes our use of the event set as the TID, and as the key
     * to look up the associated thread.  Confirmed in a test.
     *
     * If we still want to free the resources, we could register them
     * and free them all together at vm_death time.  But it might be
     * unnecessary at that time.
     */
    /* Free all memory and data structures.  EventSet must be empty. */
    /*
    if ( (retval=PAPI_destroy_eventset(&EventSet)) != PAPI_OK) {
        PAPI_ERROR_MSG_RETURN(retval, "PAPI_destroy_eventset");
    }
    */

    return;
}

static jint 
getLineNumber (jmethodID method, jlocation location,
               jvmtiLineNumberEntry *ppentries, int count)
{
    jint lineno;

    PSJP_ASSERT(method!=NULL);
    if ( location < 0 ) {
        PSJP_ASSERT(location > -4);
        return -1;
    }
    lineno = map_loc2line(location, ppentries, count);
    return lineno;
}

static inline jint
map_loc2line(jlocation location, jvmtiLineNumberEntry *table, jint count)
{
    jint line_number;

    PSJP_ASSERT(location>=0);
    PSJP_ASSERT(count>=0);

    line_number = -1;
    if ( count == 0 ) {
        return line_number;
    }
    
    {
        int low = 0;
        int high = count-1;
        int mid;
        while ((high - low) > 1) {
            mid = low + ((high - low) >> 1);
            if (table[mid].start_location < location) {
                low = mid;
            } else if (table[mid].start_location > location) {
                high = mid - 1;
            } else {
                low = mid;
                break;
            }
        }
        if (table[high].start_location < location) {
            line_number = table[high].line_number;
        } else {
            line_number = table[low].line_number;
        }
    }

    PSJP_ASSERT(line_number > 0);

    return line_number;
}

/* This function uses JVMTI functions to resolve details about a
 * method given a methodID.  It does not perform any hash table operation.
 *
 * It finds the method name, class name, source file name and
 * the line number table associated with the method.
 */
static void
jvmtiGetMethodInfo (jmethodID methodID, METHOD_DATA * methodData,
                  int called_from_method_entry)
{
    char *pname;
    char *psignature;
    char *generic_signature;
    jint pcount = 0;
    jvmtiLineNumberEntry *ppentries = NULL;
    jvmtiError error;
    jclass clazz;
    char * pclasssignature;
    char * psourceFileName;
    int i;
    int slen;

    /* Get the method name. */
    error = JVMTI_FUNC_PTR(g_jvmti,GetMethodName)
        (g_jvmti, methodID, &pname, &psignature, &generic_signature);
    methodData->name = pname;
    jvmtiDeallocate (psignature);
    jvmtiDeallocate (generic_signature);

    /* If called from method_entry, check whether the method name is "main",
     * and only if so, get the class name.
     */
    if (called_from_method_entry && (strcmp (pname, "main"))) {
        return;
    }

    /* Get the name of the declaring class. */
    error = JVMTI_FUNC_PTR(g_jvmti,GetMethodDeclaringClass)
        (g_jvmti, methodID, &clazz);
    check_jvmti_error(g_jvmti, error, "get method class");

    error = JVMTI_FUNC_PTR(g_jvmti,GetClassSignature)
        (g_jvmti, clazz, &pclasssignature, &generic_signature);
    methodData->classname = "\0";
    slen = strlen (pclasssignature);
    if ( (error == JVMTI_ERROR_NONE) &&
         (pclasssignature[0] == 'L') &&
         (pclasssignature[slen-1] == ';') ) {
        /* Skip the leading 'L' and trailing ';'.
         * Change '/' to '.'.
         */
        for (i = 1; i < slen - 1; i++) {
            if (pclasssignature[i] == '/') {
                pclasssignature[i] = '.';
            }
        }
        pclasssignature[slen-1] = 0; /* Change ';' to '\0'. */
        methodData->classname = &pclasssignature[1]; /* Skip the leading 'L'. */
    } else {
        check_jvmti_error(g_jvmti, error, "get class signature");
    }
    jvmtiDeallocate(generic_signature);

    /* Return if called from method_entry.  When it reaches here,
     * the method name must have been "main".  In this case,
     * we only need the class name for the "main" method,
     * and do not need more details.
     */
    if (called_from_method_entry) {
        return;
    }

    /* Get the source file name. */
    error = JVMTI_FUNC_PTR(g_jvmti,GetSourceFileName)
        (g_jvmti, clazz, &psourceFileName);
    methodData->sourceFileName = psourceFileName;
    if ( error == JVMTI_ERROR_ABSENT_INFORMATION ) {
        methodData->sourceFileName = "\0";
    }

    /* Get the line number table. */
    methodData->lineNumberTable = NULL;
    methodData->count = 0;
    error = JVMTI_FUNC_PTR(g_jvmti,GetLineNumberTable)
        (g_jvmti, methodID, &pcount, &ppentries);
    if ( error == JVMTI_ERROR_ABSENT_INFORMATION ) {
        error = JVMTI_ERROR_NONE;
    }
    if ( error != JVMTI_ERROR_NONE ) {
        print_jvmti_error (g_jvmti, error, "get source line numbers");
    } else {
        methodData->lineNumberTable = ppentries;
        methodData->count = pcount;
    }

    return;
}

/* Callback for JVMTI_EVENT_VM_DEATH */
static void JNICALL
vm_death (jvmtiEnv *jvmti, JNIEnv *env)
{
    int error;
    int i;

    debugputs (PS_DEBUG_VERBOSE, "Entering vm_death...");

    /* Close the pipe. */
    close (g_pipefd[0]);
    close (g_pipefd[1]);

    /* Wait for the semaphore. */
    error = sem_wait (&g_sem);
    if (error == -1) {
        perror ("sem_wait");
    }

    /* Free the resources used by PAPI. */
    PAPI_shutdown();

#ifndef NO_MUTEX
    /* Destroy the mutexes. */
    for (i = 0; i < g_max_num_threads; i++) {
        if (1 == g_threadFinished[i]) {
            MUTEX_DESTROY (&g_eventSetMutex[i]);
        }
    }
#endif

    debugputs (PS_DEBUG_VERBOSE, " Exiting vm_death...");
    return;
}

static void
jvmtiDeallocate(void *ptr)
{
    if ( ptr != NULL ) {
        jvmtiError error;

        error = JVMTI_FUNC_PTR(g_jvmti,Deallocate)
            (g_jvmti, (unsigned char*)ptr);
        check_jvmti_error(g_jvmti, error, "deallocate jvmti memory");
    }
    return;
}

static int 
get_tok(char **src, char *buf, int buflen, int sep)
{
    int len;
    char *p;
    
    buf[0] = 0;
    if ( **src == 0 ) {
	return 0;
    }
    p = strchr(*src, sep);
    if ( p==NULL ) {
	len = (int)strlen(*src);
	p = (*src) + len;
    } else {
	/*LINTED*/
	len = (int)(p - (*src));
    }
    if ( (len+1) > buflen ) {
	return 0;
    }
    (void)memcpy(buf, *src, len);
    buf[len] = 0;
    if ( *p != 0 && *p == sep ) {
	(*src) = p+1;
    } else {
	(*src) = p;
    }
    return len;
}

static void
option_error(char *description)
{
    printf ("option error: %s\n", description);
    exit (1);
}

static void
parse_options (char *command_line_options)
{
    char *options;
    int   ulen;
    int numEventConfs = 0;
    int numSeparators;  /* The number of event config separators, '+'. */
    const char EVENT_CONF_SEPARATOR_CHAR = '+';
    int i;  /* Loop index. */

    if (command_line_options == 0) {
        command_line_options = "";
    }
    
    ulen = strlen (command_line_options);
    options = (char *) malloc (ulen + 1);
    memcpy (options, command_line_options, ulen);
    options[ulen] = 0;

    debug (PS_DEBUG_INFO, "command line options: '%s'\n", options);

    /* Find out the number of the event config separation character -- '+',
     * to be used as the number of entries for memory allocation of
     * "g_eventConfs".
     */
    numSeparators = 0;
    for (i = 0; i <= ulen; i++) {
        if (EVENT_CONF_SEPARATOR_CHAR == options[i]) {
            numSeparators++;
        }
    }

    /* Allocate memory for input event configuration.
     * The "+1" is necessary.  The number of events given on
     * the command line is the number of separators plus 1.
     */
    g_eventConfs =
        (PSJP_EVT_CONF *) calloc (numSeparators+1, sizeof (PSJP_EVT_CONF));
    /* Allocate memory for the ignored thread names. */
    g_ignoredThreadNames = (const char **)
        calloc (1, g_numIgnoredThreadNames * sizeof (char *));

    while (*options) {
        /* This is to contain option names such as "events" and
         * "ignored_thread_names", and needs to be large enough.
         * Use 80 for now.  An older value of 16 caused a problem
         * for "ignored_thread_names".
         */
        char option[80];
        char suboption[FILENAME_MAX+1];

        if (!get_tok(&options, option, (int)sizeof(option), '=')) {
            option_error("general syntax error parsing options");
        }
        debug (PS_DEBUG_INFO, "get_tok got option '%s'\n", option);
        if (strcmp(option, "events") == 0) {
            /* Add the input event configuration. */
            char thresholdStr[16];
            int threshold;
            char eventsStr[4096];
            char *pevents = &eventsStr[0];
            numEventConfs = 0;

            if (!get_tok(&options, eventsStr, (int)sizeof(eventsStr), ',')) {
                option_error("events option missing ','");
            }

            while (*pevents) {
                if (!get_tok(&pevents, suboption,
                             (int)sizeof(suboption), '@')) {
                    option_error("events option missing '@'");
                }
                debug (PS_DEBUG_INFO,
                       "get_tok got suboption '%s'\n", suboption);
                if (!get_tok(&pevents, thresholdStr, 16, '+')) {
                    option_error("events option missing '+'");
                }
                debug (PS_DEBUG_INFO,
                       "get_tok got thresholdStr '%s'\n", thresholdStr);
                /* "atoi(input)" is the same as "strtol(input, NULL, 10)"
                 * except that it does not set the errno.  PAPI_overflow()
                 * uses "int" as the data type of "threshold", so also uses
                 * "int" and thus "atoi", but guard against under and over-flow.
                 */
                threshold = atoi (thresholdStr);
                if (threshold > 0 && (threshold != INT_MAX)) {
                    /* Use strdup to get persistent storage
                     * for the event names.
                     */
                    g_eventConfs[numEventConfs].event_name = strdup (suboption);
                    g_eventConfs[numEventConfs].threshold = threshold;
                    numEventConfs++;
                } else {
                    printf ("error in option threshold (%s), should be a positive integer that is less than INT_MAX (%d).\n", thresholdStr, INT_MAX);
                    exit (1);
                }
            }
        } else if (strcmp(option, "domain") == 0) {
            char eventDomain[16];
            if (!get_tok(&options, eventDomain, sizeof(eventDomain), ',')) {
                option_error("domain option missing ','");
            }
            if (strcasecmp(eventDomain, "user") == 0) {
                g_eventDomain = PAPI_DOM_USER;
            } else if (strcasecmp(eventDomain, "kernel") == 0) {
                g_eventDomain = PAPI_DOM_KERNEL;
            } else if (strcasecmp(eventDomain, "all") == 0) {
                g_eventDomain = PAPI_DOM_ALL;
            } else {
                puts ("Wrong event domain specified.  Should be 'user', 'kernel', or 'all'.");
                exit (1);
            }
            debug (PS_DEBUG_INFO,
                   "User specified event domain is: '%s' (%d).\n",
                   eventDomain, g_eventDomain);
            g_eventDomainStr = strdup (&eventDomain[0]);
            
        } else if (strcmp(option, "max_num_threads") == 0) {
            char maxNumThreadsStr[16];
            int  maxNumThreads;

            if (!get_tok(&options, maxNumThreadsStr,
                         sizeof(maxNumThreadsStr), ',')) {
                option_error("max_num_threads option missing ','");
            }
            maxNumThreads = atoi (maxNumThreadsStr);
            if ( (maxNumThreads > 0) && (maxNumThreads != INT_MAX) ) {
                g_max_num_threads = maxNumThreads;
                debug (PS_DEBUG_INFO,
                       "Setting g_max_num_threads to %d.\n", g_max_num_threads);
            } else {
                printf ("Error: max_num_threads should be >0 and <INT_MAX(%d), but '%s' was given.", INT_MAX, maxNumThreadsStr);
                exit (1);
            }
        } else if (strcmp(option, "pipe") == 0) {
            char blockingStr[16];

            if (!get_tok(&options, blockingStr, sizeof(blockingStr), ',')) {
                option_error("pipe option missing ','");
            }
            if (0 == strcmp(&blockingStr[0], "blocking")) {
                g_pipe_behavior = PSJP_PIPE_BLOCKING;
            } else if (0 == strcmp(&blockingStr[0], "non-blocking")) {
                g_pipe_behavior = PSJP_PIPE_NON_BLOCKING;
            } else {
                printf ("Error: pipe option value should be \"non-blocking\" or \"blocking\", but '%s' was given.", &blockingStr[0]);
                exit (1);
            }
        } else if (strcmp(option, "ignored_thread_names") == 0) {
            char ignoredThreadNamesStr[160];
            char * pnames = &ignoredThreadNamesStr[0];
            char nameStr[80];
            int nameCount = 0;

            if (!get_tok(&options, ignoredThreadNamesStr,
                         sizeof(ignoredThreadNamesStr), ',')) {
                option_error("ignored_thread_names option missing ','");
            }
            while (*pnames) {
                if (!get_tok(&pnames, nameStr, sizeof(nameStr), '+')) {
                    option_error("ignored_thread_names option missing '+'");
                }
                g_ignoredThreadNames[nameCount] = strdup (nameStr);
                printf ("Set g_ignoredThreadNames[%d] to '%s'\n",
                        nameCount, nameStr);
                nameCount++;
                if (nameCount >= g_numIgnoredThreadNames) {
                    break;
                }
            }
        } else {
            printf ("Error: unknown option '%s'.\n", option);
            exit (1);
        }
    }
    g_numEventConfs = numEventConfs;
    debug (PS_DEBUG_INFO, "g_numEventConfs = %d.\n", g_numEventConfs);
    if (0 == g_numEventConfs) {
        puts ("No profile event is given. Please specify the profile event(s). Examples:\n    java -agentlib:psjprof=events=PAPI_TOT_CYC@1000000000 MyClass\n    java -agentlib:psjprof=events=PAPI_TOT_CYC@1000000000+PAPI_L2_TCA@1000000 MyClass");
        exit (1);
    }
    return;
}

/* After getting the "main" method, disable the METHOD_ENTRY event
 * notification, and relinquish the capability, to remove the
 * significant overhead.
 */
static void JNICALL
method_entry (jvmtiEnv *jvmti, JNIEnv* env, jthread thread, jmethodID methodID)
{
    METHOD_DATA mdata;
    jvmtiError  error;
    jarray array;
    jvmtiCapabilities needed_capabilities;

    jvmtiGetMethodInfo (methodID, &mdata, 1);
    if (strcmp (mdata.name, "main") != 0) {
        return;
    }

    /* The classname came from GetClassSignature call, the memory
     * was allocated and never freed unless a user explicitly calls
     * Deallocate(), so no need to do strdup().
     */
    globals.main_class_name = (char *) mdata.classname;


    /* GetLocalObject's signature is:
     *     (jvmtiEnv* env, jthread thread,
     *      jint depth, jint slot, jobject* value_ptr)
     *
     * From the JVM specification,
     * the arguments were stored in the first slot (slot = 0).
     *
     * Frame 0 means the current frame.
     */
    error = JVMTI_FUNC_PTR(g_jvmti,GetLocalObject)
        (g_jvmti, thread, 0, 0, &array);
    print_jvmti_error(g_jvmti, error, "get local object");
    if (JVMTI_ERROR_NONE == error) {
        jsize sizeArgs = (*env)->GetArrayLength(env, array);
        debug (PS_DEBUG_INFO,
               "local array of 'main' method has length %d.\n", sizeArgs);
        if (sizeArgs > 0) {
            int i;
            int argsStrLen = 0;
            const char ** mainArgs =
                (const char **) malloc (sizeArgs * sizeof(char *));
            for (i = 0; i < sizeArgs; i++) {
                jstring str = (*env)->GetObjectArrayElement (env, array, i);
                mainArgs[i] = (*env)->GetStringUTFChars(env, str, 0);
                argsStrLen += (strlen (mainArgs[i]) + 1);
                debug (PS_DEBUG_INFO, "arg[%d]: %s.\n", i, mainArgs[i]);
            }
            globals.main_method_args = (char *) calloc (1, argsStrLen + 1);
            for (i = 0; i < sizeArgs; i++) {
                /* Use spaces to separate the arguments. */
                strcat (globals.main_method_args, " ");
                strcat (globals.main_method_args, mainArgs[i]);
            }
        }
    }

    /* Disable the event notification. */
    error = (*jvmti)->SetEventNotificationMode
        (jvmti, JVMTI_DISABLE, JVMTI_EVENT_METHOD_ENTRY, NULL);
    check_jvmti_error(jvmti, error, "set event notify, MethodEntry");

    /* Relinquish the capability.  Fill in only the required ones. */
    (void)memset(&needed_capabilities,0,sizeof(needed_capabilities));
    needed_capabilities.can_generate_method_entry_events = 1;
    needed_capabilities.can_access_local_variables = 1;
    error = JVMTI_FUNC_PTR(jvmti,RelinquishCapabilities)
        (jvmti, &needed_capabilities);
    check_jvmti_error(jvmti, error, "relinquish capability");

    return;
}
