/* $Id: splay.c,v 1.2 2011/07/11 20:49:19 ruiliu Exp $ */

/*
           An implementation of top-down splaying with sizes
             D. Sleator <sleator@cs.cmu.edu>, January 1994.

  This extends top-down-splay.c to maintain a size field in each node.
  This is the number of nodes in the subtree rooted there.  This makes
  it possible to efficiently compute the rank of a key.  (The rank is
  the number of nodes to the left of the given key.)  It it also
  possible to quickly find the node of a given rank.  Both of these
  operations are illustrated in the code below.  The remainder of this
  introduction is taken from top-down-splay.c.

  "Splay trees", or "self-adjusting search trees" are a simple and
  efficient data structure for storing an ordered set.  The data
  structure consists of a binary tree, with no additional fields.  It
  allows searching, insertion, deletion, deletemin, deletemax,
  splitting, joining, and many other operations, all with amortized
  logarithmic performance.  Since the trees adapt to the sequence of
  requests, their performance on real access patterns is typically even
  better.  Splay trees are described in a number of texts and papers
  [1,2,3,4].

  The code here is adapted from simple top-down splay, at the bottom of
  page 669 of [2].  It can be obtained via anonymous ftp from
  spade.pc.cs.cmu.edu in directory /usr/sleator/public.

  The chief modification here is that the splay operation works even if the
  item being splayed is not in the tree, and even if the tree root of the
  tree is NULL.  So the line:

                              t = splay(i, t);

  causes it to search for item with key i in the tree rooted at t.  If it's
  there, it is splayed to the root.  If it isn't there, then the node put
  at the root is the last one before NULL that would have been reached in a
  normal binary search for i.  (It's a neighbor of i in the tree.)  This
  allows many other operations to be easily implemented, as shown below.

  [1] "Data Structures and Their Algorithms", Lewis and Denenberg,
       Harper Collins, 1991, pp 243-251.
  [2] "Self-adjusting Binary Search Trees" Sleator and Tarjan,
       JACM Volume 32, No 3, July 1985, pp 652-686.
  [3] "Data Structure and Algorithm Analysis", Mark Weiss,
       Benjamin Cummins, 1992, pp 119-130.
  [4] "Data Structures, Algorithms, and Performance", Derick Wood,
       Addison-Wesley, 1993, pp 367-375

  The following code was written by Daniel Sleator, and is released
  in the public domain.
*/

/*
 * splay.c: Splay tree implementation (slightly changed splay() and
 *          splay_insert(), and added splay_treewalk())
 *          based on Mr. Sleator's implementation.
 *
 * Author: Rui Liu
 * Date:   2011-05
 *
 * This file is part of PerfSuite.
 *
 */

#include "splay.h"


BEGIN_C_DECLS
static SplayTreeNode * splay PARAMS((MethodLocPair i, SplayTreeNode *t));
SplayTreeNode * splay_insert PARAMS((MethodLocPair i, SplayTree * tree));
int splay_treewalk           PARAMS((SplayTree* tree, SplayTreeNode*** sarray,
                                     unsigned int *p_numNodes,
                                     int *p_totalCount));
END_C_DECLS


/* This is the comparison.
 * Sort in ascending order of methodID and location.
 */
#define compare(i,j) (memcmp(&i,&j,sizeof(MethodLocPair)))

/* Splay using the key i (which may or may not be in the tree.)
 * The starting root is t, and the tree used is defined by rat.
 * Size fields are maintained.
 */
static SplayTreeNode *
splay(MethodLocPair i, SplayTreeNode *t)
{
    SplayTreeNode N, *l, *r, *y;
    int comp;
    
    if (t == NULL) return t;
    N.left = N.right = NULL;
    l = r = &N;
 
    for (;;) {
        comp = compare(i, t->key);
        if (comp < 0) {
            if (t->left == NULL) break;
            if (compare(i, t->left->key) < 0) {
                y = t->left;                           /* rotate right */
                t->left = y->right;
                y->right = t;
                t = y;
                if (t->left == NULL) break;
            }
            r->left = t;                               /* link right */
            r = t;
            t = t->left;
        } else if (comp > 0) {
            if (t->right == NULL) break;
            if (compare(i, t->right->key) > 0) {
                y = t->right;                          /* rotate left */
                t->right = y->left;
                y->left = t;
                t = y;
                if (t->right == NULL) break;
            }
            l->right = t;                              /* link left */
            l = t;
            t = t->right;
        } else {
            break;
        }
    }

    l->right = t->left;                                /* assemble */
    r->left = t->right;
    t->left = N.right;
    t->right = N.left;

    return t;
}

/* Insert key i into the tree, if it is not already there.
 * Return a pointer to the resulting tree.
 */
SplayTreeNode *
splay_insert(MethodLocPair i, SplayTree * tree)
{
    SplayTreeNode * new;
    SplayTreeNode * t;

    if (NULL == tree) {
        return NULL;
    }
    t = tree->root;

    if (t != NULL) {
	t = splay(i,t);
	if (compare(i, t->key)==0) {
            t->count++;
            tree->root = t;
	    return t;  /* It's already there. */
	}
    }
    new = (SplayTreeNode *) malloc(sizeof (SplayTreeNode));
    if (new == NULL) {
        return NULL;
    }
    if (t == NULL) {
	new->left = new->right = NULL;
    } else if (compare(i, t->key) < 0) {
	new->left = t->left;
	new->right = t;
	t->left = NULL;
    } else {
	new->right = t->right;
	new->left = t;
	t->right = NULL;
    }
    new->key = i;
    new->count = 1;
    tree->numNodes++;
    tree->root = new;
    return new;
}


/* Create a new array data structure from the splay tree.
 *
 * "tree" is the input, the other three arguments are the output.
 * Upon return, "*p_numNodes" should equal to "tree->numNodes",
 * and "*p_totalCount" should equal to the number of total samples.
 *
 * Returns PS_SUCCESS if successful, other values if any error occurred.
 */
int
splay_treewalk(SplayTree* tree, SplayTreeNode*** sarray,
               unsigned int *p_numNodes, int *p_totalCount)
{
    unsigned int current;
    unsigned int i;
    SplayTreeNode** array;
    SplayTreeNode* node;
    int totalCount = 0;
    if (NULL == tree) {
        /* Given tree is null. */
        return PS_EINVAL;
    } else if (0 == tree->numNodes) {
        /* Given tree is empty. */
        /* Another implementation:
         *
         * return PS_EINVAL;
         */
        *p_numNodes = 0;
        *p_totalCount = 0;
        *sarray = NULL;
        return PS_SUCCESS;
    }

    array = (SplayTreeNode **) calloc(tree->numNodes, sizeof(SplayTreeNode*));
    if (NULL == array) {
        /* Memory allocation for array failed. */
        return PS_ENOMEM;
    }

    array[0] = tree->root;
    node = tree->root;
    i = 0;
    current = 0;
    totalCount += array[current]->count;

    while ( (i < tree->numNodes) && (array[i] != 0)) {
        node = array[i];
        i++;
        if (node->left != NULL) {
            array[++current] = node->left;
            totalCount += array[current]->count;
        }
        if (node->right != NULL) {
            array[++current] = node->right;
            totalCount += array[current]->count;
        }
    }

    *p_numNodes = current + 1;
    *sarray = array;
    *p_totalCount = totalCount;
    return PS_SUCCESS;
}
