/* $Id: write-xml.c,v 1.3 2011/07/05 20:25:39 ruiliu Exp $ */

/****************************************************************************
 *
 *                      University of Illinois/NCSA
 *                          Open Source License
 * 
 *            Copyright(C) 2011, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                             Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * write-xml.c : brute-force approach to writing out a Java profiling
 *               XML file in the hwpcprofilereport format, based on
 *               the code in src/libpshwpc/hwpc-xml.c.
 *
 * Author: Rui Liu
 * Date:   May 2011
 *
 * This file is part of PerfSuite.
 */


#include <perfsuite.h>  /* For the defs of BEGIN_C_DECLS, PARAMS, ... */
#include "write-xml.h"

BEGIN_C_DECLS

static int loc_cmp PARAMS((const void * left, const void * right));
static inline const char * str_xml_escape     PARAMS((const char * inStr));
static inline void xml_write_profileinfo_java PARAMS((FILE *fd, int indent,
                                                      int num_nodes,
                                                      psjp_loc_t *locs));
static inline void xml_start_java    PARAMS((FILE *fd, int standalone,
                                             char * doctype, char * dtdtype));
static inline void report_start_java PARAMS((FILE *fd, char *generator,
                                             char *doctype));
static inline void report_end_java   PARAMS((FILE *fd, char *doctype));

END_C_DECLS




/* Include the entire ".c" file for the xml_... functions. */
#include <hwpc-xml.c>


/**********************************************************************
         Routines that are used by the Java profiling feature.
 **********************************************************************/


/* ============================================================== *
 *                   EXPORTED FUNCTIONS                           *
 * ============================================================== */

/* This function is exported.  Called by the "psjprof" JVMTI agent
 * to write out "hwpcprofilereport"s.
 */
int
_ps_hwpc_xml_write_java(FILE *fd, char *generator, psjp_globals_t *globals,
                        unsigned int tid,
                        unsigned long long ticks,
                        double usersecs, double systemsecs,
                        int numEvents, psjp_thread_event_profile_t * profiles,
                        int close, const char * threadName)
{
    int indent=0;
    int standalone = DEFXMLSTANDALONE;
    int i; /* As an index variable. */

    const char *event_name;
    const char *event_type;
    unsigned long event_threshold;
    int num_nodes;
    psjp_loc_t *locs;
    unsigned int num_total_samples;


    PS_DEBUG(PS_DEBUG_INFO, puts("XML output routine entered...");)

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("Tracking function completion:");)

    xml_start_java(fd, standalone, PROFILEREPORTTAG, PROFILEREPORT_DTDTYPE);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_start...");)

    report_start_java(fd, generator, PROFILEREPORTTAG);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("report_start...");)

    indent += DEFAULTINDENT;

    xml_write_date(fd, indent);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_date...");)

    xml_write_hostname(fd, indent);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_hostname...");)

    xml_write_username(fd, indent);
    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_username...");)
        
    /* Write the thread name as an annotation. */
    do_indent(fd,indent);
    fprintf(fd, "<%s>Thread name: '%s'</%s>\n",
            ANNOTATIONTAG, threadName, ANNOTATIONTAG);

    do_indent(fd, indent);
    fprintf(fd, "<%s>\n", EXECINFOTAG);
    indent += DEFAULTINDENT;


    /* Inside of ExecutionInfo. */
    xml_write_date(fd, indent);
    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_date...");)

    xml_write_hostname(fd, indent);
    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_hostname...");)

    do_indent(fd, indent);
    /* "%d" is correct for getpid().  Its return value type--"pid_t"--is
     * defined to "int" on x86, x86-64, ppc64, and ia64.
     * getpid() is done only once, not per XML file.
     */
    fprintf(fd, "<%s>%d</%s>\n", PIDTAG, globals->pid, PIDTAG);
    do_indent(fd, indent);
    fprintf(fd, "<%s>%d</%s>\n", THREADTAG, tid, THREADTAG);
    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_ids...");)

    xml_write_username(fd, indent);
    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_username...");)

    /* md5sum of the command file is not applicable for Java profiling. */
    do_indent(fd, indent);
    fprintf(fd, "<%s>%s%s</%s>\n", COMMANDTAG,
            globals->main_class_name, globals->main_method_args, COMMANDTAG);
    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_command...");)

    xml_write_annotation(fd, indent);
    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_annotation...");)

    /* "globals" and "mhz" are currently unused in xml_write_wallclock. */
    xml_write_wallclock(fd, indent, NULL, ticks, globals->mhz);
    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_wallclock...");)

    xml_write_cputime(fd, indent, usersecs, systemsecs);
    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_cputime...");)

    xml_write_machineinfo(fd, indent, &(globals->cpuid), globals->mhz);
    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_machineinfo...");)

    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fprintf(fd, "</%s>\n", EXECINFOTAG);

    for (i = 0; i < numEvents; i++) {

        event_name        = profiles[i].event_name;
        event_type        = profiles[i].event_type;
        event_threshold   = profiles[i].event_threshold;
        num_nodes         = profiles[i].num_nodes;
        locs              = profiles[i].locs;
        num_total_samples = profiles[i].num_total_samples;

        /* Write out the profile event info. */
        do_indent(fd, indent);
        fputs("<hwpcprofile mapper=\"JVMTI\">\n", fd);
        indent += DEFAULTINDENT;

        do_indent(fd, indent);
        fprintf(fd, "<%s class=\"%s\" version=\"%s\" type=\"%s\" name=\"%s\" domain=\"%s\" />\n",
                HWPCEVENTTAG, "PAPI", globals->papi_version,
                event_type, event_name, globals->event_domain);

        do_indent(fd, indent);
        fprintf(fd, "<period>%lu</period>\n", event_threshold);

        do_indent(fd, indent);
        fprintf(fd, "<totalsamples>%u</totalsamples>\n", num_total_samples);

        /* Write out the sample info. */

        xml_write_profileinfo_java(fd, indent, num_nodes, locs);

        /* Finish it off. */
        indent -= DEFAULTINDENT;
        do_indent(fd, indent);
        fputs("</hwpcprofile>\n", fd);

    }

    report_end_java(fd, PROFILEREPORTTAG);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("report_end...");)

    if ( close ) {
        fclose(fd);
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("XML output routine complete (stream was%s closed)\n", \
                    close ? "" : " not"); \
    )

    return PS_SUCCESS;
}


/* ============================================================== *
 *                  FILE-PRIVATE FUNCTIONS                        *
 * ============================================================== */


/* Used in writing out the Java profile data as a sort comparison function.
 * In ascending order of "file", "function" and "line".
 */
static int
loc_cmp(const void * left, const void * right)
{
    int ret;
    ret = strcmp((*(psjp_loc_t *) left).file, (*(psjp_loc_t*)right).file);
    if (ret) {
        return ret;
    }
    ret = strcmp((*(psjp_loc_t *) left).function, (*(psjp_loc_t*)right).function);
    if (ret) {
        return ret;
    }
    return (*(psjp_loc_t *) left).line - (*(psjp_loc_t*)right).line;
}

/* Escape the "<" and ">" characters in the given string,
 * changing them to "&lt;" and "&gt;".
 *
 * Needed by the Java method names, such as "<clinit>".
 *
 * If there is no '<' nor '>' character, the return value points to inStr;
 * otherwise, the return value points to a newly allocated string,
 * replacing '<' with "&lt;" and '>' with "&gt;".
 */
static inline const char *
str_xml_escape(const char * inStr)
{
    char * result;
    int count = 0;  /* The count of '<' or '>' character. */
    int inLen = strlen(inStr);
    int lenReg;     /* The length of regular characters not copied. */
    int i;

    for (i = 0; i < inLen; i++) {
        if (('<' == inStr[i]) || ('>' == inStr[i])) {
            count++;
        }
    }
    if (0 == count) {
        return inStr;
    }
    /* 3 is the additional size of "&lt;" over "<".
     * 1 is for the terminating "\0".
     */
    result = calloc(1, inLen + 3 * count + 1);
    lenReg = 0;
    for (i = 0; i < inLen; i++) {
        if (('<' == inStr[i]) || ('>' == inStr[i])) {
            if (lenReg > 0) {
                strncat(result, inStr+i-lenReg, lenReg);
                lenReg = 0;
            }
            if ('<' == inStr[i]) {
                strcat(result, "&lt;");
            } else {
                strcat(result, "&gt;");
            }
        } else {
            /* Regular characters, neither '<' nor '>'. */
            lenReg++;
        }
    }
    /* Write the accumulated regular characters not written yet. */
    if (lenReg > 0) {
        strncat(result, inStr+inLen-lenReg, lenReg);
    }
    return result;
}

/* A helper function for _ps_hwpc_xml_write_java to write out
 * the profile data.
 */
static inline void
xml_write_profileinfo_java(FILE *fd, int indent,
                           int num_nodes, psjp_loc_t *locs)
{
    int i;           /* The index of the node. */
    int next;        /* The index of the next node. */
    /* The number of sample for a line, added to deal with the case
     * where there are multiple location items that map to the same
     * line number.
     */
    int num_samples = 0;

    /* Sanity check.  Write nothing if no sample is collected. */
    if (0 == num_nodes) {
        return;
    }

    /* Aggregate and sort the data according to
     * the file name, function name and line number.
     * Maybe we don't need to concern about the module.
     * I'm not sure what Java concept is equivalent to the
     * "module" concept in C.
     */
    qsort(locs, num_nodes, sizeof(psjp_loc_t), loc_cmp);
    /* Use XML escape sequences for "<" and ">", which appear in
     * the class static initializer method.
     * Change the line number "-1" to "0".
     */
    for (i = 0; i < num_nodes; i++) {
        if (locs[i].line < 0) {
            locs[i].line = 0;
        }
        locs[i].function = str_xml_escape(locs[i].function);
    }

    /* Write the file, function, line in an embedded style.
     *
     * Since we don't have an embedded map (map of map of map...)
     * implementation in C language readily available, and implementing
     * it is non-trivial, I chose an alternative, easier approach to
     * achieve the same result:
     *   As the writing proceeds, look ahead and do things accordingly.
     *
     *   Iterate through all the sample data items.
     *
     *   First write the lines of the opening "<module>", "<file>" and
     *   "<function>" tags, then loop through the next data items.
     *
     *   For the next data item:
     *   if it has a different "module",
     *       write the "<line>" line,
     *       close the "line", "function", "file" and "module" tags,
     *       break from this inner loop.
     *   else if the "module" is the same, but the "file" is different,
     *       write the "<line>" line,
     *       close the "line", "function" and "file" tags,
     *       break from this inner loop.
     *   else if the "file" is the same, but the "function" is different,
     *       write the "<line>" line,
     *       close the "line" and "function" tags,
     *       write the opening "function" tag for the next data item.
     *   else if the "file" and "function" are the same,
     *   but the "line" is different,
     *       write the "<line>" line.
     *   else if the "file", "function" and "line" are all the same,
     *       don't write the "<line>" line yet,
     *       just add up the number of samples associated with this line number.
     *
     *   After the iteration is done, finish writing out the last item.
     */
    for (i = 0; i < num_nodes; i++) {
        do_indent(fd, indent);
        fprintf(fd, "<module name=\"%s\">\n", locs[i].module);
        indent += DEFAULTINDENT;
        do_indent(fd, indent);
        fprintf(fd, "<file name=\"%s\">\n", locs[i].file);
        indent += DEFAULTINDENT;
        do_indent(fd, indent);
        fprintf(fd, "<function name=\"%s\">\n", locs[i].function);
        indent += DEFAULTINDENT;
        num_samples = locs[i].num_samples; /* Initialize it. */
        for (next = i + 1; next < num_nodes; next++, i++) {
            if (strcmp(locs[i].module, locs[next].module)) {
                /* "module" is different. */
                do_indent(fd, indent);
                fprintf(fd, "<line lineno=\"%d\">%u</line>\n",
                        locs[i].line, num_samples);
                indent -= DEFAULTINDENT;
                do_indent(fd, indent);
                fputs("</function>\n", fd);
                indent -= DEFAULTINDENT;
                do_indent(fd, indent);
                fputs("</file>\n", fd);
                indent -= DEFAULTINDENT;
                do_indent(fd, indent);
                fputs("</module>\n", fd);
                break;
            } else if (strcmp(locs[i].file, locs[next].file)) {
                /* "file" is different. */
                do_indent(fd, indent);
                fprintf(fd, "<line lineno=\"%d\">%u</line>\n",
                        locs[i].line, num_samples);
                indent -= DEFAULTINDENT;
                do_indent(fd, indent);
                fputs("</function>\n", fd);
                indent -= DEFAULTINDENT;
                do_indent(fd, indent);
                fputs("</file>\n", fd);
                break;
            } else if (strcmp(locs[i].function, locs[next].function)) {
                /* "file" is the same, but "function" is different. */
                do_indent(fd, indent);
                fprintf(fd, "<line lineno=\"%d\">%u</line>\n",
                        locs[i].line, num_samples);
                indent -= DEFAULTINDENT;
                do_indent(fd, indent);
                fputs("</function>\n", fd);
                do_indent(fd, indent);
                fprintf(fd, "<function name=\"%s\">\n", locs[next].function);
                indent += DEFAULTINDENT;
                num_samples = locs[next].num_samples; /* Re-initialize. */
            } else if (locs[i].line != locs[next].line) {
                /* "file" and "function" are the same, but "line" is different.
                 */
                do_indent(fd, indent);
                fprintf(fd, "<line lineno=\"%d\">%u</line>\n",
                        locs[i].line, num_samples);
                num_samples = locs[next].num_samples; /* Re-initialize. */
            } else {
                /* Everything is the same.
                 * Add up the number of samples. Don't write anything yet.
                 */
                num_samples += locs[next].num_samples;
            }
        }
    }

    /* This is the last item.
     * Close the tags to prepare the end of the loop.
     */
    do_indent(fd, indent);
    fprintf(fd, "<line lineno=\"%d\">%u</line>\n",
            locs[num_nodes-1].line, num_samples);
    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fputs("</function>\n", fd);

    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fputs("</file>\n", fd);

    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fputs("</module>\n", fd);

    return;
}

static inline void
xml_start_java(FILE *fd, int standalone, char * doctype, char * dtdtype)
{
    fprintf(fd,"<?xml version=\"%s\" encoding=\"%s\" standalone=\"%s\" ?>\n",
            XMLVERSION, XMLENCODING, standalone ? "yes" : "no");

    fprintf(fd,"<!DOCTYPE %s \n%cSYSTEM \"%s%c%s.dtd\">\n",
            doctype, DEFAULTINDENTCHAR, PS_DTDDIR, DIRECTORY_SEP, dtdtype);

    return;
}

static inline void
report_start_java(FILE *fd, char *generator, char *doctype)
{
    fprintf(fd,"<%s version=\"%s\" generator=\"%s\"\n",
            doctype, XMLREPORTVERSION, generator);
    do_indent(fd,DEFAULTINDENT);
    do_indent(fd,DEFAULTINDENT);
    fprintf(fd,"xmlns:perfsuite=\"%s\">\n", REPORTXMLNS);

    return;
}

static inline void
report_end_java(FILE *fd, char *doctype)
{
    fprintf(fd,"</%s>\n", doctype);

    return;
}
