/* $Id: psjrun.c,v 1.31 2009/11/05 22:03:30 rkufrin Exp $ */

/****************************************************************************
 *
 *                      University of Illinois/NCSA
 *                          Open Source License
 * 
 *          Copyright(C) 2004-2006, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                             Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/* psjrun: psrun for Java
 *
 *         This is a Java 1.5 JVMTI agent for performance monitoring
 *         using libpshwpc.
 *
 * Author: Rick Kufrin
 * Date:   12/4/2004
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#define TRACK_METHODS 0

#include <stdio.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#if HAVE_DLFCN_H
# include <dlfcn.h>
#endif

#if HAVE_JVMTI_H
# include <jvmti.h>
#endif

#include "psjrun.h"

#include <perfsuite.h>
#include <psdebug.h>


/* Handle and function pointers to access routines in libpshwpc */
static void *libpshwpc_dlhandle;

static int (*hwpc_init)(void);
static int (*hwpc_start)(void);
static int (*hwpc_stop)(char *);
static int (*hwpc_shutdown)(void);

/* Things to store away globally for the agent's use */
static jvmtiEnv *jvmti = NULL;
static jrawMonitorID lock;

#if TRACK_METHODS
static unsigned long entry_count=0, exit_count=0;
#endif

/* "Special" threads that we'll ignore */
static char *ignoredThreads[] = {
    PROCESS_REAPER_THREAD,
    DESTROY_JAVAVM_THREAD,
    SIGNAL_DISPATCHER_THREAD,
    HPROF_GC_THREAD,
    HPROF_SAMPLING_THREAD,
    NULL
};

BEGIN_C_DECLS
static inline void enter_critical_section PARAMS((jvmtiEnv *jvmti));
static inline void exit_critical_section  PARAMS((jvmtiEnv *jvmti));
static inline int ignore                  PARAMS((char *threadname));
static inline void report_error           PARAMS((jvmtiError err, char *str));
#if TRACK_METHODS
static void JNICALL method_entry          PARAMS((jvmtiEnv *jvmti, JNIEnv *env,
                                                  jthread thread, 
                                                  jmethodID method));
static void JNICALL method_exit           PARAMS((jvmtiEnv *jvmti, JNIEnv *env,
                                                  jthread thread, 
                                                  jmethodID method,
                                                  jboolean popped_by_exception,
                                                  jvalue return_value));
#endif
static void JNICALL thread_start          PARAMS((jvmtiEnv *jvmti, JNIEnv *env,
                                                  jthread thread));
static void JNICALL thread_end            PARAMS((jvmtiEnv *jvmti, JNIEnv *env,
                                                  jthread thread));
static void JNICALL vm_init               PARAMS((jvmtiEnv *jvmti, JNIEnv *env,
                                                  jthread thread));
END_C_DECLS


/*
 * @@ Should be expanded to allow options to be passed in, e.g.
 *
 *  java -agentlib:psjrun=PS_HWPC_CONFIG:config.xml,PS_HWPC_DOMAIN:kernel
 */

JNIEXPORT jint JNICALL
Agent_OnLoad(JavaVM *jvm, char *options, void *reserved)
{
    jvmtiError jvmtiret;
    jint res;
    jvmtiEventCallbacks callbacks;
#if TRACK_METHODS
    jvmtiCapabilities capabilities;
#endif
    int ret;
    char *error;

#if DEBUG
    ps_setdebuglevel();
#endif

    PS_DEBUG(PS_DEBUG_INFO, printf("Agent_OnLoad called [PID %d]\n", getpid());)
    PS_DEBUG(PS_DEBUG_INFO, printf("Options supplied: %s\n", options);)

    /* Check the JVMTI environment */

    res = (*jvm)->GetEnv(jvm, (void **) &jvmti, JVMTI_VERSION_1_0);

    if ( (res != JNI_OK) || (jvmti == NULL) ) {
        fputs("Error: unable to access JVMTI\n", stderr);
        fprintf(stderr, "Expected 0x%x, GetEnv returned %d.\n",
                JVMTI_VERSION_1, res);
        fputs("Make sure you are running JDK 1.5 or higher.\n", stderr);
        return 1;
    }

    /* Create a raw monitor to protect critical sections of code.  */

    jvmtiret = (*jvmti)->CreateRawMonitor(jvmti, "psjrun", &lock);
    if ( jvmtiret != JVMTI_ERROR_NONE ) {
        report_error(jvmtiret, "CreateRawMonitor");
        return 1;
    }

    /* Register callbacks for events of interest. */

    memset(&callbacks, 0, sizeof(callbacks));
    callbacks.VMInit        = &vm_init;
    callbacks.ThreadStart   = &thread_start;
    callbacks.ThreadEnd     = &thread_end;
#if TRACK_METHODS
    callbacks.MethodEntry   = &method_entry;
    callbacks.MethodExit    = &method_exit;
#endif

    jvmtiret = (*jvmti)->SetEventCallbacks(jvmti, &callbacks,
                                          (jint) sizeof(callbacks));
    if ( jvmtiret != JVMTI_ERROR_NONE ) {
        report_error(jvmtiret, "SetEventCallbacks");
        return 1;
    }

    jvmtiret = (*jvmti)->SetEventNotificationMode(jvmti, JVMTI_ENABLE,
                                                  JVMTI_EVENT_VM_INIT,
                                                  (jthread) NULL);
    if ( jvmtiret != JVMTI_ERROR_NONE ) {
        report_error(jvmtiret, "SetEventNotificationMode");
        return 1;
    }

#if TRACK_METHODS
    /* @@ Should be done before setting the call backs */
    memset(&capabilities, 0, sizeof(jvmtiCapabilities));
    jvmtiret = (*jvmti)->GetPotentialCapabilities(jvmti, &capabilities);
    if ( jvmtiret == JVMTI_ERROR_NONE ) {
        if ( capabilities.can_generate_method_entry_events ) {
            memset(&capabilities, 0, sizeof(jvmtiCapabilities));
            capabilities.can_generate_method_entry_events = 1;
            capabilities.can_generate_method_exit_events = 1;
            jvmtiret = (*jvmti)->AddCapabilities(jvmti, &capabilities);
            if ( jvmtiret != JVMTI_ERROR_NONE ) {
                report_error(jvmtiret, "AddCapabilities");
            }
        }
    }
#endif


    libpshwpc_dlhandle = dlopen(PSHWPC_SHLIB, RTLD_NOW);
    if ( ! libpshwpc_dlhandle ) {
        fprintf(stderr, "%s\n", dlerror());
        return 1;
    }

    hwpc_init     = dlsym(libpshwpc_dlhandle, INIT_FUNC);
    CHECK_DLSYM
    hwpc_start    = dlsym(libpshwpc_dlhandle, START_FUNC);
    CHECK_DLSYM
    hwpc_stop     = dlsym(libpshwpc_dlhandle, STOP_FUNC);
    CHECK_DLSYM
    hwpc_shutdown = dlsym(libpshwpc_dlhandle, SHUTDOWN_FUNC);
    CHECK_DLSYM

    PS_DEBUG(PS_DEBUG_INFO, \
             puts("Agent_OnLoad: function pointers initialized");)

    ret = (*hwpc_init)();
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "psjrun initialization error: %s\n", ps_strerror(ret));
        return ret;
    }

    PS_DEBUG(PS_DEBUG_INFO, puts("hwpc_init called successfully");)

    return JNI_OK;
}

JNIEXPORT void JNICALL
Agent_OnUnload(JavaVM *jvm)
{
#if 0
    int ret;
#endif

    PS_DEBUG(PS_DEBUG_INFO, puts("Agent_OnUnload called");)
#if TRACK_METHODS
    printf("Method entries: %ld\n", entry_count);
    printf("Method exits:   %ld\n", exit_count);
#endif

#if 0
    ret = (*hwpc_shutdown)();
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "psjrun shutdown error: %s\n", ps_strerror(ret));
    }
    else {
        PS_DEBUG(PS_DEBUG_INFO, puts("hwpc_shutdown called successfully");)
    }
#endif

    return;
}

static void JNICALL
vm_init(jvmtiEnv *jvmti_env, JNIEnv* jni_env, jthread thread)
{
#if TRACK_METHODS
    jvmtiCapabilities capabilities;
#endif
    jvmtiError jvmtiret;

    PS_DEBUG(PS_DEBUG_INFO, printf("Java VM initialization detected.\n");)

    enter_critical_section(jvmti);

    jvmtiret = (*jvmti)->SetEventNotificationMode(jvmti, JVMTI_ENABLE,
                                                  JVMTI_EVENT_THREAD_START,
                                                  (jthread) NULL);
    if ( jvmtiret != JVMTI_ERROR_NONE ) {
        report_error(jvmtiret, "SetEventNotificationMode");
    }

    jvmtiret = (*jvmti)->SetEventNotificationMode(jvmti, JVMTI_ENABLE,
                                                  JVMTI_EVENT_THREAD_END,
                                                  (jthread) NULL);
    if ( jvmtiret != JVMTI_ERROR_NONE ) {
        report_error(jvmtiret, "SetEventNotificationMode");
    }

#if TRACK_METHODS
    memset(&capabilities, 0, sizeof(jvmtiCapabilities));
    jvmtiret = (*jvmti)->GetCapabilities(jvmti, &capabilities);
    if ( jvmtiret != JVMTI_ERROR_NONE ) {
        report_error(jvmtiret, "GetCapabilities");
    }

    if ( capabilities.can_generate_method_entry_events  &&
         capabilities.can_generate_method_exit_events ) {
        jvmtiret = (*jvmti)->SetEventNotificationMode(jvmti, JVMTI_ENABLE,
                                                      JVMTI_EVENT_METHOD_ENTRY,
                                                      (jthread) NULL);
        if ( jvmtiret != JVMTI_ERROR_NONE ) {
            report_error(jvmtiret, "SetEventNotificationMode (method entry)");
        }
        jvmtiret = (*jvmti)->SetEventNotificationMode(jvmti, JVMTI_ENABLE,
                                                      JVMTI_EVENT_METHOD_EXIT,
                                                      (jthread) NULL);
        if ( jvmtiret != JVMTI_ERROR_NONE ) {
            report_error(jvmtiret, "SetEventNotificationMode (method exit)");
        }
    }
#endif

    exit_critical_section(jvmti);
}

#if TRACK_METHODS
static void JNICALL 
method_entry(jvmtiEnv *jvmti, JNIEnv *env, jthread thread, jmethodID method)
{
    entry_count++;
}

static void JNICALL 
method_exit(jvmtiEnv *jvmti, JNIEnv *env, jthread thread, jmethodID method,
            jboolean popped_by_exception, jvalue return_value)
{
    exit_count++;
}
#endif

static void JNICALL 
thread_start(jvmtiEnv *jvmti, JNIEnv *env, jthread thread)
{
    jvmtiError jvmtiret;
    jvmtiThreadInfo info;
    int res;

    jvmtiret = (*jvmti)->GetThreadInfo(jvmti, thread, &info);
    PS_DEBUG(PS_DEBUG_VERBOSE, \
             printf("Thread \"%s starting\".\n", info.name);)
    /*
     * @@ The thread may not have a name (NULL).
     */
    if ( ignore(info.name) ) {
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Ignoring thread \"%s\" at startup.\n", info.name);)
    }
    else {
        res = (*hwpc_start)();
        if ( res != PS_SUCCESS ) {
            fprintf(stderr, "psjrun thread start error: %s\n",
                    ps_strerror(res));
        }
    }
    jvmtiret = (*jvmti)->Deallocate(jvmti, (unsigned char *) info.name);
    if ( jvmtiret != JVMTI_ERROR_NONE ) {
        report_error(jvmtiret, "GetThreadInfo");
    }

    return;
}

static void JNICALL 
thread_end(jvmtiEnv *jvmti, JNIEnv *env, jthread thread)
{
    jvmtiError jvmtiret, jvmtiret2;
    jvmtiThreadInfo info;
    jvmtiThreadGroupInfo groupInfo;
    int res, write = 1;

    jvmtiret = (*jvmti)->GetThreadInfo(jvmti, thread, &info);
    PS_DEBUG(PS_DEBUG_VERBOSE, \
             printf("Thread \"%s\" ending.\n", info.name);)

    if ( jvmtiret == JVMTI_ERROR_NONE  ) {
        if ( info.thread_group != NULL ) {
            jvmtiret2 = (*jvmti)->GetThreadGroupInfo(jvmti, info.thread_group,
                                                     &groupInfo);
            if ( jvmtiret2 == JVMTI_ERROR_NONE ) {
                if ( strcmp(groupInfo.name, "system") == 0 ) {
                    write = 0;
                    PS_DEBUG(PS_DEBUG_INFO, \
                         printf("Disabling system thread \"%s\" output\n", \
                               info.name);) 
                }
                (*jvmti)->Deallocate(jvmti, (unsigned char *) groupInfo.name);
            }
            else {
                report_error(jvmtiret2, "GetThreadGroupInfo");
            }
        }
        else {
            write = 1;
        }
    }

    if ( ignore(info.name) ) {
        write = 0;
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Ignoring thread \"%s\" at termination.\n", info.name);)
    }
    
    if ( write ) {
        /* @@ info.name is encoded as a modified UTF string */
        res = (*hwpc_stop)(info.name);
        if ( res != PS_SUCCESS ) {
            fprintf(stderr, "psjrun thread stop error: %s\n", ps_strerror(res));
        }
    }

    (*jvmti)->Deallocate(jvmti, (unsigned char *) info.name);

    return;
}

/* Enter a critical section by doing a JVMTI Raw Monitor Enter */
static inline void
enter_critical_section(jvmtiEnv *jvmti)
{
    jvmtiError jvmtiret;

    jvmtiret = (*jvmti)->RawMonitorEnter(jvmti, lock);
    if ( jvmtiret != JVMTI_ERROR_NONE ) {
        report_error(jvmtiret, "RawMonitorEnter");
    }

    return;
}

/* Exit a critical section by doing a JVMTI Raw Monitor Exit */
static inline void
exit_critical_section(jvmtiEnv *jvmti)
{
    jvmtiError jvmtiret;

    jvmtiret = (*jvmti)->RawMonitorExit(jvmti, lock);
    if ( jvmtiret != JVMTI_ERROR_NONE ) {
        report_error(jvmtiret, "RawMonitorExit");
    }

    return;
}

static inline void
report_error(jvmtiError err, char *str)
{
    jvmtiError ret;
    char *errstring = NULL;

    ret = (*jvmti)->GetErrorName(jvmti, err, &errstring);
    fprintf(stderr, "Error [%s]: %s\n", str, errstring);
    (*jvmti)->Deallocate(jvmti, (unsigned char *) errstring);
    
    return;
}

static inline int
ignore(char *threadname)
{
    char **name;

    for ( name = ignoredThreads; *name != NULL; name++ ) {
        if ( strcmp(*name, threadname) == 0 ) {
            return 1;
        }
    }

    return 0;
}
