/* $Id: cache.c,v 1.4 2011/02/11 19:54:49 rkufrin Exp $ */

/****************************************************************************
 *
 *                      University of Illinois/NCSA
 *                          Open Source License
 * 
 *          Copyright(C) 2002-2008, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                             Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * cache.c: architecture-neutral routines for manipulating data structures
 *          relevant to the cache/TLB hierarchy.
 *
 * This source file is #include'd from the main "cpuid.c" file.
 *
 * Author: Rick Kufrin
 *
 */

#ifndef CACHE_C_INCLUDED
#define CACHE_C_INCLUDED 1

BEGIN_C_DECLS
static int add_cache             PARAMS((ps_cpu_cache_t *newcache));
static int cache_cmp             PARAMS((ps_cpu_cache_t *a,
                                         ps_cpu_cache_t *b));
static int cache_equal           PARAMS((ps_cpu_cache_t *a,
                                         ps_cpu_cache_t *b));
static int extend_cachearray     PARAMS((ps_cpu_cachearray_t *arr, int level));
static int init_cachearray       PARAMS((ps_cpu_cachearray_t *arr));
static ps_cpu_cachelist_t * insert_cache
                                 PARAMS((ps_cpu_cachelist_t *new,
                                         ps_cpu_cachearray_t *arr,
                                         int level));
static ps_cpu_cache_t *new_cache PARAMS((void));
#if DEBUG
static void print_cache          PARAMS((ps_cpu_cache_t *cache));
static void print_cachearray     PARAMS((ps_cpu_cachearray_t *cachearray));
#endif
END_C_DECLS

/*
 * Uninitialized cachearray structure has its "cachelist" member set to NULL
 * and "levels" set to zero.
 */

static int
add_cache(ps_cpu_cache_t *cache)
{
    ps_cpu_cachelist_t *new, *elt, *prev;
    int level = (int) cache->level;  /* @@ Needed due to unsigned int type */
    int insert_level, ret, done;

    /* The x86 version of the cpuid routine sets cache/tlb level to
     * -1 by default.
     * @@ Need to check what is done for ia64 and unify.
     *
     * We assume that an existing level assigned > 1 is the final
     * level to which this cache/tlb should be assigned.
     */

    /*
     * Create the initial (one-level) cache array if not yet done.
     */
    if ( cpuinfo->cachearray.levels == 0 ) {
        ret = init_cachearray(&(cpuinfo->cachearray));
        if ( ret != PS_SUCCESS ) {
            return ret;
        } 
    }

    /*
     * Check if array needs extending to higher levels if the new cache has
     * already been assigned a level higher than any seen so far.
     */
    if ( cpuinfo->cachearray.levels < level ) {
        ret = extend_cachearray(&(cpuinfo->cachearray), level);
        if ( ret != PS_SUCCESS ) {
            return ret;
        }
    }

    /*
     * Allocate a new entry for the new cache.
     */
    new = (ps_cpu_cachelist_t *) malloc(sizeof(ps_cpu_cachelist_t));
    if ( new == NULL ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("add_cache: allocation for new entry failed\n",stderr);)
        return PS_ENOMEM;
    }
    new->cache = cache;
    new->next = NULL;

    /*
     * If the new cache has already had a level assigned to it, add it to the
     * head of the list at the appropriate level, and we're done.
     *
     * @@ NEED TO CHECK IF AN EXISTING CACHE AT THIS LEVEL NEEDS TO BE
     * DISPLACED.
     */
    if ( level > 0 ) {
        new->next = cpuinfo->cachearray.cachelist[level-1];
        cpuinfo->cachearray.cachelist[level-1] = new;
        PS_DEBUG(PS_DEBUG_VERBOSE, \
                 printf("new cache inserted at array slot %d\n", level-1);)
        PS_DEBUG(PS_DEBUG_VERBOSE, \
                 printf("Added pre-assigned cache at level %d, ", level); \
                 printf("current levels: %d\n", cpuinfo->cachearray.levels);)

        /*
         * Check to see if an existing cache at this level now needs to
         * be relocated to an outer level.
         */
        prev = cpuinfo->cachearray.cachelist[level-1];
        elt = prev->next;
        done = 0;
        while ( (elt != NULL) && ! done ) {
            if ( elt->cache->type == new->cache->type ) {
                if ( cache_equal(elt->cache, new->cache) ) {
                    if ( ((int) elt->cache->level) == -1 ) {
                        prev->next = elt->next;
                        PS_DEBUG(PS_DEBUG_VERBOSE, \
                                 puts("relocating existing unassigned cache");)
                        /*
                         * @@ Change this to non-recursive, this is just
                         * a temporary workaround to "re-add" the cache
                         * that's being reassigned levels.
                         */
                        add_cache(elt->cache);
                        free(elt);
                    }
                    else {
                        PS_DEBUG(PS_DEBUG_WARNING, \
                                 fputs("Warning: multiple identical",stderr); \
                                 fputs(" preassigned caches encountered\n", \
                                       stderr);)
                    }
                }
            }
            prev = elt;
            elt = elt->next;
        }
        
#if DEBUG
        print_cachearray(&(cpuinfo->cachearray));
#endif
        return PS_SUCCESS;
    }

    /*
     * The new cache/TLB does not yet have a level assigned. Cycle through
     * the current assigned caches to determine where it should be inserted.
     */

    insert_level = 0;
    do {
        new = insert_cache(new, &(cpuinfo->cachearray), insert_level);
        insert_level++;
    } while ( (new != NULL) && (insert_level < cpuinfo->cachearray.levels) );

    if ( new != NULL ) {
        /*
         * Cache was still not inserted after looking through all levels,
         * so extend the array and add the new cache at the new level.
         */
        ret = extend_cachearray(&(cpuinfo->cachearray), insert_level+1);
        if ( ret != PS_SUCCESS ) {
            PS_DEBUG(PS_DEBUG_FATAL, fputs("failed allocation in add_cache", \
                                            stderr);)
            return ret;
        }
        cpuinfo->cachearray.cachelist[insert_level] = new;
    }

    PS_DEBUG(PS_DEBUG_VERBOSE, \
             printf("End of add_cache output, current levels: %d\n", 
                    cpuinfo->cachearray.levels);)

#if DEBUG
    print_cachearray(&(cpuinfo->cachearray));
#endif

    return PS_SUCCESS;
}


/*
 *
 */

static ps_cpu_cache_t *
new_cache(void)
{
    ps_cpu_cache_t *c;

    c = PS_MALLOC(ps_cpu_cache_t, 1);

    if ( c != NULL ) {
        memset(c, 0, sizeof(ps_cpu_cache_t));
        c->level = -1;
        c->type = UNKNOWN_CACHE;
    }

    return c;
}


/*
 * Auxiliary function used to differentiate between caches/tlbs of
 * the same type that may reside at the same level in the hierarchy.
 * Testing for caches is straightforward, and consists of a check of
 * their total size.  For TLBs, "equality" is assumed to mean an
 * identical number of pagesizes supported as well as specific pagesizes.
 *
 * This function does not try to select one or the other cache/tlb
 * supplied, but only determine whether selection is necessary.
 *
 * Return value: 1 if equal, 0 if not
 */

static int
cache_equal(ps_cpu_cache_t *a, ps_cpu_cache_t *b)
{
    unsigned int i;

    switch (a->type) {
    case ICACHE:
    case DCACHE:
    case UCACHE:
        return (a->specs.cache.size == b->specs.cache.size);
        break;
    case ITRACECACHE:
        return (a->specs.itcache.uops == b->specs.itcache.uops);
        break;
    case ITLB:
    case DTLB:
    case UTLB:
        if ( a->specs.tlb.npagesize != b->specs.tlb.npagesize ) {
            return 0;
        }
        for ( i = 0; i < a->specs.tlb.npagesize; i++ ) {
            if ( a->specs.tlb.pagesize[i] != b->specs.tlb.pagesize[i] ) {
                return 0;
            }
        }
        break;
    default:
        return 0;
        break;
    }

    return 1;
}


/*
 * Comparison function to be applied to two caches of identical type.
 * Similar to strcmp(), return values are:
 *   <0 : a is less than b
 *   >0 : a is greater than b
 *    0 : a and b are equal (should not happen)
 */

static int
cache_cmp(ps_cpu_cache_t *a, ps_cpu_cache_t *b)
{
    switch (a->type) {
    case ICACHE:
    case DCACHE:
    case UCACHE:
        return (a->specs.cache.size - b->specs.cache.size);
    case ITRACECACHE:
        return (a->specs.itcache.uops - b->specs.itcache.uops);
    case ITLB:
    case DTLB:
    case UTLB:
        return (a->specs.tlb.entries - b->specs.tlb.entries);
    default:
        return 0;
    }
}


/*
 */

static int
init_cachearray(ps_cpu_cachearray_t *arr)
{
    arr->cachelist =
        (ps_cpu_cachelist_t **) malloc(sizeof(ps_cpu_cachelist_t *));
    if (arr->cachelist == NULL) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("init_cacherray: initial cache array allocation failed\n", \
                       stderr);)
        return PS_ENOMEM;
    }

    arr->cachelist[0] = NULL;
    arr->levels       = 1;

    PS_DEBUG(PS_DEBUG_VERBOSE, \
             puts("init_cachearray: initial cache array allocated");)

    return PS_SUCCESS;
}


/*
 * Extend an existing cachearray structure to accommodate "level"
 * levels.  Retain all existing information.  Make sure to account for
 * proper initialization to NULL if a gap is introduced (e.g. only one
 * level encountered so far, but request to accommodate 3 levels).
 */

static int
extend_cachearray(ps_cpu_cachearray_t *arr, int level)
{
    int oldlevels, i;

    oldlevels = arr->levels;

    arr->cachelist = (ps_cpu_cachelist_t **)
        realloc(arr->cachelist, level * sizeof(ps_cpu_cachelist_t *));
    if (arr->cachelist == NULL) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("extend_cachelist: cache array extension failed\n", \
                       stderr);)
        return PS_ENOMEM;
    }

    for ( i = oldlevels; i < level; i++ ) {
        arr->cachelist[i] = NULL;
    }
    arr->levels = level;

    PS_DEBUG(PS_DEBUG_VERBOSE, \
             printf("extend_cachelist: extended cache array to %d levels\n", \
                    level);)

    return PS_SUCCESS;
}


/*
 * Auxiliary function to aid in insertion of a cache/TLB for which no level
 * has been assigned.
 */

static ps_cpu_cachelist_t *
insert_cache(ps_cpu_cachelist_t *new, ps_cpu_cachearray_t *arr,
                int level)
{
    ps_cpu_cachelist_t *elt, *prev;
    int ret;

    prev = NULL;
    elt = arr->cachelist[level];

    while ( elt != NULL ) {
        if ( new->cache->type == elt->cache->type ) {
            /*
             * The types match, now check specs to see if they match as well.
             *
             * If they do, then determine which of the two is "smaller"
             * (meaning depends on the cache type).  The smaller one should be
             * assigned to the current level and the larger moved up one
             * level in the hierarchy.
             *
             * If they don't, then the existing one remains at this level
             * and the new one continues to be checked against the rest of
             * the list.
             */
            if ( cache_equal(new->cache, elt->cache) ) {
                ret = cache_cmp(new->cache, elt->cache);
                if ( ret < 0 ) {
                    /*
                     * new is smaller than existing.  Assign new to this
                     * level, and return the existing cache for reassignment
                     * to a higher level.
                     */
                    if ( arr->cachelist[level] == elt ) {
                        arr->cachelist[level] = new;
                    }
                    else {
                        prev->next = new;
                    }
                    new->next = elt->next;
                    return elt;
                }
                else if ( ret > 0 ) {
                    /*
                     * new is larger than existing, so return it and let the
                     * caller try a different level.
                     */
                    return new;
                }
                else {
                    /*
                     * Should not happen: exactly matching caches.
                     */
                    PS_DEBUG(PS_DEBUG_WARNING, \
                             fputs("Warning: identical caches/TLBs found!\n", \
                                   stderr);)
                }
            }
        }
        prev = elt;
        elt = elt->next;
    }

    /*
     * No entries were found at this level that matched the cache we are
     * wanting to add.  Insert the new cache at the head of the list and return
     * NULL to indicate the cache was inserted.
     */
#if 0 /* leave this at -1 for now/devel */
    new->cache->level = level+1;
#endif
    new->next = arr->cachelist[level];
    arr->cachelist[level] = new;

    return (ps_cpu_cachelist_t *) NULL;
}


#if DEBUG
static void
print_cache(ps_cpu_cache_t *cache)
{
    unsigned int i;

    printf("Cache information:\n\t");

    printf("Level %d ", cache->level);
    switch(cache->type) {
    case ITLB:
        printf("ITLB, ");
        break;
    case DTLB:
        printf("DTLB, ");
        break;
    case UTLB:
        printf("UTLB, ");
        break;
    case ICACHE:
        printf("I-cache, ");
        break;
    case DCACHE:
        printf("D-cache, ");
        break;
    case UCACHE:
        printf("U-cache, ");
        break;
    case ITRACECACHE:
        printf("ITrace-cache, ");
        break;
    default: printf("UNKNOWN cache/tlb type"); break;
    }

    if (cache->nway == FULLY_ASSOCIATIVE) {
        printf("Fully associative, ");
    }
    else if (cache->nway == DIRECT_MAPPED) {
        printf("Direct-mapped, ");
    }
    else {
        printf("%d-way associative, ", cache->nway);
    }

    switch (cache->type) {
    case ITLB:
    case DTLB:
    case UTLB:
        printf("%d entries, page sizes: ", cache->specs.tlb.entries);
        for (i = 0; i < cache->specs.tlb.npagesize; i++) {
            printf("%d", cache->specs.tlb.pagesize[i]);
            if (i < (cache->specs.tlb.npagesize - 1)) {
                printf(", ");
            }
        }
        break;
    case ICACHE:
    case DCACHE:
    case UCACHE:
        printf("size %d KB, linesize %d B",
               cache->specs.cache.size, cache->specs.cache.linesize);
        break;
    case ITRACECACHE:
        printf("%d Kuops", cache->specs.itcache.uops);
        break;
    default:
        break;
    }

    puts("");
    return;
}

static void
print_cachearray(ps_cpu_cachearray_t *cachearray)
{
    ps_cpu_cachelist_t *elt;
    int i;

    PS_DEBUG(PS_DEBUG_VERBOSE,puts("print_cachearray:");)
    PS_DEBUG(PS_DEBUG_VERBOSE, \
             puts("**************************************************");)
    for (i=0; i<cachearray->levels; i++) {
        PS_DEBUG(PS_DEBUG_VERBOSE, printf("LEVEL %d\n", i+1);)
        elt = cachearray->cachelist[i];
        while (elt != NULL) {
            PS_DEBUG(PS_DEBUG_VERBOSE, print_cache(elt->cache);)
            elt = elt->next;
        }
    }
    PS_DEBUG(PS_DEBUG_VERBOSE, \
             puts("**************************************************");)

    return;
}
#endif /* DEBUG */

#endif /* CACHE_C_INCLUDED */
