/* $Id: cpuid-x86.c,v 1.80 2013/04/16 15:56:41 ruiliu Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *   Copyright(C) 2002-2009, 2012-2013, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *           National Center for Supercomputing Applications
 *             University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * cpuid: implement access to CPUID instruction.
 *
 *        Requires GCC-style inline assembly.
 *
 *        Specific to Intel x86 processors, 486
 *        and subsequent processors that implement the
 *        CPUID instruction and AMD x86/x86_64 processors
 *        beginning with the AM486 DX4 processor.
 *
 *        Intel support is based on information in the document
 *        "Intel Processor Identification and the CPUID
 *        Instruction", January 2011, Intel Corporation
 *        Application Note AP-485.
 *
 *        AMD support is based on information in the
 *        document "AMD Processor Recognition Application
 *        Note", AMD Corporation 20734, Rev 3.13, December
 *        2005, and "CPUID Specification", AMD Corporation 25481,
 *        Rev 2.26, July 2007.
 *
 * @@ With the addition of AMD processors to this, there is a LOT
 * of opportunity for cleanup/shortening and revision - still needs
 * to be done so this is definitely a rough-draft implementation...
 *
 * Author: Rick Kufrin
 * Date:   November 23, 2002
 *
 * This file is part of PerfSuite.
 *
 */

static int do_intel PARAMS((ps_cpuid_t *cpuinfo, unsigned int max_fn));
static int do_amd PARAMS((ps_cpuid_t *cpuinfo, unsigned int max_fn));
static int do_intel_caches PARAMS((unsigned int regs[], int nregs,
				   int leaf4_used));

#define VENDOR_ID_LEN 13

static inline void
cpuid(unsigned int op, unsigned int *eax, unsigned int *ebx,
      unsigned int *ecx, unsigned int *edx)
{
    unsigned int _op=op, _eax, _ebx, _ecx=*ecx, _edx;

# if defined(__i386__)
    asm volatile ("pushl %%ebx    \n\t"   /* save %ebx */
                  "cpuid          \n\t"
                  "movl %%ebx, %1 \n\t"   /* save what cpuid just put in %ebx */
                  "popl %%ebx     \n\t"   /* restore the old %ebx */
                  : "=a" (_eax), \
                    "=r" (_ebx), \
                    "=c" (_ecx), \
                    "=d" (_edx)  \
                  : "a" (_op), \
                    "c" (_ecx)); 

# elif defined(__x86_64__)
    asm volatile ("cpuid"
                  : "=a" (_eax), \
                    "=b" (_ebx), \
                    "=c" (_ecx), \
                    "=d" (_edx)  \
                  : "a" (_op), \
                    "c" (_ecx)); 

# else
#  error "Cannot determine arch for CPUID x86 ASM.  Please submit a bug report."
# endif

    PS_DEBUG(PS_DEBUG_INFO, printf("%x %x %x %x\n", _eax, _ebx, _ecx, _edx);)

    *eax = _eax;  *ebx = _ebx;  *ecx = _ecx;  *edx = _edx;

    return;
}


static int
_ps_cpuid_internal(ps_cpuid_t *cpuinfo)
{
    unsigned int op, eax, ebx, ecx, edx;
    unsigned int max_fn;
    char vendor[VENDOR_ID_LEN];

    reset_cpuinfo(cpuinfo);

    /* Max allowable CPUID function & vendor string */

    PS_DEBUG(PS_DEBUG_INFO, \
             puts("Retrieving max CPUID function and vendor string...");)

    op = 0;
    ecx = 0;
    cpuid(op,&eax,&ebx,&ecx,&edx);

    max_fn = eax;

    memcpy(vendor,   &ebx,4);
    memcpy(vendor+4, &edx,4);
    memcpy(vendor+8, &ecx,4);
    vendor[VENDOR_ID_LEN-1] = '\0';

    if ( strcmp(vendor, VENDOR_STRING_INTEL) == 0 ) {
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Vendor identified as Intel (%s)\n", vendor);)
        cpuinfo->vendor = INTEL;
        do_intel(cpuinfo, max_fn);
    }
    else if ( strcmp(vendor, VENDOR_STRING_AMD) == 0 ) {
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Vendor identified as AMD (%s)\n", vendor);)
        cpuinfo->vendor = AMD;
        do_amd(cpuinfo, max_fn);
    }
    else {
        PS_DEBUG(PS_DEBUG_WARNING, printf("Vendor unknown (%s)\n", vendor);)
        cpuinfo->vendor = UNKNOWN_VENDOR;
    }

    cpuinfo->arch = IA32_ARCH;

    return PS_SUCCESS;
}


/* Use IA-32 CPUID instruction to retrieve processor information 
 * from Intel processors
 */

static int
do_intel(ps_cpuid_t *cpuinfo, unsigned int max_fn)
{
    unsigned int op, eax, ebx, ecx, edx, regs[4];
    char brand_string[INTEL_BRAND_STRING_LENGTH];
    int i, off, reps, ret;
    int leaf4_used = 0;
    unsigned int partitions, sets, tmpval;
    ps_cpu_cache_t *cache;

    if ( max_fn > 0 ) {

        /**********************************************
         **            Processor signature           **
         **********************************************/
    
        PS_DEBUG(PS_DEBUG_INFO, puts("Retrieving processor signature...");)

        op = 1;
        ecx = 0;
        cpuid(op, &eax, &ebx, &ecx, &edx);

        /* Processor type */
        switch ( eax & INTEL_TYPE_MASK ) {
        case INTEL_TYPE_OEM:
            cpuinfo->type = OEM;
            break;
        case INTEL_TYPE_OVERDRIVE:
            cpuinfo->type = OVERDRIVE;
            break;
        case INTEL_TYPE_DUAL:
            cpuinfo->type = DUAL;
            break;
        default:
            cpuinfo->type = UNKNOWN_TYPE;
            break;
        }

        /* Use plain addition instead of shifting 4 the extended family value
           then adding the family value.  Intel doc text says plain addition,
           but its sample code shifted it.  The AMD result on Forge seemed to
           indicate plain addition, so changed it so for both Intel and AMD.
        */
        cpuinfo->cpuid_family = ((eax >> 20) & 0xFF) + ((eax >> 8) & 0xF);
        /* model: (eax & EXTENDED_MODEL_MASK) >> 16 << 4 == () >> 12. */
        cpuinfo->cpuid_model = ((eax >> 12) & 0xF0) + ((eax >> 4) & 0xF);
        cpuinfo->cpuid_stepping = (eax & 0xF);

        PS_DEBUG(PS_DEBUG_INFO, printf("eax = %x, ebx=%x\n", eax, ebx);)
        PS_DEBUG(PS_DEBUG_INFO, printf("family=%u, model=%u, stepping=%u.\n", \
                                       cpuinfo->cpuid_family, \
                                       cpuinfo->cpuid_model, \
                                       cpuinfo->cpuid_stepping);)

        cpuinfo->revision =  ( eax & INTEL_STEPPING_MASK );

        /**************************************************
         **           Processor brand string             **
         **************************************************/

        PS_DEBUG(PS_DEBUG_INFO, \
                 puts("Checking for implementation of brand string...");)

        op = INTEL_BRAND_STRING_DETECT;
        cpuid(op, &eax, &ebx, &ecx, &edx);

        if ( ( eax & INTEL_BRAND_STRING_DETECT ) != 0 ) {
            op = INTEL_BRAND_STRING_FETCH;  /* Starting opcode */

            /* Need to call this three times to get all 48 bytes */
            for ( i = off = 0;
                  i < 3;
                  i++, op++, off += (sizeof(unsigned int) * 4) ) {
                cpuid(op, &eax, &ebx, &ecx, &edx);
                memcpy(brand_string+off,     &eax, 4);
                memcpy(brand_string+off+4,   &ebx, 4);
                memcpy(brand_string+off+8,   &ecx, 4);
                memcpy(brand_string+off+12,  &edx, 4);
            }

            /* Strip leading spaces */
            for ( i = 0; brand_string[i] == ' '; i++ ) ;

            cpuinfo->brand_string =
                PS_MALLOC(char, (INTEL_BRAND_STRING_LENGTH-i));
            strncpy(cpuinfo->brand_string, &brand_string[i],
                    INTEL_BRAND_STRING_LENGTH-i);
        }
        else {
            cpuinfo->brand_string = NULL;
        }
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Brand string is %savailable\n", \
                        cpuinfo->brand_string == NULL ? "not " : "");)
    
    }

    /*******************************************************
     **                Cache information                  **
     *******************************************************/

    if ( max_fn > 1 ) {
        if ( max_fn > 3 ) {

            /**********************************************************
                   Deterministic cache parameters (Core and beyond)
             **********************************************************/

            PS_DEBUG(PS_DEBUG_INFO, \
                     puts("Checking deterministic cache parameters...");)

            for ( i = 0; ; i++ ) {
                op = 4;
                ecx = i;
                cpuid(op, &eax, &ebx, &ecx, &edx);
		tmpval = eax & 0x0000001f;
                if ( tmpval == 0 ) {
                    PS_DEBUG(PS_DEBUG_INFO, \
                             printf("No more caches (iteration %d)\n", i);)
                    break;
                }

		cache = new_cache();
		if ( cache == NULL ) {
		    return PS_ENOMEM;
		}

                switch ( tmpval ) {
                case 0: break; /* Should not happen */
                case 1: cache->type = DCACHE; break;
                case 2: cache->type = ICACHE; break;
                case 3: cache->type = UCACHE; break;
                default:
                    PS_DEBUG(PS_DEBUG_WARNING, \
                             fputs("ERROR: reserved value found in ",stderr);\
                             fputs("scan of deterministic cache params\n", \
                                   stderr);)
                    break;
                }

		cache->level                = (eax & 0x0000000e0) >> 5;
                cache->nway                 = ((ebx & 0xffc00000) >> 22) + 1;
                partitions                  = ((ebx & 0x003ff000) >> 12) + 1;
                cache->specs.cache.linesize = (ebx & 0x00000fff) + 1;
                sets = ecx + 1;
                cache->specs.cache.size = (cache->nway
                                          * partitions
                                          * cache->specs.cache.linesize
                                          * sets)
                                          / 1024;

		add_cache(cache);

                PS_DEBUG(PS_DEBUG_VERBOSE, \
                         printf("\nTotal # cores on this die: %d\n", \
                                ((eax & 0xfc000000) >> 26) + 1); \
                         printf("Total # threads sharing this cache: %d\n", \
                                ((eax & 0x03ffc000) >> 14) + 1); \
                         printf("Fully associative: %s\n", \
                                ((eax & 0x00000200) >> 9 ) ? "yes" : "no"); \
                         printf("Self initializing cache level: %d\n", \
                                (eax & 0x000001000) >> 8); \
                         printf("Cache level: %d\n", \
                                (eax & 0x0000000e0) >> 5); \
                         printf("Number of sets: %d\n", sets); \
                         printf("Prefetch stride: %d\n", \
                                edx & 0x000001ff); \
                         printf("Associativity: %d\n", cache->nway); \
                         printf("Physical line partitions: %d\n", \
                                partitions); \
                         printf("Cache line size: %d\n", \
                                cache->specs.cache.linesize); \
                         printf("Cache size (KB) = %d\n", \
                                cache->specs.cache.size); \
                )
            }

	    leaf4_used = 1;
	}

        /* Cache size and format */

        PS_DEBUG(PS_DEBUG_INFO, puts("Retrieving cache information...");)

        op = 2;
        cpuid(op, &eax, &ebx, &ecx, &edx);
        reps = eax & 0x000000ff;     /* Number of times required to
                                        obtain all cache info */
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Need to run CPUID %d more times to get cache info\n",\
                        reps-1);)

        regs[0] = eax & 0xffffff00;  /* Mask off reps */
        regs[1] = ebx; regs[2] = ecx; regs[3] = edx;

        ret = do_intel_caches(regs, 4, leaf4_used);
        if ( ret != PS_SUCCESS ) {
            return ret;
        }

        for ( i = 1; i < reps; i++ ) {
            cpuid(op, &eax, &ebx, &ecx, &edx);
            regs[0] = eax & 0xffffff00; /* Mask off reps */
            regs[1] = ebx; regs[2] = ecx; regs[3] = edx;
            ret = do_intel_caches(regs, 4, leaf4_used);
            if ( ret != PS_SUCCESS ) {
                return ret;
            }
        }

    }

    return PS_SUCCESS;
}


static int
do_intel_caches(unsigned int regs[], int nregs, int leaf4_used)
{
    int i, j;
    unsigned int regval;
    ps_cpu_cache_t *cache;

    /* Table lookup applied to values passed in - what a pain... */

    for ( i = 0; i < nregs; i++ ) {
        if ( regs[i] & 0x80000000 ) continue; /* Bit 31 must be zero */
        for ( j = 0; j < 4; j++ ) {
            regval = (regs[i] >> (j * 8)) & 0x000000ff;

            PS_DEBUG(PS_DEBUG_INFO, printf("Cache regval = %x\n", regval);)

            /*
             * The following are "special cases" that we test for before
             * entering the longer case statement below.  The reason is that
             * we will not allocate a ps_cpu_cache_t structure for these
             * cases.
             *
             * 0x00: Null
             * 0x40: No L2 cache (P6), or if the processor contains a
             *       valid 2nd-level cache, then no L3 cache (Pentium 4)
             * 0xf0: 64-byte prefetching (?)
             * 0xf1: 128-byte prefetching (?)
             */
            if ( ( regval == 0 ) || ( regval == 0x40 )
		  || ( regval == 0xf0 ) || ( regval == 0xf1 ) ) {
                PS_DEBUG(PS_DEBUG_INFO, \
                         printf("Skipping cache alloc for descriptor %x\n",\
                                regval);)
                continue;
            }

            cache = new_cache();
            if ( cache == NULL ) {
                return PS_ENOMEM;
            }

            /* This could be shortened by grouping things, but
             * we'll leave it explicit to make it easier to follow
             * (not to mention debug!)
             */

            /* Defer */
            cache->level = -1;

            /* @@ Not checking these mallocs */
            switch ( regval ) {

            case 0x01: /* ITLB, 4K pages, 4-way set assoc, 32 entries */
                cache->type                   = ITLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.entries      = 32;
                break;

            case 0x02: /* ITLB, 4M pages, fully assoc, 2 entries */
                cache->type                   = ITLB;
                cache->nway                   = FULLY_ASSOCIATIVE;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4096;
                cache->specs.tlb.entries      = 2;
                break;

            case 0x03: /* DTLB, 4K pages, 4-way set assoc, 64 entries */
                cache->type                   = DTLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.entries      = 64;
                break;

            case 0x04: /* DTLB, 4M pages, 4-way set assoc, 8 entries */
                cache->type                   = DTLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4096;
                cache->specs.tlb.entries      = 8;
                break;

            case 0x05: /* DTLB, 4M pages, 4-way set assoc, 32 entries */
                cache->type                   = DTLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4096;
                cache->specs.tlb.entries      = 32;
                break;

            case 0x06: /* 1st-level icache, 8K, 4-way set assoc,
                          32 byte line */
                cache->level                  = 1;
                cache->type                   = ICACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 8;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x08: /* 1st-level icache, 16K, 4-way set assoc,
                          32 byte line */
                cache->level                  = 1;
                cache->type                   = ICACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 16;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x09: /* 1st-level icache, 32K, 4-way set assoc,
                          64 byte line */
                cache->level                  = 1;
                cache->type                   = ICACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 32;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x0a: /* 1st-level dcache, 8K, 2-way set assoc,
                          32 byte line */
                cache->level                  = 1;
                cache->type                   = DCACHE;
                cache->nway                   = 2;
                cache->specs.cache.size       = 8;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x0b: /* ITLB, 4M pages, 4-way set assoc, 4 entries */
                cache->type                   = ITLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4096;
                cache->specs.tlb.entries      = 4;
                break;

            case 0x0c: /* 1st-level dcache, 16K, 4-way set assoc,
                          32 byte line */
                cache->level                  = 1;
                cache->type                   = DCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 16;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x0d: /* 1st-level dcache, 16K, 4-way set assoc,
                          64 byte line */
                cache->level                  = 1;
                cache->type                   = DCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 16;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x0e: /* 1st-level dcache, 24K, 6-way set assoc,
                          64 byte line */
                cache->level                  = 1;
                cache->type                   = DCACHE;
                cache->nway                   = 6;
                cache->specs.cache.size       = 24;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x21: /* 2nd-level ucache, 256K, 8-way set assoc,
                          64 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 256;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x22: /* 3rd-level ucache, 512K, 4-way set assoc,
                          dual-sectored line, 64-byte sector size*/
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 512;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x23: /* 3rd-level ucache, 1M, 8-way set assoc,
                          dual-sectored line, 64-byte sector size*/
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 1024;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x25: /* 3rd-level ucache, 2M, 8-way set assoc,
                          dual-sectored line, 64-byte sector size*/
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 2048;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x29: /* 3rd-level ucache, 4M, 8-way set assoc,
                          dual-sectored line, 64-byte sector size*/
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 4096;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x2c: /* 1st-level dcache, 32K, 8-way set assoc,
                          64-byte line size*/
                cache->level                  = 1;
                cache->type                   = DCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 32;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x30: /* 1st-level icache, 32K, 8-way set assoc,
                          64-byte line size*/
                cache->level                  = 1;
                cache->type                   = ICACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 32;
                cache->specs.cache.linesize   = 64;
                break;

            /*******************************************************
             THE FOLLOWING ENTRY DOES NOT APPEAR IN AP-485, Jan. 11
             *******************************************************/
            case 0x39: /* 2nd-level ucache, 128K, 4-way set assoc,
                          dual-sectored line, 64-byte sector size*/
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 128;
                cache->specs.cache.linesize   = 64;
                break;

            /*******************************************************
             THE FOLLOWING ENTRY DOES NOT APPEAR IN AP-485, Jan. 11
             *******************************************************/
            case 0x3a: /* 2nd-level ucache, 192K, 6-way set assoc,
                          64-byte line size*/
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 6;
                cache->specs.cache.size       = 192;
                cache->specs.cache.linesize   = 64;
                break;

            /*******************************************************
             THE FOLLOWING ENTRY DOES NOT APPEAR IN AP-485, Jan. 11
             *******************************************************/
            case 0x3b: /* 2nd-level ucache, 128K, 2-way set assoc,
                          dual-sectored line, 64-byte sector size*/
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 2;
                cache->specs.cache.size       = 128;
                cache->specs.cache.linesize   = 64;
                break;

            /*******************************************************
             THE FOLLOWING ENTRY DOES NOT APPEAR IN AP-485, Jan. 11
             *******************************************************/
            case 0x3c: /* 2nd-level ucache, 256K, 4-way set assoc,
                          dual-sectored line, 64-byte sector size*/
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 256;
                cache->specs.cache.linesize   = 64;
                break;

            /*******************************************************
             THE FOLLOWING ENTRY DOES NOT APPEAR IN AP-485, Jan. 11
             *******************************************************/
            case 0x3d: /* 2nd-level ucache, 384K, 6-way set assoc,
                          dual-sectored line, 64-byte sector size*/
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 6;
                cache->specs.cache.size       = 384;
                cache->specs.cache.linesize   = 64;
                break;

            /*******************************************************
             THE FOLLOWING ENTRY DOES NOT APPEAR IN AP-485, Jan. 11
             *******************************************************/
            case 0x3e: /* 2nd-level ucache, 512K, 4-way set assoc,
                          dual-sectored line, 64-byte sector size*/
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 512;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x41: /* 2nd-level ucache, 128K, 4-way set assoc,
                          32 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 128;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x42: /* 2nd-level ucache, 256K, 4-way set assoc,
                          32 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 256;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x43: /* 2nd-level ucache, 512K, 4-way set assoc,
                          32 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 512;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x44: /* 2nd-level ucache, 1M, 4-way set assoc,
                          32 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 1024;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x45: /* 2nd-level ucache, 2M, 4-way set assoc,
                          32 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 2048;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x46: /* 3rd-level ucache, 4M, 4-way set assoc,
                          64-byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 4096;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x47: /* 3rd-level cache, 8M, 8-way set assoc,
                          64-byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 8192;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x48: /* 2nd-level cache, 3M, 12-way set assoc,
                          64-byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 12;
                cache->specs.cache.size       = 3072;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x49: /* 3rd-level cache, 4M, 16-way set assoc,
                          64-byte line */
                if ( (cpuinfo->cpuid_family == 15) && (cpuinfo->cpuid_model == 6) ) {
                    cache->level              = 3;
                }
                else {
                    cache->level              = 2;
                }
                cache->type                   = UCACHE;
                cache->nway                   = 16;
                cache->specs.cache.size       = 4096;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x4a: /* 3rd-level cache, 6M, 12-way set assoc,
                          64-byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 12;
                cache->specs.cache.size       = 6144;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x4b: /* 3rd-level cache, 8M, 16-way set assoc,
                          64-byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 16;
                cache->specs.cache.size       = 8192;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x4c: /* 3rd-level cache, 12M, 12-way set assoc,
                          64-byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 12;
                cache->specs.cache.size       = 12288;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x4d: /* 3rd-level cache, 16M, 16-way set assoc,
                          64-byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 16;
                cache->specs.cache.size       = 16384;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x4e: /* 2nd-level cache, 6M, 24-way set assoc,
                          64-byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 24;
                cache->specs.cache.size       = 6144;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x4f: /* ITLB, 4K pages, NOTE: assoc unknown, 32 entries */
                cache->type                   = ITLB;
                cache->nway                   = FULLY_ASSOCIATIVE;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(1*sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.entries      = 32;
                break;

            case 0x50: /* ITLB, 4K, 2M or 4M pages, fully assoc, 64 entries */
                cache->type                   = ITLB;
                cache->nway                   = FULLY_ASSOCIATIVE;
                cache->specs.tlb.npagesize    = 3;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(3*sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.pagesize[1]  = 2048;
                cache->specs.tlb.pagesize[2]  = 4096;
                cache->specs.tlb.entries      = 64;
                break;

            case 0x51: /* ITLB, 4K, 2M or 4M pages, fully assoc, 128 entries */
                cache->type                   = ITLB;
                cache->nway                   = FULLY_ASSOCIATIVE;
                cache->specs.tlb.npagesize    = 3;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(3*sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.pagesize[1]  = 2048;
                cache->specs.tlb.pagesize[2]  = 4096;
                cache->specs.tlb.entries      = 128;
                break;

            case 0x52: /* ITLB, 4K, 2M or 4M pages, fully assoc, 256 entries */
                cache->type                   = ITLB;
                cache->nway                   = FULLY_ASSOCIATIVE;
                cache->specs.tlb.npagesize    = 3;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(3*sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.pagesize[1]  = 2048;
                cache->specs.tlb.pagesize[2]  = 4096;
                cache->specs.tlb.entries      = 256;
                break;

            case 0x55: /* ITLB, 2M or 4M pages, fully assoc, 7 entries */
                cache->type                   = ITLB;
                cache->nway                   = FULLY_ASSOCIATIVE;
                cache->specs.tlb.npagesize    = 2;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(2*sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 2048;
                cache->specs.tlb.pagesize[1]  = 4096;
                cache->specs.tlb.entries      = 7;
                break;

            case 0x56: /* L1 DTLB, 4M pages, 4-way set assoc, 16 entries */
                cache->level                  = 1;
                cache->type                   = DTLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4096;
                cache->specs.tlb.entries      = 16;
                break;

            case 0x57: /* L1 DTLB, 4K pages, 4-way set assoc, 16 entries */
                cache->level                  = 1;
                cache->type                   = DTLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.entries      = 16;
                break;

            case 0x59: /* L1 DTLB, 4K pages, fully assoc, 16 entries */
                cache->level                  = 1;
                cache->type                   = DTLB;
                cache->nway                   = FULLY_ASSOCIATIVE;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.entries      = 16;
                break;

            case 0x5a: /* L1 DTLB, 2M or 4M pages, 4-way assoc, 32 entries */
                cache->level                  = 1;
                cache->type                   = DTLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 2;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(2*sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 2048;
                cache->specs.tlb.pagesize[1]  = 4096;
                cache->specs.tlb.entries      = 32;
                break;

            case 0x5b: /* DTLB, 4K or 4M pages, fully assoc, 64 entries */
                cache->type                   = DTLB;
                cache->nway                   = FULLY_ASSOCIATIVE;
                cache->specs.tlb.npagesize    = 2;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(2*sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.pagesize[1]  = 4096;
                cache->specs.tlb.entries      = 64;
                break;

            case 0x5c: /* DTLB, 4K or 4M pages, fully assoc, 128 entries */
                cache->type                   = DTLB;
                cache->nway                   = FULLY_ASSOCIATIVE;
                cache->specs.tlb.npagesize    = 2;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(2*sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.pagesize[1]  = 4096;
                cache->specs.tlb.entries      = 128;
                break;

            case 0x5d: /* DTLB, 4K or 4M pages, fully assoc, 256 entries */
                cache->type                   = DTLB;
                cache->nway                   = FULLY_ASSOCIATIVE;
                cache->specs.tlb.npagesize    = 2;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(2*sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.pagesize[1]  = 4096;
                cache->specs.tlb.entries      = 256;
                break;

            case 0x60: /* 1st-level dcache, sectored, 16K, 8-way set assoc,
                          64 byte line */
                cache->level                  = 1;
                cache->type                   = DCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 16;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x66: /* 1st-level dcache, sectored, 8K, 4-way set assoc,
                          64 byte line */
                cache->level                  = 1;
                cache->type                   = DCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 8;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x67: /* 1st-level dcache, sectored, 16K, 4-way set assoc,
                          64 byte line */
                cache->level                  = 1;
                cache->type                   = DCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 16;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x68: /* 1st-level dcache, sectored, 32K, 4-way set assoc,
                          64 byte line */
                cache->level                  = 1;
                cache->type                   = DCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 32;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x70: /* ITraceCache,  8-way set assoc, 12K uops */
                cache->type                   = ITRACECACHE;
                cache->nway                   = 8; /* changed in Intel doc */
                cache->specs.itcache.uops     = 12;
                break;

            case 0x71: /* ITraceCache,  8-way set assoc, 16K uops */
                cache->type                   = ITRACECACHE;
                cache->nway                   = 8; /* changed in Intel doc */
                cache->specs.itcache.uops     = 16;
                break;

            case 0x72: /* ITraceCache,  8-way set assoc, 32K uops */
                cache->type                   = ITRACECACHE;
                cache->nway                   = 8; /* changed in Intel doc */
                cache->specs.itcache.uops     = 32;
                break;

            /*******************************************************
             THE FOLLOWING ENTRY DOES NOT APPEAR IN AP-485, Jan. 11
             *******************************************************/
            case 0x73: /* ITraceCache,  8-way set assoc, 64K uops */
                cache->type                   = ITRACECACHE;
                cache->nway                   = 8; /* changed in Intel doc */
                cache->specs.itcache.uops     = 64;
                break;

            case 0x76: /* ITLB, 2M or 4M pages, fully assoc, 8 entries. In AP-485, Apr. 2012. */
                cache->type                   = ITLB;
                cache->nway                   = FULLY_ASSOCIATIVE;
                cache->specs.tlb.npagesize    = 2;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(2*sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 2048;
                cache->specs.tlb.pagesize[1]  = 4096;
                cache->specs.tlb.entries      = 8;
                break;

            case 0x78: /* 2nd-level ucache, 1M, 4-way set assoc,
                          64 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 1024;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x79: /* 2nd-level ucache, sectored, 128K, 8-way set assoc,
                          64 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 128;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x7a: /* 2nd-level ucache, sectored, 256K, 8-way set assoc,
                          64 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 256;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x7b: /* 2nd-level ucache, sectored, 512K, 8-way set assoc,
                          64 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 512;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x7c: /* 2nd-level ucache, sectored, 1M, 8-way set assoc,
                          64 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 1024;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x7d: /* 2nd-level ucache, 2M, 8-way set assoc,
                          64 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 2048;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x7f: /* 2nd-level ucache, 512K, 2-way set assoc,
                          64 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 2;
                cache->specs.cache.size       = 512;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x80: /* 2nd-level ucache, 512K, 8-way set assoc,
                          64 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 512;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x82: /* 2nd-level ucache, 256K, 8-way set assoc,
                          32 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 256;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x83: /* 2nd-level ucache, 512K, 8-way set assoc,
                          32 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 512;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x84: /* 2nd-level ucache, 1M, 8-way set assoc,
                          32 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 1024;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x85: /* 2nd-level ucache, 2M, 8-way set assoc,
                          32 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 2048;
                cache->specs.cache.linesize   = 32;
                break;

            case 0x86: /* 2nd-level ucache, 512K, 4-way set assoc,
                          64 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 512;
                cache->specs.cache.linesize   = 64;
                break;

            case 0x87: /* 2nd-level ucache, 1M, 8-way set assoc,
                          64 byte line */
                cache->level                  = 2;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 1024;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xb0: /* ITLB, 4K pages, 4-way set assoc, 128 entries */
                cache->type                   = ITLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.entries      = 128;
                break;

            case 0xb1: /* ITLB, 4M pages, 4-way set assoc, 4 entries */
                       /* or 2 MB pages, 4-way set assoc, 8 entries (not used)*/
                cache->type                   = ITLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4096;
                cache->specs.tlb.entries      = 4;
                break;

            case 0xb2: /* ITLB, 4K pages, 4-way set assoc, 64 entries */
                cache->type                   = ITLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.entries      = 64;
                break;

            case 0xb3: /* DTLB, 4K pages, 4-way set assoc, 128 entries */
                cache->type                   = DTLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.entries      = 128;
                break;

            case 0xb4: /* DTLB, 4K pages, 4-way set assoc, 256 entries */
                cache->type                   = DTLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.entries      = 256;
                break;

            case 0xba: /* DTLB, 4K pages, 4-way set assoc, 64 entries */
                cache->type                   = DTLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.entries      = 64;
                break;

            case 0xc0: /* DTLB, 4K or 4M pages, 4-way set assoc, 8 entries */
                cache->type                   = DTLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 2;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.pagesize[0]  = 4096;
                cache->specs.tlb.entries      = 8;
                break;

            case 0xca: /* Shared L2 UTLB, 4K pages, 4-way set assoc,
                          512 entries */
                cache->level                  = 2;
                cache->type                   = UTLB;
                cache->nway                   = 4;
                cache->specs.tlb.npagesize    = 1;
                cache->specs.tlb.pagesize     =
                    (unsigned int *) malloc(sizeof(unsigned int));
                cache->specs.tlb.pagesize[0]  = 4;
                cache->specs.tlb.entries      = 512;
                break;

            case 0xd0: /* 3rd-level ucache, 512K, 4-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 512;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xd1: /* 3rd-level ucache, 1M, 4-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 1024;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xd2: /* 3rd-level ucache, 2M, 4-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 4;
                cache->specs.cache.size       = 2048;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xd6: /* 3rd-level ucache, 1M, 8-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 1024;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xd7: /* 3rd-level ucache, 2M, 8-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 2048;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xd8: /* 3rd-level ucache, 4M, 8-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 8;
                cache->specs.cache.size       = 4096;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xdc: /* 3rd-level ucache, 1.5M, 12-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 12;
                cache->specs.cache.size       = 1536;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xdd: /* 3rd-level ucache, 3M, 12-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 12;
                cache->specs.cache.size       = 3072;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xde: /* 3rd-level ucache, 6M, 12-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 12;
                cache->specs.cache.size       = 6144;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xe2: /* 3rd-level ucache, 2M, 16-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 16;
                cache->specs.cache.size       = 2048;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xe3: /* 3rd-level ucache, 4M, 16-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 16;
                cache->specs.cache.size       = 4096;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xe4: /* 3rd-level ucache, 8M, 16-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 16;
                cache->specs.cache.size       = 8192;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xea: /* 3rd-level ucache, 12M, 24-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 24;
                cache->specs.cache.size       = 12288;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xeb: /* 3rd-level ucache, 18M, 24-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 24;
                cache->specs.cache.size       = 18432;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xec: /* 3rd-level ucache, 24M, 24-way set assoc,
                          64 byte line */
                cache->level                  = 3;
                cache->type                   = UCACHE;
                cache->nway                   = 24;
                cache->specs.cache.size       = 24576;
                cache->specs.cache.linesize   = 64;
                break;

            case 0xf0: /* 64-byte prefetching (?)  */
                PS_DEBUG(PS_DEBUG_WARNING, \
                         puts("Warning: encountered prefetching cache");)
                break;

            case 0xf1: /* 128-byte prefetching (?) */
                PS_DEBUG(PS_DEBUG_WARNING, \
                         puts("Warning: encountered prefetching cache");)
                break;

            case 0xff:
                /* CPUID Leaf 2 does not report cache descriptor information,
                   use CPUID Leaf 4 to query cache parameters */
                PS_DEBUG(PS_DEBUG_INFO, \
                         puts("CPUID reports that leaf 2 lacks cache info");)
                         break;

            default:   /* Unknown! */
                cache->type = UNKNOWN_CACHE;
                return PS_EUNKNOWNCPU;
                break;
            }

            if ( cache->type != UNKNOWN_CACHE ) {
		switch( cache->type ) {
		case ITLB:
		case DTLB:
		case UTLB:
		case ITRACECACHE:
		    add_cache(cache);
		    break;
		case ICACHE:
		case DCACHE:
		case UCACHE:
		    if ( ! leaf4_used ) {
			add_cache(cache);
		    }
		    else {
			free(cache); /* No additional memory allocated for
				      * the cache types, so a free of the
				      * basic structure should be fine. */
			PS_DEBUG(PS_DEBUG_VERBOSE, \
				 printf("Skipped duplicate cache add (%x)\n", \
					regval);)
		    }
		    break;
		case UNKNOWN_CACHE: /* Compiler likes this */
		    break;
		}
            }
            else {
                free(cache); /* No additional memory allocated beyond base. */
            }
        }
    }

    return PS_SUCCESS;
}


/*
 * Determine the number of cache levels for an AMD processor.  This is done
 * in a very simplistic way, specifically:
 *
 * - check that cpuid instruction supports extended cpuid function 8000_0006,
 *   aka AMD_L2CACHE_FETCH defined in cpuid.h
 * - if it does, execute the instruction and examine the contents of the
 *   edx register (L3 information), assume zero corresponds to no level 3
 *   cache so return 2.
 * - if edx != 0, return 3.
 */

static int
get_amd_cachelevels(unsigned int max_extended_function)
{
    unsigned int op, eax, ebx, ecx, edx;

    if ( max_extended_function >= AMD_L2CACHE_FETCH ) {
        op = AMD_L2CACHE_FETCH;
        ecx = 0;
        cpuid(op, &eax, &ebx, &ecx, &edx);

        if ( edx ) {
            return 3;
        }
    }
    
    return 2;
}


/*
 * Map codes to associativity.
 */

static int
get_amd_l2assoc(unsigned int code)
{
    switch (code) {
    case AMD_ASSOC_L2OFF:        return 0;
    case AMD_ASSOC_DIRECTMAPPED: return 1;
    case AMD_ASSOC_2WAY:         return 2;
    case AMD_ASSOC_4WAY:         return 4;
    case AMD_ASSOC_8WAY:         return 8;
    case AMD_ASSOC_16WAY:        return 16;
    case AMD_ASSOC_32WAY:        return 32;
    case AMD_ASSOC_48WAY:        return 48;
    case AMD_ASSOC_64WAY:        return 64;
    case AMD_ASSOC_96WAY:        return 96;
    case AMD_ASSOC_128WAY:       return 128;
    case AMD_ASSOC_FULL:         return FULLY_ASSOCIATIVE;
    default:                     return -1;
    }
}

static int
do_amd(ps_cpuid_t *cpuinfo, unsigned int max_fn)
{
    unsigned int op, eax, ebx, ecx, edx, regval;
    unsigned int max_extended_function = 0;
    char brand_string[AMD_BRAND_STRING_LENGTH];
    int nlevels;
    int off, i;
    int page4k;
    ps_cpu_cache_t *cache;

    if ( max_fn <= 0 ) return -1;

    /**********************************************
     **            Processor signature           **
     **********************************************/

    PS_DEBUG(PS_DEBUG_INFO, puts("Retrieving processor signature...");)

    op = 1;
    ecx = 0;
    cpuid(op, &eax, &ebx, &ecx, &edx);

    /* According to AMD "CPUID Specification", Publication 25481, Sep 2010: */
    if ( (eax >> 8 & 0xF) < 0xF ) {
        cpuinfo->cpuid_family = ((eax >> 8) & 0xF);
        cpuinfo->cpuid_model  = ((eax >> 4) & 0xF);
    } else {
        cpuinfo->cpuid_family = ((eax >> 20) & 0xFF) + ((eax >> 8) & 0xF);
        /* model: (eax & EXTENDED_MODEL_MASK) >> 16 << 4 == () >> 12. */
        cpuinfo->cpuid_model  = ((eax >> 12) & 0xF0) + ((eax >> 4) & 0xF);
    }
    cpuinfo->cpuid_stepping = (eax & 0xF);

    PS_DEBUG(PS_DEBUG_INFO, printf("eax = %x, ebx=%x\n", eax, ebx);)
    PS_DEBUG(PS_DEBUG_INFO, printf("family=%d, model=%d, stepping=%d\n", \
                                   cpuinfo->cpuid_family, \
                                   cpuinfo->cpuid_model, \
                                   cpuinfo->cpuid_stepping);)

    cpuinfo->revision = ( eax & AMD_STEPPING_MASK );

    /* Test for extended functions */

    op = AMD_EXTENDED_FUNCTION_DETECT;
    cpuid(op, &eax, &ebx, &ecx, &edx);
    max_extended_function = eax;
    PS_DEBUG(PS_DEBUG_VERBOSE, \
             printf("Extended function detect returned eax=%x\n", \
                    max_extended_function);)

    if ( ( max_extended_function >= AMD_BRAND_STRING_DETECT ) != 0 ) {
        PS_DEBUG(PS_DEBUG_INFO, puts("Brand string is available");)
        op = AMD_BRAND_STRING_FETCH;  /* Starting opcode */

        /* Need to call this three times to get all 48 bytes */
        for ( i = off = 0;
              i < 3;
              i++, op++, off += (sizeof(unsigned int) * 4) ) {
            cpuid(op, &eax, &ebx, &ecx, &edx);
            memcpy(brand_string+off,     &eax, 4);
            memcpy(brand_string+off+4,   &ebx, 4);
            memcpy(brand_string+off+8,   &ecx, 4);
            memcpy(brand_string+off+12,  &edx, 4);
        }

        /* Strip leading spaces */
        for ( i = 0; brand_string[i] == ' '; i++ ) ;

        cpuinfo->brand_string =
            PS_MALLOC(char, (AMD_BRAND_STRING_LENGTH-i));
        strncpy(cpuinfo->brand_string, &brand_string[i],
                AMD_BRAND_STRING_LENGTH-i);
    }
    else {
        PS_DEBUG(PS_DEBUG_INFO, puts("Brand string is not available");)
        cpuinfo->brand_string = NULL;
    }

    /* Caches */

    page4k = (getpagesize()/1024 == 4);
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("Checking AMD 4K pagesize: %s\n", page4k ? "yes" : "no");)

    /* Get L1 information */

    op = AMD_L1CACHE_FETCH;
    cpuid(op, &eax, &ebx, &ecx, &edx);
    regval = (page4k ? ebx : eax);

    cache = new_cache();
    cache->level = 1;
    cache->type = DTLB;
    cache->specs.tlb.entries = AMD_L1DTLB_ENTRIES(regval);

    if ( AMD_L1DTLB_ASSOC(regval) == 0xff ) {
        cache->nway = FULLY_ASSOCIATIVE;
    }
    else {
        cache->nway = AMD_L1DTLB_ASSOC(regval);
    }

    if ( page4k ) {
        cache->specs.tlb.npagesize = 1;
        cache->specs.tlb.pagesize =
            (unsigned int *) malloc(sizeof(unsigned int));
        cache->specs.tlb.pagesize[0] = 4;
    }
    else {
        cache->specs.tlb.npagesize = 2;
        cache->specs.tlb.pagesize =
                    (unsigned int *) malloc(sizeof(unsigned int) * 2);
        cache->specs.tlb.pagesize[0] = 2048;
        cache->specs.tlb.pagesize[1] = 4096;
    }

    add_cache(cache);

    cache = new_cache();

    cache->level = 1;
    cache->type = ITLB;
    cache->specs.tlb.entries = AMD_L1ITLB_ENTRIES(regval);

    if ( AMD_L1ITLB_ASSOC(regval) == 0xff ) {
        cache->nway = FULLY_ASSOCIATIVE;
    }
    else {
        cache->nway = AMD_L1ITLB_ASSOC(regval);
    }

    if ( page4k ) {
        cache->specs.tlb.npagesize = 1;
        cache->specs.tlb.pagesize =
            (unsigned int *) malloc(sizeof(unsigned int));
        cache->specs.tlb.pagesize[0] = 4;
    }
    else {
        cache->specs.tlb.npagesize = 2;
        cache->specs.tlb.pagesize =
            (unsigned int *) malloc(sizeof(unsigned int) * 2);
        cache->specs.tlb.pagesize[0] = 2048;
        cache->specs.tlb.pagesize[1] = 4096;
    }

    add_cache(cache);

    cache = new_cache();

    cache->level = 1;
    cache->type = DCACHE;
    cache->nway = AMD_L1DCACHE_ASSOC(ecx);
    cache->specs.cache.size = AMD_L1DCACHE_SIZE(ecx);
    cache->specs.cache.linesize = AMD_L1DCACHE_LINESIZE(ecx);

    add_cache(cache);

    cache = new_cache();

    cache->level = 1;
    cache->type = ICACHE;
    cache->nway = AMD_L1ICACHE_ASSOC(edx);
    cache->specs.cache.size = AMD_L1ICACHE_SIZE(edx);
    cache->specs.cache.linesize = AMD_L1ICACHE_LINESIZE(edx);

    add_cache(cache);
    
    nlevels = get_amd_cachelevels(max_extended_function);
    if ( nlevels < 2 ) {
        return PS_SUCCESS;
    }

    /* Get L2 information */
    op = AMD_L2CACHE_FETCH;
    cpuid(op, &eax, &ebx, &ecx, &edx);

    regval = (page4k ? ebx : eax);

    cache = new_cache();

    cache->level = 2;
    cache->type = AMD_L2UTLB_P(ebx) ? UTLB: ITLB;
    cache->specs.tlb.entries = AMD_L2IORUTLB_ENTRIES(regval);
    cache->nway = get_amd_l2assoc(AMD_L2IORUTLB_ASSOC(regval));
    if ( page4k ) {
        cache->specs.tlb.npagesize = 1;
        cache->specs.tlb.pagesize =
            (unsigned int *) malloc(sizeof(unsigned int));
        cache->specs.tlb.pagesize[0] = 4;
    }
    else {
        cache->specs.tlb.npagesize = 2;
        cache->specs.tlb.pagesize =
            (unsigned int *) malloc(sizeof(unsigned int) * 2);
        cache->specs.tlb.pagesize[0] = 2048;
        cache->specs.tlb.pagesize[1] = 4096;
    }

    add_cache(cache);

    if ( ! AMD_L2UTLB_P(ebx) ) {
        cache = new_cache();

        cache->level = 2;
        cache->type = DTLB;
        cache->specs.tlb.entries = AMD_L2DTLB_ENTRIES(regval);
        cache->nway = get_amd_l2assoc(AMD_L2DTLB_ASSOC(regval));

        if ( page4k ) {
            cache->specs.tlb.npagesize = 1;
            cache->specs.tlb.pagesize =
                (unsigned int *) malloc(sizeof(unsigned int));
            cache->specs.tlb.pagesize[0] = 4;
        }
        else {
            cache->specs.tlb.npagesize = 2;
            cache->specs.tlb.pagesize =
                (unsigned int *) malloc(sizeof(unsigned int) * 2);
            cache->specs.tlb.pagesize[0] = 2048;
            cache->specs.tlb.pagesize[1] = 4096;
        }
        
        add_cache(cache);
    }

    cache = new_cache();

    cache->level = 2;
    cache->type = UCACHE;
    cache->nway = get_amd_l2assoc(AMD_L2UCACHE_ASSOC(ecx));
    cache->specs.cache.size = AMD_L2UCACHE_SIZE(ecx);
    cache->specs.cache.linesize = AMD_L2UCACHE_LINESIZE(ecx);

    add_cache(cache);

    if ( nlevels == 2 ) {
        return PS_SUCCESS;
    }

    cache = new_cache();

    cache->level = 3;
    cache->type = UCACHE;
    cache->nway = get_amd_l2assoc(AMD_L3UCACHE_ASSOC(edx));
    cache->specs.cache.size = AMD_L3UCACHE_SIZE(edx) * 512;
    cache->specs.cache.linesize = AMD_L3UCACHE_LINESIZE(edx);

    add_cache(cache);


    if ( max_extended_function >= AMD_LONGMODE_SIZE_IDS ) {
        op = AMD_LONGMODE_SIZE_IDS;
        cpuid(op, &eax, &ebx, &ecx, &edx);
        PS_DEBUG(PS_DEBUG_VERBOSE, printf("%d\n", AMD_CORES_PER_PROC(ecx) + 1);)
    }

    /* Info from AMD application note

       AMD_L1CACHE_FETCH

       EAX for 2-MB/4-MB pages ** ONLY FOR ATHLON, DURON, ATHLON64, OPTERON **
       EAX: bits 31-24: associativity (data TLB)
            bits 23-16: # entries     (data TLB)
            bits 15-8:  associativity (inst TLB)
            bits  7-0:  # entries     (inst TLB)
       EBX for 4-KB pages
       EBX: bits 31-24: associativity (data TLB)
            bits 23-16: # entries     (data TLB)
            bits 15-8:  associativity (inst TLB)
            bits  7-0:  # entries     (inst TLB)
       ECX: bits 31-24: size (KB)     (L1 DCACHE)
            bits 23-16: associativity (L1 DCACHE)
            bits 15-8:  lines per tag (L1 DCACHE)
            bits  7-0:  line size (B) (L1 DCACHE)
       EDX: bits 31-24: size (KB)     (L1 ICACHE)
            bits 23-16: associativity (L1 ICACHE)
            bits 15-8:  lines per tag (L1 ICACHE)
            bits  7-0:  line size (B) (L1 ICACHE)

       AMD_L2CACHE_FETCH

       L2 cache information only relevant for AMD-K6-III, Duron,
       Athlon, Athlon64, and Opteron.

       A unified L2 TLB is indicated by a value of 0000h in the upper
       16 bits of the EBX register (info contained in the lower 16 bits)

       EAX for 2-MB/4-MB pages
       EAX: bits 31-28: associativity (L2 data TLB)
            bits 27-16: # entries     (L2 data TLB)
            bits 15-12: associativity (L2 inst or unified TLB)
            bits 11-0:  # entries     (L2 inst or unified TLB)
       EBX for 4-KB pages
       EBX: bits 31-28: associativity (L2 data TLB)
            bits 27-16: # entries     (L2 data TLB)
            bits 15-12: associativity (L2 inst or unified TLB)
            bits 11-0:  # entries     (L2 inst or unified TLB)
       ECX: bits 31-16: size (KB)     (L2 cache)
            bits 15-12: associativity (L2 cache)
            bits 11-8:  lines per tag (L2 cache)
            bits 7-0:   line size (B) (L2 cache)
    */

    return PS_SUCCESS;
}
