/* $Id: cpuid.c,v 1.40 2013/04/16 15:56:41 ruiliu Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 * Copyright(C) 2002-2008, 2010, 2012-2013, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *           National Center for Supercomputing Applications
 *             University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * cpuid: implement access to CPUID instruction.
 *
 *
 * Author: Rick Kufrin
 * Date:   November 23, 2002
 *
 * This file is part of PerfSuite.
 *
 */

/*
 * This file contains "generic" routines.  Architecture-specific
 * code is #include'd at the end.
 *
 * static inline void reset_cpuinfo
 * insert_cache
 * print_cachelist
 * do_assign
 * assign_cache_levels
 *
 */

#define USE_PS_MALLOC 1

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if _REENTRANT
# include <pthread.h>
#endif

#include <stdio.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#include <perfsuite.h>

#include "psinternal.h"
#include "cpuid.h"
#include "psdebug.h"

BEGIN_C_DECLS
static inline int _ps_cpuid_internal PARAMS((ps_cpuid_t *cpuid));
static inline void reset_cpuinfo PARAMS((ps_cpuid_t *cpuinfo));
END_C_DECLS

#if _REENTRANT
static pthread_mutex_t cpuinfo_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

static ps_cpuid_t *cpuinfo = NULL;

#include "cache.c"

/*
 * reset_cpuinfo -
 *
 * Return value:
 *
 * Side effects:
 */

static inline void
reset_cpuinfo(ps_cpuid_t *cpuinfo)
{
    cpuinfo->arch                 = UNKNOWN_ARCH;
    cpuinfo->vendor               = UNKNOWN_VENDOR;
    cpuinfo->type                 = UNKNOWN_TYPE;
    cpuinfo->cpuid_family         = 0;
    cpuinfo->cpuid_model          = 0;
    cpuinfo->cpuid_stepping       = 0;
    cpuinfo->cachearray.cachelist = NULL; /* @@ Should be freed if non-NULL */
    cpuinfo->cachearray.levels    = 0;
    cpuinfo->revision             = -1;
    cpuinfo->extra                = -1;
    cpuinfo->brand_string         = NULL;

    return;
}

/*
 * Pull in the proper architecture-specific source file
 */

#if IA32
# include "cpuid-x86.c"
#elif IA64
# include "cpuid-ia64.c"
#elif AIX5
# include "cpuid-aix.c"
#elif (PPC32 || PPC64)
# include "cpuid-powerpc.c"
#else
# include "cpuid-unknown.c"
#endif

/*
 * _ps_cpuid -
 *
 * Return value:
 *
 * Side effects:
 */

int
_ps_cpuid(ps_cpuid_t *cpuid)
{
    int ret;
#if _REENTRANT
    int sysret;
#endif

#if _REENTRANT
    pthread_cleanup_push((void *) pthread_mutex_unlock, &cpuinfo_mutex);
    sysret = pthread_mutex_lock(&cpuinfo_mutex);
    if ( sysret != 0 ) {
        return PS_ESYS;
    }
#endif

    if ( cpuinfo == NULL ) {
#if DEBUG
        ps_setdebuglevel();
#endif
        PS_DEBUG(PS_DEBUG_VERBOSE, puts("Allocating internal cpuid data.");)
        cpuinfo = PS_MALLOC(ps_cpuid_t, 1);
        ret = _ps_cpuid_internal(cpuinfo);
        if ( ret != PS_SUCCESS ) {
            free(cpuinfo);
            cpuinfo = NULL;
        }
    }
    else {
        ret = PS_SUCCESS;
        PS_DEBUG(PS_DEBUG_VERBOSE, puts("Reusing internal cpuid data.");)
    }

#if _REENTRANT
    pthread_cleanup_pop(1);
#endif

    if ( ret == PS_SUCCESS ) {
        memcpy(cpuid, cpuinfo, sizeof(ps_cpuid_t));
    }

    return ret;
}
