/* $Id: cpuid.h,v 1.32 2013/04/16 15:56:41 ruiliu Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *   Copyright(C) 2002-2010, 2012-2013, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *           National Center for Supercomputing Applications
 *             University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

#ifndef CPUID_H_INCLUDED
#define CPUID_H_INCLUDED 1

#include <stdlib.h>
#include <stdint.h>        /* For the def of uint32_t. */
#include "constants.h"

#define VENDOR_STRING_INTEL "GenuineIntel"
#define VENDOR_STRING_AMD   "AuthenticAMD"

/* Signature bit patterns in EAX for Intel */
/* See Intel Application Note 485, Order Number 241618-038, April 2012. */

#define INTEL_EXTENDED_FAMILY_MASK 0x0ff00000
#define INTEL_EXTENDED_MODEL_MASK  0x000f0000
#define INTEL_TYPE_MASK            0x00003000
#define INTEL_FAMILY_MASK          0x00000f00
#define INTEL_MODEL_MASK           0x000000f0
#define INTEL_STEPPING_MASK        0x0000000f

#define INTEL_TYPE_OEM             0x00000000
#define INTEL_TYPE_OVERDRIVE       0x00001000
#define INTEL_TYPE_DUAL            0x00002000

#define INTEL_FAMILY_486           0x00000400
#define INTEL_FAMILY_PENTIUM       0x00000500
#define INTEL_FAMILY_P6            0x00000600
#define INTEL_FAMILY_P4            0x00000f00

#define PARTIAL_MODEL              0x33333333

/* 486 models */
#define INTEL_MODEL_I486SL         0x00000040
#define INTEL_MODEL_IDX2WBE        0x00000070
#define INTEL_MODEL_IDX4           0x00000080 /* DX4 and DX4 OverDrive */

/* Pentium models - need to inspect type too */
#define INTEL_MODEL_PENTIUM_1      0x00000010 /* 60, 66 */
#define INTEL_MODEL_PENTIUM_2      0x00000020 /* 75, 90, 100, 120, 133 
                                                 (also 150, 166, 200) */
#define INTEL_MODEL_PENTIUM_3      0x00000030
#define INTEL_MODEL_PENTIUM_4      0x00000040

/* Pentium Pro (P6) models - need to inspect type too */
#define INTEL_MODEL_P6_1           0x00000010
#define INTEL_MODEL_P6_3           0x00000030 /* Type check here */
#define INTEL_MODEL_P6_5           0x00000050
#define INTEL_MODEL_P6_6           0x00000060
#define INTEL_MODEL_P6_7           0x00000070
#define INTEL_MODEL_P6_8           0x00000080
#define INTEL_MODEL_P6_10          0x000000a0

/* Pentium 4 models need to check the extended family mask */

/* Core models */
#define INTEL_MODEL_CORE_1         0x000000e0 /* Core Duo, Core Solo */
#define INTEL_MODEL_CORE_2         0x000000f0 /* Core 2 Duo */

/* @@ None of these are used yet, and the bit masks not checked */
#define INTEL_MODEL_P4_1           0x00000000 /* P4, Xeon, 0.18 micron */
#define INTEL_MODEL_P4_2           0x00000010 /* P4, Xeon, Xeon MP,
                                                 Celeron 0.18 */
#define INTEL_MODEL_P4_3           0x00000020 /* P4, Mobile P4 -M , Xeon, 
                                                 Xeon MP, Celeron,
                                                 Mobile Celeron
                                                 0.13 micron */
#define INTEL_MODEL_P4_4           0x00000030 /* P4, Mobile P4 - M,
                                                 Xeon, Celeron
                                                 90 nm */
#define INTEL_MODEL_P4_5           0x00000040 /* P4, Mobile P4 - M,
                                                 Xeon, Celeron
                                                 90 nm */


/* Brand ID - returned in EBX when CPUID called with EAX = 1 */

#define INTEL_BRAND_MASK           0x000000ff

#define INTEL_BRAND_CELERON        0x00000001
#define INTEL_BRAND_PIII           0x00000002
#define INTEL_BRAND_PIIIXEON       0x00000003 /* Need to check signature */
#define INTEL_BRAND_PIII_2         0x00000004 /* Same as 0x02? */
#define INTEL_BRAND_MOBILE_PIII    0x00000006
#define INTEL_BRAND_MOBILE_CELERON 0x00000007
#define INTEL_BRAND_P4             0x00000008 /* Need to check signature */
#define INTEL_BRAND_P4_2           0x00000009
#define INTEL_BRAND_CELERON_2      0x0000000a
#define INTEL_BRAND_XEON_B         0x0000000b /* Need to check signature */
#define INTEL_BRAND_XEON_MP        0x0000000c
#define INTEL_BRAND_XEON_E         0x0000000e /* Need to check signature */
#define INTEL_BRAND_MOBILE_CELERON_2 0x0000000f
#define INTEL_BRAND_MOBILE         0x00000011 /* @@ None below added to psinv yet */
#define INTEL_BRAND_CELERON_M      0x00000012
#define INTEL_BRAND_MOBILE_CELERON_3 0x00000013
#define INTEL_BRAND_CELERON_3      0x00000014
#define INTEL_BRAND_MOBILE_2       0x00000015
#define INTEL_BRAND_PENTIUM_M      0x00000016
#define INTEL_BRAND_MOBILE_CELERON_4 0x00000017

/* Brand string */

#define INTEL_BRAND_STRING_DETECT  0x80000000
#define INTEL_BRAND_STRING_FETCH   0x80000002 /* 2 through 4 */
#define INTEL_BRAND_STRING_LENGTH  48


/* Signature bit patterns in EAX for AMD */

#define AMD_EXTENDED_FAMILY_MASK   INTEL_EXTENDED_FAMILY_MASK
#define AMD_EXTENDED_MODEL_MASK    INTEL_EXTENDED_MODEL_MASK
#define AMD_FAMILY_MASK            INTEL_FAMILY_MASK
#define AMD_MODEL_MASK             INTEL_MODEL_MASK
#define AMD_STEPPING_MASK          INTEL_STEPPING_MASK

/* AMD processor families */

#define AMD_FAMILY_AM_486_586      0x00000400
#define AMD_FAMILY_K5_K6           0x00000500
#define AMD_FAMILY_K7              0x00000600
#define AMD_FAMILY_K8              0x00000f00

/* AMD processor models */

#define AMD_MODEL_K5_0             0x00000000
#define AMD_MODEL_K5_1             0x00000010
#define AMD_MODEL_K5_2             0x00000020
#define AMD_MODEL_K5_3             0x00000030
#define AMD_MODEL_K6_6             0x00000060
#define AMD_MODEL_K6_7             0x00000070
#define AMD_MODEL_K6_2_8           0x00000080
#define AMD_MODEL_K6_III_9         0x00000090
#define AMD_MODEL_ATHLON_1         0x00000010
#define AMD_MODEL_ATHLON_2         0x00000020
#define AMD_MODEL_DURON_3          0x00000030
#define AMD_MODEL_ATHLON_4         0x00000040
#define AMD_MODEL_ATHLON_DURON_6   0x00000060  /* Multiple models */
#define AMD_MODEL_DURON_7          0x00000070  /* Multiple models */
#define AMD_MODEL_ATHLON_8         0x00000080  /* Multiple models */
#define AMD_MODEL_ATHLON_XP_M_8_10 0x000000a0  /* Multiple models */
#define AMD_MODEL_ATHLON64_4       0x00000040
#define AMD_MODEL_OPTERON_5        0x00000050

/* AMD extended function test */

#define AMD_EXTENDED_FUNCTION_DETECT 0x80000000

/* AMD address size and # CPU cores op */

#define AMD_LONGMODE_SIZE_IDS        0x80000008

/* AMD brand string info */

#define AMD_BRAND_STRING_DETECT      0x80000002
#define AMD_BRAND_STRING_FETCH       INTEL_BRAND_STRING_FETCH
#define AMD_BRAND_STRING_LENGTH      INTEL_BRAND_STRING_LENGTH

/* AMD cache-related extended functions and masks */

#define AMD_L1CACHE_FETCH            0x80000005
#define AMD_L1DTLB_ASSOC(r)          ((r & 0xff000000) >> 24)
#define AMD_L1DTLB_ENTRIES(r)        ((r & 0x00ff0000) >> 16)
#define AMD_L1ITLB_ASSOC(r)          ((r & 0x0000ff00) >>  8)
#define AMD_L1ITLB_ENTRIES(r)         (r & 0x000000ff)
#define AMD_L1DCACHE_SIZE(r)         ((r & 0xff000000) >> 24)
#define AMD_L1DCACHE_ASSOC(r)        ((r & 0x00ff0000) >> 16)
#define AMD_L1DCACHE_LINESIZE(r)      (r & 0x000000ff)
#define AMD_L1ICACHE_SIZE(r)         ((r & 0xff000000) >> 24)
#define AMD_L1ICACHE_ASSOC(r)        ((r & 0x00ff0000) >> 16)
#define AMD_L1ICACHE_LINESIZE(r)      (r & 0x000000ff)

#define AMD_L2CACHE_FETCH            0x80000006
#define AMD_L2UTLB_P(r)              (((r >> 16) & 0x0000ffff) == 0)
#define AMD_L2DTLB_ASSOC(r)          ((r & 0xf0000000) >> 28)
#define AMD_L2DTLB_ENTRIES(r)        ((r & 0x0fff0000) >> 16)
#define AMD_L2IORUTLB_ASSOC(r)       ((r & 0x0000f000) >> 12)
#define AMD_L2IORUTLB_ENTRIES(r)      (r & 0x00000fff)
#define AMD_L2UCACHE_SIZE(r)         ((r & 0xffff0000) >> 16)
#define AMD_L2UCACHE_ASSOC(r)        ((r & 0x0000f000) >> 12)
#define AMD_L2UCACHE_LINESIZE(r)      (r & 0x000000ff)

#define AMD_L3UCACHE_SIZE(r)         ((r & 0xfff60000) >> 18)
#define AMD_L3UCACHE_ASSOC(r)        ((r & 0x0000f000) >> 12)
#define AMD_L3UCACHE_LINESIZE(r)      (r & 0x000000ff)

/* Number of CPU cores - 1 */

#define AMD_CORES_PER_PROC(r)         (r & 0x000000ff)

/* AMD L2 cache and TLB associativity codes (L1 are their actual value) */

#define AMD_ASSOC_L2OFF              0x0
#define AMD_ASSOC_DIRECTMAPPED       0x1
#define AMD_ASSOC_2WAY               0x2
#define AMD_ASSOC_4WAY               0x4
#define AMD_ASSOC_8WAY               0x6
#define AMD_ASSOC_16WAY              0x8
#define AMD_ASSOC_32WAY              0xa
#define AMD_ASSOC_48WAY              0xb
#define AMD_ASSOC_64WAY              0xc
#define AMD_ASSOC_96WAY              0xd
#define AMD_ASSOC_128WAY             0xe
#define AMD_ASSOC_FULL               0xf

/* Cache */

/* /proc filesystem places to get information */

#ifndef PROC_CPUINFO
# define PROC_CPUINFO "/proc/cpuinfo"
#endif

#if IA64
# ifndef PROC_CACHEINFO
#   define PROC_CACHEINFO "/proc/pal/cpu0/cache_info"
# endif

# ifndef PROC_VMINIFO
#   define PROC_VMINFO "/proc/pal/cpu0/vm_info"
# endif
#endif

/*
 * cpuid information (cpuid.c)
 */

/* Recognized architectures */

typedef enum {
    UNKNOWN_ARCH = -1,
    IA32_ARCH,
    IA64_ARCH,
    POWER_RS_ARCH,
    POWER_PC_ARCH
} ps_cpu_arch_t;
/* Recognized vendors */

typedef enum {
    UNKNOWN_VENDOR = -1,
    AMD,
    INTEL,
    IBM
} ps_cpu_vendor_t;

typedef enum {
    UNKNOWN_TYPE = -1,
    OEM,
    OVERDRIVE,
    DUAL
} ps_cpu_type_t;

#if 0
/* <ptrace.h> defines this, at least on x86 */
#if defined(_I386_PTRACE_H)
#undef SS
#endif
#endif

typedef enum {
    UNKNOWN_CACHE = -1,
    ITLB,
    DTLB,
    UTLB,
    ICACHE,
    DCACHE,
    UCACHE,
    ITRACECACHE
} ps_cpu_cachetype_t;

/* Descriptors for a cache:
   All:
     Type (above)
     Associativity (n-way or fully associative)
   TLB:
     Page size [KB] (TLB - could be multiple)
     Number of entries (TLB)
   Data/Instruction Cache
     Cache size [KB]
     Line size {B}
     Cores (how many are sharing this cache?)
     Load/store latencies, attributes (not implemented)
   Instruction Trace cache
     Number of uOps [K]
 */

typedef union {
    struct {
        unsigned int npagesize;
        unsigned int *pagesize; /* kilobytes */
        unsigned int entries;
    } tlb;
    struct {
        unsigned int size;      /* kilobytes  */
        unsigned int linesize;  /*     bytes  */
        unsigned int cores;     /* 0: unknown */
        /* @@ Not yet added.
           unsigned int lines;
           unsigned int store_latency;   cycles
           unsigned int load_latency;    cycles
           ps_cache_attribute_t attributes;  write-through, write-back, etc
        */
    } cache;
    struct {
        unsigned int uops;      /* kilo-ops  */
    } itcache;
} ps_cache_specs_t;
    
/* Used in nway below */
#define FULLY_ASSOCIATIVE 0
#define DIRECT_MAPPED     1

typedef struct {
    ps_cpu_cachetype_t  type;
    unsigned int        level;     /* 1+ */
    unsigned int        nway; 
    ps_cache_specs_t    specs;
} ps_cpu_cache_t;

/*
 * Contains ALL "caches" : inst/data caches, instruction trace,
 * unified, ITLB, DTLB.  
 *
 */

/* @@ This may be better moved to internal cpuid.h header */
typedef struct ps_cpu_cachelist {
    ps_cpu_cache_t          *cache;
    struct ps_cpu_cachelist *next;
} ps_cpu_cachelist_t;

typedef struct {
    ps_cpu_cachelist_t **cachelist;
    int                levels;
} ps_cpu_cachearray_t;

/* @@ Bus speeds? */
typedef struct {
    ps_cpu_arch_t       arch;
    ps_cpu_vendor_t     vendor;
    uint32_t            cpuid_family;
    uint32_t            cpuid_model;
    uint32_t            cpuid_stepping;
    ps_cpu_type_t       type;
    ps_cpu_cachearray_t cachearray;
    int                 revision;
    int                 extra;         /* A field to store extra info.
                                          For IBM Power, this is the PVR_VER. */
    char                *brand_string;
} ps_cpuid_t;

#endif
