/* $Id: environment.c,v 1.18 2006/01/16 21:15:02 rkufrin Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *        Copyright(C) 2002-2006, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *           National Center for Supercomputing Applications
 *             University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#include <perfsuite.h>
#include "psinternal.h"
#include "psdebug.h"


/*
 * Add the value "val" to the environment variable
 * named "var", optionally separating it from an
 * existing value using the string "sep".
 *
 * The new value is placed at the beginning of the
 * environment value (@@ should be an option to
 * place at the end)
 *
 * Memory is malloc'd for the entire new value, old
 * value, and the separator string, if supplied.
 */

int
_ps_addtoenv(const char *var, const char *val, const char *sep)
{
    char *oldval, *newval;
    int newlen = 1;  /* For the terminating NULL */

    /* First check for silly use... */
    if ( ( val == NULL ) || ( var == NULL ) ) {
        return 0;
    }
    if ( strlen(var) == 0 ) {
        return 0;
    }

    /* Stash the old value or set it empty if nonexistent */
    oldval = getenv(var);

    if ( oldval == NULL ) {
        oldval = "";
    }

    /* Figure out total space */
    newlen = strlen(var) + strlen(val) + strlen(oldval) + 1;
    if ( sep != NULL ) {
        newlen += strlen(sep);
    }

    /* Allocate the space and arrange for the new value to be
     * inserted at the beginning
     */
    /*
      if ((newval = (char *) malloc(newlen+1)) == NULL) {
      return PS_ENOMEM;
      }
    */
    newval = PS_MALLOC(char, (newlen+1));
    strcpy(newval,var);
    strcat(newval,"=");
    strcat(newval,val);
    if ( sep != NULL ) {
        strcat(newval, sep);
    }

    /* Concatenate the old value on and place in the environment */
#if HAVE_PUTENV
    return ( putenv(strcat(newval,oldval)) );
#else
#error "missing putenv()"
#endif
}

/*
 * _ps_findexe: search through PATH and return the absolute
 *              name of an executable provided in exename,
 *              up to maxlen characters.  Result returned in 
 *              pathname.
 *
 * Not thoroughly tested.
 */

int
_ps_findexe(char *exename, char *pathname, int maxlen)
{
    char *path, *index;
    int len, avail;

    if ( strchr(exename, '/') != NULL ) {
        if ( access(exename, X_OK) == 0 ) {
            strncpy(pathname, exename, maxlen);
            return PS_SUCCESS;
        }
        return -1;
    }

    if ( (path = getenv("PATH")) == NULL) {
        return -1;
    }

    avail = maxlen - strlen(exename) - 1;

    /* Cycle through directories in path */
    while ( (index = strchr(path, ':')) != NULL ) {
        len = index - path;
        if ( len <= avail ) {
            strncpy(pathname, path, avail);
            pathname[len]   = '/';
            pathname[len+1] = '\0';
            strcat(pathname, exename);
            if ( access(pathname, X_OK) == 0 ) {
                return PS_SUCCESS;
            }
        }
        path = index + 1;
    }

    /* Check the final path component */
    if ( path != NULL ) {
        len = strlen(path);
        if ( len <= avail ) {
            strncpy(pathname, path, avail);
            pathname[len]   = '/';
            pathname[len+1] = '\0';
            strcat(pathname, exename);
            if ( access(pathname, X_OK) == 0 ) {
                return PS_SUCCESS;
            }
        }
    }

    return -1;
}
