/* $Id: error.c,v 1.32 2006/01/16 21:15:02 rkufrin Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *        Copyright(C) 2002-2006, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *           National Center for Supercomputing Applications
 *             University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * error.c - error-handling routines for PerfSuite
 *
 * Author: Rick Kufrin
 * Date:   5/2003
 *
 * This file is part of PerfSuite
 *
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#include <perfsuite.h>
#include <pserror.h>
#include "psinternal.h"

typedef struct {
    const ps_status_t code;
    const char *string;
} ps_error_t;

/*
 * These strings should be no longer than PS_MAXERRLEN,
 * as defined in <perfsuite.h>
 */

static const ps_error_t ps_errors[] = {
    /***********************************/
    /* System errors                   */
    /***********************************/
    {PS_ENOMEM,                  "memory allocation failed"},
    {PS_ESYS,                    "system error"},

    /***********************************/
    /* libperfsuite  (codes    1- 999) */
    /***********************************/
    {PS_ECPUNOTSUPPORTED,        "CPU not supported"},
    {PS_EUNKNOWNCPU,             "CPU is unknown"},
    {PS_ENOTFOUND,               "could not locate required item"},
    {PS_ENOTINSTALLED,           "a required software component not installed"},
    {PS_EBADXMLDOCTYPE,          "improper XML document type supplied"},
    {PS_ENOTIMPLEMENTED,         "this routine is not implemented"},
    {PS_EBUFFER,                 "internal buffer exhausted"},
    {PS_EINVAL,                  "invalid argument"},
    {PS_EEXISTS,                 "item already exists"},

    /***********************************/
    /* libpshwpc     (codes 1000-1999) */
    /***********************************/
    {PS_HWPC_EXMLPARSE,          "error parsing XML document"},
    {PS_HWPC_EEVENTSEXCEEDED,    "maximum number of hardware events exceeded"},
    {PS_HWPC_EEVENTNOTSUPPORTED, "unsupported hardware event"},
    {PS_HWPC_EBADCLASS,          "unsupported value for \"class\" attribute"},
    {PS_HWPC_ESTREAM,            "error writing output stream"},
    {PS_HWPC_EPERF,              "error reported by performance software layer"},
    {PS_HWPC_EBADEVENT,          "invalid hardware event specification"},
    {PS_HWPC_ENOSUCHEVENT,       "no such hardware event exists"},
    {PS_HWPC_EEVENTNOTAVAILABLE, "hardware event not available"},
    {PS_HWPC_ENOTSUPPORTED,      "not supported"},
    {PS_HWPC_ENOSUCHXMLDOC,      "XML document does not exist"},
    {PS_HWPC_ESEQUENCE,          "calling sequence not allowed"},
    {PS_HWPC_EMODE,              "not available in this mode"},

    /***********************************/
    /* libpspmpi     (codes 2000-2999) */
    /***********************************/

    /***********************************/
    /* No error                        */
    /***********************************/
    {PS_SUCCESS,                 NULL}
};

#define NERRORS sizeof(ps_errors)/sizeof(ps_error_t)

#define NO_ERROR      "no error"
#define INVALID_ERROR "unknown error code"

const char *
ps_strerror(int code)
{
    int lo, hi, mid;

    if ( code == PS_SUCCESS ) {
        return NO_ERROR;
    }

    for (lo = 0, hi = NERRORS - 1; lo <= hi; ) {
        mid =  lo + ((hi-lo)/2);
        if ( ps_errors[mid].code == code ) {
            return ps_errors[mid].string;
        }
        if ( code > ps_errors[mid].code ) {
            lo = mid + 1;
        }
        else {
            hi = mid - 1;
        }
    }

    return INVALID_ERROR;
}

/* Fortran wrapper */

#define PSF_STRERROR_F77     F77_FUNC_(psf_strerror,PSF_STRERROR)

void
PSF_STRERROR_F77(int *code, char *buf, int *n2copy, int *ierr, int maxlen)
{
    const char *errstring;
    int len;

    errstring = ps_strerror(*code);

    if ( strcmp(errstring, INVALID_ERROR) == 0 ) {
        *ierr = -1;
        return;
    }

    /*
     * Adjust the length copied.  *n2copy tells us how many
     * characters the user wants to copy, maxlen tells us
     * how much space is available in the destination string
     * according to the compiler.  We want to never exceed
     * maxlen, and we won't copy any more than the user asks.
     */
    if ( *n2copy > maxlen ) {
        len = maxlen;
    }
    else {
        len = *n2copy;
    }

    _ps_c2fstring(buf, strlen(errstring), errstring, len);

    *ierr = PS_SUCCESS;

    return;
}
