/* $Id: hash.c,v 1.17 2006/01/16 21:15:02 rkufrin Exp $ */

/****************************************************************************
 *
 *                      University of Illinois/NCSA
 *                          Open Source License
 * 
 *          Copyright(C) 2004-2006, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                             Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * hash.c - support for hash tables
 *
 * Author: Rick Kufrin
 * Date:   7/15/2004
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>

#if HAVE_STRING_H
# include <string.h>
#endif

#include <perfsuite.h>
#include <psdebug.h>
#include "hash.h"

#define PS_FASTLOOKUP 1

/*
 * This is purely to try to cut some corners internally for
 * speed.  It's used by _ps_hashtable_lookup to store away
 * the last hash key generated so that _ps_hashtable_insert
 * can reuse it.  See the comments in _ps_hashtable_insert.
 * This should not be relied on to remain in the library.
 * Also, it is currently not thread-safe, so disabled for
 * compilation for threads.
 */
#undef PS_FASTHASH
#if _REENTRANT
# define PS_FASTHASH 0
#else
# define PS_FASTHASH 1
#endif
#if PS_FASTHASH
static unsigned long lasthash;
#endif

/*
 * Dan Bernstein's djb2 hash function.
 */
static inline unsigned long
_ps_hash_internal(unsigned char *str)
{
    unsigned long hash = 5381;
    int c;

    while ( (c = *str++) != 0 )  {
        hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
    }
   
    return hash;
}

unsigned long
_ps_hash(char *str)
{
    return _ps_hash_internal((unsigned char *) str);
}

/*
 * Create a new hashtable of a given size.
 */

int
_ps_hashtable_create(int size, ps_hashtable_t **table)
{
    int i, j;

    /* Make sure we weren't called by a nut. */
    if ( size < 1 ) {
        return PS_EINVAL;
    }

    /* Allocate the overall table structure. */
    *table = (ps_hashtable_t *) malloc(sizeof(ps_hashtable_t));
    if ( *table == NULL ) {
        return PS_ENOMEM;
    }

    /* Allocate the individual bins. */
    (*table)->bin =
        (ps_hashchain_t **) malloc(sizeof(ps_hashchain_t *) * size);

    if ( (*table)->bin == NULL ) {
        free(table);
        return PS_ENOMEM;
    }

    /* Allocate and initialize each bin.  Set count to zero. */
    for ( i = 0; i < size; i++ ) {
        (*table)->bin[i] = (ps_hashchain_t *) malloc(sizeof(ps_hashchain_t));
        if ( (*table)->bin[i] == NULL ) {
            for ( j = 0; j < i; j++ ) {
                free((*table)->bin[j]);
            }
            free(table);
            return PS_ENOMEM;
        }
        (*table)->bin[i]->nentries = (*table)->bin[i]->size = 0;
        (*table)->bin[i]->entries  = NULL;
    }

    /* Store the table size. */
    (*table)->size = size;

    return PS_SUCCESS;
}

/*
 * Insert a data block into the hash table, indexed by the given key.
 * The key must not already exist in the table.
 */
int
_ps_hashtable_insert(ps_hashtable_t *table, char *key, void *data)
{
    int hash, ret, i;
#if ! HAVE_MEMMOVE
    int j;
#endif
    ps_hashentry_t *entry;
    ps_hashchain_t *chain;
    void *where;

    /* Check to make sure no identical key already exists in the table. */
    ret = _ps_hashtable_lookup(table, key, &where);
    if ( ret == PS_SUCCESS ) {
        return PS_EEXISTS;
    }

    /*
     * Generate the hash function and map it to an index based
     * on the table size.
     *
     * @@ We are hashing this key twice, so could save some
     * time by reusing the hash from the above call to lookup.
     * Generating the hash value seems to be the second most
     * time-consuming thing after strcmp().
     */
#if PS_FASTHASH
    hash = lasthash;
#else
    hash = _ps_hash_internal((unsigned char *) key) % table->size;
#endif

    /*
     *  See if the chain needs to be expanded.
     */
    chain = table->bin[hash];
    if ( chain->nentries == 0  ) {
        chain->entries = (ps_hashentry_t *)
            malloc(sizeof(ps_hashentry_t) * HASHCHAIN_INITIAL_SIZE);
        if ( chain->entries == NULL ) {
            return PS_ENOMEM;
        }
        chain->size = HASHCHAIN_INITIAL_SIZE;
    }
    else if ( chain->nentries == chain->size ) {
        chain->size *= 2;
        chain->entries = (ps_hashentry_t *)
            realloc(chain->entries, chain->size * sizeof(ps_hashentry_t));
        if ( chain->entries == NULL ) {
            return PS_ENOMEM;
        }
    }

    /* Allocate a new entry for this data */
    entry = (ps_hashentry_t *) malloc(sizeof(ps_hashentry_t));
    if ( entry == NULL ) {
        return PS_ENOMEM;
    }

    /* Fill in the appropriate slots. */
    entry->key       = strdup(key);
    entry->data      = data;

    /*
     * Insert the new entry in the right spot.  This could be
     * expensive due to memory moves, but hopefully additions
     * happen much less frequently than probes/lookups.
     *
     * We try to short-circuit and avoid calls to strcmp by first 
     * comparing the initial characters of each string.
     */
    for ( i = 0; i < chain->nentries; i++ ) {
       if ( chain->entries[i].key[0] >= entry->key[0] ) {
           if ( strcmp(chain->entries[i].key, entry->key) > 0 ) {
               break;
           }
       }
    }

    /*
     * "i" now points to either the first existing entry beyond
     * where the new entry should go or is at the end of the list
     */
#if HAVE_MEMMOVE
    memmove(&(chain->entries[i+1]), &(chain->entries[i]),
            (chain->nentries - i) * sizeof(ps_hashentry_t));
#else
    /*  Older version, "hand"-copying */
    for ( j = chain->nentries; j > i; j-- ) {
        memcpy(&(chain->entries[j]), &(chain->entries[j-1]),
               sizeof(ps_hashentry_t));
    }
#endif
    memcpy(&(chain->entries[i]), entry, sizeof(ps_hashentry_t));
    chain->nentries++;

    return PS_SUCCESS;
}

/*
 * Search the table to see if a given key exists in the table.
 * If it does, return the data block associated with it.
 * The data block is not returned if the data argument is NULL.
 * This can be useful just to see if a key exists.
 *
 * The current implementation uses a binary search on the
 * table, which is stored in ascending sorted order.  See
 * _ps_hashtable_insert for the management details.
 *
 */
/*
static long lookups = 0 , probes = 0;
*/
int
_ps_hashtable_lookup(ps_hashtable_t *table, char *key, void **data)
{
    int hash, lo, hi, mid, ret;
    ps_hashentry_t *entries;

/*
    lookups++;
*/

#if PS_FASTHASH
    lasthash = 
#endif
    hash = /*(int)*/ _ps_hash_internal((unsigned char *) key) % table->size;

    /*fprintf(stderr, "key %s hashes to index %d\n", key, hash);*/

    lo      = 0;
    hi      = table->bin[hash]->nentries-1;
    entries = table->bin[hash]->entries;

    while ( lo <= hi ) {
/*
        probes++;
*/
        mid = (hi + lo) / 2;
#if PS_FASTLOOKUP
        /* Try a short-circuit here, as in insert */
        ret = ( *entries[mid].key - *key );
        if ( ret == 0 ) {
            ret = strcmp(entries[mid].key, key);
        }
#else
        ret = strcmp(entries[mid].key, key);
#endif
        if ( ret == 0 ) {
            if ( data != NULL ) {
                *data = entries[mid].data;
            }
            return PS_SUCCESS;
        }
        else if ( ret < 0 ) {
            lo = mid + 1;
        }
        else {
            hi = mid - 1;
        }
    }

    return PS_ENOTFOUND;
}

/*
 * Convenience function that fills the array "entries" with
 * the number of keys that have hashed to each bin in the
 * hash table.  Caller has the responsibility to allocate
 * this array with sufficient space.
 */
int
_ps_hashtable_stats(ps_hashtable_t *table, int *entries)
{
    int bin;

    for ( bin = 0; bin < table->size; bin++ ) {
            entries[bin] = table->bin[bin]->nentries;
    }

/*
    printf("There were %ld lookups, average %.4f probes per lookup\n",
           lookups, (double)probes/(double)lookups);
 */

    return PS_SUCCESS;
}
