/* $Id: procfs.c,v 1.71 2010/09/16 20:38:46 ruiliu Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *     Copyright(C) 2002-2006, 2010, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *           National Center for Supercomputing Applications
 *             University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * procfs : assorted /proc utility routines
 *
 * Author : Rick Kufrin
 *   Date : Jan 2002
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if _REENTRANT
# if HAVE_PTHREAD
#  include <pthread.h>
# endif
#endif

#include <stdio.h>
#include <errno.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#if HAVE_FCNTL_H
# include <fcntl.h>
#endif

#if HAVE_LIMITS_H
# include <limits.h>
#endif

#if HAVE_ASSERT_H
# include <assert.h>
#endif

#if HAVE_MALLOC_H
# include <malloc.h>
#endif

#include <perfsuite.h>
#include "psinternal.h"
#include "psdebug.h"
#include "constants.h"

/* MAXTOKENLEN is a string constant that is used in sizing buffers in the
 * functions _ps_addrmap() and _ps_getargs().
 */

#if defined(PATH_MAX)
# define MAXTOKENLEN PATH_MAX
#else
# define MAXTOKENLEN 4096  /* Value from linux/limits.h */
#endif

#if defined(AIX5)
# if HAVE_PMAPI_H
#   include <pmapi.h>
# endif
#endif

#if _REENTRANT
void
_ps_pthread_cleanup_fd(void *p)
{
    int ret;
#if DEBUG
    FILE *fp;
#endif
    
    ret = fclose(*((FILE **)p));

#if DEBUG
    fp = fopen("PROCSTAT_CLEANUP", "w+");
    fprintf(fp, "cancelled in _ps_pthread_cleandup_fd. fclose returned %d\n",
            ret);
    fclose(fp);
#endif

    return;
}
#endif


/* Fortran wrappers for CPU times routines */

#define PSF_CPUSPEED_F77 F77_FUNC_(psf_cpuspeed,PSF_CPUSPEED)
#define PSF_MEMUSAGE_F77 F77_FUNC_(psf_memusage,PSF_MEMUSAGE)
#define PSF_PROCSTAT_F77 F77_FUNC_(psf_procstat,PSF_PROCSTAT)

int
ps_procstat(pid_t pid, ps_procstat_t *p)
{
    FILE *fd;
    char fname[256];
    int n;
#if _REENTRANT && (KERNEL_MINOR_VERSION >= 6)
    int ret;
#endif

    if ( pid == 0 ) {
#if _REENTRANT && (KERNEL_MINOR_VERSION >= 6)
        ret = _ps_gettid(&pid);
        if ( ret != PS_SUCCESS ) {
            return ret;
        }
#else
        pid = getpid();
#endif
    }

    sprintf(fname, "/proc/%d/stat", pid);

#if _REENTRANT
    pthread_cleanup_push((void *)_ps_pthread_cleanup_fd, &fd);
#endif

    fd = fopen(fname, "r");
    if ( fd == NULL ) {
        if ( errno == ENOENT ) {
            errno = ESRCH;
        }
        return PS_ESYS;
    }

    /* 2.6: numthreads *NEW, was constant zero previously
            RT_PRIORITY *ADDED*, POLICY *ADDED* but not here
    */
    /* @@ Should change to one call to fscanf */
    n = fscanf(fd, "%d (%[^)]) %c %d", &p->pid, p->comm, &p->state, &p->ppid);
    n = fscanf(fd, "%d %d %d %d %lu",  &p->pgrp, &p->session, &p->tty,
               &p->tpgid, &p->flags);
    n = fscanf(fd, "%lu %lu %lu %lu",  &p->minflt, &p->cminflt, &p->majflt, 
               &p->cmajflt);
    n = fscanf(fd, "%ld %ld %ld %ld",  &p->utime, &p->stime, &p->cutime,
               &p->cstime);
#if (KERNEL_MAJOR_VERSION == 2) && (KERNEL_MINOR_VERSION < 6)
    n = fscanf(fd, "%ld %ld %*d",      &p->priority, &p->nice);
#else
    n = fscanf(fd, "%ld %ld %ld",      &p->priority, &p->nice, &p->numthreads);
#endif
    n = fscanf(fd, "%lu",              &p->itrealvalue);
    n = fscanf(fd, "%lu %lu %lu %lu",  &p->starttime, &p->vsize, &p->rss,
               &p->rlim);
    n = fscanf(fd, "%lu %lu %lu %lu %lu", &p->startcode, &p->endcode,
               &p->startstack, &p->kstkesp, &p->kstkeip);
    n = fscanf(fd, "%d %d %d %d %lu",   &p->signal, &p->blocked, &p->sigignore,
               &p->sigcatch, &p->wchan);
    n = fscanf(fd, "%lu %lu %d %d",    &p->nswap, &p->cnswap, &p->exitsignal,
               &p->processor);

    fclose(fd);

#if _REENTRANT
    pthread_cleanup_pop(0);
#endif

    return PS_SUCCESS;
}

/*
  character(MAXCOMMSIZ) (comm)
  character (state)
*/

void
PSF_PROCSTAT_F77(int *pid, int *ia, char *command, char *state,
                 int *ierr, int commandlen)
{
    ps_procstat_t p;
    int i;

    *ierr = ps_procstat(*pid, &p);
    if ( *ierr != PS_SUCCESS )
        return;

    /* Fill 'er up... */

    for ( i = 0; i < commandlen; i++ ) *(command+i) = ' ';

    strncpy(command, p.comm, strlen(p.comm));

    *state             = p.state;

    ia[PS_PID]         = p.pid;
    ia[PS_PPID]        = p.ppid;
    ia[PS_PGRP]        = p.pgrp;
    ia[PS_SESSION]     = p.session;
    ia[PS_TTY]         = p.tty;
    ia[PS_TPGID]       = p.tpgid;
    ia[PS_FLAGS]       = (int) p.flags;
    ia[PS_MINFLT]      = (int) p.minflt;
    ia[PS_CMINFLT]     = (int) p.cminflt;
    ia[PS_MAJFLT]      = (int) p.majflt;
    ia[PS_CMAJFLT]     = (int) p.cmajflt;
    ia[PS_UTIME]       = (int) p.utime;
    ia[PS_STIME]       = (int) p.stime;
    ia[PS_CUTIME]      = (int) p.cutime;
    ia[PS_CSTIME]      = (int) p.cstime;
    ia[PS_PRIORITY]    = (int) p.priority;
    ia[PS_NICE]        = (int) p.nice;
    ia[PS_NUMTHREADS]  = (int) p.numthreads;
    ia[PS_ITREALVALUE] = (int) p.itrealvalue;
    ia[PS_STARTTIME]   = (int) p.starttime;
    ia[PS_VSIZE]       = (int) p.vsize;
    ia[PS_RSS]         = (int) p.rss;
    ia[PS_RLIM]        = (int) p.rlim;
    ia[PS_STARTCODE]   = (int) p.startcode;
    ia[PS_ENDCODE]     = (int) p.endcode;
    ia[PS_STARTSTACK]  = (int) p.startstack;
    ia[PS_KSTKESP]     = (int) p.kstkesp;
    ia[PS_KSTKEIP]     = (int) p.kstkeip;
    ia[PS_SIGNAL]      = p.signal;
    ia[PS_BLOCKED]     = p.blocked;
    ia[PS_SIGIGNORE]   = p.sigignore;
    ia[PS_SIGCATCH]    = p.sigcatch;
    ia[PS_WCHAN]       = (int) p.wchan;
    ia[PS_NSWAP]       = (int) p.nswap;
    ia[PS_CNSWAP]      = (int) p.cnswap;
    ia[PS_EXITSIGNAL]  = p.exitsignal;
    ia[PS_PROCESSOR]   = p.processor;

    return;
}

int
ps_memusage(pid_t pid, float *vsizemb, float *rssmb)
{
    int ret = 0, pagesize;
    ps_procstat_t p;

    ret = ps_procstat(pid, &p);
    if ( ret == PS_SUCCESS ) {
        pagesize = getpagesize();
        *vsizemb = p.vsize / (double) MEGABYTE;
        *rssmb = (p.rss * pagesize) / (double) MEGABYTE;
    }

    return ret;
}

void
PSF_MEMUSAGE_F77(int *pid, float *vsizemb, float *rssmb, int *ierr)
{
    *ierr = ps_memusage(*pid, vsizemb, rssmb);

    return;
}

#define CPUINFOFILE "/proc/cpuinfo"

/* Key to search for in CPUINFOFILE */

#if defined(IA32) || defined(IA64)
# define SPEED_KEY "cpu MHz"
#elif defined(PPC32) || defined(PPC64)
# define SPEED_KEY "clock"
#else
/*#error "Architecture not defined"*/
# define SPEED_KEY "cpu MHz"
#endif

#define CPUINFOBUFSIZ 256

int
ps_cpuspeed(double *speed)
{
#if defined(IA32) || defined(IA64) || defined(PPC32) || defined(PPC64)
    char buf[CPUINFOBUFSIZ];
    char keybuf[CPUINFOBUFSIZ], valbuf[CPUINFOBUFSIZ];
    FILE *fp;
    int found=0, len=strlen(SPEED_KEY);

#if _REENTRANT
    pthread_cleanup_push((void *)_ps_pthread_cleanup_fd, &fp);
#endif

    fp = fopen(CPUINFOFILE,"r");
    if ( fp == NULL ) {
        if ( errno == ENOENT ) {
            errno = ESRCH;
        }
        return PS_ESYS;
    }

    while ( ( ! found ) && ( fgets(buf, CPUINFOBUFSIZ, fp) != NULL ) ) {
        if ( buf[0] != '\n' ) {
            sscanf(buf,"%[^:]:%[^\n:]",keybuf,valbuf);
            if ( strncmp(keybuf,SPEED_KEY,len) == 0 ) {
                /* @@ No error-checking here */
                *speed = strtod(valbuf, (char **)NULL);
                /*printf("Found speed=%f (valbuf=%s)\n", *speed, valbuf);*/
                found = 1;
            }
        } else {
            fclose(fp);
            return PS_ENOTFOUND;
        }
    }

    fclose(fp);

#if _REENTRANT
    pthread_cleanup_pop(0);
#endif

    return PS_SUCCESS;

#elif defined(AIX5)
# if HAVE_PMAPI_H
    *speed = pm_cycles();
    if ( *speed < 1.0 ) {
        return PS_ESYS;
    }
    *speed *= 1.e-6;
    return PS_SUCCESS;
# else
    return PS_ENOTINSTALLED;
# endif
#else
# error "Unsupported architecture (ps_cpuspeed)"
#endif
}

void
PSF_CPUSPEED_F77(double *speed, int *ierr)
{
    *ierr = ps_cpuspeed(speed);

    return;
}


/***************************************************************
 *    INTERNAL ROUTINES NOT MEANT FOR END USER                 *
 ***************************************************************/

typedef struct maplist {
    ps_addrmap_t *map;
    struct maplist *next;
} maplist_t; 

static void
cleanup_tempmaps(maplist_t *list)
{
    maplist_t *next;

    while ( list != NULL ) {
        next = list->next;
        free(list->map->module);
        free(list->map);
        free(list);
        list = next;
    }

    return;
}

int
_ps_addrmap(int pid, int *mapc, ps_addrmap_t ***map)
{
    FILE *fd;
    char fname[256], modname[MAXTOKENLEN], prevname[MAXTOKENLEN];
    char perms[5], line[MAXTOKENLEN], *strret;
    char format[128];
    int scanned, i;
    caddr_t start, end, offset;
    long unsigned int istart, iend, ioffset;
    ps_addrmap_t *newmap;
    maplist_t *head = (maplist_t *) NULL, *tail = (maplist_t *) NULL, *ptr;
    int tmppid = pid;

#if DEBUG
    ps_setdebuglevel();
#endif

#if _REENTRANT
    pthread_cleanup_push((void *)_ps_pthread_cleanup_fd, &fd);
#endif

    *mapc = 0;

    if ( tmppid == 0 ) {
        tmppid = getpid();
    }

    sprintf(fname, "/proc/%d/maps", tmppid);

    fd = fopen(fname, "r");
    if ( fd == NULL ) {
        if ( errno == ENOENT ) {
            errno = ESRCH;
        }
        return PS_ESYS;
    }

    PS_DEBUG(PS_DEBUG_INFO, printf("Process load map from /proc:\n");)
#if DEBUG
    while ( (strret = fgets(line, MAXTOKENLEN - 1, fd)) != NULL ) {
        fputs(line, stdout);
    }
    errno = 0; /* Ensure this is cleared before proceeeding on */
    rewind(fd);
#endif

    /*
     * Looking for entries of the form:
     *
     * 40033000-40165000 r-xp 00000000 08:06 959106     /lib/i686/libc-2.2.4.so
     *    ^         ^     ^      ^       ^     ^                ^
     *    |         |     |      |       |     |                |
     * vm_start   vm_end perms offset  device inode           module
     *
     */

    /*
     * This format string will be used in the sscanf below.  It is written
     * to a string in order to size the final field width by MAXTOKENLEN.
     *
     * Ex.: "%lx-%lx %s %lx %*s %*u %4096[^\n]"
     */
    sprintf(format, "%%lx-%%lx %%s %%lx %%*s %%*u %%%d[^\n]", MAXTOKENLEN-1);

    strcpy(prevname, "");

    while ( 1 ) {
        /*
         * It is possible that the /proc entry was opened without error,
         * but read access is denied, so we check the return value of
         * fgets to see if an error occurred before proceeding.
         */
        strret = fgets(line, MAXTOKENLEN - 1, fd);
        if ( (strret == NULL) && (errno != 0) ) {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fprintf(stderr, "_ps_addrmap: read of %s failed (%s)\n", \
                             fname, strerror(errno));)
            if ( head != NULL ) {
                /* This should not actually happen, as permission failure
                 * should occur on first attempt to read, but play it safe.
                 */
                cleanup_tempmaps(head);
            }
            fclose(fd);
            return PS_ESYS;
        }
        modname[0] = '\0';
        /*
         * Using the same format as in linux/fs/proc/array.c
         * minus field width specifiers, that's why the reading
         * into a long int and then conversion to caddr_t for
         * our purposes.
         */
        scanned = sscanf(line, format, &istart, &iend, perms, &ioffset, modname);
        start  = (caddr_t) istart;
        end    = (caddr_t) iend;
        offset = (caddr_t) ioffset;

        PS_DEBUG(PS_DEBUG_VERBOSE, \
                 printf("scan of map entry returned %d (should be 5), offset = %lx\n", \
                        scanned, ioffset);)

        if ( feof(fd) ) break;

        /* Construct a temporary linked list of the maps we're interested in */

        /* @@ Previously, was this:
        if ( (strcmp(perms, "r-xp") == 0) || (strcmp(perms, "rwxp") == 0) ) {
         */

        if ( strcmp(perms, "r-xp") == 0 ) {
            /*
             * Don't bother with modules not associated with a file,
             * pseudo-files like [stack], [heap], [vdso].
             */
            if ( (*modname == '\0') || (*modname == '[') ) continue;

            /*
             * SGI note/mod:
             * On SLES11 there are two lines with permission "r-xp" for
             * the same module, which results in excessive memory allocation.
             * The following block will discard the second such entry and
             * keep only the first one.
             *
             * @@ This assumes, therefore, that the first such line ALWAYS
             * correspond to a program segment and could be a potential source
             * of error.
             */
            if ( strcmp(modname, prevname) == 0 ) {
                PS_DEBUG(PS_DEBUG_VERBOSE, \
                         printf("load map: discarding duplicate module \"%s\"",\
                                modname);)
                continue;
            }
            else {
                strcpy(prevname, modname);
            }

            /*
             * Here's one we want.  Allocate a new map structure
             * and populate it with the proper entries.
             */

            newmap = (ps_addrmap_t *) malloc(sizeof(ps_addrmap_t));
            if ( newmap == NULL ) {
                cleanup_tempmaps(head);
                fclose(fd);
                return PS_ENOMEM;
            } 
            PS_DEBUG(PS_DEBUG_VERBOSE, \
                     printf("Allocated memory at %p\n", newmap);)

            newmap->start = start;
            newmap->end = end;
            newmap->offset = offset;
            newmap->module = strdup(modname);
            if ( newmap->module == NULL ) {
                cleanup_tempmaps(head);
                fclose(fd);
                return PS_ENOMEM;
            }
            PS_DEBUG(PS_DEBUG_VERBOSE, \
                     printf("Allocated memory (strdup) at %p\n", newmap->module);)

            if ( head == NULL ) {
                head = tail = (maplist_t *) malloc(sizeof(maplist_t));
                if ( head == NULL ) {
                    cleanup_tempmaps(head);
                    fclose(fd);
                    return PS_ENOMEM;
                }
                PS_DEBUG(PS_DEBUG_VERBOSE, \
                         printf("Allocated memory at %p\n", head);)
                head->map = tail->map = newmap;
                head->next = NULL;
            }
            else {
                tail->next = (maplist_t *) malloc(sizeof(maplist_t));
                if ( tail->next == NULL ) {
                    cleanup_tempmaps(head);
                    fclose(fd);
                    return PS_ENOMEM;
                }
                PS_DEBUG(PS_DEBUG_VERBOSE, \
                         printf("Allocated memory at %p\n", tail->next);)
                tail = tail->next;
                tail->map = newmap;
                tail->next = NULL;
            }

            (*mapc)++;
        }
    }
    
    /* Restructure into an array for return */
    *map = (ps_addrmap_t **) malloc(*mapc * sizeof(ps_addrmap_t *));
    if ( *map == NULL) {
        cleanup_tempmaps(head);
        fclose(fd);
        return PS_ENOMEM;
    }
    PS_DEBUG(PS_DEBUG_VERBOSE, printf("Allocated memory at %p\n", *map);)

    /* Populate the return array and free temporary memory used */
    ptr = head;
    for ( i = 0; i < *mapc; i++ ) {
        (*map)[i] = ptr->map;
        head = ptr;
        ptr = ptr->next;
        free(head);
        PS_DEBUG(PS_DEBUG_VERBOSE, printf("Freeing memory at %p\n", head);)
    }
   
    fclose(fd);

#if _REENTRANT
    pthread_cleanup_pop(0);
#endif

    return PS_SUCCESS;
}

int
_ps_addrmap_free(int mapc, ps_addrmap_t **mapv)
{
    int map;

    for ( map = 0; map < mapc; map++) {
        PS_DEBUG(PS_DEBUG_VERBOSE, \
                 printf("Freeing memory at %p [\"%s\"]\n", 
                        mapv[map]->module, mapv[map]->module);)
        free(mapv[map]->module);
        PS_DEBUG(PS_DEBUG_VERBOSE, printf("Freeing memory at %p\n", mapv[map]);)
        free(mapv[map]);
    }
    PS_DEBUG(PS_DEBUG_VERBOSE, printf("Freeing memory at %p\n", mapv);)
    free(mapv);

    return PS_SUCCESS;
}

/*
 * Find the index in the address map that corresponds to
 * the current process.
 */

int
_ps_addrmap_exe(int mapc, ps_addrmap_t **mapv, int *index)
{
    char *ptr, *comm, exe[256], buf[256]; /* @@ MAXPATHLEN? */
    int commlen;
    pid_t pid;

    pid = getpid();

    sprintf(buf, "/proc/%d/exe", pid);
    commlen = readlink(buf, exe, 255);
    exe[commlen] = '\0';

    comm = strrchr(exe, '/'); /* @@ UNIX-style */
    if ( comm == NULL ) {
        comm = exe;
    }
    else {
        comm++;
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("_ps_addrmap_exe: scanning address map (%d entries) for executable \"%s\":\n", \
                    mapc, comm);)

    for ( *index = 0; *index < mapc; (*index)++ ) {
        ptr = strrchr(mapv[*index]->module, '/');
        if ( ptr != NULL ) {
             ptr++;
        } else {
             ptr = mapv[*index]->module;
        }
        assert (ptr != NULL);
        PS_DEBUG(PS_DEBUG_INFO, printf("\t%d: \"%s\"\n", *index, ptr);)

        /* @@ This is potentially a subtle bug - there could
         * be multiple modules with the same prefix and we could
         * pick up the wrong one.
         */
        if ( strncmp(ptr, comm, commlen) == 0 ) {
            return PS_SUCCESS;
        }
    }

    PS_DEBUG(PS_DEBUG_WARNING, puts("Scan for executable failed.");)

    return PS_ENOTFOUND;
}


/*
 * _ps_getargs : return the complete command line associated with a given
 *               process ID, structured in standard argc/argv form. 
 *
 * /proc/PID/exe can also be used to get the executable name.
 * We want to pick up the command-line arguments as well.
 *
 * This function was modified for PerfSuite version 0.6.2b2 to address
 * errors that were occurring with more recent versions of the Linux
 * kernel (demonstrated on kernel 2.6.23).  In particular, the documented
 * format of the /proc/PID/cmdline file (man page), which stated that
 * the contents are a series of NULL-terminated strings, appears to be
 * no longer constant across all cmdline entries.  There are some entries
 * that contain no NULLs at all, yet contain 1 or more characters that
 * can be read.
 *
 * Additionally, the function was modified to handle a different format
 * in which a cmdline file can contain multiple trailing NULLs, none of
 * which appear to be associated with an argv entry.
 *
 * Examples of these types of entries:
 *
 * Case 1. The format as documented in the proc(5) man page.
 *
 * $ od -cv /proc/self/cmdline  ; # Taken with kernel 2.6.9
 * 0000000   o   d  \0   -   c   v  \0   /   p   r   o   c   /   s   e   l
 * 0000020   f   /   c   m   d   l   i   n   e  \0
 * 0000032
 *
 * Case 2. An entry that contains extra trailing NULLs.
 *
 * $ od -cv /proc/1/cmdline  ; # Also with kernel 2.6.9
 *
 * 0000000   i   n   i   t       [   5   ]  \0  \0  \0  \0  \0  \0  \0  \0
 * 0000020  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0
 * 0000040  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0  \0
 * 0000060  \0  \0  \0  \0  \0  \0  \0
 * 0000067
 *
 * Case 3. An entry with no NULLs ; # Taken with kernel 2.6.23
 *
 * $ od -cv /proc/1/cmdline
 *
 * 0000000   /   s   b   i   n   /   i   n   i   t
 * 0000012
 *
 */

int
_ps_getargs(volatile int pid, int *argc, char ***argv)
{
    FILE *fd;
    char buf[MAXTOKENLEN]; /* @@ MAXPATHLEN? Buffer used for 2 purposes */
    int c, i, pos;
    int nulls = 0, realnulls, count = 0;

#if DEBUG
    ps_setdebuglevel();
#endif

#if _REENTRANT
    pthread_cleanup_push((void *)_ps_pthread_cleanup_fd, &fd);
#endif

    /*
     * Open the proper /proc cmdline entry, as requested by the caller.  
     * If the pid argument supplied is 0, then we open the calling process'
     * entry, otherwise open the entry corresponding to the arguemnt.
     */
    if ( pid == 0 ) {
        pid = getpid();
    }

    sprintf(buf, "/proc/%d/cmdline", pid);

    fd = fopen(buf, "r");
    if ( fd == NULL ) {
        /*
         * If errno indicates "no such file or directory",
         * we modify it to "no such process", which is hopefully more
         * meaningful to the caller of this function.
         */
        PS_DEBUG(PS_DEBUG_WARNING, \
                 fprintf(stderr, "_ps_getargs: could not open %s (%s)\n", \
                         buf, strerror(errno));)
       
        if ( errno == ENOENT ) {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fputs("Resetting errno to ESRCH\n", stderr);)
            errno = ESRCH;
        }
        return PS_ESYS;
    }

    PS_DEBUG(PS_DEBUG_VERBOSE, printf("ps_getargs: opened %s\n", buf);)

    /*
     * Read the contents of the cmdline file, noting the number of NULL
     * bytes that are encountered.  In the "standard" format as documented
     * in man pages, this will tell us how many strings contain the command
     * executed as well as its arguments.
     *
     * We also count the total number of bytes read, so that in the case
     * no NULLs are encountered, we can allocate space to treat the entry
     * as a single string that is returned to the caller.  Note that this
     * behavior may result in "fake" entries, that do not correspond to any
     * actual command line executed.
     *
     * In the case of additional trailing NULLs, this loop will end up
     * counting all of them (and the variable nulls will be larger than
     * we really want).
     */
    while (1) {
       c = fgetc(fd);
       if ( c == EOF ) {
           break;
       }
       if ( c == 0 ) {
           nulls++;
       }
       count++;
    }

    PS_DEBUG(PS_DEBUG_VERBOSE, 
             printf("%s contained %d bytes, %d NULLs\n", buf, count, nulls);)

    /*
     * Store away the actual number of NULLs counted, as the logic
     * that follows may modify the variable nulls.
     */
    realnulls = nulls;

    /*
     * Allocate space for the returned argv array of strings.
     *
     * First, we check for case 3 above (no trailing NULLs).  In this
     * case, we will force argc to be 1 so that it can contain whatever
     * the contents of /proc/PID/cmdline are.
     */
    if ( nulls > 0 ) {
        *argc = nulls;
    }
    else {
        PS_DEBUG(PS_DEBUG_VERBOSE, \
                 printf("_ps_getargs: forcing cmdline argc to 1\n");)
        *argc = nulls = 1;
    }

    /*
     * Allocate the space for the number of entries we want plus
     * an additional entry that will signal the final element of
     * the array.
     *
     * Also initialize to zero.
     *
     * In the additional trailng NULL case, we will be overallocating
     * additional entries and will deallocate them before return.
     */
    *argv = (char **) malloc(sizeof(char *) * (nulls + 1));
    if ( *argv == NULL ) {
        fclose(fd);
        return PS_ENOMEM;
    }
    memset(*argv, 0, sizeof(char **) * (nulls + 1));

    /*
     * Now rescan the /proc entry to copy the data into the return
     * array for this function.
     */

    rewind (fd);

    for ( i = 0; i < nulls; i++ ) {
        /*
         * Initialize the buffer to ensure that there is a zero byte
         * at the end of the string, since there may not be one contained
         * in the cmdline entry itself (for case 3).
         */
        memset(buf, 0, MAXTOKENLEN);
        pos = 0;
        while (1) {
            c = buf[pos++] = fgetc(fd);
            if ( (c == 0) || feof(fd) ) {
                break;
            }
            if ( pos == MAXTOKENLEN ) {
                for ( pos = 0; (*argv)[pos]; pos++ ) {
                    free((*argv)[pos]);
                }
                free(*argv);
                return PS_EBUFFER;
            }
        }
        (*argv)[i] = strdup(buf);
    }
   
    /*
     * Final correction for case 2 where trailing NULLs exist.
     * We do not want them counted as arguments.  This loop will
     * note the position of the first NULL-valued entry.
     */
    for ( nulls = 0; nulls < *argc; nulls++ ) {
        PS_DEBUG(PS_DEBUG_VERBOSE, \
                 printf("_ps_getargs: Argument %d: \"%s\"\n", \
                        nulls, (*argv)[nulls]);)
        if ( (*argv)[nulls][0] == 0 ) {
            break;
        }
    }
    *argc = nulls;

    /*
     * If the first NULL entry is found to be earlier than the
     * original NULL count, then shrink the returned argv array
     * after deallocating the excess entries that were strdup'd
     * above.
     */
    if ( nulls != realnulls ) {
        PS_DEBUG(PS_DEBUG_VERBOSE, \
                 puts("_ps_getargs: trimming trailing cmdline NULLs");)
        for ( i = nulls + 1; i < realnulls; i++ ) {
            free((*argv)[i]);
        }
        if ( realloc(*argv, sizeof(char **) * (nulls + 1)) == NULL) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("_ps_getargs: final reallocation of argv failed\n",\
                           stderr);)
            return PS_ENOMEM;
        }
    }

    /*
     * Make completely sure that the final argv returned is NULL.
     */
    (*argv)[*argc] = NULL;

    fclose(fd);

#if _REENTRANT
    pthread_cleanup_pop(0);
#endif

    PS_DEBUG(PS_DEBUG_VERBOSE, \
             printf("_ps_getargs: returning %d args to caller\n", *argc);)
    return PS_SUCCESS;
}

int
_ps_getexebase(int pid, char **exebase)
{
    int argc, i, ret;
    char *lastslash, **argv;

    ret = _ps_getargs(pid, &argc, &argv);
    if ( ret != PS_SUCCESS ) {
        return ret;
    }

    lastslash = strrchr(argv[0], '/');
    if (lastslash == NULL) {
        lastslash = argv[0];
    }
    else {
        lastslash++;
    }

    *exebase = strdup(lastslash);

    for (i = 0; i < argc; i++) {
        free(argv[i]);
    }
    free(argv);

    return PS_SUCCESS;
}

int
_ps_getexemd5(char md5out[PS_MD5LEN])
{
    char exe[256], buf[256];  /* @@ MAXPATHLEN? */
    int pid, commlen, ret;

    pid = getpid();

    sprintf(buf, "/proc/%d/exe", pid);
    commlen = readlink(buf, exe, 255);
    exe[commlen] = '\0';

    ret = _ps_md5_file(exe, md5out);

    return ret;
}

