/* $Id: psinternal.h,v 1.43 2010/09/16 20:38:46 ruiliu Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *        Copyright(C) 2002-2010, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *           National Center for Supercomputing Applications
 *             University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

#ifndef PSINTERNAL_H_INCLUDED
#define PSINTERNAL_H_INCLUDED 1

#if HAVE_CONFIG_H
# include <config.h>
#endif

/* Leaving this out for now (works on x86)
 * #define _POSIX_SOURCE 1
 */

#include <perfsuite.h>
#include <pserror.h>
#include <cpuid.h>

#if HAVE_SIGNAL_H
# include <signal.h>
#endif

#define PS_UNKNOWN_VALUE -1
#define PS_UNKNOWN_TICKS  0

#define PS_MALLOC(type, n) ((type *) _ps_malloc((n) * sizeof(type)))

#define PS_MD5LEN (16*2 + 1)

/*
 * Load module map, available through _ps_addrmap()
 */
struct ps_addrmap {
    char    *module;
    caddr_t start, end, offset;
};
typedef struct ps_addrmap ps_addrmap_t;

/*
 * Hash table data structures, available through _ps_hashtable_X()
 * functions.
 */

struct ps_hashentry {
    char *key;                    /* NULL-terminated user-supplied key */
    void *data;                   /* Pointer to the user's data        */
};
typedef struct ps_hashentry ps_hashentry_t;

struct ps_hashchain {
    int nentries;                 /* Count of current items stored       */
    int size;                     /* Current allocated space for entries */
    struct ps_hashentry *entries; /* Array of items                      */
};
typedef struct ps_hashchain ps_hashchain_t;

struct ps_hashtable {
    int size;                     /* # bins in the table */
    ps_hashchain_t **bin;         /* The bins            */
};
typedef struct ps_hashtable ps_hashtable_t;

#define _ps_hashtable_size(hashtable) (hashtable->size)
    

BEGIN_C_DECLS

/*
 * Processor detection routines (cpuid-XXX.c)
 */

extern int _ps_cpuid          PARAMS((ps_cpuid_t *cpuid));

/*
 * environment-related routines (environment.c)
 */

extern int _ps_addtoenv PARAMS((const char *var, const char *val,
                                const char *sep));
extern int _ps_findexe  PARAMS((char *exename, char *pathname, int maxlen));

/*
 * Fortran utilities (futils.c)
 */

extern void _ps_f2cstring PARAMS((const char *fstring, int n2copy,
                                  char *string, int maxlen));

extern void _ps_c2fstring PARAMS((const char *fstring, int n2copy,
                                  const char *string, int maxlen));

/*
 * Hash tables (hash.c)
 */

extern unsigned long _ps_hash   PARAMS((char *str));
extern int _ps_hashtable_create PARAMS((int size, ps_hashtable_t **table));
extern int _ps_hashtable_insert PARAMS((ps_hashtable_t *table, char *key,
                                        void *data));
extern int _ps_hashtable_lookup PARAMS((ps_hashtable_t *table, char *key,
                                        void **data));
extern int _ps_hashtable_stats  PARAMS((ps_hashtable_t *table, int *entries));


/*
 * Memory allocation wrappers (malloc.c)
 */

extern void * _ps_malloc PARAMS((size_t bytes));

/*
 * Process/thread ID  routines (process.c)
 */

extern int _ps_gettid PARAMS((pid_t *tid));

/*
 * Load map-related functions (procfs.c)
 */

extern int _ps_addrmap      PARAMS((pid_t pid,
                                    int *mapc, ps_addrmap_t ***mapv));
extern int _ps_addrmap_free PARAMS((int mapc, ps_addrmap_t **mapv));
extern int _ps_addrmap_exe  PARAMS((int mapc, ps_addrmap_t **mapv, int *index));

/*
 * Command line arguments (procfs.c)
 */

extern int _ps_getargs    PARAMS((pid_t pid, int *argc, char ***argv));
extern int _ps_getexebase PARAMS((int pid, char **exebase));
extern int _ps_getexemd5  PARAMS((char md5out[PS_MD5LEN]));

/*
 * Miscellaneous utility functions (utils.c)
 */

extern int _ps_tokenize PARAMS((char *str, char sep, char ***out));
extern int _ps_get_first_matching_value
                        PARAMS((const char * filename, const char * key,
                                char ** outstr));
extern int _ps_readuint PARAMS((const char * filename_pattern,
                                unsigned int * value));
extern int _ps_get_linux_kernel_version PARAMS((int * version));

/*
 * MD5 checksum generation (checksum.c)
 */

extern int _ps_md5_file PARAMS((char *fname, char md5out[PS_MD5LEN]));

#if _REENTRANT
/*
 * Pthread miscellaneous (procfs.c)
 */
extern void _ps_pthread_cleanup_fd PARAMS((void *p));
#endif


END_C_DECLS

#endif
