/* $Id: timers.c,v 1.50 2010/10/19 21:30:54 ruiliu Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *     Copyright(C) 2002-2006, 2010, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *           National Center for Supercomputing Applications
 *             University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * timers.c - timing related-routines for libperfsuite.
 *            The primary routines here are for access to "real-time"
 *            clock information (may not be precise, but as close
 *            as possible).
 *
 *            The main routines - ps_rtcinit() and ps_rtc() - are
 *            used frequently as the basis for calculating elapsed
 *            time, so they must be increasing if the library was
 *            configured properly for a given system.  They are
 *            advertised as returning the current value of the
 *            cycle counter.
 *
 * Author: Rick Kufrin
 * Date:   10/2002
 *
 * This file is part of PerfSuite.
 *
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if ! HAVE_LONG_LONG
#error "no support for long long data type on this system"
#endif

#if _REENTRANT
# include <pthread.h>
#endif

#if HAVE_STDIO_H
#include <stdio.h>
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>          /* for free() */
#endif

#if HAVE_UNISTD_H
# include <unistd.h>         /* for sysconf() */
#endif

#include <perfsuite.h>
#include "psdebug.h"

/**************************************************************
 * ps_rtcinit and ps_rtc - return the current cycle counter   *
 *                         or a "good approximation"          *
 *                                                            *      
 *  These are grouped according to the user configuration     *
 *  options and the system.                                   *
 **************************************************************/

/*
 * These two access methods require some initialization to
 * calculate a scale factor.
 */

#if RTC_MMTIMER || RTC_CLOCK_GET_TIME || RTC_GETTIMEOFDAY || defined(PPC32) \
    || defined(PPC64)
# if _REENTRANT
static pthread_mutex_t rtcscale_mutex = PTHREAD_MUTEX_INITIALIZER;
# endif /* _REENTRANT */
static double rtcscale;
#endif

#if RTC_MMTIMER || defined(PPC32) || defined(PPC64)
# if _REENTRANT
static inline int
rtcmutex_unlock(void)
{
    int ret;

    ret = pthread_mutex_unlock(&rtcscale_mutex);
    if ( ret != 0 ) {
        return PS_ESYS;
    }

    return ret;
}
# else
static inline int
rtcmutex_unlock(void)
{
    return PS_SUCCESS;
}
# endif /* _REENTRANT */
#endif  /* RTC_MMTIMER || defined(PPC32) || defined(PPC64) */


#if RTC_MMTIMER

/**************************************************************
 *  MMTIMER multimedia interface.  Useful on the SGI Altix    *
 *  as it is accurate, low-overhead, and maintains a clock    *
 *  that is synchronized across the nodes.                    *
 *                                                            *
 *  Thanks to Bernd Mohr of Research Centre Juelich for       *
 *  providing information about this and an example of use.   *
 **************************************************************/

# if HAVE_SYS_MMAN_H
#  include <sys/mman.h>  /* @@ Gotta check for all of these */
# endif
# if HAVE_SYS_IOCTL_H
#  include <sys/ioctl.h>
# endif
# if HAVE_FCNTL_H
#  include <fcntl.h>
# endif
# if HAVE_ERRNO_H
#  include <errno.h>
# endif
# if HAVE_LINUX_MMTIMER_H
#  include <linux/mmtimer.h>
# elif HAVE_SN_MMTIMER_H
#  include <sn/mmtimer.h>
# endif
# ifndef MMTIMER_FULLNAME
#  define MMTIMER_FULLNAME "/dev/mmtimer"
# endif

static volatile unsigned long *mmdev_timer_addr = NULL;
static unsigned long mmdev_timer_start = 0ul;

/* mmtimer variants */

int
ps_rtcinit(void)
{
    int fd, offset, ret;
    unsigned long femtosecs_per_tick = 0;
    double mhz;

#if DEBUG
    ps_setdebuglevel();
#endif
    PS_DEBUG(PS_DEBUG_INFO, puts("Using mmtimer for RTC.");)

# if _REENTRANT
    pthread_cleanup_push((void *) pthread_mutex_unlock, &rtcscale_mutex);
    ret = pthread_mutex_lock(&rtcscale_mutex);
    if ( ret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 perror("ps_rtcinit: error locking mutex");)
        return PS_ESYS;
    }
# endif

    if ( mmdev_timer_addr == NULL ) {
        fd = open(MMTIMER_FULLNAME, O_RDONLY);
        if ( fd == -1 ) {
            PS_DEBUG(PS_DEBUG_FATAL, perror("ps_rtcinit: cannot open mmtimer");)
            rtcmutex_unlock();
            return PS_ESYS;
        }

        offset = ioctl(fd, MMTIMER_GETOFFSET, 0);
        if ( offset == -ENOSYS ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     perror("ps_rtcinit: ioctl failure (offset)");)
            close(fd);
            rtcmutex_unlock();
            return PS_ESYS;
        }

        mmdev_timer_addr = mmap(0, getpagesize(), PROT_READ, MAP_SHARED, fd, 0);
        if ( mmdev_timer_addr == NULL ) {
            PS_DEBUG(PS_DEBUG_FATAL, perror("ps_rtcinit: mmap failure");)
            close(fd);
            rtcmutex_unlock();
            return PS_ESYS;
        }

        mmdev_timer_addr += offset;
        mmdev_timer_start = *mmdev_timer_addr;

        /* @@ check error return - is it -1? */
        ret = ioctl(fd, MMTIMER_GETRES, &femtosecs_per_tick);
        if ( ret == -1 ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     perror("ps_rtcinit: ioctl failure (resolution)");)
            close(fd);
            rtcmutex_unlock();
            return PS_ESYS;
        }

        close(fd);

        ret = ps_cpuspeed(&mhz);
        if ( ret != PS_SUCCESS ) {
            rtcmutex_unlock();
            return ret;
        }

        rtcscale = (mhz * 1.0e-9) * femtosecs_per_tick;

        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Clicks per tick (rtcscale) = %.20f, mmdev_timer_start = %ld\n", \
                        rtcscale, mmdev_timer_start);)
    }

# if _REENTRANT
    pthread_cleanup_pop(1);
# endif

    return PS_SUCCESS;
}

int
ps_rtc(unsigned long long *rtcval)
{
    PS_DEBUG(PS_DEBUG_VERBOSE, printf("mmtimer val: %ld\n", *mmdev_timer_addr);)

    *rtcval = ((long long) rtcscale * (*mmdev_timer_addr - mmdev_timer_start));

    return PS_SUCCESS;
}

#elif RTC_CLOCK_GETTIME

/**************************************************************
 **************************************************************/

# if ! HAVE_CLOCK_GETTIME
#  error "clock_gettime() required for ps_rtc as configured"
# endif

#if HAVE_TIME_H
# include <time.h>
#endif

/* clock_gettime() variants */
int
ps_rtcinit(void)
{
    int ret = PS_SUCCESS;

#if DEBUG
    ps_setdebuglevel();
#endif
    PS_DEBUG(PS_DEBUG_INFO, puts("Using clock_gettime() for RTC.");)

# if _REENTRANT
    pthread_cleanup_push((void *) pthread_mutex_unlock, &rtcscale_mutex);
    ret = pthread_mutex_lock(&rtcscale_mutex);
    if ( ret != 0 ) {
        return PS_ESYS;
    }
# endif

    ret = ps_cpuspeed(&rtcscale);
    rtcscale /= 1000.0;

# if _REENTRANT
    pthread_cleanup_pop(1);
# endif

    return ret;
}

int
ps_rtc(unsigned long long *rtcval)
{
    struct timespec tv;

    if ( clock_gettime(CLOCK_REALTIME, &tv) != 0 ) {
        PS_DEBUG(PS_DEBUG_WARNING, \
                 fputs("Warning: clock_gettime() error\n", stderr);)
        return PS_ESYS;
    }

    /* 
     * Have had link time problems with Intel 7.1 compiler if this
     * is cast to unsigned long long.
     */

    *rtcval = ((long long) ((((double)tv.tv_sec * 1000000000L) + tv.tv_nsec) * rtcscale));

    return PS_SUCCESS;
}

#elif RTC_GETTIMEOFDAY

/**************************************************************
 **************************************************************/

# if ! HAVE_GETTIMEOFDAY
#  error "gettimeofday() required for ps_rtc as configured"
# endif

#if HAVE_SYS_TIME_H
# include <sys/time.h>
#endif

/* gettimeofday() variants */
int
ps_rtcinit(void)
{
    int ret = PS_SUCCESS;

#if DEBUG
    ps_setdebuglevel();
#endif
    PS_DEBUG(PS_DEBUG_INFO, puts("Using gettimeofday() for RTC.");)

# if _REENTRANT
    pthread_cleanup_push((void *) pthread_mutex_unlock, &rtcscale_mutex);
    ret = pthread_mutex_lock(&rtcscale_mutex);
    if ( ret != 0 ) {
        return PS_ESYS;
    }
# endif

    ret = ps_cpuspeed(&rtcscale);

# if _REENTRANT
    pthread_cleanup_pop(1);
# endif

    return ret;
}

int
ps_rtc(unsigned long long *rtcval)
{
    struct timeval tv;

    if ( gettimeofday(&tv, NULL) != 0 ) {
        PS_DEBUG(PS_DEBUG_WARNING, \
                 fputs("Warning: gettimeofday() error\n", stderr);)
        return PS_ESYS;
    }

    /* 
     * Have had link time problems with Intel 7.1 compiler if this
     * is cast to unsigned long long.
     */

    /*
     * @@ This should be adjusted to count from a reference value obtained
     * at rtc_init() time.  This would make cross-process comparisons
     * meaningless, but that functionality is not intended.
     *
     * If my calculations are correct, this version will mess up when
     * the number of ticks since the UNIX epoch (1/1/1970) exceeds the
     * available space in a signed 64-bit integer.  Since the number of
     * ticks vary on a per-CPU basis, it's not uniform.
     */

    *rtcval = ((long long) ((((double)tv.tv_sec * 1000000L) + tv.tv_usec) * rtcscale));

    return PS_SUCCESS;
}

#elif defined(IA32) || defined(IA64)

/**************************************************************
 **************************************************************/

/* Assembly-level access to the timestamp counter */

#include "timers.h"

int
ps_rtcinit(void)
{
#if DEBUG
    ps_setdebuglevel();
#endif
    PS_DEBUG(PS_DEBUG_INFO, puts("Using timestamp counter for RTC.");)

    return PS_SUCCESS;
}

int
ps_rtc(unsigned long long *rtcval)
{
    unsigned long long tmp;

    rdtscll(tmp);
    *rtcval = tmp;

    return PS_SUCCESS;
}

#elif defined(AIX5)

/**************************************************************
 **************************************************************/

# include <sys/systemcfg.h>

/* AIX variants */
int
ps_rtcinit(void)
{
#if DEBUG
    ps_setdebuglevel();
#endif
    PS_DEBUG(PS_DEBUG_INFO, puts("Using read_real_time() counter for RTC.");)

    return PS_SUCCESS;
}

int
ps_rtc(unsigned long long *rtcval)
{
    /* @@ not tested at all! */
    timebasestruct_t t;
    int aixret;

    aixret = read_real_time(&t, TIMEBASE_SZ); 
    *rtcval = ((unsigned long long) t.high * 1000000000L) + t.low;

    return PS_SUCCESS;
}

#elif defined(PPC32) || defined(PPC64)

/**************************************************************
 *  Use the PowerPC instruction "mftb".                       *
 **************************************************************/

/*
  To get the update frequency of the timebase register:

  for PPC32, read from device tree.
  for PPC64,  since kernel 2.6.5, read from /proc/cpuinfo;
             before kernel 2.6.5, read from device tree.
 */

#if ! defined(CPUINFOFILE)
  #define CPUINFOFILE "/proc/cpuinfo"
#endif
#define TIMEBASE_KEY "timebase"
#define TIMEBASE_FREQ_FILE_PATTERN \
    "/proc/device-tree/cpus/PowerPC,POWER*@0/timebase-frequency"

int
ps_rtcinit(void)
{
    int ret;
    double mhz;
    char *    timebase_freq_str;
    unsigned int timebase_freq;
    int use_device_tree = 0;
    int kern_ver;

#if DEBUG
    ps_setdebuglevel();
#endif
    PS_DEBUG(PS_DEBUG_INFO, puts("Using timebase register (mftb) for RTC.");)

    ret = ps_cpuspeed(&mhz);

# if _REENTRANT
    pthread_cleanup_push((void *) pthread_mutex_unlock, &rtcscale_mutex);
    ret = pthread_mutex_lock(&rtcscale_mutex);
    if ( ret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, perror("ps_rtcinit: error locking mutex");)
        return PS_ESYS;
    }
# endif

#if defined(PPC32)
    use_device_tree = 1;
#elif defined(PPC64)
    ret = _ps_get_linux_kernel_version(&kern_ver);
    if (ret != PS_SUCCESS) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("ps_rtcinit: error reading kernel version", stderr);)
        rtcmutex_unlock();
        return ret;
    }
    if (kern_ver < 0x02060500) {
        /* Use device tree to get timebase frequency
           if kernel version < 2.6.5 and it is PPC64. */
        use_device_tree = 1;
    }
#endif
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("use_device_tree is %d.\n", use_device_tree);)

    if (! use_device_tree) {
        /* Get the frequency of the timebase register
           by parsing the "^timebase" line in /proc/cpuinfo.
           The value is in Hz. An example line is:
           "timebase\t: 512000000". */

        ret = _ps_get_first_matching_value(CPUINFOFILE, TIMEBASE_KEY,
                                           &timebase_freq_str);
        if (ret != PS_SUCCESS) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("ps_rtcinit: error getting timebase frequency in /proc/cpuinfo", stderr);)
            rtcmutex_unlock();
            return ret;
        }
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Timebase frequency string in /proc/info is: '%s'.\n", timebase_freq_str);)

        timebase_freq = atoi(timebase_freq_str);
        free(timebase_freq_str);
    } else {
        /* Use device tree. */
        ret = _ps_readuint(TIMEBASE_FREQ_FILE_PATTERN, &timebase_freq);
        if (ret != PS_SUCCESS) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("ps_rtcinit: error reading timebase frequency from device tree", stderr);)
            rtcmutex_unlock();
            return ret;
        }
    }
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("Found timebase frequency: %u.\n", timebase_freq);)

    /* Check whether timebase_freq is 0 before attempting to divide by it. */
    if (timebase_freq == 0) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("ps_rtcinit: timebase frequency is 0.", stderr);)
        rtcmutex_unlock();
        return PS_EINVAL;
    }
    rtcscale = mhz * 1000000.0 / timebase_freq;

# if _REENTRANT
    pthread_cleanup_pop(1);
# endif

    return PS_SUCCESS;
}

int
ps_rtc(unsigned long long *rtcval)
{
    /* For now, use "unsigned long long", later could use a better
       way to ensure it's uint64_t. */
    unsigned long long x;

#if defined(PPC64)
    asm volatile ("mftb %[x]" : [x] "=r" (x) );
#else /* PPC32 */
    unsigned int upper, lower, new_upper;
    do
    {
        asm volatile (
            "mftbu   %[Upper];"
            "mftb    %[Lower];"
            "mftbu   %[NewUpper];"

            : [Upper] "=r" (upper), [Lower] "=r" (lower), [NewUpper] "=r" (new_upper)
            :
            : "cr0"
         );
    } while (upper != new_upper);
    x = (((unsigned long long) upper) << 32) | ((unsigned long long) lower);
#endif

    *rtcval = x * rtcscale;

    return PS_SUCCESS;
}

#else

/**************************************************************
 **************************************************************/

/* Everything else */

int
ps_rtcinit(void)
{
#if DEBUG
    ps_setdebuglevel();
#endif
    PS_DEBUG(PS_DEBUG_WARNING, fputs("No RTC configured.", stderr);)
    return PS_ENOTIMPLEMENTED;
}

int
ps_rtc(unsigned long long *rtcval)
{
    *rtcval = PS_UNKNOWN_TICKS;
    return PS_EUNKNOWNCPU;
}
#endif /* End of ps_rtcinit() and ps_rtc() functions */

/* Fortran wrappers */

#define PSF_RTCINIT_F77 F77_FUNC_(psf_rtcinit,PSF_INIT)
#define PSF_RTC_F77     F77_FUNC_(psf_rtc,PSF_RTC)

void
PSF_RTCINIT_F77(int *ierr)
{
    *ierr = ps_rtcinit();

    return;
}

void
PSF_RTC_F77(unsigned long long *rtcval, int *ierr)
{
    *ierr = ps_rtc(rtcval);

    return;
}

/*****************************************************************
 *  End of real-time clock functions.                            *
 *****************************************************************/

/*
 * Not grouping this among the procfs routines, because
 * we may use a different method of getting CPU time
 * of the process in the future (but that's what we're
 * using right now).
 *
 */

int
ps_cpuusage(pid_t pid, ps_time_t *utime, ps_time_t *stime)
{
    ps_procstat_t p;
    int ret;
    unsigned long resolution, msec_conv;

    resolution = sysconf(_SC_CLK_TCK);
    msec_conv = 1000.0/resolution;

    ret = ps_procstat(pid, &p);
    if ( ret != PS_SUCCESS ) {
        return ret;
    }

    utime->secs = p.utime/resolution;
    utime->msecs = (p.utime - utime->secs) * msec_conv;
    stime->secs = p.stime/resolution;
    stime->msecs = (p.stime - stime->secs) * msec_conv;

    return PS_SUCCESS;
}

#define PSF_CPUUSAGE_F77 F77_FUNC_(psf_cpuusage,PSF_CPUUSAGE)
void
PSF_CPUUSAGE_F77(pid_t *pid,
		 int *utime_secs, int *utime_msecs,
		 int *stime_secs, int *stime_msecs, int *ierr)
{
    ps_time_t utime, stime;

    *ierr = ps_cpuusage(*pid, &utime, &stime);

    if ( *ierr == 0 ) {
        *utime_secs = utime.secs;
        *utime_msecs = utime.msecs;
        *stime_secs = stime.secs;
        *stime_msecs = stime.msecs;
    }

    return;
}
