/* $Id: utils.c,v 1.6 2010/09/16 20:38:46 ruiliu Exp $ */

/****************************************************************************
 *
 *                      University of Illinois/NCSA
 *                          Open Source License
 * 
 *       Copyright(C) 2005-2006, 2010, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                             Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * utils.c - miscellaneous utility functions
 *
 * Author: Rick Kufrin
 * Date:   October 26, 2005
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if HAVE_STDIO_H
#include <stdio.h>
#endif

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#if HAVE_GLOB_H
# include <glob.h>
#endif

#if HAVE_ERRNO_H
# include <errno.h>
#endif

#include <perfsuite.h>
#include <psdebug.h>

#include "psinternal.h"

/*
 * _ps_tokenize: split a string into individual tokens separated
 *               by the character supplied in the input parameter "sep"
 *               The output is a NULL-terminated array of character
 *               strings returned in "out".  It is the caller's
 *               responsibility to free these strings as well as
 *               the array.
 *
 *               Note that an empty string "" can be returned as a token.
 *
 */

int
_ps_tokenize(char *str, char sep, char ***out)
{
    int i, j, count;
    size_t len;
    char *ptr, sepstr[2];

    PS_DEBUG(PS_DEBUG_VERBOSE, \
             printf("_ps_tokenize: string \"%s\", separator \"%c\"\n", \
                    str, sep);)

    /* First find how many instances of the separator there are
     * in the string.
     */

    for ( ptr = str, count = 0; *ptr != '\0'; ptr++ ) {
        if ( *ptr == sep ) {
            count++;
        }
    }

    /*
     * Assume there will be one more token returned than the number
     * of separator characters.  Allocate pointers for these strings
     * plus an extra one for the terminating NULL pointer.
     */
    count++;
    *out = (char **) malloc( (count + 1) * sizeof (char *));
    if ( *out == NULL ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("malloc failure in _ps_tokenize\n", stderr);)
        return PS_ENOMEM;
    }

    for ( i = 0; i <= count; i++ ) {
        (*out)[i] = NULL;
    }

    sprintf(sepstr, "%c", sep);
    for ( ptr = str, i = 0; i < count; i++ ) {
        len = strcspn(ptr, sepstr);
        (*out)[i] = (char *) malloc(len+1);
        if ( (*out)[i] == NULL ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("malloc failure in _ps_tokenize\n", stderr);)
            for ( j = 0; j < i; j++ ) {
                free((*out)[j]);
            }
            free(*out);
            return PS_ENOMEM;
        }
        memset((*out)[i], 0x0, len+1);
        memcpy((*out)[i], ptr, len);
        PS_DEBUG(PS_DEBUG_VERBOSE,printf("Token %d = \"%s\"\n", i, (*out)[i]);)
        ptr += len;
        if ( *ptr == sep ) ptr++;
    }

    return PS_SUCCESS;
}

/*
 * _ps_get_first_matching_value:
 *         Given a filename, and a <key> string, find the first line
 *         containing "<key>[\s*]:<value>", and set the value in the
 *         output string.  Memory is dynamically allocated. It is the
 *         caller's responsibility to free it using free().
 *
 *         Note that the line buffer size is limited to LINEBUFSIZE,
 *         so if the line is longer than that, the first matching value
 *         might not be found.
 */

#define LINEBUFSIZE 256

int
_ps_get_first_matching_value
(const char * filename, const char * key, char ** outstr)
{
    FILE *fp;
    char buf[LINEBUFSIZE];
    char valbuf[LINEBUFSIZE];
    const int FORMAT_STR_LEN = 40;
    char      format_str[FORMAT_STR_LEN];
    int found = 0;
    int ret;

#if _REENTRANT
    pthread_cleanup_push((void *)_ps_pthread_cleanup_fd, &fp);
#endif

    /* Initialize to NULL. */
    *outstr = NULL;

    fp = fopen(filename,"r");
    if ( fp == NULL ) {
        if ( errno == ENOENT ) {
            errno = ESRCH;
        }
        return PS_ESYS;
    }

    /* Use a width before [^\n] to defend against a buffer overflow. */
    snprintf(format_str, FORMAT_STR_LEN, "%s : %%%d[^\n]", key, LINEBUFSIZE-1);

    while ( fgets(buf, LINEBUFSIZE, fp) != NULL ) {
        if ( buf[0] != '\n' ) {
            if (sscanf(buf, format_str, valbuf) == 1) {
                *outstr = strdup(valbuf);
                found = 1;
                break;
            }
        }
    }

    fclose(fp);

#if _REENTRANT
    pthread_cleanup_pop(0);
#endif

    if (found == 1) {
        ret = PS_SUCCESS;
    } else {
        ret = PS_ENOTFOUND;
    }

    return ret;
}

/*
 *  getline_or_bytes:
 *      Get the first line (size limit is LINEBUFSIZE),
 *      or the first specified number of bytes.
 *
 *      If *numbytes == 0, read the first line using fgets();
 *      else if *numbytes > 0,  read the first *numbytes, and set
 *           the number of units (in *numbytes bytes) read in numbytes.
 *      else return PS_EINVAL.
 */
static int
getline_or_bytes(const char * filename, char * outstr, int * numbytes)
{
    FILE *fp;
    int ret;

    if ( (*numbytes < 0) || (*numbytes > LINEBUFSIZE) ) {
        return PS_EINVAL;
    }

#if _REENTRANT
    pthread_cleanup_push((void *)_ps_pthread_cleanup_fd, &fp);
#endif

    /* Initialize it to '\0'. */
    *outstr = 0;

    fp = fopen(filename,"r");
    if ( fp == NULL ) {
        if ( errno == ENOENT ) {
            errno = ESRCH;
        }
        return PS_ESYS;
    }

    if (*numbytes == 0) {
        if (fgets(outstr, LINEBUFSIZE, fp) == NULL) {
            fclose(fp);
            return PS_ENOTFOUND;
        }
    } else if (*numbytes > 0) {
        ret = fread(outstr, *numbytes, 1, fp);
        if (ret == 0 ) {
            fclose(fp);
            return PS_ENOTFOUND;
        }
        *numbytes = ret;
    }

    fclose(fp);

#if _REENTRANT
    pthread_cleanup_pop(0);
#endif

    return PS_SUCCESS;
}

/*
 * _ps_readuint: read an unsigned integer from a given file name
 *               pattern.  The pattern will be expanded using glob().
 *               This function uses getline_or_bytes().
 */
int
_ps_readuint(const char * filename_pattern, unsigned int * value)
{
    int ret;
    int UNIT = sizeof(unsigned int);
    int numbytes;
    glob_t globbuf;

    *value = 0;

#if ! HAVE_GLOB_H
    return PS_ENOTINSTALLED;
#endif

    ret = glob(filename_pattern, 0, NULL, &globbuf);
    if (ret != 0) {
        return PS_ENOTFOUND;
    }
    PS_DEBUG(PS_DEBUG_VERBOSE, \
             printf("glob gave the file name: '%s'.\n", globbuf.gl_pathv[0]);)

    numbytes = UNIT;
    ret = getline_or_bytes(globbuf.gl_pathv[0], (char *)value, &numbytes);
    globfree(&globbuf);

    if (ret != PS_SUCCESS) {
        return ret;
    }
    PS_DEBUG(PS_DEBUG_VERBOSE, \
             printf("getline_or_bytes got %d bytes; the first 4: %08x.\n",
                    numbytes*UNIT, *value);)

    return PS_SUCCESS;
}


/*
 * _ps_get_linux_kernel_version: Get linux kernel version from
 *                               /proc/sys/kernel/osrelease and set it
 *                               in the integer pointer passed in.
 */

#define KERNEL_VERSION_FILE "/proc/sys/kernel/osrelease"

int
_ps_get_linux_kernel_version(int * version)
{
    char buf[LINEBUFSIZE];
    int numbytes = 0;
    int major, minor, rev;
    int ret;

    ret = getline_or_bytes(KERNEL_VERSION_FILE, buf, &numbytes);
    if (ret != PS_SUCCESS) {
        return ret;
    }

    ret = sscanf(buf, "%d.%d.%d", &major, &minor, &rev);
    if (ret != 3) {
        /* Need to match 3 entries, otherwise there's an error. */
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "error parsing linux kernel version: need to match 3 entries, only matched %d.\n", ret);)
        return PS_ENOTFOUND;
    }
    *version = ((major & 0xFF) << 24) + ((minor & 0xFF) << 16) + (rev << 8);
    /* No need for a "\n" after the "%s", since buf has it at the end. */
    PS_DEBUG(PS_DEBUG_INFO, printf("Linux kernel version line is: %s", buf);)
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("Linux kernel version parsed is: 0x%08X.\n", *version);)
    return PS_SUCCESS;
}
