/* $Id: dyn-conf.c,v 1.3 2013/04/30 16:34:06 ruiliu Exp $ */

/****************************************************************************
 *
 *                      University of Illinois/NCSA
 *                          Open Source License
 * 
 *            Copyright(C) 2013, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                             Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if HAVE_STDIO_H
# include <stdio.h>
#endif

#if STDC_HEADERS
# include <string.h>
#endif

#include <psinternal.h>
#include <psdebug.h>

#if (PPC32 || PPC64)
# include "cpuid-powerpc.h"
#endif

char * PS_HWPC_DEF_CONFIGFILE = "uninitialized";

/* Dynamically set the default configuration file at run time,
 * instead of using the file determined at configure time.
 */

int set_def_config() {
    int ret1 = 0;
    PS_HWPC_DEF_CONFIGFILE = "null.xml";
    ps_cpuid_t cpuinfo;

#if HAVE_PERFMON
    PS_HWPC_DEF_CONFIGFILE = "pfm_ipc.xml";
#elif (! HAVE_PAPI)
    /* no PAPI, no Perfmon */
    PS_HWPC_DEF_CONFIGFILE = "profil.xml";
#else /* HAVE_PAPI */
    /* Use the PAPI_TOT_INS and PAPI_TOT_CYC as the default events,
     * as they are available on most CPU types. */
    PS_HWPC_DEF_CONFIGFILE = "papi2_any-null.xml";

    int ret = _ps_cpuid (&cpuinfo);
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("arch=%d, family=%u, model=%u, stepping=%u, extra=%d.\n", \
             cpuinfo.arch, \
             cpuinfo.cpuid_family, cpuinfo.cpuid_model, \
             cpuinfo.cpuid_stepping, cpuinfo.extra);)

    /* Determine whether it's IA32, IA64 (Itanium), or PowerPC.
     * If IA32, call _ps_cpuid () to find out the values of the cpuid family
     *     and model, and set the def file;
     * else if IA64, use the /proc/cpuinfo  "family" line to determine whether
     *     it's an Itanium or Itanium2 CPU, and set the def file;
     * else if PowerPC, use the - "cpuinfo.extra" field - to determine whether
     *     it's a Power5, Power6, or Power7 CPU, and set the def file.
     */

    if (cpuinfo.arch == IA32_ARCH) {
        if (cpuinfo.vendor == INTEL) {
            switch (cpuinfo.cpuid_family) {
            case 5:
                PS_HWPC_DEF_CONFIGFILE = "papi3_p6.xml";
                break;
            case 6:
                if (cpuinfo.cpuid_model < 14) {
                    PS_HWPC_DEF_CONFIGFILE = "papi3_p6.xml";
                } else {
                    switch (cpuinfo.cpuid_model) {
                    case 14:
                    case 15:
                    case 23:
                    case 29:
                        PS_HWPC_DEF_CONFIGFILE = "papi3_core.xml";
                        break;
                    case 26:
                    case 30:
                    case 31:
                    case 46:
                        PS_HWPC_DEF_CONFIGFILE = "papi3_nhm.xml";
                        break;
                    case 37:
                    case 44:
                    case 47:
                        PS_HWPC_DEF_CONFIGFILE = "papi3_wsm.xml";
                        break;
                    case 42:
                    case 45:
                        PS_HWPC_DEF_CONFIGFILE = "papi_sandybridge.xml";
                        break;
                    }
                }
                break;
            case 15:
                PS_HWPC_DEF_CONFIGFILE = "papi3_p4.xml";
                break;
            default:
                break;
            }
        } else if (cpuinfo.vendor == AMD) {
            switch (cpuinfo.cpuid_family) {
            case 6:
                PS_HWPC_DEF_CONFIGFILE = "papi3_k7.xml";
                break;
            case 15:
                PS_HWPC_DEF_CONFIGFILE = "papi3_k8.xml";
                break;
            case 16:
            case 17:
                PS_HWPC_DEF_CONFIGFILE = "papi3_k10.xml";
                break;
            case 21:
                PS_HWPC_DEF_CONFIGFILE = "papi_interlagos.xml";
                break;
            default:
                break;
            }
        }
    } else if (cpuinfo.arch == IA64_ARCH) {
        int ret;
        char * cpuinfo_family_line;
        ret = _ps_get_first_matching_value("/proc/cpuinfo", "family",
                                           &cpuinfo_family_line);
        if (ret != PS_SUCCESS) {
            return ret;
        }

        if ( (!strcmp (cpuinfo_family_line, "Itanium")) ||
             (!strcmp (cpuinfo_family_line, "7")) ) {
            PS_HWPC_DEF_CONFIGFILE = "papi3_itanium.xml";
        } else if ( (!strcmp (cpuinfo_family_line, "Itanium 2")) ||
             (!strcmp (cpuinfo_family_line, "31")) ||
             (!strcmp (cpuinfo_family_line, "32")) ) {
            PS_HWPC_DEF_CONFIGFILE = "papi3_itanium2.xml";
        } 
    } else if (cpuinfo.arch == POWER_PC_ARCH) {

#if (PPC32 || PPC64)
        int pvr_ver = cpuinfo.extra;

        switch (pvr_ver) {

        case PVR_POWER6:
            PS_HWPC_DEF_CONFIGFILE = "papi_power6.xml";
            break;
        case PVR_POWER7:
            PS_HWPC_DEF_CONFIGFILE = "papi_power7.xml";
            break;
        case PVR_POWER4:
        case PVR_POWER4P:
        case PVR_POWER5:
        case PVR_POWER5P:
            PS_HWPC_DEF_CONFIGFILE = "papi_power5.xml";
        default:
            break;
        }
#endif

    } else {
        PS_HWPC_DEF_CONFIGFILE = "null.xml";
    }

# endif

    return PS_SUCCESS;
}
