/* $Id: hwpc-itimer.c,v 1.18 2013/05/07 21:51:49 ruiliu Exp $ */

/****************************************************************************
 *
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *       Copyright(C) 2004-2006, 2013, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * itimer: interval timer layer for libpshwpc.
 *
 * Author: Rick Kufrin
 * Date:   12/27/2004
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#if HAVE_SYS_TIME_H
# include <sys/time.h>
#endif

#if HAVE_SIGNAL_H
# include <signal.h>
#endif

#include <perfsuite.h>
#include "hwpc.h"
#include "hwpc-itimer.h"
#include "profile.h"

static struct {
    char *name;
    int  code;
    int  sig;
} itimer_types[3] = {
    {"ITIMER_REAL",    ITIMER_REAL,    SIGALRM},
    {"ITIMER_VIRTUAL", ITIMER_VIRTUAL, SIGVTALRM},
    {"ITIMER_PROF",    ITIMER_PROF,    SIGPROF}
};

static int timer=-1;
static struct itimerval itval;

int
get_num_counters_itimer()
{
    return 1;
}

int
get_max_counters_itimer(ps_hwpc_globals_t *globals)
{
    return get_num_counters_itimer();
}

int
init_counters_itimer(ps_hwpc_globals_t *globals)
{
    ps_hwpc_eventarray_t *events = &globals->events;
    int i, ret;

    for ( i = 0; i < 3; i++ ) {
        if ( strcmp(itimer_types[i].name, events->events[0].val) == 0 ) {
            timer = i;
            PS_DEBUG(PS_DEBUG_INFO,printf("Found %s\n", itimer_types[i].name);)
            break;
        }
    }

    if ( timer == -1 ) {
        return PS_HWPC_ENOSUCHEVENT;
    }

    globals->version = ITIMER_VERSION;

    events->events[0].derived = 0;

    switch ( itimer_types[timer].code ) {
    case ITIMER_REAL:
    case ITIMER_PROF:
        globals->domain = ALL;
        break;
    case ITIMER_VIRTUAL:
        globals->domain = USER;
        break;
    default:
        globals->domain = UNKNOWN;
        PS_DEBUG(PS_DEBUG_WARNING, \
                 fprintf(stderr, "Unknown ITIMER type (%d)\n", \
                        itimer_types[timer].code);)
        break;
    }

    /* The threshold for itimer class is provided in microseconds */
    itval.it_interval.tv_usec = itval.it_value.tv_usec =
        globals->threshold % 1000000;
    itval.it_interval.tv_sec =  itval.it_value.tv_sec =
        globals->threshold / 1000000;

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("itimer interval: sec = %ld, usec = %ld\n", \
                    itval.it_interval.tv_sec, itval.it_interval.tv_usec);)

    /* Let the signal-handling support know about global data */
    ret = _ps_hwpc_setprofglobals(globals, NULL);
    if ( ret != PS_SUCCESS ) {
        return ret;
    }

    return PS_SUCCESS;
}

int
start_counters_itimer(ps_hwpc_globals_t *globals)
{
    return PS_SUCCESS;
}

int
read_counters_itimer(ps_hwpc_value_t *values)
{
    return PS_SUCCESS;
}

int
stop_counters_itimer(ps_hwpc_value_t *values)
{
    return PS_SUCCESS;
}

int
destroy_counters_itimer()
{
    return PS_SUCCESS;
}

int
init_profiling_itimer(ps_hwpc_globals_t *globals, ps_hwpc_sample_t **samples)
{
    int ret;
    struct sigaction act;

    memset(&act, 0, sizeof(struct sigaction));
    act.sa_sigaction = _ps_hwpc_profhandler;
    act.sa_flags     = (SA_SIGINFO | SA_RESTART);

    ret = sigaction(itimer_types[timer].sig, &act, NULL);
    if ( ret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error in init_profiling_itimer: %s\n", \
                         strerror(ret));)
        return PS_ESYS;
    }

    /* Give the signal-handling layer the sampling buffers to use */
    ret = _ps_hwpc_setprofbuf(samples);
    if ( ret != PS_SUCCESS ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error setting sampling buffers: %s\n", \
                         ps_strerror(ret));)
        return ret;
    }

    return PS_SUCCESS;
}

int
start_profiling_itimer(ps_hwpc_globals_t *globals, ps_hwpc_sample_t **samples)
{
    int ret;

    ret = setitimer(itimer_types[timer].code, &itval, 0);
    if ( ret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error in start_profiling_itimer: %s\n", \
                         strerror(ret));)
        return PS_ESYS;
    }

    return PS_SUCCESS;
}

/*
 * Handlers are an all-or-nothing proposition for the process.
 * If we are in threaded execution, we can't shut them off in
 * one thread without affecting others, so we don't.
 *
 * @@ This should return an appropriate error code.
 */

int
suspend_profiling_itimer(ps_hwpc_globals_t *globals, ps_hwpc_sample_t **samples)
{
#if ! _REENTRANT
    struct itimerval itval;
    int ret;

    itval.it_interval.tv_usec = itval.it_value.tv_usec =
        itval.it_interval.tv_sec =  itval.it_value.tv_sec = 0;

    ret = setitimer(itimer_types[timer].code, &itval, 0);
    if ( ret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr,"Error in suspend/stop_profiling_itimer: %s\n",\
                         strerror(ret));)
        return PS_ESYS;
    }
#endif

    return PS_SUCCESS;
}

int
stop_profiling_itimer(ps_hwpc_globals_t *globals, ps_hwpc_sample_t **samples)
{
#if ! _REENTRANT
    struct sigaction act;
    int ret;

    ret = suspend_profiling_itimer(globals, samples);
    if ( ret != PS_SUCCESS ) {
        return ret;
    }

    /*PS_DEBUG(PS_DEBUG_INFO, printf("Itimer samples: %ld\n", nsamples);)*/

    memset(&act, 0, sizeof(struct sigaction));
    act.sa_handler = SIG_IGN;

    ret = sigaction(itimer_types[timer].sig, &act, NULL);
    if ( ret != 0 ) {
        return PS_ESYS;
    }
#endif

    return PS_SUCCESS;
}
