/* $Id: hwpc-papi.c,v 1.103 2013/12/18 15:58:45 ruiliu Exp $ */

/****************************************************************************
 *
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *       Copyright(C) 2002-2006, 2013, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * PAPI-specific hardware counting management.
 *
 * The only responsibilities of this layer are to configure, start, 
 * stop the counters, return their current values, and terminate
 * counting support.
 *
 * The generic software handles it from there, including writing out the
 * final data.
 *
 * @@ Lots of work remaining to clean up this source.
 *
 * Author: Rick Kufrin
 * Date:   10/2002
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if _REENTRANT
# if HAVE_PTHREAD
#  include <pthread.h>
# else
#  error "_REENTRANT defined but Pthreads support not available"
# endif
#endif

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

# include <papi.h>
# include <perfsuite.h>
# include <pshwpc.h>
# include <psinternal.h>
# include <pshwpcpapi.h>
# include "hwpc.h"
# include "hwpc-papi.h"
# include "hwpc-xml.h"

/* @@ Must be guarded in the thread case? */
static int *event_codes = NULL;
static int numEvents = 0;

#if _REENTRANT
/*
 * TSD key and destructor.
 *
 * We only use a single event set, but there should be a different 
 * instance for each thread.  Thread-specific data holds eventsets
 * and profiling buffers as well as the most recent error code 
 * encountered. 
 *
 * Note that the TSD is not available until after ps_hwpc_start()
 * has been called which in turn calls start_counters_papi() or
 * init_profiling_papi().
 */

static pthread_key_t tsdkey;
#else
static int *eventset;
static PAPI_sprofil_t *sprofils;
#endif /* _REENTRANT */


/**********************************************************************
         Prototypes for internal functions used in this module.
         The functions are defined near the end of this file.
 **********************************************************************/

BEGIN_C_DECLS

static int do_dryrun       PARAMS((int *event_codes, int numEvents,
                                  int *dompx));
static int PAPI_to_ps_hwpc PARAMS((ps_hwpc_globals_t *globals, int eventset));

#if _REENTRANT
static void free_tsd       PARAMS((void *data));
#endif

END_C_DECLS

/*
 * This is a PAPI setup routine with operations common to counting
 * and profiling.  It's called from start_counters_papi() and
 * init_profiling_papi().
 */

static inline int
setup_counters_internal(struct ps_hwpc_globals *globals)
{
    int papiret;
    int compIndex = 0;
#if _REENTRANT
    hwpc_papi_tsd_t *tsd;
    int *eventset;
#endif /* _REENTRANT */
#if DEBUG
    PAPI_option_t options;
#endif

    PS_DEBUG(PS_DEBUG_INFO, puts("Entering setup_counters_internal");)

    /* Has the global initializer been called? */
    if ( event_codes == NULL ) {
        PS_DEBUG(PS_DEBUG_WARNING, \
                 fputs("setup_counters_internal: event_codes is NULL\n", \
                       stderr);)
        return -1;
    }

#if _REENTRANT
    tsd = (hwpc_papi_tsd_t *) malloc(sizeof(hwpc_papi_tsd_t));
    if ( tsd == NULL ) {
        return PS_ENOMEM;
    }
    eventset = tsd->eventset = (int *) malloc(sizeof(int));
    if ( tsd->eventset == NULL ) {
        return PS_ENOMEM;
    }
    tsd->sprofils = NULL;
    if ( pthread_setspecific(tsdkey, tsd) != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Error setting thread-specific data in ", stderr); \
                 fputs("setup_counters_internal\n", stderr);)
        free(eventset);
        return PS_ESYS;
    }
#else /* ! _REENTRANT */
    eventset = (int *) malloc(sizeof(int));
    /* @@ This is redundant */
    if ( eventset == NULL ) {
        return PS_ENOMEM;
    }
#endif /* _REENTRANT */

    if ( eventset == NULL ) {
        PS_DEBUG(PS_DEBUG_FATAL,\
                 fputs("Error allocating or retrieving eventset ", stderr); \
                 fputs("in setup_counters_internal\n", stderr);)
        return -1;
    }
    PS_DEBUG(PS_DEBUG_INFO, printf("Eventset address: %p\n", eventset);)

    *eventset = PAPI_NULL;

    papiret = PAPI_create_eventset(eventset);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL,\
                 fputs("Error in setup_counters_internal ", stderr); \
                 fprintf(stderr, " (PAPI_create_eventset: %s)\n", \
                         PAPI_strerror(papiret));)
#if ! _REENTRANT
        free(eventset);
#endif /* ! _REENTRANT */
        return PS_HWPC_EPERF;
    }

    PS_DEBUG(PS_DEBUG_INFO, printf("Created an eventset (%d).\n", *eventset);)

    /* @@ Should be placed in separate routine for reuse */
    PS_DEBUG(PS_DEBUG_INFO, \
             options.domain.eventset = *eventset; \
             papiret = PAPI_get_opt(PAPI_GET_DOMAIN, &options); \
             if ( papiret != PAPI_OK ) { \
                 fprintf(stderr, "error retrieving domain: %s\n", \
                         PAPI_strerror(papiret)); \
             } \
             else { \
                 fputs("Domain of eventset: ", stdout); \
                 switch ( options.domain.domain ) { \
                 case PAPI_DOM_USER:   puts("user"); break; \
                 case PAPI_DOM_KERNEL: puts("kernel"); break; \
                 case PAPI_DOM_ALL:    puts("all"); break; \
                 default: printf("unknown (%d)\n", options.domain.domain); \
                          break; \
                 } \
             })

    /* Return "invalid argument" error if no event is given,
     * else add one event.
     */
    if (numEvents < 1) {
        PS_DEBUG(PS_DEBUG_FATAL,\
                 fputs("Error in setup_counters_internal: no event is given.", stderr);)
        return PS_EINVAL;
    } else {
        papiret = PAPI_add_event(*eventset, event_codes[0]);
        if ( papiret != PAPI_OK ) {
            PS_DEBUG(PS_DEBUG_FATAL,\
                     fputs("Error in setup_counters_internal ", stderr); \
                     fprintf(stderr, "(PAPI_add_event: code %d, string '%s')\n", \
                             papiret, PAPI_strerror(papiret));)
        } else {
            PS_DEBUG(PS_DEBUG_INFO, puts("setup_counters_internal add_event succeeded.");)
        }
    }
    if (1 == numEvents) {
        PS_DEBUG(PS_DEBUG_INFO, printf("Added %d events to eventset %d\n",\
                                       numEvents, *eventset);)
        return PS_SUCCESS;
    }

    /*
     * Multiplexing should only occur when counting.
     */
    if ( globals->multiplexing ) {
#if PAPI3
        papiret = PAPI_set_multiplex(*eventset);
#else
        papiret = PAPI_set_multiplex(eventset);
#endif
        if ( papiret != PAPI_OK ) {
            /* Get the eventset component index to check whether these are
             * CPU events (CPU events' component index is always 0).  For
             * non-CPU events, they might not support or have the concept
             * of multiplexing, and in this case, this error code
             * (PAPI_EINVAL for Cray Gemini NPU events) can be ignored.
             */
            compIndex = PAPI_get_eventset_component(*eventset);
            if ( 0 == compIndex ) {
                PS_DEBUG(PS_DEBUG_FATAL,\
                     fputs("Error in setup_counters_internal ", stderr); \
                     fprintf(stderr, "(PAPI_set_multiplex: code %d, string '%s')\n", \
                             papiret, PAPI_strerror(papiret));)
#if ! _REENTRANT
                free(eventset);
#endif /* ! _REENTRANT */
                return PS_HWPC_EPERF;
            } else {
                PS_DEBUG(PS_DEBUG_INFO,printf("PAPI_set_multiplex returned %d ('%s') for the eventset's component index %d.  Ignored since it's a non-CPU component...\n", \
                         papiret, PAPI_strerror(papiret), compIndex);)
            }
        }
        PS_DEBUG(PS_DEBUG_INFO, printf("Multiplexing set for eventset %d.\n", \
                                       *eventset);)
    }

#if PAPI3
    papiret = PAPI_add_events(*eventset, &event_codes[1], numEvents-1);
#else
    papiret = PAPI_add_events(eventset, &event_codes[1], numEvents-1);
#endif
    if ( papiret != PAPI_OK) {
        PS_DEBUG(PS_DEBUG_FATAL,\
                 fputs("Error in setup_counters_internal ", stderr);
                 fprintf(stderr, "(PAPI_add_events: %s)\n",\
                         PAPI_strerror(papiret));)
#if ! _REENTRANT
        free(eventset);
#endif /* ! _REENTRANT */
        return PS_HWPC_EPERF;
    }

    PS_DEBUG(PS_DEBUG_INFO, printf("Added %d events to eventset %d\n",\
                                   numEvents, *eventset);)

    return PS_SUCCESS;
}

/**********************************************************************
         Routines that are used by the generic layer.
 **********************************************************************/

int
get_num_counters_papi(void)
{
    return PAPI_get_opt(PAPI_GET_MAX_HWCTRS, NULL);
}

int
get_max_counters_papi(struct ps_hwpc_globals *globals)
{
    /*
     * This is disabled at present because current version of PAPI
     * returns 0 for MAX_MPX_CTRS option (at least on Xeon/perfctr/perfmon
     * install).
     */

    int papiret = 0;
    int event_code = 0;
    int comp_index = 0;
    char * event = NULL;
    papiret = PAPI_library_init(PAPI_VER_CURRENT);
    if ( papiret != PAPI_VER_CURRENT ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("PAPI version check failed", stderr); \
                 fprintf(stderr, " (expected %d, got %d)\n", \
                                 PAPI_VER_CURRENT, papiret);)
        return PS_HWPC_EPERF;
    }
    PS_DEBUG(PS_DEBUG_INFO, puts("PAPI initialized successfully");)

    /* Need to get the component index of an event, using the first for now. */
    event = globals->events.events[0].val;
    papiret = PAPI_event_name_to_code(event, &event_code);
    if (papiret != PAPI_OK) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error in PAPI_event_name_to_code for event '%s', error code %d, string '%s')\n", \
                         event, papiret, PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }
    comp_index = PAPI_get_event_component(event_code);
    if (comp_index < 0) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error in PAPI_get_event_component, returned %d < 0.\n", \
                         comp_index);)
        return PS_HWPC_EPERF;
    } else {
        PS_DEBUG(PS_DEBUG_INFO, printf("event '%s', component index %d.\n",
                 event, comp_index);)
    }

    papiret = PAPI_get_cmp_opt(PAPI_MAX_MPX_CTRS, NULL, comp_index);
    if ( papiret <= 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "PAPI_MAX_MPX_CTRS is <=0: %d\n", papiret);)
        return PS_HWPC_EPERF;
    } else {
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf ("PAPI_MAX_MPX_CTRS = %d.\n", papiret);)
    }

    return papiret;
}

/*
 * This function sets up everything globally, but it
 * does not start counting.  That will be done
 * the first time through "start_counters" for each
 * thread.
 *
 * @@ This assumes that derived events consume
 * exactly two counters.
 */

int
init_counters_papi(struct ps_hwpc_globals *globals)
{
    int papiret, num_counters, num_derived, i;
    int domain = PAPI_DOM_USER;
    ps_hwpc_eventarray_t *events = &globals->events;
    PAPI_preset_info_t info;
#if defined(PAPI_DEF_MPX_USEC)
    PAPI_option_t option;
#endif
    char version[16];

    papiret = PAPI_library_init(PAPI_VER_CURRENT);
    /* @@ This needs to be updated for PAPI3 */
    if ( papiret < PS_PAPIVERSION ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("PAPI/PerfSuite version check failed", stderr); \
                 fprintf(stderr, " (require at least %d, got %d)\n", \
                                 PS_PAPIVERSION, papiret);)
        return PS_HWPC_EPERF;
    }
    PS_DEBUG(PS_DEBUG_INFO, puts("PAPI initialized successfully");)
    if ( papiret != PAPI_VER_CURRENT ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("PAPI version check failed", stderr); \
                 fprintf(stderr, " (expected %d, got %d)\n", \
                                 PAPI_VER_CURRENT, papiret);)
        return PS_HWPC_EPERF;
    }

#if PAPI3
# if defined(PAPI_LIB_VERSION)
    papiret = PAPI_get_opt(PAPI_LIB_VERSION, NULL);
# endif
    sprintf(version, "%d.%d.%d.%d", PAPI_VERSION_MAJOR(papiret),
                                    PAPI_VERSION_MINOR(papiret),
                                    PAPI_VERSION_REVISION(papiret),
                                    PAPI_VERSION_INCREMENT(papiret));
#else
    sprintf(version, "%d", papiret);
#endif
    globals->version = strdup(version);
    PS_DEBUG(PS_DEBUG_INFO, printf("PAPI version in use: %s\n", \
                                   globals->version);)

#if _REENTRANT
# if PAPI3
    papiret = PAPI_thread_init((unsigned long (*)(void))(pthread_self));
# else
    papiret = PAPI_thread_init((unsigned long (*)(void))(pthread_self),0);
# endif
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL,\
                 fputs("Error initializing PAPI POSIX threads support\n", \
                       stderr);)
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "PAPI error: %s\n", PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }
    PS_DEBUG(PS_DEBUG_INFO, puts("Initialized PAPI thread support");)

    if ( pthread_key_create(&tsdkey, free_tsd) != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Error creating PAPI thread-specific data key\n",
                       stderr);)
        return PS_ESYS;
    }
#endif /* _REENTRANT */

    numEvents = events->numEvents;
    num_derived = 0;
    event_codes = (int *) malloc(numEvents * sizeof(int));
    if ( event_codes == NULL ) {
        return PS_ENOMEM;
    }

    /*
     * Cycle through the list of events supplied, checking where
     * possible for validity and storing in the event code array.
     */
    for ( i = 0; i < numEvents; i++ ) {
        events->events[i].derived = 0; /* Assume not and check if preset */
        switch ( events->events[i].type ) {
        case PRESET:
            /*
             * Is the counter name bogus? 
             */
            if ( PAPI_event_name_to_code(events->events[i].val,
                                         &event_codes[i]) != PAPI_OK ) {
                PS_DEBUG(PS_DEBUG_WARNING,\
                         printf("Event \"%s\" is not a valid preset\n",\
                                events->events[i].val);)
                free(event_codes);
                return PS_HWPC_ENOSUCHEVENT;
            }

#if ! PAPI3
            /*
             * Is it a rate counter?  If so, reject.
             */
            if ( (event_codes[i] == PAPI_FLOPS) ||
                 (event_codes[i] == PAPI_IPS) ) {
                PS_DEBUG(PS_DEBUG_WARNING,\
                         fputs("\tPAPI \"rate\" events (such as ", stdout);\
                         printf("\"%s\") are not supported.\n", \
                                events->events[i].val);\
                         fputs("\tPlease use the underlying events ", stdout);\
                         puts("(PAPI_FP_INS/PAPI_TOT_INS) instead.");)
                free(event_codes);
                return PS_HWPC_EEVENTNOTSUPPORTED;
            }
#endif

            /*
             * Is it supported on this CPU?
             *
             * @@ PAPI3 allows us to retrieve the exact information
             * about derived events through the new interface.  Need
             * to update for "real" PAPI3, instead of using the
             * compatibility routines.
             */
            if ( PAPI_query_event_verbose(event_codes[i],&info) != PAPI_OK ) {
                PS_DEBUG(PS_DEBUG_WARNING,\
                         printf("Event \"%s\" is not available on this machine\n",\
                                events->events[i].val);)
                free(event_codes);
                return PS_HWPC_EEVENTNOTAVAILABLE;
            }

            /*
             * Is it derived?  
             */
            if ( info.flags & PAPI_DERIVED ) {
                events->events[i].derived = 1; 
                num_derived++;
            }
            break;
        case NATIVE:
#if PAPI3
            /*
             * Is the counter name bogus? 
             */
            if ( PAPI_event_name_to_code(events->events[i].val,
                                         &event_codes[i]) != PAPI_OK ) {
                PS_DEBUG(PS_DEBUG_WARNING,\
                         printf("Event \"%s\" is not a valid native event\n",\
                                events->events[i].val);)
                free(event_codes);
                return PS_HWPC_ENOSUCHEVENT;
            }
#else
            return PS_HWPC_EBADEVENT;
#endif
            break;
        default:
            PS_DEBUG(PS_DEBUG_WARNING, puts("Bad PAPI event type!");)
            free(event_codes);
            return PS_HWPC_EBADEVENT;
        }
    }

    num_counters = get_num_counters_papi();
    if ( num_counters <= 0 ) {
        return -1;
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("Available counters: %d\n", num_counters); \
             printf("Events requested:   %d\n", numEvents); \
             printf("Derived events:     %d\n", num_derived);)

    /*
     * Turn on multiplexing if we have to.
     * Here's where the assumption that a derived event 
     * takes exactly two counters is made.
     *
     * First adjust multiplexing interval if the version
     * of PAPI supports it.  This will be in effect for
     * the remainder of execution.
     */
#if defined(PAPI_DEF_MPX_USEC)
    if ( globals->mpx_usec > 0 ) {
        option.multiplex.us = globals->mpx_usec;
        papiret = PAPI_set_opt(PAPI_DEF_MPX_USEC, &option);
        if ( papiret != PAPI_OK ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("Multiplex interval adjustment failed. ", stderr); \
                     fprintf(stderr, "PAPI error: %s\n", \
                                     PAPI_strerror(papiret));)
            return PS_HWPC_EPERF;
        }
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Multiplexing interval set to %d microseconds.\n", \
                        option.multiplex.us);)
    }
#endif
    if ( ( num_derived + numEvents ) > num_counters ) {
            papiret = PAPI_multiplex_init();
            if ( papiret != PAPI_OK ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("Multiplex initialization failed. ", stderr); \
                         fprintf(stderr, "PAPI error: %s\n", \
                                         PAPI_strerror(papiret));)
                return PS_HWPC_EPERF;
            }
            globals->multiplexing = 1;
            PS_DEBUG(PS_DEBUG_INFO,\
                     puts("Initialized multiplexing");)
    }

    if ( ! (globals->multiplexing || globals->profiling) ) {
        /* Double check on multiplexing to be sure */
        if ( do_dryrun(event_codes, numEvents, &(globals->multiplexing))
                 != PS_SUCCESS ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("Dry run mpx check failed.\n", stderr);)
            return PS_HWPC_EPERF;
        }
        PS_DEBUG(PS_DEBUG_INFO, \
                 fputs("Dry run mpx check completed successfully. ", stdout); \
                 printf("Multiplexing will%s be used.\n", \
                        globals->multiplexing ? "" : " not");)
    }

    switch ( globals->domain ) {
    case USER:
    case UNKNOWN_DOMAIN:
        domain = PAPI_DOM_USER;
        break;
    case KERNEL:
        domain = PAPI_DOM_KERNEL;
        break;
    case ALL:
        domain = PAPI_DOM_ALL;
        break;
    }

    papiret = PAPI_set_domain(domain);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL,\
                 fprintf(stderr, "Error setting PAPI domain: %s\n", \
                         PAPI_strerror(papiret));) 
        free(event_codes);
        return PS_HWPC_EPERF;
    }

    return PS_SUCCESS;
}

/*
 * This should only be called once by each thread.  The model is
 * that the counters are started once, read many times, and
 * stopped once.  This routine is responsible for initializing
 * thread-specific counters and starting the counting.
 */

int
start_counters_papi(struct ps_hwpc_globals *globals)
{
    int papiret, ret;
#if _REENTRANT
    hwpc_papi_tsd_t *tsd;
    int *eventset;
#endif /* _REENTRANT */

    ret = setup_counters_internal(globals);
    if ( ret != PS_SUCCESS ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error %d: setup_counters_internal\n", ret);)
        return ret;
    }

#if _REENTRANT
    tsd = (hwpc_papi_tsd_t *) pthread_getspecific(tsdkey);
    eventset = tsd->eventset;
#endif

    papiret = PAPI_start(*eventset);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL,\
                 fprintf(stderr, \
                         "Error in start_counters_papi (PAPI_start: %s)\n",\
                         PAPI_strerror(papiret));)
#if ! _REENTRANT
        free(eventset);
#endif /* ! _REENTRANT */
        return PS_HWPC_EPERF;
    }

    PS_DEBUG(PS_DEBUG_INFO, printf("Started counting eventset %d\n",\
                                   *eventset);)

    return PS_SUCCESS;
}

/* Return the current values of the counters */

int
read_counters_papi(ps_hwpc_value_t *values)
{
    int papiret;
#if DEBUG
    int i;
#endif
#if _REENTRANT
    int *eventset;
    hwpc_papi_tsd_t *tsd;
    tsd = (hwpc_papi_tsd_t *) pthread_getspecific(tsdkey);
    eventset = tsd->eventset;
#endif /* _REENTRANT */

    PS_DEBUG(PS_DEBUG_INFO, printf("Reading eventset %d\n", *eventset);)

    papiret = PAPI_read(*eventset, values);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL,\
                 fprintf(stderr, \
                         "Error in read_counters_papi (PAPI_read: %s)\n",\
                         PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             puts("PAPI layer read:"); \
             for ( i = 0; i < numEvents; i++ ) { \
                 printf("Event %d: %lld\n", i, values[i]); \
             } )

    return PS_SUCCESS;
}

/*
 * This should only be called once by each thread.  The model is
 * that the counters are started once, read many times, and
 * stopped once.
 */

int
stop_counters_papi(ps_hwpc_value_t *values)
{
    int state, papiret;
#if _REENTRANT
    int *eventset;
    hwpc_papi_tsd_t *tsd;
    tsd = (hwpc_papi_tsd_t *) pthread_getspecific(tsdkey);
    eventset = tsd->eventset;
#endif /* _REENTRANT */

   /*
    * Before we actually try to stop the eventset, check to make sure it
    * hasn't already been stopped (stop_profiling_papi will have already
    * done this so that it can call PAPI_sprofil() to turn off overflow).
    */
    papiret = PAPI_state(*eventset, &state);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "stop_counters_papi: PAPI state error %s\n", \
                         PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }

    if ( state & PAPI_RUNNING ) {
        PS_DEBUG(PS_DEBUG_INFO, printf("Stopping eventset %d\n", *eventset);)

        papiret = PAPI_stop(*eventset, values);
        if ( papiret != PAPI_OK ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fprintf(stderr, \
                             "Error in stop_counters_papi (PAPI_stop: %s, eventset %d)\n",\
                             PAPI_strerror(papiret), *eventset);)
            return PS_HWPC_EPERF;
        }
    }
    else {
        PS_DEBUG(PS_DEBUG_INFO, \
                 puts("stop_counters_papi: eventset was stopped, no action\n");)
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("Cleaning eventset %d\n", *eventset);)
#if PAPI3
    papiret = PAPI_cleanup_eventset(*eventset);
#else
    papiret = PAPI_cleanup_eventset(eventset);
#endif
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, \
                         "Error in stop_counters_papi (PAPI_cleanup_eventset: %s)\n",\
                         PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("Destroying eventset %d\n", *eventset);)

    papiret = PAPI_destroy_eventset(eventset);
    if ( papiret != PAPI_OK ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fprintf(stderr, \
                             "Error in stop_counters_papi (PAPI_destroy_eventset: %s)\n",\
                             PAPI_strerror(papiret));)
            return PS_HWPC_EPERF;
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             puts("Exiting stop_counters_papi successfully");)

    return PS_SUCCESS;
}

/* Shut down the shop */

int
destroy_counters_papi(void)
{
    PAPI_shutdown();

    PS_DEBUG(PS_DEBUG_INFO, puts("PAPI_shutdown called");)

    return PS_SUCCESS;
}

/*
 * Profiling support
 *
 * PAPI seems to require the following model for profiling control:
 *
 * PAPI_sprofil
 * PAPI_start
 * PAPI_stop
 * PAPI_sprofil
 *
 * i.e., PAPI_sprofil cannot be called while counting is ongoing
 *
 */

int
init_profiling_papi(ps_hwpc_globals_t *globals, ps_hwpc_sample_t **samples)
{
    int papiret, ret, sprofil, eventcode, maxsize = 1;
#if _REENTRANT
    hwpc_papi_tsd_t *tsd;
    int *eventset;
    PAPI_sprofil_t *sprofils;
#endif /* _REENTRANT */
#if DEBUG
    char name[PAPI_MAX_STR_LEN];
#endif

    PS_DEBUG(PS_DEBUG_INFO, puts("Initializing profiling in PAPI layer");)

    ret = setup_counters_internal(globals);
    if ( ret != PS_SUCCESS ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error %d: setup_counters_internal\n", ret);)
        return ret;
    }

#if _REENTRANT
    tsd = (hwpc_papi_tsd_t *) pthread_getspecific(tsdkey);
    eventset = tsd->eventset;
#endif

    /* @@ Could change to avoid call to list_events but this is simple for now.
     * This is also where the restriction of only one event in a
     * config document is coming from.
     */
    papiret = PAPI_list_events(*eventset, &eventcode, &maxsize);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL, fprintf(stderr, "PAPI error: %s\n", \
                                         PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }

    PS_DEBUG(PS_DEBUG_INFO, PAPI_event_code_to_name(eventcode, name); \
                            printf("Profiling using PAPI event code %x (%s)\n",\
                                  eventcode, name);)

    sprofils = (PAPI_sprofil_t *) malloc(globals->profiles.nmaps * 
                                         sizeof(PAPI_sprofil_t));
    if ( sprofils == NULL ) {
        return PS_ENOMEM;
    }
#if _REENTRANT
    tsd->sprofils = sprofils;
#endif
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("PAPI_sprofil_t structure allocated at %p\n", \
                    sprofils);)

    for ( sprofil = 0; sprofil < globals->profiles.nmaps; sprofil++ ) {
        sprofils[sprofil].pr_base  = samples[sprofil];
        sprofils[sprofil].pr_size  = globals->profiles.profmap[sprofil].nbins;
        sprofils[sprofil].pr_off   =
#if PAPI3
                            globals->profiles.profmap[sprofil].offset;
#else
            (unsigned long) globals->profiles.profmap[sprofil].offset;
#endif
        sprofils[sprofil].pr_scale = globals->profiles.profmap[sprofil].scale;
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             puts("Calling PAPI_sprofil with arguments"); \
             printf("sprofils  = %p\n", sprofils); \
             printf("nmaps     = %d\n", globals->profiles.nmaps); \
             printf("*eventset = %d\n", *eventset); \
             printf("eventcode = %x\n", eventcode); \
             printf("threshold = %d\n", globals->threshold); \
             printf("flags     = %d\n", PAPI_PROFIL_POSIX);)

    /* @@ Add the ability to control profiling style beyond POSIX */
    papiret = PAPI_sprofil(sprofils, globals->profiles.nmaps, *eventset,
                                  eventcode, globals->threshold,
                                  PAPI_PROFIL_POSIX);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL, fprintf(stderr, "PAPI_sprofil error: %s\n", \
                                         PAPI_strerror(papiret));)
#if ! _REENTRANT
        free(sprofils);
#endif
        return PS_HWPC_EPERF;
    }
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("Profiling initialized successfully (PAPI_sprofil: %d buffers)\n", \
                    globals->profiles.nmaps);)

    return PS_SUCCESS;
}

/*
 * Profiling control - note that unlike start_counters_papi(), this
 * can be called any number of times by the generic layer (only
 * init_profiling is called once).
 */

int
start_profiling_papi (ps_hwpc_globals_t *globals,
                      ps_hwpc_sample_t **samples)
{
    int papiret;
#if _REENTRANT
    int *eventset;
    hwpc_papi_tsd_t *tsd;
    tsd = (hwpc_papi_tsd_t *) pthread_getspecific(tsdkey);
    eventset = tsd->eventset;
#endif /* _REENTRANT */

    PS_DEBUG(PS_DEBUG_INFO, puts("Starting profiling");)

    papiret = PAPI_start(*eventset);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL,\
                 fprintf(stderr, \
                         "Error in start_profiling_papi (PAPI_start: %s)\n",\
                         PAPI_strerror(papiret));)
#if ! _REENTRANT
        free(eventset);
#endif /* ! _REENTRANT */
        return PS_HWPC_EPERF;
    }
    return PS_SUCCESS;
}

/*
 * Can also be called multiple times by the generic layer.
 */

int
suspend_profiling_papi (ps_hwpc_globals_t *globals,
                        ps_hwpc_sample_t **samples)
{
    int papiret;
    ps_hwpc_value_t value;
#if _REENTRANT
    int *eventset;
    hwpc_papi_tsd_t *tsd;
    tsd = (hwpc_papi_tsd_t *) pthread_getspecific(tsdkey);
    eventset = tsd->eventset;
#endif /* _REENTRANT */

    PS_DEBUG(PS_DEBUG_INFO, printf("Stopping eventset %d to suspend profiling\n",
                                   *eventset);)

    papiret = PAPI_stop(*eventset, &value);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Error in suspend_profiling_papi ", stderr); \
                 fprintf(stderr, "(PAPI_stop: %s, eventset %d)\n", \
                         PAPI_strerror(papiret), *eventset);)
        return PS_HWPC_EPERF;
    }

    return PS_SUCCESS;
}

/* 
 * Only called once by the generic layer.
 */

int
stop_profiling_papi (ps_hwpc_globals_t *globals,
                     ps_hwpc_sample_t **samples)
{
    int papiret, ret;
    ps_hwpc_value_t value;
#if _REENTRANT
    int *eventset;
    PAPI_sprofil_t *sprofils;
    hwpc_papi_tsd_t *tsd;
    tsd = (hwpc_papi_tsd_t *) pthread_getspecific(tsdkey);
    eventset = tsd->eventset;
    sprofils = tsd->sprofils;
#endif /* _REENTRANT */

    PS_DEBUG(PS_DEBUG_INFO, puts("Stopping profiling");)

    /* @@ The "1" below is just a dummy value.  This assumes PAPI
     * doesn't actually check it for validity.
     */
    papiret = PAPI_stop(*eventset, &value);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "stop_profiling_papi: PAPI_stop error %s\n", \
                         PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }

    papiret = PAPI_sprofil(sprofils, 1, *eventset, event_codes[0], 0,
                           PAPI_PROFIL_POSIX);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, \
                         "stop_profiling_papi: PAPI_sprofil error %s\n", \
                         PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }

    ret = stop_counters_papi(&value);
    return ret;  /* Will be PS_SUCCESS if all's OK */
}


/*
 * Perform a dry run of creating the eventset and starting it.
 * This is intended to double-check whether multiplexing is
 * going to be needed.  On Pentium 4, for example, although
 * there are 18 counters, multiplexing may still be needed
 * with event sets that use less because of resource conflicts.
 * This adds a little overhead, but we'll take the hit since
 * it's only done once (besides, we have no choice... :)
 *
 * We don't worry about threads here, since only one thread is
 * supposed to call ps_hwpc_init(), and that is where this
 * routine is called.  However, if compiled with _REENTRANT,
 * PAPI_thread_init() will already have been called before
 * entering this routine.
 *
 * @@ This routine can be shortened considerably by pulling out 
 * common source.  As it is, it's a honker.
 *
 */

static int
do_dryrun(int *event_codes, int numEvents, int *dompx)
{
    int tmp_eventset, papiret;
    ps_hwpc_value_t *tmp_values;
#if DEBUG
    unsigned long long startticks, stopticks;
#endif

    PS_DEBUG(PS_DEBUG_INFO, \
             puts("Starting dry run check of event set."); \
             ps_rtc(&startticks);)

    tmp_eventset = PAPI_NULL;
    papiret = PAPI_create_eventset(&tmp_eventset);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Dry run non-mpx eventset creation failed.\n", stderr);)
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "PAPI error: %s\n", PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             puts("Created dry run non-mpx event set successfully.");)

#if PAPI3
    papiret = PAPI_add_events(tmp_eventset, event_codes, numEvents);
#else
    papiret = PAPI_add_events(&tmp_eventset, event_codes, numEvents);
#endif
    if ( papiret != PAPI_OK ) {
#if PAPI3
        /*
         * PAPI 3's add_events can "partially succeed".  For us, it's
         * all or nothin.
         */
        if ( (papiret == PAPI_ECNFLCT) || (papiret > 0) ) {
#else
        if ( papiret == PAPI_ECNFLCT ) {
#endif
            PS_DEBUG(PS_DEBUG_INFO, \
                     fputs("Dry run event addition non-mpx failed. ", stdout); \
                     puts("Retrying with multiplexing.");)
#if PAPI3
            papiret = PAPI_cleanup_eventset(tmp_eventset);
#else
            papiret = PAPI_cleanup_eventset(&tmp_eventset);
#endif
            if ( papiret != PAPI_OK ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("Dry run non-mpx cleanup eventset failed.\n", \
                               stderr);)
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fprintf(stderr, \
                                 "PAPI error: %s\n", PAPI_strerror(papiret));)
                return PS_HWPC_EPERF;
            }

            PS_DEBUG(PS_DEBUG_INFO, \
                         puts("Cleaned up dry run non-mpx event set.");)

            papiret = PAPI_destroy_eventset(&tmp_eventset);
            if ( papiret != PAPI_OK ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("Dry run destroy non-mpx eventset failed.\n", \
                               stderr);)
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fprintf(stderr, \
                                 "PAPI error: %s\n", PAPI_strerror(papiret));)
                return PS_HWPC_EPERF;
            }

            PS_DEBUG(PS_DEBUG_INFO, \
                     puts("Destroyed dry run non-mpx event set.");)

            /* 
             * Multiplexing has not yet been initialized (otherwise
             * we wouldn't be executing this routine).  Go for it...
             * @@ Note that multiplex_init() is called twice.  Need
             * to verify that this is OK, perhaps write a test case.
             */
            papiret = PAPI_multiplex_init();
            if ( papiret != PAPI_OK ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("Dry run mpx initialization failed.\n", stderr);)
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fprintf(stderr, \
                                 "PAPI error: %s\n", PAPI_strerror(papiret));)
                return PS_HWPC_EPERF;
            }
            PS_DEBUG(PS_DEBUG_INFO, \
                     puts("Dry run mpx initialization successful.");)

            tmp_eventset = PAPI_NULL;    
            papiret = PAPI_create_eventset(&tmp_eventset);
            if ( papiret != PAPI_OK ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("Dry run mpx eventset creation failed.\n", \
                               stderr);)
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fprintf(stderr, \
                                 "PAPI error: %s\n", PAPI_strerror(papiret));)
                return PS_HWPC_EPERF;
            }
            /* Add the first event to get the correct component assigned.
             * This is required by PAPI_set_multiplex().
             */
            if (numEvents <= 1) {
                PS_DEBUG(PS_DEBUG_FATAL,\
                         fprintf(stderr, "Dry run error: numEvents (%d) <= 1.", numEvents);)
                return PS_EINVAL;
            }
            papiret = PAPI_add_event(tmp_eventset, event_codes[0]);
            if ( papiret != PAPI_OK ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("Dry run PAPI_add_event failed.\n", stderr);)
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fprintf(stderr, "PAPI error code %d, string '%s'\n", \
                                 papiret, PAPI_strerror(papiret));)
                return PS_HWPC_EPERF;
            }
#if PAPI3
            papiret = PAPI_set_multiplex(tmp_eventset);
#else
            papiret = PAPI_set_multiplex(&tmp_eventset);
#endif
            if ( papiret != PAPI_OK ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("Dry run PAPI_set_multiplex failed.\n", stderr);)
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fprintf(stderr, \
                                 "PAPI error: %s\n", PAPI_strerror(papiret));)
                return PS_HWPC_EPERF;
            }
            else {
                PS_DEBUG(PS_DEBUG_INFO, \
                         puts("Dry run PAPI_set_multiplex succeeded.");)
            }

#if PAPI3
            papiret = PAPI_add_events(tmp_eventset, &event_codes[1], numEvents-1);
#else
            papiret = PAPI_add_events(&tmp_eventset, &event_codes[1], numEvents-1);
#endif
            if ( papiret != PAPI_OK ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("Add events to dry run mpx set failed.\n", \
                               stderr); \
                         fprintf(stderr, \
                                 "PAPI error: %s.\n", PAPI_strerror(papiret));)
                return PS_HWPC_EPERF;
            }
            else {
                PS_DEBUG(PS_DEBUG_INFO, \
                         printf("Added %d events to mpx dry run event set successfully.\n", \
                                numEvents);)
            }
            *dompx = 1;
        }
    }
    else {
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Added %d events to dry run non-mpx event set successfully.\n", numEvents);)
    }

    papiret = PAPI_start(tmp_eventset);
    if ( papiret == PAPI_ECNFLCT ) {
        PS_DEBUG(PS_DEBUG_INFO, \
                 puts("Dry run start failed with PAPI_ECNFLCT.");)
        return PS_HWPC_EPERF;
    }
    else {
        if ( papiret != PAPI_OK ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fprintf(stderr, \
                             "Dry run PAPI_start failed != ECNFLCT (%s).\n", \
                             PAPI_strerror(papiret));)
            return PS_HWPC_EPERF;
        }
    }

    PS_DEBUG(PS_DEBUG_INFO, puts("Started dry run check successfully.");)

    /* In the threaded case, we'll install a cleanup handler just in
     * case PAPI_stop is cancelled (unlikely, but playing it safe)
     */
    tmp_values = (ps_hwpc_value_t *) malloc(numEvents * sizeof(ps_hwpc_value_t));
    if ( tmp_values == NULL ) {
        PS_DEBUG(PS_DEBUG_FATAL, fputs("Dry run malloc failed.\n", stderr);)
        return PS_ENOMEM;
    }
#if _REENTRANT
    pthread_cleanup_push((void (*)())free, (void *) tmp_values);
#endif

    PS_DEBUG(PS_DEBUG_INFO, \
             puts("Allocated space for dry run counter values.");)

    papiret = PAPI_stop(tmp_eventset, tmp_values);
    if ( papiret != PAPI_OK ) {
        free(tmp_values);
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Dry run PAPI_stop failed.\n", stderr);)
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "PAPI error: %s\n", PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }

    PS_DEBUG(PS_DEBUG_INFO, puts("Stopped dry run test successfully.");)

#if _REENTRANT
    pthread_cleanup_pop(0);
#endif
    free(tmp_values);

    PS_DEBUG(PS_DEBUG_INFO, puts("Deallocated dry run counter values."); \
                            ps_rtc(&stopticks);
                            printf("Wallclock ticks elapsed in dry run = %lld\n", \
                                   stopticks-startticks);)

#if PAPI3
    papiret = PAPI_cleanup_eventset(tmp_eventset);
#else
    papiret = PAPI_cleanup_eventset(&tmp_eventset);
#endif
    if ( papiret != PAPI_OK ) {
        free(tmp_values);
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Dry run PAPI_cleanup_eventset failed.\n", stderr);)
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "PAPI error: %s\n", PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }

    PS_DEBUG(PS_DEBUG_INFO, puts("Cleaned up dry run event set");)

    papiret = PAPI_destroy_eventset(&tmp_eventset);
    if ( papiret != PAPI_OK ) {
        free(tmp_values);
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Dry run PAPI_destroy_eventset failed.\n", stderr);)
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "PAPI error: %s\n", PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }

    PS_DEBUG(PS_DEBUG_INFO, puts("Destroyed dry run event set.");)

    return PS_SUCCESS;
}

#if _REENTRANT
static void
free_tsd(void *data)
{
    hwpc_papi_tsd_t *tsd = (hwpc_papi_tsd_t *) data;
    if ( tsd->eventset != NULL ) {
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Deallocating TSD eventset at %p\n", tsd->eventset);)
        free(tsd->eventset);
    }
    if ( tsd->sprofils != NULL ) {
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Deallocating TSD PAPI_sprofil_t structure at %p\n", \
                        tsd->sprofils);)
        free(tsd->sprofils);
    }
    PS_DEBUG(PS_DEBUG_INFO, printf("Deallocating TSD at %p\n", tsd);)
    free(tsd);
    pthread_setspecific(tsdkey, NULL);

    return;
}
#endif /* _REENTRANT */


/* ============================================================== *
 *                PACKAGE INTERFACE ROUTINES                      *
 * ============================================================== */

/*
 * Convenience routine to provide a way for PAPI-collected events
 * to be written out in ps_hwpc format.
 *
 * @@ This has not been tested extensively.
 */

int
ps_hwpc_PAPI_write(const char *fname, int eventset,
                   ps_hwpc_value_t *values, unsigned long long ticks)
{
    FILE *fout;
    int ret, close;
    char *exebase;
    ps_hwpc_values_t _values;
    ps_hwpc_globals_t globals;

    _ps_hwpc_init_globals(&globals);

    ret = PAPI_to_ps_hwpc(&globals, eventset);
    if ( ret != 0 ) {
        PS_DEBUG(PS_DEBUG_WARNING, \
                 fprintf(stderr, \
                 "ps_hwpc_PAPI_write: error return in PAPI_to_ps_hwpc (%d)\n", \
                 ret);)
        return -1;
    };

#if _REENTRANT
    _values.myid = (unsigned int) PAPI_thread_id();
#endif
    _values.total = values;

    ret = _ps_getexebase(0, &exebase);
    if ( ret != PS_SUCCESS ) {
        exebase = NULL;
    }

    fout = _ps_hwpc_select_stream(fname, PS_SUFFIX_XML, exebase, &close);

    return (_ps_hwpc_xml_write(fout, LIBRARY_NAME, &globals, &_values, ticks,
            (double) PS_UNKNOWN_VALUE, (double) PS_UNKNOWN_VALUE, exebase,
            close));
}

/*
 *  @@ This assumes that the PAPI high level always uses event set zero. 
 *  Alternate approach would be to check integers up to some limit to
 *  see if there is an event set associated with it.
 */

int
ps_hwpc_PAPI_hl_write(const char *fname, int *eventcodes, int eventcodeslen,
                      ps_hwpc_value_t *values, unsigned long long ticks)
{
    return(ps_hwpc_PAPI_write(fname, 0, values, ticks));
}

void
PSF_HWPC_PAPI_WRITE_F77(const char *fname, int *eventset,
                        ps_hwpc_value_t *values, long long *ticks,
                        int *ierr, int fname_len)
{
    char cname[256]; /*@@ MAXPATHLEN? */
		       
    _ps_f2cstring(fname, fname_len, cname, 255);

    *ierr = ps_hwpc_PAPI_write(cname, *eventset, values, 
                               (unsigned long long) *ticks);

    return;
}

void
PSF_HWPC_PAPI_HL_WRITE_F77(const char *fname,
                           int *eventcodes, int *eventcodeslen,
                           ps_hwpc_value_t *values, long long *ticks,
                           int *ierr, int fname_len)
{
    char cname[256]; /*@@ MAXPATHLEN? */
		       
    _ps_f2cstring(fname, fname_len, cname, 255);

    *ierr = ps_hwpc_PAPI_hl_write(cname, eventcodes, *eventcodeslen,
                                  values, *ticks);

    return;
}


/*
 * Converts a PAPI event set to the pshwpc global
 * descriptor (as much as possible)
 */

static int
PAPI_to_ps_hwpc(ps_hwpc_globals_t *globals, int eventset)
{
    PAPI_option_t opt;
    PAPI_preset_info_t eventinfo;
    int state, event, *events, papiret;
    int max_mpx_counters;

#if DEBUG
    ps_setdebuglevel();
#endif

    max_mpx_counters = PAPI_get_opt(PAPI_MAX_MPX_CTRS, NULL);
    if (max_mpx_counters < 0) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error retrieving PAPI_MAX_MPX_CTRS option: %s\n", \
                         PAPI_strerror(max_mpx_counters));)
        return PS_HWPC_EPERF;
    }
    events = (int *) malloc (max_mpx_counters * sizeof(int));
    if ( events == NULL ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Error allocating memory for events in PAPI_to_ps_hwpc\n", stderr);)
        return PS_ENOMEM;
    }

    memset(&opt, 0, sizeof(PAPI_option_t));
    opt.domain.eventset = eventset;
    papiret = PAPI_get_opt(PAPI_GET_DOMAIN, &opt);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error retrieving PAPI domain: %s\n", \
                         PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }

    switch ( opt.domain.domain ) {
    case PAPI_DOM_USER:
        PS_DEBUG(PS_DEBUG_INFO, puts("PAPI domain identifed as USER");)
        globals->domain = USER;
        break;
    case PAPI_DOM_KERNEL:
        PS_DEBUG(PS_DEBUG_INFO, puts("PAPI domain identifed as KERNEL");)
        globals->domain = KERNEL;
        break;
    case PAPI_DOM_ALL:
        PS_DEBUG(PS_DEBUG_INFO, puts("PAPI domain identifed as ALL");)
        globals->domain = ALL;
        break;
    default:
        PS_DEBUG(PS_DEBUG_WARNING, puts("PAPI domain not identifed");)
        globals->domain = UNKNOWN_DOMAIN;
        break;
    }

    papiret = PAPI_state(eventset, &state);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error retrieving PAPI state: %s\n", \
                         PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }
    globals->multiplexing = (state & PAPI_MULTIPLEXING);
    PS_DEBUG(PS_DEBUG_INFO,\
             printf("PAPI eventset identified as%s multiplexing\n",\
             globals->multiplexing ? "" : " not");)

    /* Pull out the event information from the event set */
    globals->events.numEvents = max_mpx_counters;
    papiret = PAPI_list_events(eventset, events, 
                               (int *) &globals->events.numEvents);
    if ( papiret != PAPI_OK ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error in PAPI_list_events: %s\n", \
                         PAPI_strerror(papiret));)
        return PS_HWPC_EPERF;
    }

    globals->events.class = PAPI;

    /* @@ Not freed after done using */
    globals->events.events = (ps_hwpc_event_t *)
        malloc(globals->events.numEvents * sizeof(ps_hwpc_event_t));
    if ( globals->events.events == NULL ) {
        return -1;
    }

    /* @@ Note: this assumes the eventset only contains PAPI presets */
    /* This also over-allocates for the event names (PAPI_MAX_STR_LEN)
     * but is simple */

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("Found %d events in the PAPI event set\n",\
                    globals->events.numEvents);)

    for ( event = 0; event < globals->events.numEvents; event++ ) {
        /* @@ Update for PAPI3 */
        PAPI_query_event_verbose(events[event], &eventinfo);
        globals->events.events[event].type = PRESET;
        globals->events.events[event].derived = 
            (eventinfo.flags & PAPI_DERIVED) ? 1 : 0;
        globals->events.events[event].val =
             (char *) malloc(PAPI_MAX_STR_LEN * sizeof(char));
        if ( globals->events.events[event].val == NULL ) {
            free(globals->events.events);
            return PS_ENOMEM;
        }
        papiret = PAPI_event_code_to_name(events[event],
                                          globals->events.events[event].val);
        if ( papiret != PAPI_OK ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fprintf(stderr,
                             "_PAPI_to_ps_hwpc: event name lookup error: %s\n",\
                             PAPI_strerror(papiret));)
            free(globals->events.events);
            return PS_HWPC_EPERF;
        }
    }

    return PS_SUCCESS;
}
