/* $Id: hwpc-text.c,v 1.31 2012/05/11 15:11:11 ruiliu Exp $ */

/****************************************************************************
 *
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *       Copyright(C) 2004-2010, 2012, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * hwpc-text.c : write out plain text version of libpshwpc data
 *
 * This is a rough-draft version and is a placeholder to be redone.
 *
 * Author: Rick Kufrin
 * Date:   5/3/2004
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if HAVE_TIME_H
# include <time.h>
#endif

#if HAVE_PWD_H
# include <pwd.h>
#endif

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#include <perfsuite.h>
#include <psinternal.h>
#include "hwpc.h"
#include "hwpc-text.h"

#ifdef MEGABYTE
#undef MEGABYTE
#endif
#define MEGABYTE 1048576.0
#define KILOBYTE 1024

static char *
vendorstring(ps_cpuid_t *p)
{
    switch ( p->vendor ) {
    case INTEL: return "Intel"; break;
    case AMD:   return "AMD"; break;
    default:    return "unknown"; break;
    }
}

/*
 * Main driver for writing out libpshwpc plain text output.
 *
 */

static inline void
output_banner(FILE *fd, char *banner)
{
    int i;

    fprintf(fd, "%s\n", banner);
    for (i = 0; i < MAJORSEPLENGTH; i++) {
        fprintf(fd, "%c", MAJORSEPCHAR);
    }
    fputs("\n", fd);

    return;
}

static inline void
output_minorsep(FILE *fd)
{
    int i;

    for (i = 0; i < MINORSEPLENGTH; i++) {
        fprintf(fd, "%c", MINORSEPCHAR);
    }
    fputs("\n", fd);

    return;
}

static inline void
output_value_nonl(FILE *fd, char *label, char *string)
{
    int len = strlen(label);
    int i;

    fprintf(fd, "%s", label);

    for (i = len; i < LABELSEPPOS; i++) {
        fputs(" ", fd);
    }

    fprintf(fd, "%c ", LABELSEPCHAR);

    fprintf(fd, "%s", string);

    return;
}

static inline void
output_value(FILE *fd, char *label, char *string)
{
    output_value_nonl(fd, label, string);

    fputs("\n", fd);

    return;
}

static inline void
output_fill(FILE *fd, char *used)
{
    int i;
    
    for ( i = strlen(used); i < NAMEFIELDLEN; i++ ) {
        fputs(".", fd);
    }

}

/* @@ There is no performance layer identification/version output */
static inline void
output_events(FILE *fd, ps_hwpc_globals_t *globals, ps_hwpc_values_t *values)
{
    int i;
    char buf[32];

    output_banner(fd, EVENTBANNER);

    for ( i = 0; i < globals->events.numEvents; i++ ) {
        fprintf(fd, "%5d %s", i, globals->events.events[i].val);
        output_fill(fd, globals->events.events[i].val);
        fprintf(fd, " %20lld\n", values->total[i]);
    }

    /* Pad out to line up with event labels */
    fprintf(fd, "%6s", "");
    fputs(DOMAINLABEL, fd);
    output_fill(fd, DOMAINLABEL);
    
    switch( globals->domain ) {
    case USER:
        sprintf(buf, "user");
        break;
    case KERNEL:
        sprintf(buf, "kernel");
        break;
    case ALL:
        sprintf(buf, "all");
        break;
    default:
        sprintf(buf, "unknown");
        break;
    }

    fprintf(fd, "%21s\n", buf);

    fprintf(fd, "%6s", "");
    fputs(MPXLABEL, fd);
    output_fill(fd, MPXLABEL);
    fprintf(fd, "%21s", globals->multiplexing ? "yes" : "no");
    if ( globals->mpx_usec > 0 ) {
        fprintf(fd, " (%d microseconds)", globals->mpx_usec);
    }
    fputs("\n", fd);

    return;
}

static inline void
output_profile(FILE *fd, ps_hwpc_globals_t *globals, ps_hwpc_values_t *values)
{
    int map, pct, i, j;
    long totalsamples;
    char buf[32], address[32];

    output_banner(fd, PROFILEBANNER);

    switch ( globals->events.class ) {
    case PAPI:
        sprintf(buf, "%-21s", "PAPI");
        break;
    case HWPCNULL:
        sprintf(buf, "%-21s", "profil");
        break;
    case ITIMER:
        sprintf(buf, "%-21s", "itimer");
        break;
    default:
        sprintf(buf, "%-21s", "unknown");
        break;
    }
    output_value(fd, CLASSLABEL, buf);

    if ( globals->version != NULL ) {
        sprintf(buf, "%-21s", globals->version);
        output_value(fd, VERSIONLABEL, buf);
    }

    switch ( globals->domain ) {
    case USER:
        sprintf(buf, "%-21s", "user");
        break;
    case KERNEL:
        sprintf(buf, "%-21s", "kernel");
        break;
    case ALL:
        sprintf(buf, "%-21s", "all");
        break;
    default:
        sprintf(buf, "%-21s", "unknown");
        break;
    }
    output_value(fd, DOMAINLABEL, buf);

    switch ( globals->events.class ) {
    case PAPI:
    case ITIMER:
        sprintf(buf, "%-21s", globals->events.events[0].val);
        output_value(fd, EVENTLABEL, buf);
        sprintf(buf, "%-21d", globals->threshold);
        break;
    case HWPCNULL:
        output_value(fd, EVENTLABEL, "milliseconds");
        sprintf(buf, "%-21d", 10);  /* @@ Not correct */
        break;
    default:
        sprintf(buf, "unknown");
        break;
    }
    output_value(fd, THRESHOLDLABEL, buf);

    totalsamples = 0;
    for ( map = 0; map < globals->profiles.nmaps; map++ ) {
        for ( i = 0; i < globals->profiles.profmap[map].nbins; i++ ) {
            totalsamples += values->samples[map][i];
        }
    }

    sprintf(buf, "%-21ld", totalsamples);
    output_value(fd, TOTSAMPLABEL, buf);
    fputs("\n", fd);

    output_banner(fd, SAMPLESBANNER);

    sprintf(buf, "%-d", globals->profiles.nmaps);
    output_value(fd, NMODULESLABEL, buf);

    for ( map = 0; map < globals->profiles.nmaps; map++ ) {
        output_minorsep(fd);
        sprintf(buf, "%-s", globals->profiles.profmap[map].name);
        output_value(fd, MODULELABEL, buf);
        sprintf(buf, "%-p", globals->profiles.profmap[map].offset);
        output_value(fd, OFFSETLABEL, buf);
        for ( i = 0; i < globals->profiles.profmap[map].nbins; i++ ) {
            if ( values->samples[map][i] ) {
                /*
                 * @@ Here is where the scaling factor is hardcoded
                 * in (i*2).  Must be adapted to be more flexible,
                 * see also hwpc-xml.c.
                 */
                sprintf(address, "[%p]",
                        globals->profiles.profmap[map].offset+(i*2));
                sprintf(buf, "%-6hd", values->samples[map][i]);
                output_value_nonl(fd, address, buf);
                /*
                 * Add some asterisks for another visual display of
                 * relative count for each address.  We have about 40
                 * characters to play with on the line, so each asterisk
                 * counts for 2.5% of the total samples collected.
                 */
                pct = (int) ((values->samples[map][i] / (totalsamples * 0.025)))
                            + 0.5;
                fputs(" ", fd);
                for ( j = 1; j <= pct; j++ ) {
                    fputs("*", fd);
                }
                fputs("\n", fd);
            }
        }
    }

    return;
}

/*
 * @@ A bunch of this stuff can get factored out into routines
 * that can be used by multiple output formatters.
 */

int
_ps_hwpc_text_write(FILE *fd, char *generator, ps_hwpc_globals_t *globals,
		    ps_hwpc_values_t *values, unsigned long long ticks,
                    double usersecs, double systemsecs, char *command,
                    int close)
{
    char buf[256], *envval;
    ps_cpuid_t *p = &(globals->cpuid);
    time_t tp;
/*
    struct passwd pw;
    struct passwd *pwp = NULL;
    const int BUFLEN = 4096; // @@ sysconf(_SC_GETPW_R_SIZE_MAX)
    char buf[BUFLEN];
*/
    char *username = NULL;
    double cpuspeed;
    int ret;
    ps_cpu_cachearray_t c;
    ps_cpu_cachelist_t *cachelist;
    ps_cpu_cache_t *cache;
    int level;

    PS_DEBUG(PS_DEBUG_INFO, puts("Text output routine entered...");)

    time(&tp);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("Retrieved current time...");)

    ret = ps_cpuspeed(&cpuspeed);
    if ( ret != PS_SUCCESS ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("ps_cpuspeed failed\n", stderr);)
        return ret;
    }

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("Retrieved CPU speed...");)

    fprintf(fd, "%s\n", REPORTBANNER);
    fputs("\n", fd);
    sprintf(buf, "%s", TEXTREPORTVERSION);
    output_value(fd, VERSIONLABEL, buf);
#if _REENTRANT
    ctime_r(&tp, buf);
#else
    strcpy(buf, ctime(&tp));
#endif
    buf[strlen(buf)-1] = '\0';

    output_value(fd, CREATEDLABEL, buf);

    sprintf(buf, "%s (version %s)\n", LIBRARY_NAME, PACKAGE_VERSION);

    output_value(fd, GENERATORLABEL, buf);

    output_banner(fd, EXECUTIONBANNER);

#if HAVE_GETHOSTNAME
    gethostname(buf, 256);
    output_value(fd, HOSTLABEL, buf);
#endif

    sprintf(buf, "%d", getpid());
    output_value(fd, PIDLABEL, buf);
#if _REENTRANT
    sprintf(buf, "%d", values->myid);
#else
    sprintf(buf, "0");
#endif
    output_value(fd, THREADLABEL, buf);

    /* @@ Not written properly when using getpwuid().  Play it
     * safe and just use LOGNAME for now.
     */
    username = getenv("LOGNAME");
    if (username == NULL) {
        username = getenv("USER");
/*
        if (username == NULL) {
            getpwuid_r(geteuid(), &pw, &buf[0], BUFLEN, &pwp);
            if (pwp != NULL) {
                username = pw.pw_name;
            }
        }
*/
    }
    if (username == NULL) {
        username = "UNKNOWN";
    }
    sprintf(buf, "%s", username);
    output_value(fd, USERLABEL, buf);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("Retrieved user name...");)

    if ( command != NULL ) {
        sprintf(buf, "%s", command);
        output_value(fd, COMMANDLABEL, buf);
    }

    if ( globals->md5sum != NULL ) {
        sprintf(buf, "%s", globals->md5sum);
        output_value(fd, MD5LABEL, buf);
    }

    envval = getenv(ANNOTATIONFROMENV);
    if ( envval != NULL ) {
        output_value(fd, ANNOTATIONLABEL, envval);
    }

    /* @@ This is not printed by psprocess */
    if ( ticks != PS_UNKNOWN_TICKS ) {
        sprintf(buf, "%lld", ticks);
        output_value(fd, WALLCLOCKLABEL, buf);
    }

    fputs("\n", fd);

    output_banner(fd, PROCSYSBANNER);

    sprintf(buf, "%ld", sysconf(_SC_NPROCESSORS_ONLN));
    output_value(fd, NODECPUSLABEL, buf);

    sprintf(buf, "%s", vendorstring(p));
    output_value(fd, VENDORLABEL, buf);

    if ( p->brand_string != NULL ) {
        sprintf(buf, "%s", p->brand_string);
        output_value(fd, BRANDLABEL, buf);
    }

    sprintf(buf, "family %u, model %u, stepping %u",
            p->cpuid_family, p->cpuid_model, p->cpuid_stepping);
    output_value(fd, CPUIDINFOLABEL, buf);

    sprintf(buf, "%d", p->revision);
    output_value(fd, REVISIONLABEL, buf);

    sprintf(buf, "%.3f", cpuspeed);
    output_value(fd, CLOCKLABEL, buf);

    sprintf(buf, "%.2f",
            ((double)sysconf(_SC_PHYS_PAGES)*(double)getpagesize())/MEGABYTE);
    output_value(fd, MEMORYLABEL, buf);

    sprintf(buf, "%d",getpagesize()/KILOBYTE);
    output_value(fd, PAGESIZELABEL, buf);

    fputs("\n", fd);

    output_banner(fd, CACHEBANNER);

    c = p->cachearray;

    sprintf(buf, "%d", c.levels);
    output_value(fd, CACHELEVELSLABEL, buf);

    /* Ignore TLBs */
    for ( level = 0; level < c.levels; level++ ) {
        cachelist = c.cachelist[level];
        while ( cachelist != NULL ) {
            cache = cachelist->cache;
            switch ( cache->type ) {
            case ICACHE:
                output_minorsep(fd);
                sprintf(buf, "%s", ICACHEVALUE);
                output_value(fd, CACHETYPELABEL, buf);
                sprintf(buf, "%d", cache->specs.cache.size);
                output_value(fd, ICACHESIZELABEL, buf);
                sprintf(buf, "%d", cache->nway);
                output_value(fd, CACHEASSOCLABEL, buf);
                break;
            case ITRACECACHE:
                output_minorsep(fd);
                sprintf(buf, "%s", ITRACECACHEVALUE);
                output_value(fd, CACHETYPELABEL, buf);
                sprintf(buf, "%d", cache->specs.cache.size);
                output_value(fd, ITRACECACHESIZELABEL, buf);
                sprintf(buf, "%d", cache->nway);
                output_value(fd, CACHEASSOCLABEL, buf);
                break;
            case DCACHE:
                output_minorsep(fd);
                sprintf(buf, "%s", DCACHEVALUE);
                output_value(fd, CACHETYPELABEL, buf);
                sprintf(buf, "%d", cache->specs.cache.size);
                output_value(fd, DCACHESIZELABEL, buf);
                sprintf(buf, "%d", cache->nway);
                output_value(fd, CACHEASSOCLABEL, buf);
                break;
            case UCACHE:
                output_minorsep(fd);
                sprintf(buf, "%s", UCACHEVALUE);
                output_value(fd, CACHETYPELABEL, buf);
                sprintf(buf, "%d", cache->specs.cache.size);
                output_value(fd, UCACHESIZELABEL, buf);
                sprintf(buf, "%d", cache->nway);
                output_value(fd, CACHEASSOCLABEL, buf);
                break;
            default:
                break;
            }
            cachelist = cachelist->next;
        }
    }

    fputs("\n", fd);

    if ( globals->profiling ) {
        output_profile(fd, globals, values);
    }
    else {
        output_events(fd, globals, values);
    }

    /*
     * @@ This is an unresolved bug.  Closing the output
     * stream here can cause segfaults.
     */
    return PS_SUCCESS;

    if ( close ) {
        fclose(fd);
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("Text output routine complete (stream was%s closed)\n", \
                    close ? "" : " not"); \
    )

    return PS_SUCCESS;
}
