/* $Id: hwpc-xml.c,v 1.83 2013/04/30 16:44:44 ruiliu Exp $ */

/****************************************************************************
 *
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *       Copyright(C) 2003-2010, 2012, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * hwpc-xml.c : brute-force approach to writing out an XML file.
 *
 *              Could probably be done much more cleanly with
 *              XML libraries and a DOM tree built in memory,
 *              but this works for now.
 *
 * Author: Rick Kufrin
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if _REENTRANT
# if HAVE_PTHREAD
#  include <pthread.h>
# endif
#endif

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if HAVE_TIME_H
# include <time.h>
#endif

#if HAVE_PWD_H
# include <pwd.h>
#endif

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#include <perfsuite.h>
#include <psinternal.h>
#include "hwpc.h"
#include "hwpc-xml.h"

#ifdef MEGABYTE
#undef MEGABYTE
#endif
#define MEGABYTE 1048576.0
#define KILOBYTE 1024

static inline void
xml_start(FILE *fd, int standalone)
{
    fprintf(fd,"<?xml version=\"%s\" encoding=\"%s\" standalone=\"%s\" ?>\n",
            XMLVERSION, XMLENCODING, standalone ? "yes" : "no");

    fprintf(fd,"<!DOCTYPE %s \n%cSYSTEM \"%s%c%s.dtd\">\n",
            REPORTTAG, DEFAULTINDENTCHAR, PS_DTDDIR, DIRECTORY_SEP, REPORTDTD);

    return;
}

static inline void
report_start(FILE *fd, ps_hwpc_globals_t *globals, char *generator)
{
    fprintf(fd,"<%s version=\"%s\" generator=\"%s\" ",
            REPORTTAG, XMLREPORTVERSION, generator);
    fprintf(fd, "mode=\"%s\"\n", 
            globals->profiling ? PROFILEMODETAG : COUNTMODETAG);
    do_indent(fd,DEFAULTINDENT);
    do_indent(fd,DEFAULTINDENT);
    fprintf(fd,"xmlns:perfsuite=\"%s\">\n", REPORTXMLNS);

    return;
}

static inline void
report_end(FILE *fd)
{
    fprintf(fd,"</%s>\n", REPORTTAG);

    return;
}

static inline void
xml_write_date(FILE *fd, int indent)
{
    time_t tp;
    char buf[64];

    time(&tp);

    do_indent(fd,indent);
    fprintf(fd,"<%s>", DATETAG);
#if _REENTRANT
    ctime_r(&tp,buf);
#else
    strcpy(buf, ctime(&tp));
#endif
    buf[strlen(buf)-1] = '\0';
    fprintf(fd,"%s", buf);
    fprintf(fd,"</%s>\n", DATETAG);

    return;
}

static inline void
xml_write_hostname(FILE *fd, int indent)
{
    char hostname[256];
    char domainname[256];
    int ret;

#if HAVE_GETHOSTNAME
    ret = gethostname(hostname, 256);
#endif
#if HAVE_GETDOMAINNAME
    ret = getdomainname(domainname, 256);
#endif

    do_indent(fd, indent);
    fprintf(fd, "<%s>", HOSTNAMETAG);
    fprintf(fd, "%s", hostname);
    fprintf(fd, "</%s>\n", HOSTNAMETAG);

    do_indent(fd, indent);
    fprintf(fd, "<%s>", DOMAINNAMETAG);
    fprintf(fd, "%s", domainname);
    fprintf(fd, "</%s>\n", DOMAINNAMETAG);

    return;
}

static inline void
xml_write_ids(FILE *fd, ps_hwpc_values_t *values, int indent)
{
    do_indent(fd, indent);
    fprintf(fd, "<%s>", PIDTAG);
    fprintf(fd, "%d", getpid());
    fprintf(fd, "</%s>\n", PIDTAG);

    do_indent(fd, indent);
    fprintf(fd, "<%s>", THREADTAG);
#if _REENTRANT
    fprintf(fd, "%d", values->myid);
#else
    fprintf(fd, "0");
#endif
    fprintf(fd, "</%s>\n", THREADTAG);

    return;
}

/*
 * @@ This should use getpwuid_r() for _REENTRANT case.
 * Unfortunately, have had many problems with that function
 * (IA-64 especially) so $LOGNAME is used here.  When
 * called from hwpc.c, access is serialized through a mutex
 * so hopefully a reasonable workaround.
 */

static inline void
xml_write_username(FILE *fd, int indent)
{
/*
    struct passwd pw;
    struct passwd *pwp = NULL;
    const int BUFLEN = 4096; // @@ sysconf(_SC_GETPW_R_SIZE_MAX)
    char buf[BUFLEN];
*/
    char * username = NULL;

    if ( getenv(ANONYMOUSFROMENV) == NULL ) {
        username = getenv("LOGNAME");
        if (username == NULL) {
            username = getenv("USER");
/*
            if (username == NULL) {
                getpwuid_r(geteuid(), &pw, &buf[0], BUFLEN, &pwp);
                if (pwp != NULL) {
                    username = pw.pw_name;
                }
            }
*/
        }
        if (username == NULL) {
            username = "UNKNOWN";
        }
        do_indent(fd,indent);
        fprintf(fd, "<%s>", USERNAMETAG);
        fprintf(fd, "%s", username);
        fprintf(fd, "</%s>\n", USERNAMETAG);
    }

    return;
}

static inline void
xml_write_command(FILE *fd, int indent, char *command, char* md5sum)
{
    if ( command != NULL ) {
        do_indent(fd, indent);
        fprintf(fd, "<%s", COMMANDTAG);
        if ( md5sum != NULL ) {
            fprintf(fd, " %s=\"%s\"", COMMANDMD5ATT, md5sum);
        }
        fprintf(fd, ">%s</%s>\n", command, COMMANDTAG);
    }

    return;
}

static inline void
xml_write_annotation(FILE *fd, int indent)
{
    char *envval;

    envval = getenv(ANNOTATIONFROMENV);
    if ( envval != NULL ) {
        do_indent(fd,indent);
        fprintf(fd, "<%s>", ANNOTATIONTAG);
        fprintf(fd, "%s", envval);
        fprintf(fd, "</%s>\n", ANNOTATIONTAG);
    }

    return;
}

static inline void
xml_write_wallclock(FILE *fd, int indent, ps_hwpc_globals_t *globals,
                    unsigned long long ticks, double cpuspeed)
{

    if ( ticks != PS_UNKNOWN_TICKS ) {
        do_indent(fd, indent);
        fprintf(fd, "<%s units=\"ticks\">", WALLCLOCKTAG);
        fprintf(fd, "%lld", ticks);
        fprintf(fd, "</%s>\n", WALLCLOCKTAG);
    }

    return;
}

static inline void
xml_write_cputime(FILE *fd, int indent, double usersecs, double systemsecs)
{
    do_indent(fd, indent);
    fprintf(fd, "<%s units=\"seconds\">\n", CPUTIMETAG);
    indent += DEFAULTINDENT;
    do_indent(fd, indent);
    fprintf(fd, "<%s>%.6f</%s>\n", CPUTIMEUSERTAG, usersecs, CPUTIMEUSERTAG);
    do_indent(fd, indent);
    fprintf(fd, "<%s>%.6f</%s>\n", CPUTIMESYSTEMTAG, systemsecs,
                                  CPUTIMESYSTEMTAG);
    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fprintf(fd, "</%s>\n", CPUTIMETAG);

    return;
}


static inline void
xml_write_machinecpucount(FILE *fd, int indent, ps_cpuid_t *p)
{
    do_indent(fd, indent);
    fprintf(fd, "<%s>", CPUCOUNTTAG);
    fprintf(fd, "%ld", sysconf(_SC_NPROCESSORS_ONLN));
    fprintf(fd, "</%s>\n", CPUCOUNTTAG);

    return;
}

static inline void
xml_write_cachespecs(FILE *fd, int indent, int level, ps_cpu_cache_t *cache)
{
    do_indent(fd, indent);
    indent += DEFAULTINDENT;
    fprintf(fd, "<%s type=\"", CACHETAG);
    switch ( cache->type ) {
    case ICACHE:      fputs("instruction", fd); break;
    case ITRACECACHE: fputs("instruction trace", fd); break;
    case DCACHE:      fputs("data", fd); break;
    case UCACHE:      fputs("unified", fd); break;
    default: break;
    }
    fprintf(fd, "\" level=\"%d\">\n", level);

    switch ( cache->type ) {
    case ICACHE:
    case DCACHE:
    case UCACHE:
        do_indent(fd, indent);
        fprintf(fd, "<%s units=\"KB\">%d</%s>\n",
                CACHESIZETAG, cache->specs.cache.size, CACHESIZETAG);
        do_indent(fd, indent);
        fprintf(fd, "<%s units=\"B\">%d</%s>\n",
                LINESIZETAG, cache->specs.cache.linesize, LINESIZETAG);
        break;
    case ITRACECACHE:
        do_indent(fd, indent);
        fprintf(fd, "<%s units=\"KuOps\">%d</%s>\n",
                CACHESIZETAG, cache->specs.cache.size, CACHESIZETAG);
        break;
    default:
        break;
    }
    do_indent(fd, indent);
    fprintf(fd, "<%s type=\"%s\">%d</%s>\n", ASSOCIATIVITYTAG,
            cache->nway == FULLY_ASSOCIATIVE?"full":"set",
            cache->nway, ASSOCIATIVITYTAG);

    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fprintf(fd, "</%s>\n", CACHETAG);

    return;
}

static inline void
xml_write_machinecachespecs(FILE *fd, int indent, ps_cpu_cachearray_t caches)
{
    ps_cpu_cachelist_t *cachelist;
    ps_cpu_cache_t *cache;
    int level;

    do_indent(fd, indent);
    fprintf(fd, "<%s>\n", CACHEINFOTAG);

    indent += DEFAULTINDENT;
    do_indent(fd, indent);
    fprintf(fd, "<%s>", CACHELEVELSTAG);
    fprintf(fd, "%d", caches.levels);
    fprintf(fd, "</%s>\n", CACHELEVELSTAG);

    for ( level=0; level < caches.levels; level++ ) {
        cachelist = caches.cachelist[level];
        while ( cachelist != NULL ) {
            cache = cachelist->cache;
            switch ( cache->type ) {
            case ICACHE:
            case ITRACECACHE:
            case DCACHE:
            case UCACHE:
                xml_write_cachespecs(fd, indent, level+1, cache);
                break;
            default:
                break;
            }
            cachelist = cachelist->next;
        }
    }

    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fprintf(fd, "</%s>\n", CACHEINFOTAG);

    return;
}

static inline void
xml_write_tlbpagesizes(FILE *fd, int indent, ps_cpu_cache_t *cache)
{
    unsigned int size;

    do_indent(fd, indent);
    fprintf(fd, "<%s>\n", PAGESIZESTAG);
  
    indent += DEFAULTINDENT;
    for ( size = 0; size < cache->specs.tlb.npagesize; size++ ) {
        do_indent(fd, indent);
        fprintf(fd, "<%s units=\"KB\">%d</%s>\n",
                PAGESIZETAG, cache->specs.tlb.pagesize[size], PAGESIZETAG);
    }
    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fprintf(fd, "</%s>\n", PAGESIZESTAG);

    return;
}

static inline void
xml_write_tlbspecs(FILE *fd, int indent, int level, ps_cpu_cache_t *cache)
{
    do_indent(fd, indent);
    indent += DEFAULTINDENT;
    fprintf(fd, "<%s type=\"", TLBTAG);
    switch ( cache->type ) {
    case ITLB: fputs("instruction", fd); break;
    case DTLB: fputs("data", fd); break;
    case UTLB: fputs("unified", fd); break;
    default: break;
    }
    fprintf(fd, "\" level=\"%d\">\n", level);

    do_indent(fd, indent);
    fprintf(fd, "<%s>%d</%s>\n", 
            TLBENTRIESTAG, cache->specs.tlb.entries, TLBENTRIESTAG);

    do_indent(fd, indent);
    fprintf(fd, "<%s type=\"%s\">%d</%s>\n", ASSOCIATIVITYTAG,
            cache->nway == FULLY_ASSOCIATIVE?"full":"set",
            cache->nway, ASSOCIATIVITYTAG);

    /*
      xml_write_tlbpagesizes(fd, indent, cache);
    */

    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fprintf(fd, "</%s>\n", TLBTAG);

    return;
}

static inline void
xml_write_machinetlbspecs(FILE *fd, int indent, ps_cpu_cachearray_t caches)
{
    ps_cpu_cachelist_t *cachelist;
    ps_cpu_cache_t *cache;
    int level;

    do_indent(fd, indent);
    fprintf(fd, "<%s>\n", TLBINFOTAG);

    indent += DEFAULTINDENT;

    for ( level = 0; level < caches.levels; level++ ) {
        cachelist = caches.cachelist[level];
        while ( cachelist != NULL ) {
            cache = cachelist->cache;
            switch ( cache->type ) {
            case ITLB:
            case DTLB:
            case UTLB:
                xml_write_tlbspecs(fd, indent, level + 1, cache);
                break;
            default:
                break;
            }
            cachelist = cachelist->next;
        }
    }

    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fprintf(fd, "</%s>\n", TLBINFOTAG);
    return;
}

/* Same information as printed in psinv */
static inline void
xml_write_machinecpuinfo(FILE *fd, int indent, ps_cpuid_t *p, double cpuspeed)
{
    do_indent(fd, indent);
    fprintf(fd, "<%s>\n", CPUINFOTAG);

    indent += DEFAULTINDENT;

    do_indent(fd, indent);
    fprintf(fd, "<%s>", VENDORTAG);
    switch ( p->vendor ) {
    case INTEL: fputs("Intel", fd); break;
    case AMD:   fputs("AMD", fd); break;
    case IBM:   fputs("IBM", fd); break;
    default:    fputs("unknown", fd); break;
    }
    fprintf(fd, "</%s>\n", VENDORTAG);

    if ( p->brand_string != NULL ) {
        do_indent(fd, indent);
        fprintf(fd, "<%s>%s</%s>\n", BRANDTAG, p->brand_string, BRANDTAG);
    }

    do_indent(fd, indent);
    fprintf(fd, "<%s>family: %u, model: %u, stepping: %u</%s>\n", CPUIDINFOTAG, p->cpuid_family, p->cpuid_model, p->cpuid_stepping, CPUIDINFOTAG);

    do_indent(fd, indent);
    fprintf(fd, "<%s>%d</%s>\n", REVISIONTAG, p->revision, REVISIONTAG);

    do_indent(fd, indent);
    fprintf(fd, "<%s units=\"MHz\">%.3f</%s>\n",
            CLOCKSPEEDTAG, cpuspeed, CLOCKSPEEDTAG);

    xml_write_machinecachespecs(fd, indent, p->cachearray);
    xml_write_machinetlbspecs(fd, indent, p->cachearray);

    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fprintf(fd, "</%s>\n", CPUINFOTAG);
}

static inline void
xml_write_machinememoryspecs(FILE *fd, int indent)
{
    do_indent(fd, indent);
    fprintf(fd, "<%s units=\"MB\">", MEMORYINFOTAG);
    fprintf(fd, "%.2f", ((double)sysconf(_SC_PHYS_PAGES)*(double)getpagesize())/MEGABYTE);
    fprintf(fd, "</%s>\n", MEMORYINFOTAG);

    return;
}

static inline void
xml_write_syspagesize(FILE *fd, int indent)
{
    do_indent(fd, indent);
    fprintf(fd, "<%s units=\"KB\">", SYSPAGESIZETAG);
    fprintf(fd, "%d", getpagesize()/KILOBYTE);
    fprintf(fd, "</%s>\n", SYSPAGESIZETAG);

    return;
}

static inline void
xml_write_machineinfo(FILE *fd, int indent, ps_cpuid_t *p, double cpuspeed)
{
    do_indent(fd, indent);
    fprintf(fd, "<%s>\n", MACHINEINFOTAG);

    indent += DEFAULTINDENT;

    xml_write_machinecpucount(fd, indent, p);
    xml_write_machinecpuinfo(fd, indent, p, cpuspeed);
    /*xml_write_machinecachespecs(fd, indent, p->cacheinfo);
      xml_write_machinetlbspecs(fd, indent, p->cacheinfo);*/
    xml_write_machinememoryspecs(fd, indent);
    xml_write_syspagesize(fd, indent);

    indent -= DEFAULTINDENT;

    do_indent(fd, indent);
    fprintf(fd, "</%s>\n", MACHINEINFOTAG);

    return;
}

static inline void
xml_write_profileinfo(FILE *fd, int indent, ps_hwpc_globals_t *globals,
                      ps_hwpc_values_t *values)
{
    int i, map;
    long totalsamples;

    do_indent(fd, indent);
    fprintf(fd, "<%s class=\"", PROFILEDATATAG);
    switch( globals->events.class ) {
    case PAPI:
        fputs("PAPI\"", fd);
        break;
    case PERFMON:
        fputs("perfmon\"", fd);
        break;
    case HWPCNULL:
        fputs("profil\"", fd);
        break;
    case ITIMER:
        fputs("itimer\"", fd);
        break;
    case UNKNOWN:
    default:
        fputs("unknown\"", fd);
        break;
    }

    if ( globals->version != NULL ) {
        fprintf(fd, " version=\"%s\"", globals->version);
    }

    fputs(">\n", fd);

    if ( ( globals->events.class == PAPI ) || 
         ( globals->events.class == PERFMON ) ||
         ( globals->events.class == HWPCNULL ) ||
         ( globals->events.class == ITIMER) ) {
        indent += DEFAULTINDENT;
        do_indent(fd, indent);
        fprintf(fd, "<%s>\n", PROFILEINFOTAG);
        indent += DEFAULTINDENT;
        do_indent(fd, indent);
        switch ( globals->events.class ) {
        case PAPI:
        case PERFMON:
        case ITIMER:
            /* @@ This has to change to allow for natives */
            fprintf(fd, "<%s", PROFILEEVENTTAG);
            fputs(" domain=\"", fd);
            switch(globals->domain) {
            case USER:
                fputs("user", fd);
                break;
            case KERNEL:
                fputs("kernel", fd);
                break;
            case ALL:
                fputs("all", fd);
                break;
            default:
                fputs("unknown", fd);
                break;
            }
            fputs("\"", fd);
            fputs(" type=\"preset\"", fd);
            fprintf(fd, " name=\"%s\"", globals->events.events[0].val);
            /* @@ Assumes only one event allowed for profiling */
            fprintf(fd, " derived=\"%s\"",
                    globals->events.events[0].derived ? "yes" : "no");
            fputs(">\n", fd);
            indent += DEFAULTINDENT;
            do_indent(fd, indent);
            fprintf(fd, "<%s>%d</%s>\n",
                    THRESHOLDTAG, globals->threshold, THRESHOLDTAG);
            indent -= DEFAULTINDENT;
            do_indent(fd, indent);
            fprintf(fd, "</%s>\n", PROFILEEVENTTAG);
            break;
        case HWPCNULL:
            fprintf(fd, "<%s", PROFILEEVENTTAG);
            fputs(" domain=\"all\" type=\"preset\" name=\"ITIMER_PROF\"", fd);
            fprintf(fd, " derived=\"no\">\n");
            indent += DEFAULTINDENT;
            do_indent(fd, indent);
            fprintf(fd, "<%s>%d</%s>\n", THRESHOLDTAG, 
#if ( IA32 || PPC32 || PPC64)
                                         10,   /* milliseconds */
#elif IA64
                                          1,   /* milliseconds */
#else
# error "No profiling rate defined for this machine"
#endif
                                         THRESHOLDTAG);
            indent -= DEFAULTINDENT;
            do_indent(fd, indent);
            fprintf(fd, "</%s>\n", PROFILEEVENTTAG);
            break;
        default:
            fputs("**** ERROR ****\n", fd);
            break;
        }
        indent -= DEFAULTINDENT;
        do_indent(fd, indent);
        fprintf(fd, "</%s>\n", PROFILEINFOTAG);
    }

    totalsamples = 0;
    for ( map = 0; map < globals->profiles.nmaps; map++ ) {
        for ( i = 0; i < globals->profiles.profmap[map].nbins; i++ ) {
            totalsamples += values->samples[map][i];
        }
    }

    do_indent(fd, indent);
    fprintf(fd, "<%s>%ld</%s>\n",
            TOTALSAMPLESTAG, totalsamples, TOTALSAMPLESTAG);

    for ( map = 0; map < globals->profiles.nmaps; map++ ) {
        do_indent(fd, indent);
        fprintf(fd, "<%s %s=\"%s\" %s=\"%lx\">\n", MODULETAG,
                MODULEFILEATT, globals->profiles.profmap[map].name,
                MODULEOFFSETATT,
                (long unsigned int) globals->profiles.profmap[map].offset);
        indent += DEFAULTINDENT;
        do_indent(fd, indent);
        fprintf(fd, "<%s pcbase=\"16\">\n", SAMPLESTAG);
        indent += DEFAULTINDENT;

        for ( i = 0; i < globals->profiles.profmap[map].nbins; i++ ) {
            if ( values->samples[map][i] ) {
                do_indent(fd, indent);
                /* 
                 * @@ Here is where the scaling/compression factor
                 * is hard-coded (i*2).  This must be made more
                 * flexible, and the factor included in the XML/DTD.
                 */
                /* @@ Will %lx format always work? (see also hwpc.c) */
                fprintf(fd, "<%s pc=\"%lx\">%hu</%s>\n", SAMPLETAG,
                        (long unsigned int) globals->profiles.profmap[map].offset+(i*2),
                        values->samples[map][i], SAMPLETAG);
            }
        }

        indent -= DEFAULTINDENT;
        do_indent(fd, indent);
        fprintf(fd, "</%s>\n", SAMPLESTAG);
        indent -= DEFAULTINDENT;
        do_indent(fd, indent);
        fprintf(fd, "</%s>\n", MODULETAG);
    }

    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fprintf(fd, "</%s>\n", PROFILEDATATAG);

    return;
}

static inline void
xml_write_eventinfo(FILE *fd, int indent, ps_hwpc_globals_t *globals,
                    ps_hwpc_values_t *values)
{
    int i;

    do_indent(fd, indent);
    fprintf(fd,"<%s ", HWPCEVENTDATATAG);
    fputs("class=\"", fd);
    switch( globals->events.class ) {
    case PAPI:
        fputs("PAPI\" ", fd);
        break;
    case PERFMON:
        fputs("perfmon\" ", fd);
        break;
    case HWPCNULL:
        fputs("null\" ", fd);
        break;
    case UNKNOWN:
    default:
        fputs("unknown\" ", fd);
        break;
    }

    if ( globals->version != NULL ) {
        fprintf(fd, "version=\"%s\" ", globals->version);
    }

    fputs("domain=\"", fd);
    switch( globals->domain ) {
    case USER:
        fputs("user\" ", fd);
        break;
    case KERNEL:
        fputs("kernel\" ", fd);
        break;
    case ALL:
        fputs("all\" ", fd);
        break;
    default:
        fputs("unknown\" ", fd);
        break;
    }

    fprintf(fd, "multiplexed=\"%s\"", globals->multiplexing ? "yes" : "no");
    if ( globals->mpx_usec > 0 ) {
        fprintf(fd, " multiplex_usec=\"%d\"", globals->mpx_usec);
    }
    fputs(">\n", fd);

    indent += DEFAULTINDENT;

    for ( i = 0; i < globals->events.numEvents; i++ ) {
        do_indent(fd, indent);
        fprintf(fd, "<%s ", HWPCEVENTTAG);
        fputs("type=\"", fd);
        switch ( globals->events.events[i].type ) {
        case PRESET:
            fputs("preset\" ", fd);
            fprintf(fd, "name=\"%s\" ", globals->events.events[i].val);
            fprintf(fd, "derived=\"%s\"",
                    globals->events.events[i].derived ? "yes" : "no");
            break;
        case NATIVE:
            fputs("native\" ", fd);
            fprintf(fd, "name=\"%s\" ", globals->events.events[i].val);
            fputs("derived=\"no\"", fd);
            break;
        default:
            fputs("unknown\" ", fd);
            break;
        }
        fputs(">", fd);
        fprintf(fd, "%lld", values->total[i]);
        fprintf(fd, "</%s>\n", HWPCEVENTTAG);
    }

    indent -= DEFAULTINDENT;
    do_indent(fd, indent);
    fprintf(fd, "</%s>\n", HWPCEVENTDATATAG);

    return;
}

/*
 * Main driver for writing out hwpcreport XML documents.
 *
 * >>> ANY CHANGES TO THIS ROUTINE SHOULD ALSO BE <<<
 * >>> REFLECTED IN AND TESTED WITH hwpc-dtdgen.c <<<
 * >>>      SO THAT THE DTD REMAINS IN SYNC       <<<
 */

int
_ps_hwpc_xml_write(FILE *fd, char *generator, ps_hwpc_globals_t *globals,
		   ps_hwpc_values_t *values, unsigned long long ticks,
                   double usersecs, double systemsecs, char *command,
                   int close)
{
    int indent=0;
    int standalone = DEFXMLSTANDALONE;

    PS_DEBUG(PS_DEBUG_INFO, puts("XML output routine entered...");)

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("Tracking function completion:");)

    xml_start(fd, standalone);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_start...");)

    report_start(fd, globals, generator);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("report_start...");)

    indent += DEFAULTINDENT;

    xml_write_date(fd, indent);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_date...");)

    xml_write_hostname(fd, indent);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_hostname...");)

    xml_write_ids(fd, values, indent);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_ids...");)

    xml_write_username(fd, indent);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_username...");)

    xml_write_command(fd, indent, command, globals->md5sum);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_command...");)

    xml_write_annotation(fd, indent);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_annotation...");)

    xml_write_wallclock(fd, indent, globals, ticks, globals->mhz);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_wallclock...");)

    xml_write_cputime(fd, indent, usersecs, systemsecs);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_cputime...");)

    xml_write_machineinfo(fd, indent, &(globals->cpuid), globals->mhz);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_machineinfo...");)

    if ( globals->profiling ) {
        xml_write_profileinfo(fd, indent, globals, values);
        PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_profileinfo...");)
    }
    else {
        xml_write_eventinfo(fd, indent, globals, values);
        PS_DEBUG(PS_DEBUG_VERBOSE, puts("xml_write_eventinfo...");)
    }

    report_end(fd);

    PS_DEBUG(PS_DEBUG_VERBOSE, puts("report_end...");)

    if ( close ) {
        fclose(fd);
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("XML output routine complete (stream was%s closed)\n", \
                    close ? "" : " not"); \
    )

    return PS_SUCCESS;
}
