/* $Id: hwpc.c,v 1.215 2013/05/07 21:51:49 ruiliu Exp $ */

/****************************************************************************
 *
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *     Copyright(C) 2002-2010, 2012-2013, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * Generic support for hardware performance counting and profiling.
 *
 * The idea is to handle the user specifications, and rely on underlying
 * counter-package specific support to actually do all the work under the
 * covers. The underlying support is routed through a standard small set of
 * routines that are accessed through function pointers.
 *
 * All user requirements are communicated either through an XML configuration
 * file or environment variables.
 *
 * Final results are returned in an XML document (default) or plain text
 * output.
 *
 * Author: Rick Kufrin
 *
 * @@ There is little error-checking for correct sequence of calls right now.
 *
 * @@ New calls to add:
 * int ps_hwpc_xml_write(char *filename, char *annotation, 
 *                       int mode, ps_hwpc_value_t *values,
 *                       long long ticks, double *usersecs, 
 *                       double *systemsecs);
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if _REENTRANT
# if HAVE_PTHREAD
#  include <pthread.h>
# else
#  error "_REENTRANT defined but Pthread support not available"
# endif
#endif

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#if HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif

#if HAVE_FCNTL_H
# include <fcntl.h>
#endif

#if HAVE_ERRNO_H
# include <errno.h>
#endif

#if HAVE_SIGNAL_H
# include <signal.h>
#endif

#if DEBUG
# include "psdebug.h"
#endif

#include <psinternal.h>
#include "hwpc.h"
#include "hwpc-xml.h"
#include "hwpc-text.h"
#include "eventlist.h"
#include "exclude.h"

#include "null.h"

#if HAVE_SETITIMER
# include "hwpc-itimer.h"
#endif

#if HAVE_PAPI
# include "hwpc-papi.h"
#endif

#if HAVE_PERFMON
# include "hwpc-perfmon.h"
#endif

#if HAVE_PERFMON2
# include "hwpc-perfmon2.h"
#endif

/**********************************************************************
           Variables used in this module.
 **********************************************************************/

/*
 * Global variables that can be shared among threads.
 * These should only be written by the main thread.
 */

static ps_hwpc_globals_t globals;
static short library_initialized = 0;
static char *pidfile = NULL;

#if _REENTRANT

/* Condition variable and associated mutex for initialization control */
static pthread_cond_t  initialized_condvar  = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t initialized_mutex    = PTHREAD_MUTEX_INITIALIZER;

/* Global count of threads */
static unsigned int    max_thread_num       = 0;
static pthread_mutex_t max_thread_num_mutex = PTHREAD_MUTEX_INITIALIZER;

/*
 * Array of pthread IDs, used when signal handling is in effect.
 * These are protected by the same mutex for max_thread_num and
 * are allocated in chunks of size THREAD_ID_INCR.
 */

#define THREAD_ID_INCR 64
static pthread_t       *thread_ids          = NULL;
static pthread_t       main_thread_id;
static sigset_t        thread_sigmask;

/* I/O mutex */
static pthread_mutex_t io_mutex             = PTHREAD_MUTEX_INITIALIZER;

#else /* ! _REENTRANT */

/*
 * The counter values for the single-threaded case.
 */

static ps_hwpc_values_t *values = NULL;

#endif /* _REENTRANT */

/**********************************************************************
         Prototypes for internal functions used in this module.
         The functions are defined near the end of this file.
 **********************************************************************/

BEGIN_C_DECLS

#if HAVE_ON_EXIT
static void ps_hwpc_psrun_fini      PARAMS((int status, void *arg));
#else
static void ps_hwpc_psrun_fini      PARAMS((void));
#endif

static int allocate_values          PARAMS((ps_hwpc_values_t *values));
static int create_pidfile           PARAMS((void));
static int delete_pidfile           PARAMS((void));
static FILE *generate_fd            PARAMS((const char *base,
                                            const char *suffix));
static int read_config              PARAMS((char **eventfile));
static void set_domain              PARAMS((void));
static int set_enabled              PARAMS((int mapc, ps_addrmap_t **mapv,
                                            char *db));
static inline void free_tokens      PARAMS((char **tokens));
static int set_exitsignals          PARAMS((void));
static RETSIGTYPE exit_handler      PARAMS((int signum, siginfo_t *si,
                                            void *p));
static int set_loadmap              PARAMS((int mapc, ps_addrmap_t **mapv));
static void set_md5                 PARAMS((void));
static int set_package              PARAMS((char *eventfile));


#if _REENTRANT
static void allocate_key            PARAMS((void));
static void free_values             PARAMS((void *data));
static ps_hwpc_values_t *get_values PARAMS((void));
static inline int register_thread_id PARAMS((pthread_t id));
static inline int set_thread_response PARAMS((void));
#endif /* _REENTRANT */

/* @@ Shouldn't do this, need to fix includes/declarations */
extern int _ps_parseEventListFromStream PARAMS((FILE *fin,
                                                ps_hwpc_globals_t *globals));
extern char * PS_HWPC_DEF_CONFIGFILE;

END_C_DECLS

/* ============================================================== *
 *                   EXPORTED FUNCTIONS                           *
 * ============================================================== */


/*
 * Initialize everything in preparation for monitoring, but don't do anything
 * else.  Select the particular package that will be used.
 *
 * In the threaded case, this routine should only be called by the main
 * (initial) thread, before other threads attempt to access any of the other
 * routines in this library.
 * @@ Verification still needs to be done to handle/enforce access
 * correctly.  Issues to consider are fork/signal safety.
 *
 * The only thread-specific things done in this routine are the creation of the
 * values key and a lock/release of a mutex to flag library initialization.
 * If signal-handling is enabled, a signal mask is also created to be used by
 * all threads except the main thread, which assumes the role of the designated
 * signal handler.
 */

int
ps_hwpc_init(void)
{
    char *env, *excludedbfileenv, *eventfile;
    ps_addrmap_t **mapv;
    int mapc, ret;
#if _REENTRANT
    int pthreadret;
#endif
  
#if 0
    if ( library_initialized ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("sequence error detected in ps_hwpc_init\n", stderr);)
        return PS_HWPC_ESEQUENCE;
    }
#endif

#if DEBUG
    ps_setdebuglevel();
#endif

    /*
     * First, are we actually monitoring? If not, let's get outta here.
     * set_enabled() checks both the PS_HWPC environment variable as
     * well as the exclusion database if supplied.
     */
    mapv = NULL;
    excludedbfileenv = getenv(EXCLUDEDBFROMENV);
    if ( excludedbfileenv != NULL ) {
        ret = _ps_addrmap(0, &mapc, &mapv);
        if ( ret != PS_SUCCESS ) {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fputs("error retrieving address map for exclusion\n", \
                           stderr);)
            excludedbfileenv = NULL;
            mapv = NULL;
        }
    }
    ret = set_enabled(mapc, mapv, excludedbfileenv);
    if ( ret == 0 ) {
        return PS_SUCCESS;
    }

    /* Initialize the real time clock */
    PS_DEBUG(PS_DEBUG_INFO, puts("Initializing real time clock...");)
    ret = ps_rtcinit();
    if ( ret != PS_SUCCESS ) {
        PS_DEBUG(PS_DEBUG_FATAL, fputs("RTC initialization failed.\n", stderr);)
        return ret;
    }

    /* Initialize global data structures */
    _ps_hwpc_init_globals(&globals);

    PS_DEBUG(PS_DEBUG_INFO, puts("Initialized global data");)

    /* Arrange for signal handling */
    ret = set_exitsignals();
    if ( ret != PS_SUCCESS ) {
        return ret;
    }

    PS_DEBUG(PS_DEBUG_INFO, printf("Signal handling %s.\n", \
                                   globals.signaling ? "on" : "off");)

    /* Dynamically set the default event config file at run time. */
    ret = set_def_config();
    if ( ret != PS_SUCCESS ) {
        return ret;
    }

    /* See what we want to count */
    ret = read_config(&eventfile);
    if ( ret != PS_SUCCESS ) {
        return ret;
    }

    PS_DEBUG(PS_DEBUG_INFO, puts("Configuration file read");)
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("Read %d events\n", globals.events.numEvents);)

    /*
     * If the user didn't specify any events, then switch
     * to the "do-nothing" hwpc-null layer.  Note that this
     * is different from turning off the library completely
     * by setting PS_HWPC to "off" in the environment.  The
     * difference is that an XML file will be written that
     * contains all non-hardware performance data (machine
     * arch, wallclock ticks, etc).
     */
    if ( globals.events.numEvents == 0 ) {
        globals.events.class = HWPCNULL;
        PS_DEBUG(PS_DEBUG_INFO, \
                 puts("No events specified, using NULL layer...");)
    }

    /* Assign the package-specific function pointers */
    ret = set_package(eventfile);
    free(eventfile);
    if ( ret != PS_SUCCESS ) {
        return ret;
    }

    /* Check that we haven't exceeded the allowable number of events */
    if ( globals.control.get_max_counters != NULL ) {
        ret = (*globals.control.get_max_counters)(&globals);
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Performance software supports %d events max.\n", \
                 ret);)
        /* profil() will return 0, so check for negative return */
        if ( ret < 0 ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fprintf(stderr, "FATAL: get_max_counters returned %d\n", \
                             ret);)
            return PS_HWPC_EPERF;
        }
        if ( globals.events.numEvents > ret ) {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     printf("Too many events requested (%d)!\n", \
                            globals.events.numEvents);)
            return PS_HWPC_EEVENTSEXCEEDED;
        }
    }
    else {
        PS_DEBUG(PS_DEBUG_WARNING, \
                 fputs("Warning: get_max_counters() function is NULL\n", \
                       stderr);)
    }

    /* See if the user changed the default counting domain */
    set_domain();

    /* See if the user has requested MD5 checksums */
    set_md5();

    if ( globals.profiling ) {
        if ( globals.control.start_profiling == NULL ) {
            return PS_HWPC_ENOTSUPPORTED;
        }

        if ( globals.events.numEvents > 1 ) {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fputs("Only one event allowed when profiling ", stdout); \
                     printf("(%d requested)!\n", globals.events.numEvents);)
            free(eventfile);
            return PS_HWPC_EEVENTSEXCEEDED;
        }

        /* Has a profiling threshold been set in the environment? */
        env = getenv(THRESHOLDFROMENV);
        if ( env != NULL ) {
            globals.threshold = atoi(env);
        }

        /*
         * Prepare load map information for profiling.
         * First, acquire the load map if not already done.
         */
        if ( mapv == NULL ) {
            ret = _ps_addrmap(0, &mapc, &mapv);
            if ( ret != PS_SUCCESS ) {
                free(eventfile);
                return ret;
            }
        }

        ret = set_loadmap(mapc, mapv);
        if ( ret != PS_SUCCESS ) {
            return ret;
        }
    }
    else {
        /* Free the address map if not done already */
        if ( mapv != NULL ) {
            _ps_addrmap_free(mapc, mapv);
            mapv = NULL;
        }

        /* See if the user specified a multiplexing interval */
        env = getenv(MPXUSECFROMENV);
        if ( env != NULL ) {
            globals.mpx_usec = atoi(env);
        }
    }

#if _REENTRANT
    PS_DEBUG(PS_DEBUG_INFO, \
             puts("Allocating values TSD key (pthread_once)");)

    /* This key is never deleted. */
    allocate_key();

    PS_DEBUG(PS_DEBUG_INFO, \
             puts("Values TSD allocation (pthread_once) done");)
#endif

    /*
     * This is the only call to the underlying package from
     * the generic layer in this function.  Thread-specific
     * data has not been allocated yet, so can't use it.
     */
    PS_DEBUG(PS_DEBUG_INFO, puts("Calling package init routine");)
    ret = (*globals.control.init_counters)(&globals);

    /*
     * @@ Should this be locked first thing in this function and
     * unlocked at the very end?  We have to be really careful
     * here because of forked programs - note the commenting out
     * of the check for initialized at the beginning.
     */
    if ( ret == PS_SUCCESS ) {
#if _REENTRANT
        pthreadret = pthread_mutex_lock(&initialized_mutex);
        if ( pthreadret != 0 ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fprintf(stderr, "Error locking mutex: %d\n", pthreadret);)
            return PS_ESYS;
        }
        PS_DEBUG(PS_DEBUG_INFO, puts("ps_hwpc_init: acquired init_mutex");)

        /* Stash the main thread ID away for possible use in signal handler */
        main_thread_id = pthread_self();
#endif
        create_pidfile();
        library_initialized = 1;
#if _REENTRANT
        pthreadret = pthread_mutex_unlock(&initialized_mutex);
        if ( pthreadret != 0 ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fprintf(stderr, "Error unlocking mutex: %d\n", \
                             pthreadret);)
            return PS_ESYS;
        }
        PS_DEBUG(PS_DEBUG_INFO, puts("ps_hwpc_init: released init_mutex");)
        pthreadret = pthread_cond_broadcast(&initialized_condvar);
        if ( pthreadret != 0 ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fprintf(stderr, "Error in cond_broadcast: %d\n", \
                             pthreadret);)
            return PS_ESYS;
        }
        PS_DEBUG(PS_DEBUG_INFO, \
                 puts("ps_hwpc_init: broadcast init signal");)
#endif
    }

    return ret;
}


/*
 * Start tracking counters by storing the current values.
 * Unlike the package-specific start_counters routines,
 * this one can be called many times by each thread.
 *
 * The result of this routine is that the "lastval"
 * member of each values array is set to the value of
 * the counters at the time of call.  In the case of
 * the first call, it would be zero.  Subsequent calls
 * read the values from the read_counters_PACKAGE
 * interface.
 */

int
ps_hwpc_start()
{
    int ret;
#if _REENTRANT
    ps_hwpc_values_t *values;
    int pthreadret;
#endif

    if ( ! globals.on ) {
        return PS_SUCCESS;
    }

#if _REENTRANT
    values = get_values(); /* @@ No error check */
    if ( values->started == 0 ) {
        /*
         * Wait on a condition variable that flags whether the
         * library has been initialized in ps_hwpc_init().  This is
         * added in as an extra safety check for proper sequence
         * of calls.
         */
        pthreadret = pthread_mutex_lock(&initialized_mutex);
        if ( pthreadret != 0 ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fprintf(stderr, \
                     "ps_hwpc_start: error acquiring init_mutex: %d\n", \
                     pthreadret);)
            return PS_ESYS;
        }
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("ps_hwpc_start [PID %d]: acquired init_mutex\n", \
                        getpid());)
        while ( ! library_initialized ) {
            pthreadret = pthread_cond_wait(&initialized_condvar,
                                           &initialized_mutex);
        }
        pthreadret = pthread_mutex_unlock(&initialized_mutex);
        if ( pthreadret != 0 ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fprintf(stderr, \
                     "ps_hwpc_start: error releasing init_mutex: %d\n", \
                     pthreadret);)
            return PS_ESYS;
        }
        PS_DEBUG(PS_DEBUG_INFO, puts("ps_hwpc_start: released init_mutex");)

        values->started = 1;
    }
#else
    if ( ! library_initialized ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("sequence error detected in ps_hwpc_start\n", stderr);)
        return PS_HWPC_ESEQUENCE;
    }

    if ( values == NULL ) {
        /*
         * These are currently freed in ps_hwpc_shutdown().
         * They are not freed in ps_hwpc_stop() because it
         * is not considered freeing a resource and also
         * there are instances (N.B. psrun) where stop is
         * explicitly not called but shutdown is.
         */
        values = (ps_hwpc_values_t *) malloc(sizeof(ps_hwpc_values_t));
        if ( values == NULL ) {
            PS_DEBUG(PS_DEBUG_FATAL,
                     fputs("Error allocating memory in ps_hwpc_start\n", \
                           stderr);)
            return PS_ENOMEM;
        }

        ret = allocate_values(values);
        if ( ret != PS_SUCCESS ) {
            PS_DEBUG(PS_DEBUG_FATAL,
                     fputs("Error reported in allocate_values in ps_hwpc_start\n", stderr);)
            free(values);
            return ret;
        }
    }
#endif /* ! _REENTRANT */

    /*
     * Start the real time clock first.  Since we start timing prior to
     * activating the performance measurement, we choose to include
     * the measurement setup time in the overall wall clock (i.e., we
     * "time the counters", rather than "count the timers").
     *
     * All subsequent operations should follow this sequence:
     *   On activation:   start the RTC, then the monitoring
     *   On deactivation: stop the monitoring, then the RTC
     */

    ps_rtc(&(values->starttime));
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("hwpc_start: start RTC set at %lld\n", values->starttime);)

    PS_DEBUG(PS_DEBUG_INFO, printf("hwpc_start: values->state = %d\n", \
                                   values->state);)
    if ( values->state == UNINITIALIZED ) {
        /* First time through - set things up */
        PS_DEBUG(PS_DEBUG_INFO, puts("hwpc_start: initializing");)

        if ( globals.profiling ) {
            /* Initialize for profiling */
            ret = (*globals.control.init_profiling)(&globals, values->samples);
            if ( ret != PS_SUCCESS ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("Error return from init_profiling\n", stderr);)
                return ret;
            }
        }
        else {
            /* Start aggregate counting */
            PS_DEBUG(PS_DEBUG_INFO, puts("hwpc_start: starting counters");)
            ret = (*globals.control.start_counters)(&globals);
            if ( ret != PS_SUCCESS ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("Error return from start_counters\n", stderr);)
                return ret;
            }
        }
    }

    if ( globals.profiling ) {
        /* Actually start profiling */
        ret = (*globals.control.start_profiling)(&globals, values->samples);
        if ( ret != PS_SUCCESS ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("ps_hwpc_start: can't start profiling!\n", stderr);)
            return ret;
        }
    }
    else {
        /* Subsequent times for counting, just update the starting point */
        /* @@ The state should be SUSPENDED here, but we don't
         * currently check.
         * @@ Note that successive calls to ps_hwpc_start() by the
         * user without a call to ps_hwpc_suspend() will result in
         * the data corresponding to the first call being lost.
         */
        ret = (*globals.control.read_counters)(values->current);
        if ( ret != PS_SUCCESS ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("ps_hwpc_start: can't read counters!\n", stderr);)
            return ret;

        }
        memcpy(values->startval, values->current, 
               globals.events.numEvents * sizeof(ps_hwpc_value_t));
    }

    values->state = RUNNING;

    return PS_SUCCESS;
}


/*
 * Read the counters and add the difference since the last time
 * started to the running total.
 *
 * @@ Should suspension of counters that are not running be an error?
 * Right now, it isn't - we just return.
 *
 * @@ Extension would be to add counter "groups" (distinct
 * collections of counters)
 */

int
ps_hwpc_suspend(void)
{
#if _REENTRANT
    ps_hwpc_values_t *values;
#endif
    unsigned long long currenttime;
    int i, ret;

    if ( ! globals.on ) {
        return PS_SUCCESS;
    }

#if _REENTRANT
    values = get_values();
    if ( values->state == UNINITIALIZED ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("ps_hwpc_suspend: sequence error detected", stderr); \
                 fprintf(stderr, " by thread %d\n", values->myid);)
        return PS_HWPC_ESEQUENCE;
    }
#else
    /*
     * Have to separately test values itself, because if ps_hwpc_start
     * hasn't been called, no memory will have been allocated yet.
     */
    if ( values == NULL ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("ps_hwpc_suspend: sequence error\n", stderr); )
        return PS_HWPC_ESEQUENCE;
    }
#endif

    if ( values->state == RUNNING ) {
        if ( globals.profiling ) {
            ret = (*globals.control.suspend_profiling)(&globals,
                                                       values->samples);
            if ( ret != PS_SUCCESS ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("ps_hwpc_suspend: can't suspend ", stderr); \
                         fputs("profiling!\n", stderr);)
                return(ret);
            }
        }
        else {
            ret = (*globals.control.read_counters)(values->current); 
            if ( ret != PS_SUCCESS ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("ps_hwpc_suspend: can't read ", stderr); \
                         fputs("counters!\n", stderr);)
                return ret;
            }

            PS_DEBUG(PS_DEBUG_INFO, puts("Read counters:");)
            for ( i = 0; i < globals.events.numEvents; i++ ) {
                PS_DEBUG(PS_DEBUG_INFO, \
                         printf("Event %d\n", i); \
                         printf("\t  Prior [suspend]: %lld ", \
                                values->startval[i]); \
                         printf("(last) %lld (current) %lld (total)\n", \
                                values->current[i], values->total[i]);)
                values->total[i] += values->current[i] - values->startval[i];
                values->startval[i] = values->current[i];
                PS_DEBUG(PS_DEBUG_INFO, \
                         printf("\tUpdated [suspend]: %lld ", \
                                values->startval[i]); \
                         printf("(last) %lld (current) %lld (total)\n", \
                                 values->current[i], values->total[i]);)
            }
        }

        /* Update the wall clock time (RTC) running total */
        ps_rtc(&currenttime);
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("hwpc_suspend: RTC current is %lld\n", currenttime);)
        values->totaltime += currenttime - values->starttime; 
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("\tRTC updated [suspend]: %lld\n", values->totaltime);)
    }

    values->state = SUSPENDED;

    return PS_SUCCESS;
}


/*
 * Return the current values of the counters (counting mode only).
 * The counters remain running and no changes to internally-maintained
 * counter values are made.
 *
 * @@ Using when in profiling mode is currently an error, but this
 * may be possible to support.
 */

int
ps_hwpc_read(ps_hwpc_value_t *uservals)
{
#if _REENTRANT
    ps_hwpc_values_t *values;
#endif
    int i, ret;

    if ( ! globals.on ) {
        return PS_SUCCESS;
    }

    if ( globals.profiling ) {
        return PS_HWPC_EMODE;
    }

#if _REENTRANT
    values = get_values();
    if (values->state == UNINITIALIZED) {
        PS_DEBUG(PS_DEBUG_WARNING, \
                 fputs("warning: hwpc_read called but internal values ", \
                       stdout); \
                 puts("not initialized");)
        return PS_HWPC_ESEQUENCE;
    }
#else
    if ( values == NULL ) {
        PS_DEBUG(PS_DEBUG_WARNING, \
                 fputs("warning: hwpc_read called but internal values ", \
                       stdout); \
                 puts("not initialized");)
        return PS_HWPC_ESEQUENCE;
    }
#endif

    ret = (*globals.control.read_counters)(uservals); 
    /* @@ This was previously checked within the PS_DEBUG statement.
     * Was there a particular reason?
     */
    if ( ret != PS_SUCCESS ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("ps_hwpc_read: can't read counters!\n", stderr);)
        return ret;
    }

    for ( i = 0; i < globals.events.numEvents; i++ ) {
        uservals[i] = values->total[i] + uservals[i] - values->startval[i];
    }

    return ret;
}


/*
 * Write out the result file. Note that this does NOT
 * perform a final read of the counters unless they
 * are currently running.
 *
 * Close down the (possibly per-thread) counters.
 */

int
ps_hwpc_stop(const char *fname)
{
    int ret, minsecs;
    FILE *fout;
    unsigned long long currenttime;
    char *env, *suffix = PS_SUFFIX_DEFAULT;
    double cpusecs, usersecs, systemsecs;
    ps_time_t usertime, systime;
    volatile enum { XML, TEXT } outformat = XML;

#if _REENTRANT
    ps_hwpc_values_t *values;
#endif
    int i, close = 1;

    if ( ! globals.on ) {
        return PS_SUCCESS;
    }

#if _REENTRANT
    values = get_values();
    if ( values->state == UNINITIALIZED ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("ps_hwpc_stop: sequence error detected ", stderr); \
                 fprintf(stderr, "by thread %d\n", values->myid);)
        return PS_HWPC_ESEQUENCE;
    }
#else
    /* @@ See above. */
    if ( values == NULL ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("ps_hwpc_stop: sequence error detected\n", stderr);)
        return PS_HWPC_ESEQUENCE;
    }
#endif

    /*
     * Always stop if counting, regardless of current status (running or not).
     * The reason is that the PerfSuite model is to start the counters only
     * once (for efficiency reasons), and work with deltas from that point on.
     * Therefore, even if the current state is suspended, underneath the
     * counters are still running, so we stop them before a final read.
     */
    if ( ! globals.profiling ) {
        ret = (*globals.control.stop_counters)(values->current);
        if ( ret != PS_SUCCESS ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("ps_hwpc_stop: error in stop_counters\n", stderr);)
            return ret;
        }
    }

    /*
     * Update performance values only if measurement is currently running.
     */
    if ( values->state == RUNNING ) {
        if ( globals.profiling ) {
            ret = (*globals.control.stop_profiling)(&globals, values->samples);
            if ( ret != PS_SUCCESS ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("ps_hwpc_stop: can't stop ", stderr); \
                         fputs("profiling!\n", stderr);)
                return ret;
            }
            /* @@ Add debugging output to show sampling statistics here */
        }
        else {
            PS_DEBUG(PS_DEBUG_INFO, puts("FINAL read counters:");)
            for ( i = 0; i < globals.events.numEvents; i++ ) {
                PS_DEBUG(PS_DEBUG_INFO, \
                         printf("Event %d\n", i); \
                         printf("\t  Prior [stop]: %lld ", \
                                values->startval[i]); \
                         printf("(last) %lld (current) %lld (total)\n", \
                                values->current[i], values->total[i]);)
                values->total[i] += values->current[i] - values->startval[i];
                values->startval[i] = values->current[i];
                PS_DEBUG(PS_DEBUG_INFO, \
                         printf("\tUpdated [stop]: %lld (last) ", \
                                values->startval[i]); \
                         printf("%lld (current) %lld (total)\n", \
                                values->current[i], values->total[i]);)
            }
        }

        /* Note the final RTC value and add to the running total */
        ps_rtc(&currenttime);
        values->totaltime += currenttime - values->starttime;
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("hwpc_stop: current RTC is %lld\n", currenttime);)
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("clock ticks elapsed = %lld\n", values->totaltime);)
    }
    else {
        PS_DEBUG(PS_DEBUG_INFO, \
                 fputs("ps_hwpc_stop: counters not running ", stdout); \
                 puts("(no copy of final read)");)
    }

    values->state = TERMINATED;

    /*
     * Determine the CPU usage of this process.
     */
    if ( ps_cpuusage(0, &usertime, &systime) != PS_SUCCESS ) {
        PS_DEBUG(PS_DEBUG_WARNING, \
                 puts("Error in retrieving process time");)
    }
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("CPU time secs/msecs (user: %ld/%ld, sys: %ld/%ld)\n", \
                    usertime.secs, usertime.msecs, systime.secs, \
                    systime.msecs);)
    usersecs = usertime.secs + (usertime.msecs*1.e-6);
    systemsecs = systime.secs + (systime.msecs*1.e-6);
    cpusecs = usersecs + systemsecs;
    PS_DEBUG(PS_DEBUG_INFO, printf("CPU time (secs) = %.3f\n", cpusecs);)

    /*
     * Check to see if CPU time used meets minimum requirement for output.
     */
    env = getenv(TIMELIMITFROMENV);
    if ( env != NULL ) {
        minsecs = atoi(env);
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Using time limit from environment: %d seconds\n", \
                        minsecs);)
        if ( cpusecs < minsecs ) {
            PS_DEBUG(PS_DEBUG_INFO, \
                puts("Minimum CPU time not reached. Output disabled.");)
            return PS_SUCCESS;
        }
    }

    /*
     * Determine the correct output format.
     */
    env = getenv(FORMATFROMENV);
    if ( env != NULL ) {
        if ( strcasecmp(env, PS_FORMATSTRING_TEXT) == 0 ) {
            outformat = TEXT;
            suffix = PS_SUFFIX_TEXT;
        }
        else if ( strcasecmp(env, PS_FORMATSTRING_XML) == 0 ) {
            outformat = XML;
            suffix = PS_SUFFIX_XML;
        }
        else {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fprintf(stderr, \
                             "Warning: environment variable %s value ", \
                             FORMATFROMENV); \
                     fprintf(stderr, "\"%s\" unrecognized.  Using XML.\n", \
                             env);)
        }
    }

#if _REENTRANT
    /* Serialize the I/O to try avoiding filesystem problems. */
    pthread_cleanup_push((void *) pthread_mutex_unlock, &io_mutex);
    if ( pthread_mutex_lock(&io_mutex) != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, fputs("Error locking I/O mutex.\n", stderr);)
        return PS_ESYS;
    }
#endif

    /*
     * Obtain the file descriptor to be used for output.
     */
    fout = _ps_hwpc_select_stream(fname, suffix, globals.exebase, &close);
    if ( fout == NULL ) {
        PS_DEBUG(PS_DEBUG_WARNING, \
                 puts("ps_hwpc_stop: error opening output stream");)
#if _REENTRANT
        if ( pthread_mutex_unlock(&io_mutex) != 0 ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("Error unlocking I/O mutex.\n", stderr);)
            return PS_ESYS;
        }
#endif
        return PS_HWPC_ESTREAM;
    }

    /*
     * Use the appropriate output routine based on format.
     */
    switch ( outformat ) {
    case XML:
        ret = _ps_hwpc_xml_write(fout, LIBRARY_NAME, &globals, values, 
                                 values->totaltime, usersecs, systemsecs,
                                 globals.exebase, close);
        break;
    case TEXT:
        ret = _ps_hwpc_text_write(fout, LIBRARY_NAME, &globals, values, 
                                  values->totaltime, usersecs, systemsecs,
                                  globals.exebase, close);
        break;
    default:
        ret = 0;
        PS_DEBUG(PS_DEBUG_WARNING,  puts("Unknown output format: no I/O done");)
        break;
    }

#if _REENTRANT
    /* Unlock the mutex before returning. */
    pthread_cleanup_pop(1);
#endif

    return ret;
}


/*
 * Clean everything up and shutdown the counter software.
 */

int
ps_hwpc_shutdown(void)
{
#if _REENTRANT
    int pthreadret;
#endif

    if ( ! globals.on ) {
        return PS_SUCCESS;
    }

    (*globals.control.destroy_counters)();

    PS_DEBUG(PS_DEBUG_INFO, \
             puts("ps_hwpc_shutdown: returned from destroy counters");)

#if _REENTRANT
    pthreadret = pthread_mutex_lock(&initialized_mutex);
    if ( pthreadret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("ps_hwpc_shutdown: error acquiring ", stderr); \
                 fprintf(stderr, "initialization mutex: %d\n", pthreadret);)
        return PS_ESYS;
    }
#endif
    delete_pidfile();
    /* @@ Previously only done in threaded version */
    library_initialized = 0;
#if _REENTRANT
    pthreadret = pthread_mutex_unlock(&initialized_mutex);
    if ( pthreadret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("ps_hwpc_shutdown: error releasing ", stderr); \
                 fprintf(stderr, "initialization mutex: %d\n", pthreadret);)
        return PS_ESYS;
    }
#endif

#if _REENTRANT
    free(thread_ids);
    thread_ids = NULL;
#endif

#if ! _REENTRANT
    free(values);
    values = NULL;
#endif

    return PS_SUCCESS;

}


/*
 * Return the current "state" of the library.  This will probably
 * be expanded in the future, for right now returns the current
 * value of the initialization flag.
 */

int
ps_hwpc_state(int *state)
{
#if _REENTRANT
    int ret;

    ret = pthread_mutex_lock(&initialized_mutex);
    if ( ret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error locking mutex: %d\n", ret);)
        return PS_ESYS;
    }
#endif

    *state = library_initialized;

#if _REENTRANT
    ret = pthread_mutex_unlock(&initialized_mutex);
    if ( ret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fprintf(stderr, "Error locking mutex: %d\n", ret);)
        return PS_ESYS;
    }
#endif

    return PS_SUCCESS;
}


/*
 * Hey... could this get any easier? 
 */

int
ps_hwpc_psrun(void)
{
    int ret;

#if ! (HAVE_ATEXIT || HAVE_ON_EXIT)
    return PS_ENOTIMPLEMENTED;
#endif

    ret = ps_hwpc_init();
    if ( ret != PS_SUCCESS ) {
        return ret;
    }
    PS_DEBUG(PS_DEBUG_INFO, \
             puts("ps_hwpc_psrun: initialized successfully.");)

    ret = ps_hwpc_start();
    if ( ret != PS_SUCCESS ) {
        return ret;
    }
    PS_DEBUG(PS_DEBUG_INFO, \
             puts("ps_hwpc_psrun: started successfully.");)

#if HAVE_ON_EXIT
    ret = on_exit(ps_hwpc_psrun_fini, NULL);
#elif HAVE_ATEXIT
    ret = atexit(ps_hwpc_psrun_fini);
#endif
    if ( ret != 0 ) {
        return PS_ESYS;
    }

    return PS_SUCCESS;
}


static void
#if HAVE_ON_EXIT
ps_hwpc_psrun_fini(int status, void *arg)
#else
ps_hwpc_psrun_fini(void)
#endif
{
    int ret;

    ret = ps_hwpc_stop(PS_HWPC_DEF_FNAME);
    if ( ret == PS_SUCCESS ) {
        ret = ps_hwpc_shutdown();
    }

    return;
    /* It seems that an exit handler should not call exit() in itself,
     * see "man atexit".  So Rui changed this to return. 2012-11-14.
     */

#if HAVE_ON_EXIT
    exit(status);
#else
    exit(ret);
#endif
}


/*
 * Inquiry routines.
 *
 * @@ Should it be an error if the library has not been initialized?
 * Currently, it is mainly to encourage correct library use but
 * strictly speaking these routines could just return zero event
 * counts and no change to the names vector.
 */

int
ps_hwpc_numevents(int *numevents)
{
    if ( ! library_initialized ) {
        return PS_HWPC_ESEQUENCE;
    }

    if ( ! globals.on ) {
        *numevents = 0;
        return PS_SUCCESS;
    }

    *numevents = globals.events.numEvents;

    return PS_SUCCESS;
}


int
ps_hwpc_eventnames(char ***names)
{
    int event;

    if ( ! library_initialized ) {
        return PS_HWPC_ESEQUENCE;
    }

    if ( ! globals.on ) {
        return PS_SUCCESS;
    }

    if ( *names == NULL ) {
        *names = (char **) malloc(globals.events.numEvents * sizeof(char *));
        if ( *names == NULL ) {
            return PS_ENOMEM;
        }
        PS_DEBUG(PS_DEBUG_INFO, \
                 puts("ps_hwpc_eventnames: allocating eventname memory");)
    }
    else {
        PS_DEBUG(PS_DEBUG_INFO, \
                 puts("ps_hwpc_eventnames: no memory allocation needed");)
    }

    for ( event = 0; event < globals.events.numEvents; event++ ) {
        (*names)[event] = strdup(globals.events.events[event].val);
    }

    return PS_SUCCESS;
}


int
ps_hwpc_version(int *major, int *minor, int *release, int *type)
{
    if ( major != NULL ) {
        *major = LIBRARY_MAJOR_VERSION;
    }

    if ( minor != NULL ) {
        *minor = LIBRARY_MINOR_VERSION;
    }

    if ( release != NULL ) {
        *release = LIBRARY_RELEASE_LEVEL;
    }

    if ( type != NULL ) {
        /**type = PS_RELEASE_TYPE;*/
    }

    return PS_SUCCESS;
}


/* ============================================================== *
 *                  FILE-PRIVATE FUNCTIONS                        *
 * ============================================================== */


/*
 * Initialize all global data to default values or to values
 * that will be used by all threads during execution.
 *
 * Non-static to allow access to the global initializer by
 * package-specific layers.
 */

void
_ps_hwpc_init_globals(ps_hwpc_globals_t *globals)
{
    int ret;

    globals->on           = 1;
    globals->domain       = USER;
    globals->multiplexing = 0;
    globals->profiling    = 0;
    globals->signaling    = 0;
    globals->mpx_usec     = 0;  /* Zero indicates default is used */
    globals->threshold    = PS_HWPC_DEFAULT_THRESHOLD;
    globals->version      = NULL;
    ret = _ps_getexebase(0, &(globals->exebase));
    if ( ret != PS_SUCCESS ) {
        globals->exebase = NULL;
    }
    globals->md5sum       = NULL;
    _ps_cpuid(&(globals->cpuid));
    ps_cpuspeed(&(globals->mhz));

    return;
}


static int
allocate_values(ps_hwpc_values_t *values)
{
    int map;
    unsigned long profmem;
    ps_hwpc_sample_t *currentaddr;

    /* @@ startval, current, and total are really only used
     * when in counting mode.  This should be adjusted and
     * also in the free_values() routine.
     */
    values->startval = (ps_hwpc_value_t *) 
        malloc(globals.events.numEvents * sizeof(ps_hwpc_value_t));
    if ( values->startval == NULL ) {
        return PS_ENOMEM;
    }
    memset(values->startval, 0,
           globals.events.numEvents * sizeof(ps_hwpc_value_t));

    values->current = (ps_hwpc_value_t *) 
        malloc(globals.events.numEvents * sizeof(ps_hwpc_value_t));
    if ( values->current == NULL ) {
        free(values->startval);
        return PS_ENOMEM;
    }

    values->total = (ps_hwpc_value_t *) 
        malloc(globals.events.numEvents * sizeof(ps_hwpc_value_t));
    if ( values->total == NULL ) {
        free(values->current); free(values->startval);
        return PS_ENOMEM;
    }
    memset(values->total, 0,
           globals.events.numEvents * sizeof(ps_hwpc_value_t));

    if ( globals.profiling ) {
        values->samples = (ps_hwpc_sample_t **)
            malloc(globals.profiles.nmaps * sizeof(ps_hwpc_sample_t *));
        if ( values->samples == NULL ) {
            free(values->current); free(values->startval); free(values->total);
            return PS_ENOMEM;
        }
        profmem = 0;
        for ( map = 0; map < globals.profiles.nmaps; map++ ) {
            profmem += globals.profiles.profmap[map].nbins * 
                            sizeof(ps_hwpc_sample_t);
        }
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Allocating %lu bytes for profiling\n", profmem);)
        values->samples[0] = (ps_hwpc_sample_t *) malloc(profmem);
        if ( values->samples[0] == NULL ) {
            free(values->current); free(values->startval);
            free(values->total); free(values->samples);
            return PS_ENOMEM;
        }
        memset(values->samples[0], 0, profmem);
        currentaddr = values->samples[0] + globals.profiles.profmap[0].nbins;
        PS_DEBUG(PS_DEBUG_INFO, printf("Buffer address for map 0: %p\n", \
                                       values->samples[0]);)
        for ( map = 1; map < globals.profiles.nmaps; map++ ) {
            values->samples[map] = currentaddr;
            currentaddr += globals.profiles.profmap[map].nbins;
            PS_DEBUG(PS_DEBUG_INFO, printf("Buffer address for map %d: %p\n", \
                                           map, values->samples[map]);)
        }
        PS_DEBUG(PS_DEBUG_INFO, printf("End of profiling buffer range: %p\n", \
                                       ((char *) values->samples[0]) + \
                                                 profmem - 1);)
        values->overflow = (unsigned long *) 
            malloc(globals.profiles.nmaps * sizeof(unsigned long));
        if ( values->overflow == NULL ) {
            free(values->current); free(values->startval);
            free(values->total); free(values->samples);
            return PS_ENOMEM;
        }
        memset(values->overflow, 0,
               globals.profiles.nmaps * sizeof(unsigned long));
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Allocated and zeroed overflow memory for %d maps\n", \
                 globals.profiles.nmaps);)
        values->outofrange = 0;
        PS_DEBUG(PS_DEBUG_INFO, \
                 puts("Set out-of-range sample counter to zero");)
    }

    values->totaltime = 0;
    values->state = UNINITIALIZED;

    return PS_SUCCESS;
}

#if _REENTRANT

/*
 * Counter values are maintained as thread-specific data.
 * Allocate a key for access.  This is called as a
 * pthread_once routine from ps_hwpc_init().
 */

static void
allocate_key(void)
{
    PS_DEBUG(PS_DEBUG_INFO, fputs("Allocating TSD key... ", stdout);)
    PS_DEBUG(PS_DEBUG_VERBOSE, printf("initial value %d ", globals.datakey);)
    if ( pthread_key_create(&globals.datakey, free_values) != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Error initializing thread-specific data key\n", \
                       stderr);)
    }
    PS_DEBUG(PS_DEBUG_INFO, puts("successful.");)
    PS_DEBUG(PS_DEBUG_VERBOSE, printf("New value %d ", globals.datakey);)

    return;
}


/*
 * Thread-specific data destructor function
 */

static void
free_values(void *data)
{
    ps_hwpc_values_t *values = (ps_hwpc_values_t *) data;

    free(values->startval);
    free(values->current);
    free(values->total);
    if ( globals.profiling ) {
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("hwpc: freeing TSD profile buffer at %p\n", \
                        values->samples[0]);)
        free(values->samples[0]);
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("hwpc: freeing TSD profile offsets at %p\n", \
                        values->samples);)
        free(values->samples);
    }
    free(values);

    pthread_setspecific(globals.datakey, NULL);

    return;
}


/*
 * Register an internal thread ID in a global list.  This routine should
 * be called while holding a mutex.
 */

static inline int
register_thread_id(pthread_t id)
{
    /* 
     * Memory allocated in this routine is freed in ps_hwpc_shutdown().
     */
    if ( (max_thread_num % THREAD_ID_INCR) == 0 ) {
        /* @@ Assumes that realloc(NULL) behaves like straight malloc() */
        thread_ids = (pthread_t *)
            realloc(thread_ids,
                    (max_thread_num + THREAD_ID_INCR) * sizeof(pthread_t));
        if ( thread_ids != NULL ) {
            PS_DEBUG(PS_DEBUG_INFO, \
                     printf("Thread %d reallocated space for %d IDs\n",\
                            max_thread_num, max_thread_num + THREAD_ID_INCR);)
        }
        else {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fputs("Error allocating memory for thread IDs\n",\
                           stderr);)
            return PS_ENOMEM;
        }
    }

    thread_ids[max_thread_num] = id;
    PS_DEBUG(PS_DEBUG_VERBOSE, \
             fprintf(stderr, "Thread %d ID entered as %lu\n", \
                     max_thread_num, thread_ids[max_thread_num]);)

    return PS_SUCCESS;
}


/*
 */

static inline int
set_thread_response(void)
{
    int ret;

    ret = pthread_sigmask(SIG_BLOCK, &thread_sigmask, NULL);
    if ( ret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Error setting thread signal mask.\n", stderr);)
        return PS_ESYS;
    }

    ret = pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
    if ( ret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Error setting thread canceltype.\n", stderr);)
        return PS_ESYS;
    }


    return PS_SUCCESS;
}


/*
 * This function does one of two things, depending on whether
 * it has been called before.  If it has, it returns the
 * thread-specific values data.  If it hasn't, it allocates
 * the data, sets it to zero, increments the total thread 
 * count, and stores the thread-specific ID away.
 */

static ps_hwpc_values_t *
get_values(void)
{
    ps_hwpc_values_t *values;
    pthread_t me;

    values = (ps_hwpc_values_t *) pthread_getspecific(globals.datakey);
    PS_DEBUG(PS_DEBUG_VERBOSE, printf("get_values: TSD data = %p\n", values);)
    PS_DEBUG(PS_DEBUG_VERBOSE, printf("get_values: datakey = %x\n", \
                                      globals.datakey);)
    if ( values == NULL ) { 
        values = (ps_hwpc_values_t *) malloc(sizeof(ps_hwpc_values_t));
        if ( values == NULL ) {
            return (ps_hwpc_values_t *) NULL;
        }
        pthread_setspecific(globals.datakey, values);
        PS_DEBUG(PS_DEBUG_VERBOSE, \
                 printf("get_values: setting TSD data to %p\n", values);)

        /* Increment the number of threads that have called. */
        if ( pthread_mutex_lock(&max_thread_num_mutex) != 0 ) {
            PS_DEBUG(PS_DEBUG_FATAL, fputs("Error locking mutex.\n", stderr);)
            return (ps_hwpc_values_t *) NULL;
        }

        /*
         * Store the pthread ID if a signal handler has been installed 
         * and set the cancellation type to asynchronous so that
         * we will respond immediately on cancellation.
         *
         */
         /* Changed to always register the thread, to be used later
          * in psrun_fini() to cancel the children threads in case
          * they were not terminated by pthread_exit() or other threads
          * calling pthread_cancel(). - Rui, 2012-11-14.
          */
        //if ( globals.signaling ) {
        if ( 1 ) {
            me = pthread_self();
            if ( register_thread_id(me) != PS_SUCCESS ) {
                PS_DEBUG(PS_DEBUG_WARNING, \
                         fputs("register_thread_id() failed\n", stderr);)
            }

            /*
             * Set up cancellation state/signal masks for all but
             * the main thread.
             */
            if ( pthread_equal(me, main_thread_id) == 0 ) {
                if ( set_thread_response() != PS_SUCCESS ) {
                    PS_DEBUG(PS_DEBUG_WARNING, \
                             fputs("set_thread_response() failed\n", stderr);)
                }
            }

        }

        values->myid = max_thread_num++;
  
        if ( pthread_mutex_unlock(&max_thread_num_mutex) != 0 ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("Error unlocking mutex.\n", stderr);)
            return (ps_hwpc_values_t *) NULL;
        }

        if ( allocate_values(values) != PS_SUCCESS ) {
            free(values);
            return (ps_hwpc_values_t *) NULL;
        }

        values->started = 0;
    }
    else {
        PS_DEBUG(PS_DEBUG_VERBOSE, \
                 printf("get_values: TSD exists for thread %d, returning\n", \
                        values->myid);)
    }

    return values;
}

#endif /* _REENTRANT */


/*
 * Create a PID/hostname sentinel file.
 *
 * The format of this filename is "PID.hostname"
 */

static int
create_pidfile(void)
{
    char hostname[256], *dir;
    int pidfilefd, ret;

    dir = getenv(PIDDIRFROMENV);
    if ( dir == NULL ) {
        PS_DEBUG(PS_DEBUG_INFO, puts("No PID file requested.");)
        return PS_SUCCESS;
    }

    /*
     * Check that directory is given as an absolute path name.
     * If not, it is possible that the application could change
     * directories and invalidate our notion of where the file sits.
     */
    if ( dir[0] != DIRECTORY_SEP ) {
        PS_DEBUG(PS_DEBUG_WARNING, \
                 fprintf(stderr, "PID file directory \"%s\" is not ", dir); \
                 fputs("an absolute path.  No PID file created.\n", stderr);)
        return PS_EINVAL;
    }

    pidfile = malloc(PATH_MAX);
    if ( pidfile == NULL ) {
        return PS_ENOMEM;
    }
    PS_DEBUG(PS_DEBUG_VERBOSE, printf("Initial PID file bytes: %d\n",PATH_MAX);)

    ret = gethostname(hostname, 255);
    if ( ret == 0 ) {
        /* Avoid "//" */
        if ( strlen(dir) == 1) {
            sprintf(pidfile, "%c%d.%s", DIRECTORY_SEP, getpid(), hostname);
        }
        else {
            sprintf(pidfile, "%s%c%d.%s",
                    dir, DIRECTORY_SEP, getpid(), hostname);
        }
        pidfilefd = creat(pidfile, 0644);
        if ( pidfilefd != -1 ) {
            close(pidfilefd);
            PS_DEBUG(PS_DEBUG_INFO, \
                     printf("Created PID file \"%s\"\n", pidfile);)
            pidfile = realloc(pidfile, strlen(pidfile) + 1);
            if ( pidfile == NULL ) {
                PS_DEBUG(PS_DEBUG_WARNING, \
                         fputs("Error: realloc of PID file space.\n", stderr);)
            }
            else {
                PS_DEBUG(PS_DEBUG_INFO, printf("Final PID file bytes: %d\n", \
                                               strlen(pidfile) + 1);)
            }
        }
        else {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fprintf(stderr, \
                             "Error creating PID file \"%s\": %s\n", \
                                           pidfile, strerror(errno));)
            free(pidfile);
            pidfile = NULL;
            return PS_ESYS;
        }
    }

    return PS_SUCCESS;
}


/*
 * Remove the PID sentinel file previously created.
 */

static int
delete_pidfile(void)
{
    int ret;
    pid_t pid, original_pid;
    char *ptr;

    /* 
     * Check first that this corresponds to our PID (could be
     * called after a fork(), as done by psrun).
     */
    if ( pidfile != NULL ) {
        ptr = strrchr(pidfile, DIRECTORY_SEP);
        if ( ptr != NULL ) {
            ptr++;
            ret = sscanf(ptr, "%d.%*s", &original_pid);
            if ( ret == 1 ) {
                pid = getpid();
                if ( original_pid != pid ) {
                    PS_DEBUG(PS_DEBUG_INFO, \
                             printf("PIDfile: not unlinking (%d != %d)\n",
                                    original_pid, pid);)
                    return PS_SUCCESS;
                }
            }
            else {
                PS_DEBUG(PS_DEBUG_WARNING, \
                         fprintf(stderr, \
                                 "Warning: could not extract PID: \"%s\"\n", \
                                 pidfile);)
            }
        }
        else {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fprintf(stderr, "Warning: malformed PID file name: %s\n", \
                             pidfile);)
        }
        ret = unlink(pidfile);
        if ( ret != 0 ) {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fprintf(stderr, "Error unlinking PID file: %s\n", \
                                     strerror(errno));)
            return PS_ESYS;
        }
        else {
            PS_DEBUG(PS_DEBUG_INFO, \
                     printf("Successfully unlinked PID file \"%s\"\n",pidfile);)
        }
        free(pidfile);
        pidfile = NULL;
    }
    else {
        PS_DEBUG(PS_DEBUG_INFO, puts("No PID file scheduled for deletion.");)
    }

    return PS_SUCCESS;
}


/*
 * Set the measurement domain based on environment variable.
 */

static void
set_domain(void)
{
    char *env;

    env = getenv(DOMAINFROMENV);
    if ( env != NULL ) {
        if ( strcasecmp(env, PS_DOMAINSTRING_KERNEL) == 0 ) {
            globals.domain = KERNEL;
        }
        else if ( strcasecmp(env, PS_DOMAINSTRING_ALL) == 0 ) {
            globals.domain = ALL;
        }
        else if ( strcasecmp(env, PS_DOMAINSTRING_USER) == 0 ) {
            globals.domain = USER;
        }
        else {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fputs("Invalid counting domain specified in ", stdout); \
                     printf("%s: \"%s\" ", DOMAINFROMENV, env); \
                     fputs("(defaulting to \"PS_DOMAINSTRING_DEFAULT\" ", \
                           stdout); \
                     puts("or config spec).");)
        }
    }

    return;
}


/*
 * Calculate MD5 checksum based on environment variable.
 */

static void
set_md5(void)
{
    char *env, md5[PS_MD5LEN];
    int ret;

    env = getenv(MD5FROMENV);
    if ( env != NULL ) {
        ret = _ps_getexemd5(md5);
        if ( ret == PS_SUCCESS ) {
            globals.md5sum = strdup(md5);
            PS_DEBUG(PS_DEBUG_INFO, \
                     printf("MD5 checksum requested: \"%s\"\n", \
                            globals.md5sum);)
        }
        else {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fprintf(stderr, "Error in MD5 generation: %s\n", \
                             ps_strerror(ret));)
        }
    }

    return;
}


/*
 * Determine whether measurement should be enabled for this run.
 */

static int
set_enabled(int mapc, ps_addrmap_t **mapv, char *db)
{
    char *env;
    int exclude, map, ret;

    /* First check: the value of PS_HWPC in the environment */
    env = getenv(CONTROLFROMENV);
    if ( env != NULL ) {
        if ( ( strcasecmp(env,"off") == 0 ) ||
             ( strcasecmp(env, "no") == 0 ) ) {
            globals.on = 0;
            PS_DEBUG(PS_DEBUG_INFO, puts("library is disabled (environment)");)
            return 0;
        }
    }

    /* Second check: disabled through an exclusion database */
    if ( ! ( (db == NULL) || (mapv == NULL) )  ) {
        ret = _ps_addrmap_exe(mapc, mapv, &map);
        if ( ret == PS_SUCCESS ) {
            exclude = _ps_hwpc_exclude_exe(mapv[map]->module, db);
            if ( exclude ) {
                globals.on = 0;
                PS_DEBUG(PS_DEBUG_INFO, \
                         puts("library is disabled (exclusion database)");)
                return 0;
            }
        }
        else {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     puts("set_enabled: could not locate executable in map");)
        }
    }
    
    return 1;
}


static inline void
free_tokens(char **tokens)
{
    int i;

    for ( i = 0; tokens[i] != NULL; i++ ) {
        PS_DEBUG(PS_DEBUG_VERBOSE, \
                 printf("Freeing memory for token \"%s\"\n", tokens[i]);)
        free(tokens[i]);
    }
    PS_DEBUG(PS_DEBUG_VERBOSE, puts("Freeing memory for token array");)
    free(tokens);

    return;
}


/*
 * Install a signal handler on user request that will be caught and
 * terminate the process (and cancel threads in the reentrant case)
 * in order to obtain performance data output for the case that the
 * application terminates by receipt of a signal.
 */

static int
set_exitsignals(void)
{
    struct sigaction act;
    char *env, **sigstrings;
    int signo, ret, i;

    env = getenv(SIGNALFROMENV);
    if ( env != NULL ) {
        PS_DEBUG(PS_DEBUG_INFO, printf("%s found in environment: \"%s\"\n", \
                                       SIGNALFROMENV, env);)
        act.sa_sigaction = exit_handler;
        act.sa_flags     = SA_SIGINFO; /* @@ ONESHOT too? */

#if _REENTRANT
        ret = sigemptyset(&thread_sigmask);
        if ( ret != 0 ) {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fprintf(stderr, "Error clearing thread signal mask\n");)
            return PS_ESYS;
        }
#endif

        ret = _ps_tokenize(env, EXITSIGNALSEP, &sigstrings);
        if ( ret != PS_SUCCESS ) {
            return ret;
        }

        for ( i = 0; sigstrings[i] != NULL; i++ ) {
            if ( sscanf(sigstrings[i], "%d", &signo) > 0 ) {
                ret = sigaction(signo, &act, NULL);
                if ( ret != 0 ) {
                    PS_DEBUG(PS_DEBUG_FATAL, \
                             fprintf(stderr, "sigaction error: \"%s\"\n", \
                                     strerror(ret));)
                    free_tokens(sigstrings);
                    return PS_ESYS;
                }
#if _REENTRANT
                ret = sigaddset(&thread_sigmask, signo);
                if ( ret != 0 ) {
                    PS_DEBUG(PS_DEBUG_WARNING, \
                             fprintf(stderr, \
                                     "Error adding signal %d to ", signo); \
                             fputs("thread signal mask\n", stderr);)
                    free_tokens(sigstrings);
                    return PS_ESYS;
                }
#endif
                globals.signaling = 1;
                PS_DEBUG(PS_DEBUG_INFO, \
                         printf("Installed signal handler (signal %d)\n",\
                                signo);)
            }
            else {
                    PS_DEBUG(PS_DEBUG_WARNING, \
                             fprintf(stderr, "Bad signal no. given: \"%s\"\n",\
                                     sigstrings[i]);)
            }
        }
        free_tokens(sigstrings);
    }


    return PS_SUCCESS;
}


/*
 * The signal handler itself
 */

static RETSIGTYPE
exit_handler (int signum, siginfo_t *si, void *p)
{
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("Caught signal %d, shutting down.\n", signum);)

    exit(signum + 128);
}

void
cancel_children_threads(void)
{
#if _REENTRANT
    pthread_t me;
    unsigned int thread;
    int ret;
    FILE *procfile = NULL;
    char procfilename[256];
    long unsigned int num_threads = 0;

    me = pthread_self();
    PS_DEBUG(PS_DEBUG_INFO, printf("Entering cancel_children_threads() in thread %lu\n", me);)

    /* Freeze the thread ID list by grabbing the mutex. */
    /*
     * @@ Additional threads may still have been created, but not
     * yet registered with the library, so this is not foolproof.
     *
     * @@ What are the issues regarding obtaining a mutex within
     * a signal handler?
     */
    ret = pthread_mutex_lock(&max_thread_num_mutex);
    if ( ret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, fputs("Error locking mutex.\n", stderr);)
        return;
    }

    /* Cancel every thread except this one. */

    for ( thread = 0; thread < max_thread_num; thread++ ) {
        ret = pthread_equal(me, thread_ids[thread]);
        if ( ret == 0 ) {
            PS_DEBUG(PS_DEBUG_VERBOSE, \
                     fprintf(stderr, "Cancelling thread %d\n", thread);)
            ret = pthread_cancel(thread_ids[thread]);
            if ( ret != 0 ) {
                PS_DEBUG(PS_DEBUG_INFO, \
                         printf("Notice: could not cancel thread %d\n", thread);)
            }
            /* There seems to be a bug in libc where calling pthread_join in
             * a shared library destructor with a dl lock open could cause
             * problems.  In the tests pthread_join always returned EINVL,
             * while the threads were honoring the cancel request.
             * As a workaround, check whether the number of threads is 1
             * in /proc/self/stat instead, as in below.
             * So comment out the following code that uses pthread_join.
             * - Rui, 2012-11-15.
             */
            /*
            ret = pthread_join(thread_ids[thread], NULL);
            PS_DEBUG(PS_DEBUG_VERBOSE, \
                     if ( ret != 0 ) { \
                         fprintf(stderr, "ERROR! %s", strerror(ret)); \
                     } \
                     else { \
                         fputs("done.", stderr); \
                     } \
                     fputs("\n", stderr); \
            )
            */
        }
        else {
            PS_DEBUG(PS_DEBUG_VERBOSE, \
                     fprintf(stderr, "NOT cancelling thread %d\n", thread);)
        }
    }

    /* Wait for all cancelled threads to complete */
    //snprintf(procfilename, 256, "/proc/%d/stat", getpid());
    sprintf(procfilename, "%s", "/proc/self/stat");
    procfile = fopen(procfilename, "r");
    if (NULL == procfile) {
        PS_DEBUG(PS_DEBUG_FATAL, printf("Error opening file %s.\n", procfilename);)
        return;
    }
    do {
       ret = fscanf(procfile, "%*d %*s %*c %*d %*d %*d %*d %*d %*u %*u %*u %*u %*u %*u %*u %*d %*d %*d %*d %ld", &num_threads);
        if (ret != 1) {
            if ((ret == EOF) && (errno == 0)) {
                // Could consider as pass, so set num_threads to 1,
                // otherwise it was observed to be 0 (the value in 
                // last iteration in the do loop) in the tests.
                num_threads = 1;
            } else {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         printf("Error in fscanf to parse the line in %s, ret = %d, errno = %d, error string = '%s'.\n", procfilename, ret, errno, strerror(errno));)
                return;
            }
        }
        if (num_threads != 1) {
            PS_DEBUG(PS_DEBUG_VERBOSE, \
                     printf ("num_threads = %ld, sleeping for 1 sec...\n", \
                     num_threads);)
            sleep(1);
            //usleep(500000);
        }
    } while (num_threads != 1);
    fclose(procfile);
    PS_DEBUG(PS_DEBUG_INFO, puts("All threads except main terminated in cancel_children_threads().");)

/*
    for ( thread = 0; thread < max_thread_num; thread++ ) {
        ret = pthread_equal(me, thread_ids[thread]);
        if ( ret == 0 ) {
              PS_DEBUG(PS_DEBUG_VERBOSE, \
                       fprintf(stderr, "Joining with thread %d [%lu]... ", \
                               thread, thread_ids[thread]);)
              ret = pthread_join(thread_ids[thread], NULL);
              PS_DEBUG(PS_DEBUG_VERBOSE, \
                       if ( ret != 0 ) { \
                           fprintf(stderr, "ERROR! %s", strerror(ret)); \
                       } \
                       else { \
                           fputs("done.", stderr); \
                       } \
                       fputs("\n", stderr); \
              )
        }
    }
*/

    /* Release the thread ID list (probably not necessary, since we
     * are about to exit!)
     */

    ret = pthread_mutex_unlock(&max_thread_num_mutex);
    if ( ret != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Error unlocking mutex.\n", stderr);)
    }

#endif

    return;
}


/*
 * Initialize the process load map.  This may be altered, depending
 * on the selected measurement, for example profil().
 */

static int
set_loadmap(int mapc, ps_addrmap_t **mapv)
{
    int map, tmpmapc, ret;
    ps_hwpc_scale_t scale = PS_HWPC_DEFAULT_SCALE;
    ps_addrmap_t *tmpmapv;

#if DEBUG
    unsigned int profbufmem = 0;
#endif

    /* Save the original map count for use in deallocation later */
    tmpmapc = mapc;

    /* For the HWPCNULL class, which uses profil(), profiling will
     * not be useful if the main program region is not the first one
     * in the entry list.  Adjusted for this and also to save on 
     * memory needed for sampling buffers.
     */
    if ( globals.events.class == HWPCNULL ) {
        ret = _ps_addrmap_exe(mapc, mapv, &map);
        if ( ret == PS_SUCCESS ) {
            if ( map != 0 ) {
                tmpmapv   = mapv[0];
                mapv[0]   = mapv[map];
                mapv[map] = tmpmapv;
                PS_DEBUG(PS_DEBUG_INFO, \
                         fputs("NULL profiling: adjusting maps to ", stdout);\
                         printf("single (swapped map 0 with %d)\n", map);)
            }
            else {
                PS_DEBUG(PS_DEBUG_INFO, \
                         puts("NULL profiling (no map swap needed)");)
            }
        }
        else {
            PS_DEBUG(PS_DEBUG_WARNING, \
                     puts("could not find executable in address map");)
        }
        mapc = 1;
    }

    /*
     * "scale" has a different meaning for the PerfSuite profiling scheme
     * @@ Needs to be checked with Perfmon
     */
    if ( (globals.events.class == ITIMER) ||
         (globals.events.class == PERFMON) ) {
        scale = 1;
    }

    globals.profiles.nmaps = mapc;

    globals.profiles.profmap = (ps_hwpc_profmap_t *)
        malloc(globals.profiles.nmaps * sizeof(ps_hwpc_profmap_t));

    /*
     * @@ This way over-allocates the number of bins needed
     * (and therefore the memory required).  Needs to be adjusted
     * properly while keeping in synch with other portions of
     * PerfSuite.
     * @@ Also need to ensure that we don't cause alignment problems
     * here.
     */
    for ( map = 0; map < mapc; map++ ) {
        globals.profiles.profmap[map].name = strdup(mapv[map]->module);
        globals.profiles.profmap[map].offset = mapv[map]->start;
        globals.profiles.profmap[map].nbins = 
            mapv[map]->end - mapv[map]->start;
        globals.profiles.profmap[map].scale = scale;
#if DEBUG
        profbufmem += (mapv[map]->end - mapv[map]->start);
#endif
    }

    /* @@ Will %lx format always work? */
    PS_DEBUG(PS_DEBUG_INFO, \
             puts("==> Profiling details <=="); \
             printf("%18s %18s %10s %10s %s\n", \
                    "Start", "End", "# Bins", "Scale", "Module"); \
             for ( map = 0; map < mapc; map++ ) { \
                 printf("%18p %18p %10d %10d %s\n", \
                        globals.profiles.profmap[map].offset, \
                        mapv[map]->end, \
                        globals.profiles.profmap[map].nbins, \
                        globals.profiles.profmap[map].scale, \
                        globals.profiles.profmap[map].name); \
             } \
             fputs("Total profiling memory buffer space required ", stdout); \
             printf("(bytes) = %u\n", \
                    profbufmem * sizeof(ps_hwpc_sample_t)); \
             printf("Profiling period: %d\n", globals.threshold);)

    _ps_addrmap_free(tmpmapc, mapv);
    mapv = NULL;

    return PS_SUCCESS;
}


/*
 * Assign the package-specific function pointers.
 */

static int
set_package(char *eventfile)
{
    switch ( globals.events.class ) {
    case HWPCNULL:
        PS_DEBUG(PS_DEBUG_INFO, puts("Initializing HWPCNULL class...");)
        globals.control.get_num_counters  = get_num_counters_null;
        globals.control.get_max_counters  = get_max_counters_null;
        globals.control.init_counters     = init_counters_null;
        globals.control.start_counters    = start_counters_null;
        globals.control.read_counters     = read_counters_null;
        globals.control.stop_counters     = stop_counters_null;
        globals.control.destroy_counters  = destroy_counters_null;
        globals.control.init_profiling    = init_profiling_null;
        globals.control.start_profiling   = start_profiling_null;
        globals.control.suspend_profiling = suspend_profiling_null;
        globals.control.stop_profiling    = stop_profiling_null;
        break;
    case PAPI:
#if HAVE_PAPI
        PS_DEBUG(PS_DEBUG_INFO, puts("Initializing PAPI class...");)
        globals.control.get_num_counters  = get_num_counters_papi;
        globals.control.get_max_counters  = get_max_counters_papi;
        globals.control.init_counters     = init_counters_papi;
        globals.control.start_counters    = start_counters_papi;
        globals.control.read_counters     = read_counters_papi;
        globals.control.stop_counters     = stop_counters_papi;
        globals.control.destroy_counters  = destroy_counters_papi;
# if defined(IA32) || defined(IA64) || defined(PPC32) || defined(PPC64)
        globals.control.init_profiling    = init_profiling_papi;
        globals.control.start_profiling   = start_profiling_papi;
        globals.control.suspend_profiling = suspend_profiling_papi;
        globals.control.stop_profiling    = stop_profiling_papi;
# else
        globals.control.init_profiling    = 
        globals.control.start_profiling   = 
        globals.control.suspend_profiling = 
        globals.control.stop_profiling    = NULL;
# endif /* defined(IA32) || defined(IA64) */
#else
        return PS_ENOTINSTALLED;
#endif /* HAVE_PAPI */
        break;
    case PERFMON:
#if HAVE_PERFMON
        PS_DEBUG(PS_DEBUG_INFO, puts("Initializing Perfmon class...");)
#elif HAVE_PERFMON2
        PS_DEBUG(PS_DEBUG_INFO, puts("Initializing Perfmon2 class...");)
#endif
#if HAVE_PERFMON || HAVE_PERFMON2
        globals.control.get_num_counters  = get_num_counters_perfmon;
        globals.control.get_max_counters  = get_max_counters_perfmon;
        globals.control.init_counters     = init_counters_perfmon;
        globals.control.start_counters    = start_counters_perfmon;
        globals.control.read_counters     = read_counters_perfmon;
        globals.control.stop_counters     = stop_counters_perfmon;
        globals.control.destroy_counters  = destroy_counters_perfmon;
        globals.control.init_profiling    = init_profiling_perfmon;
        globals.control.start_profiling   = start_profiling_perfmon;
        globals.control.suspend_profiling = suspend_profiling_perfmon;
        globals.control.stop_profiling    = stop_profiling_perfmon;
#else
        return PS_ENOTINSTALLED;
#endif /* HAVE_PERFMON */
        break;
    case ITIMER:
#if HAVE_SETITIMER
        PS_DEBUG(PS_DEBUG_INFO, puts("Initializing ITIMER class...");)
        globals.control.get_num_counters  = get_num_counters_itimer;
        globals.control.get_max_counters  = get_max_counters_itimer;
        globals.control.init_counters     = init_counters_itimer;
        globals.control.start_counters    = start_counters_itimer;
        globals.control.read_counters     = read_counters_itimer;
        globals.control.stop_counters     = stop_counters_itimer;
        globals.control.destroy_counters  = destroy_counters_itimer;
        globals.control.init_profiling    = init_profiling_itimer;
        globals.control.start_profiling   = start_profiling_itimer;
        globals.control.suspend_profiling = suspend_profiling_itimer;
        globals.control.stop_profiling    = stop_profiling_itimer;
        break;
#else
        return PS_ENOTINSTALLED;
#endif /* HAVE_SETITIMER */
        break;
    case UNKNOWN:
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Fatal error:\n", stderr); \
                 fputs("\tUnsupported/unknown hardware counter ", stderr); \
                 fputs("package (or none supplied)\n", stderr); \
                 fputs("\tin event configuration file:", stderr); \
                 fprintf(stderr, "\n\n\t\t\"%s\"", eventfile); \
                 fputs("\n\n\tCheck the \"class\" attribute in element ", \
                       stderr); \
                 fprintf(stderr, "\"%s\"\n\tin the above document.\n\n", \
                         EVENTLISTTAG); \
                 fputs("\tAccepted classes (case is significant) are:\n", \
                       stderr);)
#if HAVE_PAPI
        PS_DEBUG(PS_DEBUG_FATAL, fputs("\t\tPAPI\n", stderr);)
#endif
#if HAVE_PERFMON
        PS_DEBUG(PS_DEBUG_FATAL, fputs("\t\tperfmon\n", stderr);)
#endif
#if HAVE_SETITIMER
        PS_DEBUG(PS_DEBUG_FATAL, fputs("\t\titimer\n", stderr);)
#endif
        PS_DEBUG(PS_DEBUG_FATAL, fputs("\t\tnull\n", stderr);)
        return PS_HWPC_EBADCLASS;
    default:
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Internal error: class out of range.\n", stderr);)
        return PS_HWPC_EBADCLASS;
    }

    return PS_SUCCESS;
}


/*
 * Read the performance measurement configuration file.
 */

static int
read_config(char **eventfile)
{
    char *eventfileenv, *sysconfigdir;
    int len, ret; /* @@ MAXPATHLEN? */
    FILE *fin;

    eventfileenv = getenv(CONFIGFROMENV);
    if ( eventfileenv == NULL ) {
        /* User hasn't specified an event file, so use the default */
        eventfileenv = PS_HWPC_DEF_CONFIGFILE;
        len = strlen(PS_DATADIR) + strlen(eventfileenv) + 2;
        *eventfile = (char *) malloc(len);
        if ( *eventfile == NULL ) {
            return PS_ENOMEM;
        }
        sprintf(*eventfile, "%s%c%s", PS_DATADIR, DIRECTORY_SEP, eventfileenv);
    }
    else {
        len = strlen(eventfileenv) + 1;
        *eventfile = (char *) malloc(len);
        if ( *eventfile == NULL) {
            return PS_ENOMEM;
        }
        strncpy(*eventfile, eventfileenv, len);
    }

    PS_DEBUG(PS_DEBUG_INFO, printf("Configuration file set to %s\n", \
                                   *eventfile);)

    /* 
     * See if we can access the configuration file.  If not, try
     * the user-supplied or system default directory (only if requested).
     *
     * Errors will be returned either here or at the time the
     * file open is attempted.
     */
    ret = access(*eventfile, R_OK);
    if ( ret != 0  ) {
        PS_DEBUG(PS_DEBUG_INFO, printf("\"%s\" is not locally accessible\n", \
                                       *eventfile);)
        sysconfigdir = getenv(SYSCONFIGFROMENV);
        if ( sysconfigdir != NULL ) {
            /* If no directory is specified, use the default */
            if ( strlen(sysconfigdir) == 0 ) {
                sysconfigdir = PS_DATADIR;
            }
            /* Restrict to relative path names  */
            if ( *eventfile[0] != DIRECTORY_SEP ) {
                PS_DEBUG(PS_DEBUG_INFO, \
                         printf("Trying default directory \"%s\" for \"%s\"\n",
                                sysconfigdir, *eventfile);)
                len = strlen(sysconfigdir) + strlen(eventfileenv) + 2;
                *eventfile = realloc(*eventfile, len);
                if ( *eventfile == NULL ) {
                    return PS_ENOMEM;
                }
                sprintf(*eventfile, "%s%c%s", sysconfigdir, DIRECTORY_SEP,
                                              eventfileenv);
                ret = access(*eventfile, R_OK);
                if ( ret != 0 ) {
                    PS_DEBUG(PS_DEBUG_FATAL, \
                             fprintf(stderr, "\"%s\" is not readable\n", \
                                     *eventfile);)
                    return PS_HWPC_ENOSUCHXMLDOC;
                }
            }
        }
        else {
            return PS_HWPC_ENOSUCHXMLDOC;
        }
    }

    PS_DEBUG(PS_DEBUG_INFO, printf("Attempting to parse \"%s\"\n", *eventfile);)

    /*
     * Parse the run configuration file.
     */
    if ( strcmp(*eventfile, PS_STDIN) == 0 ) {
        fin = stdin;
    }
    else {
        fin = fopen(*eventfile, "r");
        if ( fin == NULL ) {
            /* @@ free eventfile here? */
            return PS_ESYS;
        }
    }

    ret = _ps_parseEventListFromStream(fin, &globals);
    if ( ret != PS_SUCCESS ) {
        free(*eventfile);
        return ret;
    }

    if ( fin != stdin ) {
        fclose(fin);
    }

   return PS_SUCCESS;
}


/*
 * Figure out what file to write the data to.
 * Precedence (highest to lowest) is:
 * - Environment ("stdout" and "stderr" are recognized)
 * - Non-NULL argument supplied
 * - Basename of the current executable
 * - Default value (defined in hwpc.h)
 *
 * Non-static to allow access to output stream by package-specific
 * layers.
 */

FILE *
_ps_hwpc_select_stream(const char *fname, const char *suffix,
                       const char *exebase, int *close)
{
    FILE *fout;
    char *env_fname;

    env_fname = getenv(FNAMEFROMENV);
    if ( env_fname != NULL ) {
        fname = env_fname;
    }
    else if ( fname == NULL ) {
        fname = ( exebase != NULL  ? exebase : PS_HWPC_DEF_FNAME);
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("_ps_hwpc_select_stream: using filename base \"%s\"\n", \
                    fname); \
             printf("_ps_hwpc_select_stream: using filename suffix \"%s\"\n", \
                    suffix);)

    if ( strcmp(fname, PS_STDOUT) == 0 ) {
        fout = stdout;
        *close = 0;
    }
    else if ( strcmp(fname, PS_STDERR) == 0 ) {
        fout = stderr;
        *close = 0;
    }
    else {
        fout = generate_fd(fname, suffix);
        *close = 1;
    }

    return fout;
}


/*
 * Construct a file name to be used for output, open the file and
 * return its file descriptor.
 */

static FILE *
generate_fd(const char *base, const char *suffix)
{
#if _REENTRANT
    ps_hwpc_values_t *values;
#endif
    char filename[256], hostname[256], *firstdot;
    int usepid = 1, usehost = 1;

    if ( getenv(NOPIDFROMENV) != NULL ) {
        usepid = 0;
    }

    if ( getenv(NOHOSTNAMEFROMENV) != NULL ) {
        usehost = 0;
    }

    if ( gethostname(hostname, 256) == -1 ) {
        strcpy(hostname, "nohost");
    }

    firstdot = strchr(hostname, '.');
    if ( firstdot != NULL ) {
        *firstdot = '\0';
    }

#if _REENTRANT
    values = get_values();
    if ( usepid ) {
        if ( usehost ) {
            sprintf(filename, "%s.%d.%d.%s.%s", base, values->myid, getpid(),
                              hostname, suffix);
        }
        else {
            sprintf(filename, "%s.%d.%d.%s", base, values->myid, getpid(),
                              suffix);
        }
    }
    else {
        if ( usehost ) {
            sprintf(filename, "%s.%d.%s.%s", base, values->myid, hostname,
                              suffix);
        }
        else {
            sprintf(filename, "%s.%d.%s", base, values->myid, suffix);
        }
    }
#else
    if ( usepid ) {
        if ( usehost ) {
            sprintf(filename, "%s.%d.%s.%s", base, getpid(), hostname,
                              suffix);
        }
        else {
            sprintf(filename, "%s.%d.%s", base, getpid(), suffix);
        }
    }
    else {
        if ( usehost ) {
            sprintf(filename, "%s.%s.%s", base, hostname, suffix);
        }
        else {
            sprintf(filename, "%s.%s", base, suffix);
        }
    }
#endif

    PS_DEBUG(PS_DEBUG_VERBOSE, printf("generate_fd: file is %s\n", filename);)
    return (fopen(filename,"w"));
}


/*
 * Fortran wrappers
 */

void
PSF_HWPC_INIT_F77(int *ierr)
{
    *ierr = ps_hwpc_init();
    return;
}

void
PSF_HWPC_START_F77(int *ierr)
{
    *ierr = ps_hwpc_start();
    return;
}

void
PSF_HWPC_SUSPEND_F77(int *ierr)
{
    *ierr = ps_hwpc_suspend();
    return;
}

void
PSF_HWPC_READ_F77(ps_hwpc_value_t *values, int *ierr)
{
    *ierr = ps_hwpc_read(values);
    return;
}

void
PSF_HWPC_SHUTDOWN_F77(int *ierr)
{
    *ierr = ps_hwpc_shutdown();
    return;
}

void
PSF_HWPC_STOP_F77(const char *fname, int *ierr, int fname_len)
{
    char cname[256]; /*@@ MAXPATHLEN? */
		       
    _ps_f2cstring(fname, fname_len, cname, 255);

    *ierr = ps_hwpc_stop(cname);
    return;
}

void
PSF_HWPC_STATE_F77(int *state, int *ierr)
{
    *ierr = ps_hwpc_state(state);
    return;
}

void
PSF_HWPC_NUMEVENTS_F77(int *numevents, int *ierr)
{
    *ierr = ps_hwpc_numevents(numevents);
    return;
}

void
PSF_HWPC_EVENTNAMES_F77(char *feventnames, int *ierr, int namelen)
{
    int i;
    char *name;

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("psf_hwpc_eventnames: namelen=%d\n", namelen);)

    /*
     * Assumes Fortran arrays of character strings of length N
     * are laid out linearly, as in:
     * 
     * String 1    String 2   String 3
     * 12345......N12345.....N12345.....N
     */

    for ( i = 0; i < globals.events.numEvents; i++ ) {
        name = globals.events.events[i].val;
        _ps_c2fstring(feventnames, strlen(name), name, namelen);
        feventnames += namelen; 
    }

    return;
}

void
PSF_HWPC_PSRUN_F77(int *ierr)
{
    *ierr = ps_hwpc_psrun();
    return;
}
