/* $Id: hwpc.h,v 1.80 2013/05/07 21:51:49 ruiliu Exp $ */

/****************************************************************************
 *
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *     Copyright(C) 2003-2009, 2012-2013, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

#ifndef HWPC_H_INCLUDED
#define HWPC_H_INCLUDED 1

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if ! HAVE_LONG_LONG
# error "long long data type is not supported by this compiler"
#endif

#if _REENTRANT
# include <pthread.h>
#endif

#include <stdio.h>

#if HAVE_LIMITS_H
# include <limits.h>
#endif

#include <pshwpc.h>
#include <cpuid.h>
#include <psdebug.h>
#include "eventlist.h"

#define LIBRARY_NAME "libpshwpc"

#define LIBRARY_MAJOR_VERSION 0
#define LIBRARY_MINOR_VERSION 6
#define LIBRARY_RELEASE_LEVEL 2

/* Environment variable names */

#define CONTROLFROMENV    "PS_HWPC"
#define ANNOTATIONFROMENV "PS_HWPC_ANNOTATION"
#define ANONYMOUSFROMENV  "PS_HWPC_ANONYMOUS"
#define CONFIGFROMENV     "PS_HWPC_CONFIG"
#define DOMAINFROMENV     "PS_HWPC_DOMAIN"
#define EXCLUDEDBFROMENV  "PS_HWPC_EXCLUDEDB"
#define FNAMEFROMENV      "PS_HWPC_FILE"
#define FORMATFROMENV     "PS_HWPC_FORMAT"
#define MD5FROMENV        "PS_HWPC_MD5"
#define MPXUSECFROMENV    "PS_HWPC_MPX_USEC"
#define NOPIDFROMENV      "PS_HWPC_NOPID"
#define NOHOSTNAMEFROMENV "PS_HWPC_NOHOSTNAME"
#define PIDDIRFROMENV     "PS_HWPC_PIDDIR"
#define PMPIFROMENV       "PS_HWPC_PMPI"
#define SIGNALFROMENV     "PS_HWPC_SIGNAL"
#define SYSCONFIGFROMENV  "PS_HWPC_SYSCONFIG"
#define THRESHOLDFROMENV  "PS_HWPC_THRESHOLD"
#define TIMELIMITFROMENV  "PS_HWPC_TIME"

#ifndef PS_HWPC_DEF_FNAME
# define PS_HWPC_DEF_FNAME "pshwpc"
#endif

#ifndef PS_DATADIR
# define PS_DATADIR "."
#endif

#ifndef PS_STDIN
# define PS_STDIN  "stdin"
#endif

#ifndef PS_STDOUT
# define PS_STDOUT "stdout"
#endif

#ifndef PS_STDERR
# define PS_STDERR "stderr"
#endif

/* @@ Assumes UNIX-style directory separator */

#ifndef DIRECTORY_SEP
# define DIRECTORY_SEP '/'
#endif

/*
 * Syntax for specifying multiple signal numbers
 * to be caught via SIGNALFROMENV.
 */

#define EXITSIGNALSEP     ','

/* Possible counting domains */

typedef enum {
    UNKNOWN_DOMAIN = -1,
    USER,
    KERNEL,
    ALL
} ps_hwpc_domain_t;

/* 
 * Environment values for domain selection 
 * @@ There should be a verification method
 * based on the support package.
 */

#ifndef PS_DOMAINSTRING_DEFAULT
# define PS_DOMAINSTRING_DEFAULT "user"
#endif

#ifndef PS_DOMAINSTRING_ALL
# define PS_DOMAINSTRING_ALL     "all"
#endif

#ifndef PS_DOMAINSTRING_KERNEL
# define PS_DOMAINSTRING_KERNEL  "kernel"
#endif

#ifndef PS_DOMAINSTRING_USER
# define PS_DOMAINSTRING_USER    "user"
#endif

/*
 * Environment variable for output format.
 */

#ifndef PS_FORMATSTRING_TEXT
# define PS_FORMATSTRING_TEXT    "text"
#endif

#ifndef PS_FORMATSTRING_XML
# define PS_FORMATSTRING_XML     "xml"
#endif

#ifndef PS_SUFFIX_TEXT
# define PS_SUFFIX_TEXT          "txt"
#endif

#ifndef PS_SUFFIX_XML
# define PS_SUFFIX_XML           "xml"
#endif

#ifndef PS_SUFFIX_DEFAULT
# define PS_SUFFIX_DEFAULT       PS_SUFFIX_XML
#endif

/*
 * Default multiplexing interval (microseconds)
 * This is not currently used.
 */

#define PS_HWPC_DEFAULT_MPX_USEC 10000

/*
 * Base size of a profiling sample, allowable threshold/period,
 * types for memory addresses, and scaling factor.  Note that the 
 * specific use of the scaling factor can be package-dependent.
 */

typedef unsigned short        ps_hwpc_sample_t;
typedef int                   ps_hwpc_threshold_t;
typedef caddr_t               ps_hwpc_address_t;
typedef caddr_t               ps_hwpc_offset_t;
typedef unsigned int          ps_hwpc_scale_t;

/*
 * Maximum number of samples that can be collected in any one profile bin.
 */

#define HWPC_SAMPLE_MAX       USHRT_MAX

/*
 * Default profiling threshold.
 */

#define PS_HWPC_DEFAULT_THRESHOLD 100000

/*
 * The scaling factor, as defined by profil(), that determines
 * the compression factor of the address range to be profiled.
 */

#define PS_HWPC_DEFAULT_SCALE 0x10000  /* 65536 */

/* Used for multiple address ranges */
struct ps_hwpc_profmap {
    char                 *name;
    ps_hwpc_offset_t     offset;
    ps_hwpc_scale_t      scale;
    /*unsigned*/ int         nbins;
};
typedef struct ps_hwpc_profmap ps_hwpc_profmap_t;

struct ps_hwpc_profinfo {
    ps_hwpc_profmap_t    *profmap;
    int                  nmaps;
};
typedef struct ps_hwpc_profinfo ps_hwpc_profinfo_t;

/* 
 * This is the data that describes the configuration of the counters.
 * There will be only one copy of this.
 */

struct ps_hwpc_globals {
    int                   on;        /* @@ These should be bit fields */
    int                   multiplexing;
    int                   profiling;
    int                   signaling;
    int                   mpx_usec;  /* Multiplexing interval         */
    ps_hwpc_domain_t      domain;
    ps_hwpc_eventarray_t  events;
    ps_hwpc_threshold_t   threshold; /* Profiling overflow threshold  */
    ps_hwpc_profinfo_t    profiles;
    ps_cpuid_t            cpuid;     /* CPU information               */
    double                mhz;       /* Clock speed of the machine    */
    char                  *version;  /* Of performance software layer */
    char                  *exebase;  /* Base name of executable       */
    char                  *md5sum;   /* MD5 checksum of executable    */
#if _REENTRANT
    pthread_key_t         datakey;
#endif
    struct {
        /* Function pointers assigned based on the package */
        int               (*get_num_counters)(void);
        int               (*get_max_counters)(struct ps_hwpc_globals *);
        int               (*init_counters)(struct ps_hwpc_globals *);
        int               (*start_counters)(struct ps_hwpc_globals *);
        int               (*read_counters)(ps_hwpc_value_t *);
        int               (*stop_counters)(ps_hwpc_value_t *);
        int               (*destroy_counters)(void);
        /* @@ init_profiling and stop_profiling can have out-of-range
         * and overflow added to them.
         */
        int               (*init_profiling)(struct ps_hwpc_globals *,
                                            ps_hwpc_sample_t **);
        int               (*start_profiling)(struct ps_hwpc_globals *,
                                             ps_hwpc_sample_t **);
        int               (*suspend_profiling)(struct ps_hwpc_globals *,
                                               ps_hwpc_sample_t **);
        int               (*stop_profiling)(struct ps_hwpc_globals *,
                                            ps_hwpc_sample_t **);
    } control;
};
typedef struct ps_hwpc_globals ps_hwpc_globals_t;

typedef enum {
    UNINITIALIZED,
    RUNNING,
    SUSPENDED,
    TERMINATED
} ps_hwpc_state_t;

/*
 * The counter data itself (one copy per thread)
 */

typedef struct {
#if _REENTRANT
    unsigned int        myid;       /* Integer thread ID (count)   */
    unsigned int        started;    /* Flag set at thread start    */
#endif
    ps_hwpc_state_t     state ;     /* Current counter state       */
    ps_hwpc_value_t     *startval;  /* Value when last activated   */
    ps_hwpc_value_t     *current;   /* Current value               */
    ps_hwpc_value_t     *total;     /* Running total               */
    unsigned long long  starttime;  /* RTC value when last started */
    unsigned long long  totaltime;  /* Running RTC total           */
    ps_hwpc_sample_t    **samples;  /* Profiling sample bins       */
    unsigned long       outofrange; /* Samples not assigned        */
    unsigned long       *overflow;  /* Samples overflowed per map  */
} ps_hwpc_values_t;


BEGIN_C_DECLS

void _ps_hwpc_init_globals   PARAMS((ps_hwpc_globals_t *globals));
FILE *_ps_hwpc_select_stream PARAMS((const char *base, const char *suffix,
                                     const char *exebase, int *close));
void cancel_children_threads PARAMS((void));

END_C_DECLS

/* Fortran wrappers */

#define PSF_HWPC_INIT_F77       F77_FUNC_(psf_hwpc_init,PSF_HWPC_INIT)
#define PSF_HWPC_EVENTNAMES_F77 F77_FUNC_(psf_hwpc_eventnames,PSF_HWPC_EVENTNAMES)
#define PSF_HWPC_NUMEVENTS_F77  F77_FUNC_(psf_hwpc_numevents,PSF_HWPC_NUMEVENTS)
#define PSF_HWPC_PSRUN_F77      F77_FUNC_(psf_hwpc_psrun,PSF_HWPC_PSRUN)
#define PSF_HWPC_READ_F77       F77_FUNC_(psf_hwpc_read,PSF_HWPC_READ)
#define PSF_HWPC_SHUTDOWN_F77   F77_FUNC_(psf_hwpc_shutdown,PSF_HWPC_SHUTDOWN)
#define PSF_HWPC_START_F77      F77_FUNC_(psf_hwpc_start,PSF_HWPC_START)
#define PSF_HWPC_STATE_F77      F77_FUNC_(psf_hwpc_state,PSF_HWPC_STATE)
#define PSF_HWPC_STOP_F77       F77_FUNC_(psf_hwpc_stop,PSF_HWPC_STOP)
#define PSF_HWPC_SUSPEND_F77    F77_FUNC_(psf_hwpc_suspend,PSF_HWPC_SUSPEND)

#endif /* HWPC_H_INCLUDED */
