/* $Id: profile.c,v 1.25 2013/04/30 16:44:44 ruiliu Exp $ */

/****************************************************************************
 *
 *                      University of Illinois/NCSA
 *                          Open Source License
 * 
 *          Copyright(C) 2004-2006, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                             Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * profile.c: profile support routines for libpshwpc
 *
 *            !!NOTE!!: this depends on the Pthreads library delivering
 *            signals in a non-standard way - that is, signals are
 *            delivered to the individual threads.  The standard
 *            states that signals can be sent to any thread that is
 *            not blocking it.
 *
 * Author: Rick Kufrin
 * Date:   12/27/2004
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if _REENTRANT
# if HAVE_PTHREAD
#  include <pthread.h>
# endif
#endif

#include <stdio.h>

#include <stdlib.h>

#if HAVE_STDDEF_H
# include <stddef.h>
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if HAVE_SIGNAL_H
# include <signal.h>
#endif

#if HAVE_SYS_TIME_H
# include <sys/time.h>
#endif

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#include <perfsuite.h>
#include <signals.h>
#include "psinternal.h"
#include "hwpc.h"
#include "profile.h"

#if HAVE_PERFMON
# include <perfmon/pfmlib.h>
# include "hwpc-perfmon.h"
#endif

#if HAVE_PERFMON2
# include "hwpc-perfmon2.h"
#endif

static ps_hwpc_globals_t *globals;
static void *extra;

/*
 * Generic thread-specific data for sampling buffers.
 */
#if _REENTRANT
static pthread_key_t tsdkey;
#else
static ps_hwpc_sample_t **samples;
#endif /* _REENTRANT */

/*
 * For the perfmon profiling layer, we use another TSD key to store
 * the data related to context and profiling buffers.
 */
#if HAVE_PERFMON
# if _REENTRANT
static pthread_key_t perfmon_tsdkey;
# else
static hwpc_perfmon_tsd_t *perfmon_tsd;
# endif /* _REENTRANT */
#endif  /* HAVE_PERFMON */

BEGIN_C_DECLS

#if _REENTRANT
static void free_tsd PARAMS((void *data));
#endif

END_C_DECLS

int
_ps_hwpc_setprofglobals(ps_hwpc_globals_t *g, void *e)
{
#if _REENTRANT
    if ( pthread_key_create(&tsdkey, free_tsd) != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Error creating TSD key in _ps_hwpc_setprofglobals\n", \
                       stderr);)
        return PS_ESYS;
    }
#endif
    globals = g;
    extra = e;

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("_ps_hwpc_setprofglobals: globals set to %p\n", globals);)

    return PS_SUCCESS;
}

int
_ps_hwpc_setprofbuf(ps_hwpc_sample_t **s)
{
#if DEBUG
    int pid;
#endif
#if _REENTRANT
    void *tsd;

    tsd = (void *) s;
    if ( pthread_setspecific(tsdkey, tsd) != 0 ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("Error setting TSD data in _ps_hwpc_setprofbuf\n", \
                       stderr);)
        return PS_ESYS;
    }
#else
    samples = s;
#endif
    if ( s == NULL ) {
        printf("WARNING: _ps_hwpc_setprofglobals: sample buffer is NULL!!!!\n");
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             _ps_gettid(&pid); \
             printf("Thread ID %d sampling buffer located at %p\n", pid, s);)

    return PS_SUCCESS;
}

static inline int
bin_pc(caddr_t pc, int *map, ptrdiff_t *bin)
{
    ptrdiff_t pcoff;
    int m;

#if _REENTRANT
    if ( pthread_getspecific(globals->datakey) == NULL ) {
        /*
         * Probably in termination, no TSD to access.
         *
         * @@ It is not clear to me if this check is necessary,
         * with the addition of a similar mod in _ps_hwpc_profhandler,
         * but leaving it in for now.  Concern is the amount of
         * overhead associated with pthread_getspecific(), since
         * this function (bin_pc) is called frequently.
         * 
         */
        return 1;
    }
#endif

    for ( m = 0; m < globals->profiles.nmaps; m++ ) {
        pcoff = pc - globals->profiles.profmap[m].offset;
        if ( pcoff < 0 ) {
            PS_DEBUG(PS_DEBUG_VERBOSE, \
                     printf("Address %p not in map %d (before offset)\n", \
                            pc, m);)
            continue;
        }
        /* @@ This is where scale comes into it */
        pcoff >>= globals->profiles.profmap[m].scale;
        if ( pcoff > globals->profiles.profmap[m].nbins ) {
            PS_DEBUG(PS_DEBUG_VERBOSE, \
                     printf("Address %p not in map %d (exceeds nbins)\n", \
                            pc, m);)
            continue;
        }
/*
        printf("Assigning address %p to bin %ld in map %d\n", 
               pc, (long) pcoff, m);
*/

        *map = m;
        *bin = pcoff;

        return 0;
    }

    return 1;
}

RETSIGTYPE
_ps_hwpc_profhandler(int signum, siginfo_t *si, void *p)
{
    int map, ret;
    ptrdiff_t pcoff;
    caddr_t pc;
#if _REENTRANT
    ps_hwpc_sample_t **samples;
#endif
#if DEBUG
    pid_t pid;
#endif

    pc = get_pc(p);

#if _REENTRANT
    samples = (ps_hwpc_sample_t **) pthread_getspecific(tsdkey);
    if ( samples == NULL ) {
        /*
         * Likely the thread has received a spurious signal before
         * it has had an opportunity to initialize and allocate
         * sampling buffers.  Throw away this sample.
         */
        return;
/*
        fputs("Fatal profiling error: cannot access TSD data in handler.\n", 
              stderr);
        fprintf(stderr, 
                "Please report this as an internal PerfSuite bug to %s.\n",
                PACKAGE_BUGREPORT);
        exit(1);
*/
    }
#endif

    PS_DEBUG(PS_DEBUG_VERBOSE, _ps_gettid(&pid); \
             printf("In library handler (TID %d, buffer %p, pc %p)\n", \
                    pid, samples, pc);)

    ret = bin_pc(pc, &map, &pcoff);
    if ( ret !=  0 ) {
        /* @@ It's actually the values structure, of which samples is
         * a part, that holds out-of-range and overflow counts.
         */
        PS_DEBUG(PS_DEBUG_INFO, printf("Sample at PC %p out of range\n", pc);)
        return;
    }

    if ( samples[map][pcoff] == HWPC_SAMPLE_MAX ) {
        PS_DEBUG(PS_DEBUG_INFO, printf("Overflow in map %d\n", map);)
    }
    else {
        samples[map][pcoff]++;
    }

    return;
}

#if HAVE_PERFMON

/*
 * Establish the thread-specific data for perfmon context, to be
 * used within the profiling handler.
 */

#if _REENTRANT
void
_ps_hwpc_perfmon_setprofile_tsdkey(pthread_key_t key)
{
    perfmon_tsdkey = key;
}
#else
void
_ps_hwpc_perfmon_setprofile_tsd(void *tsd)
{
    perfmon_tsd = (hwpc_perfmon_tsd_t *) tsd;
}
#endif /* _REENTRANT */

void
_ps_hwpc_perfmon_processbuffer(void)
{
    caddr_t pc;
    ptrdiff_t pcoff;
    perfmon_smpl_hdr_t *hdr;
    perfmon_smpl_entry_t *entry;
    unsigned long pos;
    int map, ent, ret;
#if _REENTRANT
    hwpc_perfmon_tsd_t *perfmon_tsd;
    ps_hwpc_sample_t **samples;
#endif

#if _REENTRANT
    perfmon_tsd = pthread_getspecific(perfmon_tsdkey);
    samples = (ps_hwpc_sample_t **) pthread_getspecific(tsdkey);
    if ( samples == NULL ) {
        fputs("Fatal profiling error: cannot access TSD data in handler.\n", 
              stderr);
        fprintf(stderr, 
                "Please report this as an internal PerfSuite bug to %s.\n",
                PACKAGE_BUGREPORT);
        exit(1);
    }
#endif

    hdr = (perfmon_smpl_hdr_t *) perfmon_tsd->smpl_vaddr;

    pos = (unsigned long) (hdr + 1);
    for ( ent = 0; ent < hdr->hdr_count; ent++ ) {
        entry = (perfmon_smpl_entry_t *) pos;

/*
        printf("Sampling buffer entry %d\n", ent);
        printf("PID: %d\n",          entry->pid);
        printf("CPU: %d\n",          entry->cpu);
        printf("Timestamp: 0x%lx\n", entry->stamp);
        printf("IP: %p\n",           entry->ip);
*/
        pc = (caddr_t) entry->ip;

        ret = bin_pc(pc, &map, &pcoff);
        if ( ret == 0 ) {
            samples[map][pcoff]++;
        }
        else {
            PS_DEBUG(PS_DEBUG_INFO, printf("Sample at PC %p out of range\n", pc);)
        }

        pos += hdr->hdr_entry_size;
    }

    return;
}

RETSIGTYPE
_ps_hwpc_perfmon_handler(int signum, struct pfm_siginfo *info,
                         struct sigcontext *sc)
{
    pid_t pid;
    caddr_t pc;
    perfmon_smpl_hdr_t *hdr;

#if _REENTRANT
    hwpc_perfmon_tsd_t *perfmon_tsd;
    ps_hwpc_sample_t **samples;
    perfmon_tsd = pthread_getspecific(perfmon_tsdkey);
#endif

    _ps_gettid(&pid);
    PS_DEBUG(PS_DEBUG_VERBOSE, \
             printf("In perfmon handler (TID %d)\n", pid);)

    /* @@ This is extra data needed to be passed in */
    hdr = (perfmon_smpl_hdr_t *) perfmon_tsd->smpl_vaddr;

    /* @@ Can this be done outside of the handler? */
    if ( PFM_VERSION_MAJOR(hdr->hdr_version) !=
         PFM_VERSION_MAJOR(PFM_SMPL_VERSION) ) {
        fprintf(stderr, "Error: Perfmon sampling format mismatch\n");
        exit(1);
    }

    if ( info->sy_code != PROF_OVFL ) {
        PS_DEBUG(PS_DEBUG_WARNING, \
                 fprintf(stderr, "Non-perfmon generated SIGPROF (%d)\n", \
                         info->sy_code);)
        return;
    }

    _ps_hwpc_perfmon_processbuffer();

    if (perfmonctl(pid, PFM_RESTART,NULL, 0) == -1) {
        perror("PFM_RESTART");
        exit(1);
    }

    return;
}

#endif /* HAVE_PERFMON */

#if _REENTRANT
static void
free_tsd(void *data)
{
    pthread_setspecific(tsdkey, NULL);
}
#endif
