/* $Id: xmlparse.c,v 1.47 2009/11/05 22:13:26 rkufrin Exp $ */

/****************************************************************************
 *
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *          Copyright(C) 2002-2006, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * eventListParse: a parser for PerfSuite XML format performance 
 *                 counter specification lists
 *
 * This reader uses Expat and does (very) minimal error-handling.
 * It supports the PerfSuite ps_hwpc_eventlist DTD, but isn't a proper
 * validating parser.
 *
 * There should be nothing in this file specific to a particular
 * hardware event supporting software layer (i.e., needing
 * header files, libraries, etc).
 *
 * @@ This parser uses the "val" data for both the name of a
 * PAPI preset (e.g., "PAPI_FP_INS") and the cdata code of
 * a PAPI native event.  This should be changed so that native
 * events can also have a description.
 *
 * Can you tell this is my first attempt at using Expat??? :)
 *
 * Author: Rick Kufrin
 * Date:   10/2002
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if _REENTRANT
# if HAVE_PTHREAD
#  include <pthread.h>
# else
#  error "_REENTRANT defined but Pthread support not available"
# endif
#endif

#include <stdio.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#include <expat.h>

#include <psdebug.h>

#include "eventlist.h"
#include "hwpc.h"

BEGIN_C_DECLS
static void cdata PARAMS((void *data, const XML_Char *s, int len));
static void end   PARAMS((void *data, const char *el));
static void start PARAMS((void *data, const char *el, const char **attr));
END_C_DECLS

/*
 * Apparently older versions of Expat don't seem to define this
 */
#ifndef XML_STATUS_ERROR
# define XML_STATUS_ERROR 0
#endif

#ifndef BUFSIZ
# define BUFSIZ 4096
#endif

/* 
 * We only have one state at the moment (PARSINGEVENT) 
 * This state indicates that we're interested in the cdata
 * associated with this element.
 * Might expand later.
 */

#define PARSINGEVENTLIST 0x0001
#define PARSINGEVENT     0x0002

/*
 * This is the data that we pass around among
 * the parser handlers.  This is either stuff that
 * helps keep track of what we're doing or things
 * that we want to save from the document.
 */

typedef struct idata {
    ps_hwpc_class_t     class;
    unsigned int        depth;
    ps_hwpc_domain_t    domain;
    ps_hwpc_event_t     *events;
    unsigned short      numevents;
    unsigned short      profiling;
    ps_hwpc_threshold_t threshold;
    unsigned short      state;
    unsigned short      validdoc;
} idata_t;

static void
start(void *data, const char *el, const char **attr)
{
    idata_t *idata = (idata_t *) data;
    int i, isPreset = 0;
    const char *eventName = NULL;

    if ( strcmp(el,EVENTTAG) == 0 ) {

        /*
         * Check to see if we've consumed all the available event
         * space.  If so, expand by increments of MAXEVENTS.
         * This memory is deallocated before exiting the parser.
         */
        if ( (idata->numevents > 0) &&
             ((idata->numevents % MAXEVENTS) == 0) ) {
            idata->events = (ps_hwpc_event_t *) 
                realloc(idata->events,
                        (idata->numevents+MAXEVENTS)*sizeof(ps_hwpc_event_t));
            if ( idata->events == NULL ) {
                PS_DEBUG(PS_DEBUG_FATAL, \
                         fputs("XML parser memory realloc failed\n",stderr);)
            }
            PS_DEBUG(PS_DEBUG_INFO, \
                     printf("XML parser: event limit (%d) reached.\n", \
                             idata->numevents);)
            PS_DEBUG(PS_DEBUG_INFO, \
                     printf("XML parser: reallocating for %d events.\n", \
                             idata->numevents + MAXEVENTS);)
        }

        /* 
         * The only valid cdata is in a native event.  For the
         * rest, we rely on event name mnemonics (the distinction
         * between native and other events only makes sense in the
         * context of PAPI). 
         */

        for ( i = 0; attr[i]; i += 2 ) {
            if ( strcmp(attr[i], EVENTTAGTYPEATT) == 0 ) {
                if ( strcmp(attr[i+1], PRESETATTVAL) == 0 ) {
                    idata->events[idata->numevents].type = PRESET;
                    isPreset = 1;
                }
                else if ( strcmp(attr[i+1], NATIVEATTVAL) == 0 ) {
                    /*
                     * Set the flag to indicate we want to grab
                     * the cdata.
                     */
                    idata->state |= PARSINGEVENT;
                    idata->events[idata->numevents].type = NATIVE;
                    /*
                     * @@ The following break removed to avoid premature
                     *    drop out of for loop, which was shown to
                     *    cause possible ignoring of threshold values
                     *    when profiling.  Leaving the statement and
                     *    location in for now to allow further exercise.
                     *
                     * break;
                     */
                }
                else {
                    PS_DEBUG(PS_DEBUG_WARNING, \
                             fputs("XML parser: warning - encountered ", stderr); \
                             fputs("unrecognized type", stderr); \
                             fprintf(stderr,"\"%s=%s\"\" (ignored)\n",  \
                                     attr[i], attr[i+1]);)
                }
            }
            else if ( strcmp(attr[i], EVENTTAGNAMEATT ) == 0 ) {
                idata->events[idata->numevents].val = strdup(attr[i+1]);
                eventName = attr[i+1];
            }
            else if ( strcmp(attr[i], THRESHOLDATT ) == 0 ) {
                idata->threshold = strtol(attr[i+1], (char **) NULL, 10);
                PS_DEBUG(PS_DEBUG_INFO, \
                         printf("XML parser: found profiling threshold: %d\n",
                                idata->threshold);)
            }
            else {
                PS_DEBUG(PS_DEBUG_WARNING, \
                         fputs("XML parser: warning - encountered ", stderr); \
                         fputs("unrecognized attribute", stderr); \
                         fprintf(stderr,"\"%s=%s\"\" (ignored)\n",  \
                                 attr[i], attr[i+1]);)
            }
        }

        /*
         * If we found a preset event, we need to make sure that
         * an event name was also supplied as an attribute.
         * This is pretty much the only "validating" that's done
         * in this parser.  All events should be presets except
         * when using the PAPI layer.
         */

        if ( isPreset ) {
            if ( eventName != NULL ) {
                PS_DEBUG(PS_DEBUG_INFO, \
                         printf("XML parser: found preset: \"%s\"\n", \
                                eventName);)
            } else {
                PS_DEBUG(PS_DEBUG_WARNING, \
                         fputs("XML parser: warning - invalid input:",stderr); \
                         fputs("\n\tpreset event", stderr); \
                         fputs(" specified with no \"name\" attribute.\n", \
                               stderr);)
            }
        }

        /* Increment here, since there may be no end tag. */
        idata->numevents++;
    } 
    else if ( (strcmp(el, EVENTLISTTAG) == 0) ||
              (strcmp(el,   PROFILETAG) == 0) ) {
        /* Check to make sure that these tags are the root element */
        if ( idata->depth == 0 ) {
            idata->validdoc = 1; 
        }
        else {
            /* @@ This is a little confusing as it will indicate an invalid
             * XML document, which is true but seems overloaded with
             * an invalid root element (not EVENTLISTTAG or PROFILETAG)
             */
            idata->validdoc = 0;
            PS_DEBUG(PS_DEBUG_WARNING, \
                     fputs("XML parser: error - top-level element ", stderr); \
                     fprintf(stderr, "%s found at depth %d.\n", \
                             el, idata->depth);)
        }
        if ( strcmp(el, PROFILETAG) == 0 ) {
            idata->profiling = 1;
        }
        for ( i = 0; attr[i]; i++ ) {
            if ( strcmp(attr[i], EVENTLISTCLASSATT ) == 0 ) {
                if ( strcmp(attr[i+1], PAPICLASSATTVAL) == 0 ) {
                    idata->class = PAPI;
                }
                else if ( strcmp(attr[i+1], NULLCLASSATTVAL) == 0 ) {
                    idata->class = HWPCNULL;
                }
                else if ( strcmp(attr[i+1], PERFMONCLASSATTVAL) == 0 ) {
                    idata->class = PERFMON;
                }
                else if ( strcmp(attr[i+1], ITIMERCLASSATTVAL) == 0 ) {
                    idata->class = ITIMER;
                }
                else {
                    idata->class = UNKNOWN;
                }
                PS_DEBUG(PS_DEBUG_INFO, \
                         fputs("XML parser: parsing expected XML ", stdout); \
                         printf("document element %s (class = \"", el); \
                         switch (idata->class) { \
                         case PAPI:       fputs("PAPI", stdout); break; \
                         case HWPCNULL:   fputs("null", stdout); break; \
                         case PERFMON:    fputs("perfmon", stdout); break; \
                         case ITIMER:     fputs("itimer", stdout); break; \
                         case UNKNOWN:    fputs("unknown", stdout); break; \
                         } \
                         puts("\")");)
            }
            else if ( strcmp(attr[i], EVENTLISTDOMAINATT ) == 0 ) {
                if ( strcmp(attr[i+1], USERDOMAINATTVAL) == 0 ) {
                    idata->domain = USER;
                }
                else if ( strcmp(attr[i+1], KERNELDOMAINATTVAL) == 0 ) {
                    idata->domain = KERNEL;
                }
                else if ( strcmp(attr[i+1], ALLDOMAINATTVAL) == 0 ) {
                    idata->domain = ALL;
                }
                else {
                    idata->domain = UNKNOWN_DOMAIN;
                    PS_DEBUG(PS_DEBUG_WARNING, \
                             fputs("XML parser: unknown domain ", stderr); \
                             fprintf(stderr, "attribute: %s\n", \
                                     attr[i+1]);)
                }
                PS_DEBUG(PS_DEBUG_INFO, \
                         fputs("XML parser: domain set in XML ", stdout); \
                         fputs("configuration file as \"", stdout); \
                         switch ( idata->domain ) { \
                         case USER:           fputs("user", stdout); break; \
                         case KERNEL:         fputs("kernel", stdout); break; \
                         case ALL:            fputs("all", stdout); break; \
                         case UNKNOWN_DOMAIN: fputs("unknown", stdout); break; \
                         } \
                         puts("\"");) 
            }
        }
    }

    idata->depth++;

    return;
}

static void
end(void *data, const char *el)
{
    idata_t *idata = (idata_t *) data;

    if ( ( idata->state & PARSINGEVENT ) && ( ! strcmp(el,EVENTTAG) ) ) {
        idata->state &= ~PARSINGEVENT;
    }

    idata->depth--;

    return;
}

static void
cdata(void *data, const XML_Char *s, int len)
{
    int i;
    idata_t *idata = (idata_t *) data;
    /* 
     * The start tag already advanced the count, so adjust.
     */
    int event = idata->numevents-1;

    if ( idata->state & PARSINGEVENT ){
        idata->events[event].val =
            (char *) malloc((len+1) * sizeof(char));
        if ( idata->events[event].val == NULL ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("XML parser: malloc of event val failed\n", stderr);)
        }
        for ( i = 0; i < len; i++ ) {
            idata->events[event].val[i] = s[i];
        }
        idata->events[event].val[len] = '\0';
        PS_DEBUG(PS_DEBUG_INFO, printf("XML parser: parsed XML CDATA: %s\n", \
                                       idata->events[event].val);)
    }

    return;
}

/*
 * "Validation mode" (globals == NULL) not tested or complete (not
 * even sure I want it...)
 */
int
_ps_parseEventListFromStream(FILE *fin, ps_hwpc_globals_t *globals)
{
    char *buffer;
    XML_Parser p;
    int done, len;
    idata_t idata;
    int event;

    /*
     * Allocate a buffer for Expat's use during parsing.
     */

    buffer = (char *) malloc(BUFSIZ);
    if ( buffer == NULL ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("XML parser: memory allocation error\n", stderr);)
        return PS_ENOMEM;
    }

    /*
     * Create a new parser, and set handlers for the
     * elements, character data, and private data.
     */

    p = XML_ParserCreate(NULL);

    if ( ! p ) {
        PS_DEBUG(PS_DEBUG_FATAL, \
                 fputs("XML parser: couldn't allocate memory for parser\n", \
                       stderr);)
        free(buffer);
        return PS_ENOMEM;
    }

    XML_SetElementHandler(p, start, end);

    XML_SetCharacterDataHandler(p, cdata);

    XML_SetUserData(p, &idata);

    /*
     * Initialize the private data that will be passed along
     * to the handlers as we parse.
     */

    idata.state     = 0x0;
    idata.class     = UNKNOWN;
    idata.depth     = 0;
    idata.domain    = UNKNOWN_DOMAIN;
    idata.events    = (ps_hwpc_event_t *) 
        malloc(MAXEVENTS * sizeof(ps_hwpc_event_t));
    if ( idata.events == NULL ) {
        return PS_ENOMEM;
    }
    idata.numevents = 0;
    idata.profiling = 0;
    idata.threshold = 0;
    idata.validdoc  = 0;

    /*
     * Here's where the work gets done by Expat.  We just
     * hand off a buffer of data at a time until the
     * file is read or we run into a parse error.
     */

    for ( ;; ) {
        len = fread(buffer, 1, BUFSIZ, fin);
        if ( ferror(fin) ) {
            PS_DEBUG(PS_DEBUG_FATAL, fputs("XML parser: read error\n", stderr);)
            free(buffer);
            free(idata.events);
            fclose(fin);
            return PS_ESYS;
        }
        done = feof(fin);

        /* 
         * Give the parser another chunk.  If it's the last one,
         * we're done after parsing it.
         */

        if ( XML_Parse(p, buffer, len, done) == XML_STATUS_ERROR ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fprintf(stderr, \
                             "XML parser: parsing error at line %d:\n%s\n", \
                             (int) XML_GetCurrentLineNumber(p), \
                             XML_ErrorString(XML_GetErrorCode(p)));)
            free(buffer);
            for ( event = 0; event < idata.numevents; event++ ) {
                free(idata.events[event].val);
            }
            free(idata.events);
            fclose(fin);
            return PS_HWPC_EXMLPARSE;
        }

        if ( done )
            break;
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("XML parser: found %d events:\n", idata.numevents); \
             for ( event = 0; event < idata.numevents; event++ ) { \
                 printf("\t%s\n", idata.events[event].val); \
             })

    free(buffer);

    if ( globals != NULL ) {
        /*
         * Place all the information we found in the return array 
         */

        globals->events.events = (ps_hwpc_event_t *) 
               malloc(idata.numevents * sizeof(ps_hwpc_event_t));
        if ( globals->events.events == NULL ) {
            PS_DEBUG(PS_DEBUG_FATAL, \
                     fputs("XML parser: malloc failed in return allocation\n",\
                           stderr);)
        } else {
            for ( event = 0; event < idata.numevents; event++ ) {
                globals->events.events[event].type = idata.events[event].type;
                globals->events.events[event].val = idata.events[event].val;
            }
        }
        globals->events.class     = idata.class;
        globals->events.numEvents = idata.numevents;
        if ( idata.domain != UNKNOWN_DOMAIN ) {
            globals->domain       = idata.domain;
        }
        globals->profiling        = idata.profiling;
        if ( idata.threshold ) {
            globals->threshold    = idata.threshold;
        }
    }

    /*
     * We won't free the memory allocated for the event names/values
     * because they have been stored in the return vector unless
     * this is just a validation call (events == NULL).
     *
     * @@ NOTE: freeing for validation calls still needs to be done.
     */

    free(idata.events);

    /* All done with the parser, too. */

    XML_ParserFree(p);

    if ( ! idata.validdoc ) {
        return PS_EBADXMLDOCTYPE;
    }
  
    return PS_SUCCESS;
}
