/* $Id: map.c,v 1.20 2010/10/19 21:44:10 ruiliu Exp $ */

/****************************************************************************
 *
 *                      University of Illinois/NCSA
 *                          Open Source License
 * 
 *          Copyright(C) 2006-2010, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                             Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * Author: Rick Kufrin
 * Date: October 15, 2006
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#if HAVE_LIMITS_H
# include <limits.h>
#endif

#include <perfsuite.h>
#include <psdebug.h>
#include <psx.h>

/*
 * Force failure of all _psx_bfd_XX functions if use of BFD has been
 * disabled for any reason. All entry points return PS_ENOTINSTALLED.
 */
#if (! BFD_ENABLED) && defined(HAVE_BFD_H)
# undef HAVE_BFD_H
#endif

#if HAVE_BFD_H
# include <bfd.h>

static void lookup_address(bfd *abfd, asection *section, PTR pdata);

/*
 * This is an internal data structure that will be passed to the BFD
 * routine bfd_map_over_sections and bfd_find_nearest_line.
 */

typedef struct _lookup_data {
    bfd_vma addr;               /* Input        */
    asymbol **symtab;           /* Input        */
    int found;                  /* Input/Output */
    int line;                   /* Output       */
    const char *file;           /* Output       */
    const char *function;       /* Output       */
    struct bfd_hash_table *ht;  /* Input/Output */
} lookup_data_t;

/*
 * Internal data structure that the user represents through the opaque
 * type "ps_bfd_t".
 */

typedef struct _ps_bfd_internal {
    char *file;                 /* The actual binary file in use    */
    bfd *abfd;                  /* The corresponding BFD            */
    lookup_data_t data;         /* Internal data to/from BFD        */
    struct bfd_hash_table ht;   /* BFD hash table of file/functions */
} ps_bfd_internal_t;



/*
 * Initialize mapping library.
 */

int
_psx_bfd_init(void)
{
    bfd_init();

    return PS_SUCCESS;
}


/*
 * _psx_bfd_open: open the binary file passed in by absolute or relative
 *                pathname "file" (NULL indicates the current executable
 *                is to be used).
 *
 * Returns an opaque handle through the output parameter "ps_bfd".
 * Successful completion indicated by return value PS_SUCCESS.
 * 
 */

int
_psx_bfd_open(char *file, ps_bfd_t *ps_bfd)
{
    char *exe, *buf;
    int commlen, ret;
    pid_t pid;
    ps_bfd_internal_t *ps_bfd_ptr;
    bfd *abfd;
    unsigned int symsize;
    long numsyms;
   
    exe = malloc(PATH_MAX);
    if ( file == NULL ) {
        buf = malloc(PATH_MAX);
        pid = getpid();
        sprintf(buf, "/proc/%d/exe", pid);
        commlen = readlink(buf, exe, PATH_MAX-1);
        if ( commlen < 0 ) {
            free(exe);
            free(buf);
            return PS_ESYS;
        }
        exe[commlen] = '\0';
        free(buf);
    }
    else {
        strcpy(exe,file);
    }

    PS_DEBUG(PS_DEBUG_VERBOSE, printf("opening \"%s\n", exe);)

    ps_bfd_ptr = (ps_bfd_internal_t *) malloc(sizeof(ps_bfd_internal_t));

    ps_bfd_ptr->abfd = NULL;

    ps_bfd_ptr->abfd = abfd = bfd_openr(exe, NULL);
    if ( abfd == NULL ) {
        free(exe);
        return PS_ESYS;
    }

    ret = bfd_check_format(abfd, bfd_archive);
    if ( ret != 0 ) {
        bfd_close(abfd);
        free(exe);
        ps_bfd_ptr->abfd = abfd = NULL;
        return PS_ESYS;
    }

    ret = bfd_check_format_matches(abfd, bfd_object, NULL);
    if ( ret == 0 ) {
        bfd_close(abfd);
        free(exe);
        ps_bfd_ptr->abfd = abfd = NULL;
        return PS_ESYS;
    }

    ps_bfd_ptr->data.symtab = (asymbol **) NULL;
    symsize = numsyms = 0;

    if ( bfd_get_file_flags(abfd) & HAS_SYMS ) {
        numsyms = bfd_read_minisymbols(abfd, FALSE,
                                       (void *) &(ps_bfd_ptr->data.symtab),
                                       &symsize);
        if ( numsyms == 0 ) {
            numsyms = bfd_read_minisymbols(abfd, TRUE,
                                           (void *) &(ps_bfd_ptr->data.symtab),
                                           &symsize);
        }
        if ( numsyms < 0 ) {
            bfd_close(abfd);
            free(exe);
            ps_bfd_ptr->abfd = abfd = NULL;
            return PS_ESYS;
        }

        /* @@ check for return value */
#if BFD_HASH_TABLE_INIT_HAS_ENTSIZE
        ret = bfd_hash_table_init(&(ps_bfd_ptr->ht), bfd_hash_newfunc,
                                  sizeof(const char *));
#else
        ret = bfd_hash_table_init(&(ps_bfd_ptr->ht), bfd_hash_newfunc);
#endif
    }

    ps_bfd_ptr->file = exe;

    *ps_bfd = ps_bfd_ptr;

    return PS_SUCCESS;
}


#if DEBUG
static bfd_boolean
print_entry(struct bfd_hash_entry *entry, void *ptr)
{
   printf("Entry: %s\n", entry->string);
   return TRUE;
}
#endif /* DEBUG */


/*
 * _psx_bfd_close: terminate use of a ps_bfd binary file and free
 *                 associated resources.
 *
 */

int
_psx_bfd_close(ps_bfd_t ps_bfd)
{
    ps_bfd_internal_t * abfd = (ps_bfd_internal_t *) ps_bfd;

    /*
     * If file flags did not indicate HAS_SYMS, then the symbol
     * table will not have been allocated on open, so avoid trying
     * to call free() or access the file/function hash table.
     */

    if ( abfd->data.symtab != NULL ) {
        free(abfd->data.symtab);

#if DEBUG
        bfd_hash_traverse(&(abfd->ht), print_entry, NULL);
#endif /* DEBUG */

        bfd_hash_table_free(&(abfd->ht));
    }

    free(abfd->file);

    bfd_close(abfd->abfd);

    free(abfd);

    return PS_SUCCESS;
}


/*
 * _psx_bfd_map: perform IP->source code mapping for an array of
 *               IP values contained in "ip".  The mapping is done
 *               with respect to the previously-opened binary file
 *               descriptor "ps_bfd".
 *
 *               Results are placed in the output argument "map".
 *               IT IS THE CALLERS RESPONSIBILITY TO ENSURE THAT
 *               SPACE HAS BEEN ALLOCATED TO HOLD "nips" RESULTS
 *               IN THIS ARRAY.
 *
 *               If mapping fails for any provided IP, the results
 *               will be NULL for file and function, and 0 for line
 *               number.
 */

int
_psx_bfd_map(ps_bfd_t ps_bfd, caddr_t *ip, int nips, ps_ip_map_t *map)
{
    ps_bfd_internal_t * abfd = (ps_bfd_internal_t *) ps_bfd;
    int i;

    abfd->data.ht = &(abfd->ht);
    for (i = 0; i < nips; i++) {
        abfd->data.addr = (bfd_vma) ip[i];
        abfd->data.found = abfd->data.line = 0;
        abfd->data.file = abfd->data.function = NULL;

        PS_DEBUG(PS_DEBUG_VERBOSE, \
               printf("Looking up address %x\n", (unsigned int) abfd->data.addr);)

        bfd_map_over_sections(abfd->abfd, lookup_address, (PTR) &(abfd->data));

        map[i].ip = ip[i];
        map[i].file = abfd->data.file;
        map[i].function = abfd->data.function;
        map[i].line = abfd->data.line;

        PS_DEBUG(PS_DEBUG_VERBOSE, printf("File: %s, Function: %s, Line: %d\n", \
                 abfd->data.file, abfd->data.function, abfd->data.line);)

    }

    return PS_SUCCESS;
}

/*
 * lookup_address: iterate over a section within the binary to determine
 *                 if the address passed in via the pdata argument is
 *                 within the section's address range.  If it is, return
 *                 the corresponding file, function, and line as given
 *                 by bfd_find_nearest_line.  If the return values are
 *                 valid (non-NULL), hash the file and function names
 *                 and return pointers to them as well as the line number
 *                 in the pdata structure.
 */

static void
lookup_address(bfd *abfd, asection *section, PTR pdata)
{
    lookup_data_t *data = (lookup_data_t *) pdata;
    bfd_vma vm_address;
    bfd_size_type section_size;
    const char *file = NULL, *function = NULL;
    unsigned int line = 0;
    int found;
    struct bfd_hash_entry *htentry;

    if ( data->found ) {
        return;
    }

    if ( (bfd_get_section_flags(abfd, section) & SEC_ALLOC) == 0 ) {
        return;
    }

    vm_address = bfd_get_section_vma(abfd, section);

    PS_DEBUG(PS_DEBUG_VERBOSE, \
             printf("lookup_address: data->addr=%x, vm_address=%x\n", \
             (unsigned int) data->addr, (unsigned int) vm_address);)

    if ( data->addr < vm_address ) {
        return;
    }

    section_size = bfd_section_size(abfd, section);

    if ( data->addr >= (vm_address + section_size) ) {
        return;
    }

    found = bfd_find_nearest_line(abfd, section, data->symtab,
                                  data->addr - vm_address,
                                  &file, &function, &line);

    if ( ! data->found ) {
        data->found = found;
        if ( file != NULL ) {
            htentry = bfd_hash_lookup(data->ht, file, TRUE, TRUE);
            data->file = htentry->string;
            /*data->file = strdup(file);*/
        }
        if ( function != NULL ) {
            htentry = bfd_hash_lookup(data->ht, function, TRUE, TRUE);
            data->function = htentry->string;
            /* data->function = strdup(function); */
        }
        data->line = line;
    }

    return;
}

/*
  _psx_bfd_inquire:

  int
  _psx_bfd_inquire(ps_bfd_t ps_bfd, psx_bfd_file_flag inflag,
                   int *retValue)

  Return value:
      1) PS_SUCCESS, if there is no error;
      2) PS_EINVALID, if argument ps_bfd is 0, or it points to a zero-valued
         abfd, or argument "inflag" is a non-supported format;
      3) PS_ENOTINSTALLED, if libbfd is not available.
         See code below in #else section, close to the bottom of this file;

  Side effect:
      *retValue will be initialized to 0. For the supported file flags,
      *retValue will be set to 1 if the flag matches.

  Supported file flags are:
      "dynamic", "executable", "has_reloc", "has_lineno",
      "has_debug", and "has_syms",
      captialized and with the prefix "PSX_BFD_FILE_FLAG_", such as:
          PSX_BFD_FILE_FLAG_DYNAMIC.
*/
int
_psx_bfd_inquire(ps_bfd_t ps_bfd, psx_bfd_file_flag inflag,
                 int *retValue)
{
    int flag = 0;
    flagword flags;
    ps_bfd_internal_t * abfd = (ps_bfd_internal_t *) ps_bfd;

    *retValue = 0;
    if (0 == abfd || abfd->abfd == 0) {
        return PS_EINVAL;
    }

    if (PSX_BFD_FILE_FLAG_DYNAMIC == inflag) {
        flag = DYNAMIC;
    } else if (PSX_BFD_FILE_FLAG_EXECUTABLE == inflag) {
        flag = EXEC_P;
    } else if (PSX_BFD_FILE_FLAG_HAS_RELOC == inflag) {
        flag = HAS_RELOC;
    } else if (PSX_BFD_FILE_FLAG_HAS_LINENO == inflag) {
        flag = HAS_LINENO;
    } else if (PSX_BFD_FILE_FLAG_HAS_DEBUG == inflag) {
        flag = HAS_DEBUG;
    } else if (PSX_BFD_FILE_FLAG_HAS_SYMS == inflag) {
        flag = HAS_SYMS;
    } else {
        return PS_EINVAL;
    }

    flags = bfd_get_file_flags (abfd->abfd);
    *retValue = ((flags & flag) != 0);
    return PS_SUCCESS;
}

#else /* ! HAVE_BFD_H */

int
_psx_bfd_init(void)
{
    return PS_ENOTINSTALLED;
}

int
_psx_bfd_open(char *file, ps_bfd_t *ps_bfd)
{
    return PS_ENOTINSTALLED;
}

int
_psx_bfd_close(ps_bfd_t ps_bfd)
{
    return PS_ENOTINSTALLED;
}

int
_psx_bfd_map(ps_bfd_t ps_bfd, caddr_t *ip, int nips, ps_ip_map_t *map)
{
    return PS_ENOTINSTALLED;
}

int
_psx_bfd_inquire(ps_bfd_t ps_bfd, psx_bfd_file_flag inflag,
                 int *errorCode)
{
    return PS_ENOTINSTALLED;
}

#endif
