/* $Id: Papi_init.c,v 1.15 2006/01/16 21:21:37 rkufrin Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *          Copyright(C) 2003-2006, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 *
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications 
 *               University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * papi.c: Tcl interface to PAPI
 *
 * Author: Rick Kufrin
 * Date:   5/23/2003
 *
 * This file is part of PerfSuite.
 *
 * @@ PAPI error strings should be propagated back in interp->result.
 */

#include "Papi.h"

static int initialize          PARAMS((Tcl_Interp *interp));
static int initializeStdEvents PARAMS((Tcl_Interp *interp));
static int initializeHWInfo    PARAMS((Tcl_Interp *interp));

int
Pspapi_Init(Tcl_Interp *interp)
{
    if ( Tcl_InitStubs(interp, TCL_REQUIRED_VERSION, 0) == NULL ) {
        return TCL_ERROR;
    }

    /*
     * We must create at least one command before calling initialize
     * to get the ::papi namespace created.
     */
    Tcl_CreateObjCommand(interp, "::papi::time",
                         PAPI_timeObjCmd,
                         (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

    if ( initialize(interp) != TCL_OK ) {
        return TCL_ERROR;
    }

    Tcl_PkgProvide(interp, PSTCL_PACKAGE_NAME, PSTCL_PACKAGE_VERSION);

    return TCL_OK;
}

static int
initialize(Tcl_Interp *interp)
{
    int papiret;
    Tcl_Obj *papiversion, *papipatchlevel, *varname, *varobj;
    char buffer[16];

    papiret = PAPI_library_init(PAPI_VER_CURRENT);
    if ( papiret != PAPI_VER_CURRENT ) {
        Tcl_SetObjResult(interp,
                         Tcl_NewStringObj("error: PAPI version mismatch", 28));
        return TCL_ERROR;
    }

#if PAPI3
# if defined(PAPI_LIB_VERSION)
    papiret = PAPI_get_opt(PAPI_LIB_VERSION, NULL);
# endif
    sprintf(buffer, "%d.%d", PAPI_VERSION_MAJOR(papiret),
                             PAPI_VERSION_MINOR(papiret));
#else /* ! PAPI3 */
    sprintf(buffer, "%d", papiret);
#endif
    papiversion = Tcl_NewStringObj(buffer, strlen(buffer));
    varname = Tcl_NewStringObj(PSTCL_PAPIVERSION, strlen(PSTCL_PAPIVERSION));

    varobj = Tcl_ObjSetVar2(interp, varname, NULL, papiversion,
                            TCL_LEAVE_ERR_MSG);
    if ( varobj == NULL ) {
        return TCL_ERROR;
    }

#if PAPI3
    sprintf(buffer, "%d.%d.%d", PAPI_VERSION_MAJOR(papiret),
                                PAPI_VERSION_MINOR(papiret),
                                PAPI_VERSION_REVISION(papiret));
#else
    sprintf(buffer, "%d", papiret);
#endif
    papipatchlevel = Tcl_NewStringObj(buffer, strlen(buffer));
    varname = Tcl_NewStringObj(PSTCL_PAPIPATCHLEVEL, 
                               strlen(PSTCL_PAPIPATCHLEVEL));

    varobj = Tcl_ObjSetVar2(interp, varname, NULL, papipatchlevel,
                            TCL_LEAVE_ERR_MSG);
    if ( varobj == NULL ) {
        return TCL_ERROR;
    }


    if ( initializeStdEvents(interp) != TCL_OK ) {
        return TCL_ERROR;
    }

    if ( initializeHWInfo(interp) != TCL_OK ) {
        return TCL_ERROR;
    }

    return TCL_OK;
}

/*
 * initializeStdEvents: create an associative array that describes
 *                      all PAPI standard events.  Each element is indexed
 *                      by its PAPI name (e.g., PAPI_TOT_CYC) and is a
 *                      list with the following elements:
 *
 *                      code description label status derived note
 *
 *                      code: unsigned int (stored as a long)
 *                      description: string
 *                      label: string
 *                      status: "available" or "unavailable"
 *                      derived: "derived" or "nonderived"
 *                      note: string
 */

static int
initializeStdEvents(Tcl_Interp *interp)
{
    const PAPI_preset_info_t * eventinfo;
    int event, numevents;
    Tcl_Obj *arrayname, *element, *list, **available, **derived,
        *nullstring, **objv;

    eventinfo = PAPI_query_all_events_verbose();

    if ( eventinfo == NULL ) {
        Tcl_SetObjResult(interp,
                         Tcl_NewStringObj("error retrieving event information",
                                          34));
        return TCL_ERROR;
    }

    /* Allocate string objects that we'll using */
    arrayname    = Tcl_NewStringObj(PSTCL_EVENTARRAY,
                                    strlen(PSTCL_EVENTARRAY));
    derived      = (Tcl_Obj **) Tcl_Alloc(2 * sizeof(Tcl_Obj *));
    derived[0]   = Tcl_NewStringObj(NONDERIVED, strlen(NONDERIVED));
    derived[1]   = Tcl_NewStringObj(DERIVED, strlen(DERIVED));

    available    = (Tcl_Obj **) Tcl_Alloc(2 * sizeof(Tcl_Obj *));
    available[0] = Tcl_NewStringObj(UNAVAILABLE, strlen(UNAVAILABLE));
    available[1] = Tcl_NewStringObj(AVAILABLE, strlen(AVAILABLE));

    nullstring   = Tcl_NewStringObj("", 0);

    /* Count the number of preset events */
    for (numevents=0; eventinfo[numevents].event_name; numevents++) ;

    for (event=0; event<numevents; event++) {
        objv = (Tcl_Obj **) Tcl_Alloc(PSTCL_EVENTARRAYLEN * sizeof(Tcl_Obj *));

        element = Tcl_NewStringObj(eventinfo[event].event_name,
                                   strlen(eventinfo[event].event_name));

        objv[0] = Tcl_NewLongObj(eventinfo[event].event_code);
        objv[1] = Tcl_NewStringObj(eventinfo[event].event_descr, 
                                   strlen(eventinfo[event].event_descr));
        objv[2] = Tcl_NewStringObj(eventinfo[event].event_label, 
                                   strlen(eventinfo[event].event_label));
        objv[3] = available[eventinfo[event].avail];
        if ( eventinfo[event].avail ) {
            objv[4] = derived[eventinfo[event].flags & PAPI_DERIVED];
        }
        else {
            objv[4] = nullstring;
        }
        if (eventinfo[event].event_note) {
            objv[5] = Tcl_NewStringObj(eventinfo[event].event_note, 
                                       strlen(eventinfo[event].event_note));
        }
        else {
            objv[5] = nullstring;
        }

        list = Tcl_NewListObj(PSTCL_EVENTARRAYLEN, objv);

        if (Tcl_ObjSetVar2(interp, arrayname, element, list,
                           TCL_LEAVE_ERR_MSG) == NULL) {
            return TCL_ERROR;
        }
    }

    return TCL_OK;
}

static int
initializeHWInfo(Tcl_Interp *interp)
{
    const PAPI_hw_info_t *hwinfo;
    Tcl_Obj *obj, *arrayname, *element, *nullstring;

    hwinfo = PAPI_get_hardware_info();

    if ( hwinfo == NULL ) {
        Tcl_SetObjResult(interp,
                         Tcl_NewStringObj("error retrieving hardware information",
                                          37));
        return TCL_ERROR;
    }

    arrayname  = Tcl_NewStringObj(PSTCL_HWINFOARRAY, strlen(PSTCL_HWINFOARRAY));
    nullstring = Tcl_NewStringObj("", 0);

    element = Tcl_NewStringObj(NCPUS, strlen(NCPUS));
    obj = Tcl_NewIntObj(hwinfo->ncpu);
    if ( Tcl_ObjSetVar2(interp, arrayname, element, obj,
                        TCL_LEAVE_ERR_MSG) == NULL ) {
        return TCL_ERROR;
    }

    element = Tcl_NewStringObj(NNODES, strlen(NNODES));
    obj = Tcl_NewIntObj(hwinfo->nnodes);
    if ( Tcl_ObjSetVar2(interp, arrayname, element, obj,
                       TCL_LEAVE_ERR_MSG) == NULL ) {
        return TCL_ERROR;
    }

    element = Tcl_NewStringObj(TOTALCPUS, strlen(TOTALCPUS));
    obj = Tcl_NewIntObj(hwinfo->totalcpus);
    if ( Tcl_ObjSetVar2(interp, arrayname, element, obj,
                        TCL_LEAVE_ERR_MSG) == NULL ) {
        return TCL_ERROR;
    }

    element = Tcl_NewStringObj(VENDOR, strlen(VENDOR));
    obj = Tcl_NewIntObj(hwinfo->vendor);
    if ( Tcl_ObjSetVar2(interp, arrayname, element, obj,
                        TCL_LEAVE_ERR_MSG) == NULL ) {
        return TCL_ERROR;
    }

    element = Tcl_NewStringObj(VENDORSTRING, strlen(VENDORSTRING));
    if ( hwinfo->vendor_string ) {
        obj = Tcl_NewStringObj(hwinfo->vendor_string,
                               strlen(hwinfo->vendor_string));
    } else {
        obj = nullstring;
    }
    if ( Tcl_ObjSetVar2(interp, arrayname, element, obj,
                        TCL_LEAVE_ERR_MSG) == NULL ) {
        return TCL_ERROR;
    }

    element = Tcl_NewStringObj(MODEL, strlen(MODEL));
    obj = Tcl_NewIntObj(hwinfo->model);
    if ( Tcl_ObjSetVar2(interp, arrayname, element, obj,
                        TCL_LEAVE_ERR_MSG) == NULL ) {
        return TCL_ERROR;
    }

    element = Tcl_NewStringObj(MODELSTRING, strlen(MODELSTRING));
    if ( hwinfo->model_string ) {
        obj = Tcl_NewStringObj(hwinfo->model_string,
                               strlen(hwinfo->model_string));
    }
    else {
        obj = nullstring;
    }
    if ( Tcl_ObjSetVar2(interp, arrayname, element, obj,
                        TCL_LEAVE_ERR_MSG) == NULL ) {
        return TCL_ERROR;
    }

    element = Tcl_NewStringObj(REVISION, strlen(REVISION));
    obj = Tcl_NewIntObj(hwinfo->revision);
    if ( Tcl_ObjSetVar2(interp, arrayname, element, obj,
                        TCL_LEAVE_ERR_MSG) == NULL ) {
        return TCL_ERROR;
    }

    element = Tcl_NewStringObj(MHZ, strlen(MHZ));
    obj = Tcl_NewDoubleObj(hwinfo->mhz);
    if ( Tcl_ObjSetVar2(interp, arrayname, element, obj,
                        TCL_LEAVE_ERR_MSG) == NULL ) {
        return TCL_ERROR;
    }

    element = Tcl_NewStringObj(NCOUNTERS, strlen(NCOUNTERS));
    obj = Tcl_NewIntObj(PAPI_get_opt(PAPI_GET_MAX_HWCTRS, 
                                     (PAPI_option_t *) NULL));
    if ( Tcl_ObjSetVar2(interp, arrayname, element, obj,
                        TCL_LEAVE_ERR_MSG) == NULL ) {
        return TCL_ERROR;
    }

    return TCL_OK;
}
