/* $Id: Bfd_control.c,v 1.15 2009/07/01 16:16:43 rkufrin Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *         Copyright(C) 2003-2006, The Board of Trustees of the
 *             University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications 
 *              University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * Bfd_control.c: Tcl interface to BFD
 *
 * Author: Rick Kufrin
 * Date:   5/23/2003
 *
 * This file is part of PerfSuite.
 *
 */

#include "Bfd.h"

#if BFD_ENABLED

int
BFD_openObjCmd(ClientData clientData, Tcl_Interp *interp,
               int objc, Tcl_Obj *CONST objv[])
{
    bfdinfo_t *bfdinfo = (bfdinfo_t *) clientData;
    const char *errmsg;
    int ret;

    if ( objc != 2 ) {
        Tcl_WrongNumArgs(interp, 1, objv, "object-file");
        return TCL_ERROR;
    }

    if ( bfdinfo->abfd != NULL ) {
        Tcl_SetObjResult(interp, Tcl_NewStringObj(PSBFD_EACTIVE,
                                                  strlen(PSBFD_EACTIVE)));
        return TCL_ERROR;
    }

    bfdinfo->exe = strdup(Tcl_GetString(objv[1]));

    bfdinfo->abfd = bfd_openr(bfdinfo->exe, NULL);
    if ( bfdinfo->abfd == NULL ) {
        errmsg = bfd_errmsg(bfd_get_error());
        free(bfdinfo->exe);
        Tcl_SetObjResult(interp, Tcl_NewStringObj(errmsg, strlen(errmsg)));
        return TCL_ERROR;
    }

    ret = bfd_check_format(bfdinfo->abfd, bfd_archive);
    if ( ret != 0 ) {
        /* @@ Package up these operations into a common function */
        errmsg = bfd_errmsg(bfd_get_error());
        free(bfdinfo->exe);
        bfd_close(bfdinfo->abfd);
        bfdinfo->abfd = NULL;
        Tcl_SetObjResult(interp, Tcl_NewStringObj(errmsg, strlen(errmsg)));
        return TCL_ERROR;
    }

    ret = bfd_check_format_matches(bfdinfo->abfd, bfd_object, NULL);
    if ( ret == 0 ) {
        /* @@ Use common function above */
        errmsg = bfd_errmsg(bfd_get_error());
        free(bfdinfo->exe);
        bfd_close(bfdinfo->abfd);
        bfdinfo->abfd = NULL;
        Tcl_SetObjResult(interp, Tcl_NewStringObj(errmsg, strlen(errmsg)));
        return TCL_ERROR;
    }

    bfdinfo->symtab  = (asymbol **) NULL;
    bfdinfo->symsize = 0;
    bfdinfo->numsyms = 0;

    if ( bfd_get_file_flags(bfdinfo->abfd) & HAS_SYMS ) {
        bfdinfo->numsyms = bfd_read_minisymbols(bfdinfo->abfd, FALSE,
                                                (void *) &(bfdinfo->symtab),
                                                &(bfdinfo->symsize));
        if ( bfdinfo->numsyms == 0 ) {
            bfdinfo->numsyms = bfd_read_minisymbols(bfdinfo->abfd, TRUE,
                                                    (void *) &(bfdinfo->symtab),
                                                    &(bfdinfo->symsize));
        }
        if ( bfdinfo->numsyms < 0 ) {
            errmsg = bfd_errmsg(bfd_get_error());
            Tcl_SetObjResult(interp, Tcl_NewStringObj(errmsg, strlen(errmsg)));
            return TCL_ERROR;
        }
    }

    Tcl_SetObjResult(interp, Tcl_NewStringObj(bfdinfo->exe,
                                              strlen(bfdinfo->exe)));

    return TCL_OK;
}

int
BFD_closeObjCmd(ClientData clientData, Tcl_Interp *interp,
                int objc, Tcl_Obj *CONST objv[])
{
    bfdinfo_t *bfdinfo = (bfdinfo_t *) clientData;

    if (bfdinfo->abfd != NULL) {
        free(bfdinfo->symtab);
        bfd_close(bfdinfo->abfd);
        bfdinfo->abfd = NULL;
        free(bfdinfo->exe);
    }

    return TCL_OK;
}

#else /* ! BFD_ENABLED */
int
BFD_openObjCmd(ClientData clientData, Tcl_Interp *interp,
               int objc, Tcl_Obj *CONST objv[])
{
    Tcl_SetObjResult(interp, Tcl_NewStringObj(PSBFD_ENOTINSTALLED,
                                              strlen(PSBFD_ENOTINSTALLED)));
    return TCL_ERROR;
}


int
BFD_closeObjCmd(ClientData clientData, Tcl_Interp *interp,
                int objc, Tcl_Obj *CONST objv[])
{
    Tcl_SetObjResult(interp, Tcl_NewStringObj(PSBFD_ENOTINSTALLED,
                                              strlen(PSBFD_ENOTINSTALLED)));
    return TCL_ERROR;
}

#endif
