/* $Id: Bfd_init.c,v 1.10 2009/07/01 16:16:43 rkufrin Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *         Copyright(C) 2003-2006, The Board of Trustees of the
 *             University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications 
 *              University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * Bfd_init.c: Tcl interface to BFD
 *
 * This is the initialization routine for the PerfSuite "psbfd" Tcl
 * extension.  It registers the new commands exported to the Tcl 
 * interpreter by the extension.
 *
 * The extension doesn't follow an "object-oriented" approach to
 * working with BFDs (where, for example, an "open" call would return
 * a handle to a particular BFD that is then used to perform
 * other tasks specific to that BFD).  Instead, there is always
 * only zero or one BFD open/active at any time.  The reason why
 * this design choice was made was to conserve on memory and
 * to keep implementation and usage as simple as possible.
 *
 * This extension implements a very minimal subset of the facilities
 * offered by BFD, currently only those necessary to perform address
 * to source code mappings and to do some basic inquiries about
 * binary files.
 *
 * Author: Rick Kufrin
 * Date:   5/23/2003
 *
 * This file is part of PerfSuite.
 *
 */

#include "Bfd.h"

int
Psbfd_Init(Tcl_Interp *interp)
{
#if BFD_ENABLED
    bfdinfo_t *bfdinfo;
    char fullversion[16];
    Tcl_Obj *version, *varname, *varobj;

    if ( Tcl_InitStubs(interp, TCL_REQUIRED_VERSION, 0) == NULL ) {
        return TCL_ERROR;
    }

    bfd_init();

    bfdinfo = (bfdinfo_t *) Tcl_Alloc(sizeof(bfdinfo_t));
    bfdinfo->abfd = NULL;

    Tcl_CreateObjCommand(interp, "::psbfd::open",
                         BFD_openObjCmd,
                         (ClientData) bfdinfo, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateObjCommand(interp, "::psbfd::inquire",
                         BFD_inquireObjCmd,
                         (ClientData) bfdinfo, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateObjCommand(interp, "::psbfd::lookup",
                         BFD_lookupObjCmd,
                         (ClientData) bfdinfo, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateObjCommand(interp, "::psbfd::close",
                         BFD_closeObjCmd,
                         (ClientData) bfdinfo, (Tcl_CmdDeleteProc *) NULL);

    varname = Tcl_NewStringObj(PSBFD_VERSION_VARNAME,
                               strlen(PSBFD_VERSION_VARNAME));
    version = Tcl_NewStringObj(PSBFD_PACKAGE_VERSION,
                               strlen(PSBFD_PACKAGE_VERSION));
    varobj = Tcl_ObjSetVar2(interp, varname, NULL, version,
                            TCL_LEAVE_ERR_MSG);
    if ( varobj == NULL ) {
        return TCL_ERROR;
    }

    sprintf(fullversion, "%s.%s", PSBFD_PACKAGE_VERSION, PSBFD_PATCHLEVEL);
    varname = Tcl_NewStringObj(PSBFD_PATCHLEVEL_VARNAME,
                               strlen(PSBFD_PATCHLEVEL_VARNAME));
    version = Tcl_NewStringObj(fullversion, strlen(fullversion));
    varobj = Tcl_ObjSetVar2(interp, varname, NULL, version,
                            TCL_LEAVE_ERR_MSG);
    if ( varobj == NULL ) {
        return TCL_ERROR;
    }

    Tcl_PkgProvide(interp, PSBFD_PACKAGE_NAME, fullversion);

    return TCL_OK;

#else /* ! BFD_ENABLED */

    Tcl_SetObjResult(interp, Tcl_NewStringObj(PSBFD_ENOTINSTALLED,                                               strlen(PSBFD_ENOTINSTALLED)));
    return TCL_ERROR;

#endif
}
