/* $Id: Bfd_lookup.c,v 1.21 2011/02/11 19:39:29 rkufrin Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *         Copyright(C) 2003-2006, The Board of Trustees of the
 *             University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * Bfd_lookup.c: perform a lookup on an address using an active BFD.
 *
 * Author: Rick Kufrin
 * Date:   5/23/2003
 *
 * This file is part of PerfSuite.
 */

#include "Bfd.h"

#if BFD_ENABLED

static void lookup_address PARAMS((bfd *abfd, asection *section,
                                   PTR bfdinfo));

/*
 * BFD_lookupObjCmd : the object procedure publicly-accessible through
 *                    the psbfd Tcl extension.
 */

int
BFD_lookupObjCmd(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[])
{
    bfdinfo_t *bfdinfo = (bfdinfo_t *) clientData;
    char *addr;
    Tcl_Obj *list, **listv, *nullstring;

    /* Check to make sure a BFD is active/opened */
    if ( bfdinfo->abfd == NULL ) {
        Tcl_SetObjResult(interp, Tcl_NewStringObj(PSBFD_ENOBFD,
                                                  strlen(PSBFD_ENOBFD)));
        return TCL_ERROR;
    }

    /* 
     * Verify that the right number of arguments were supplied (this
     * command takes one argument: an address to lookup).
     */
    if ( objc != 2 ) {
        Tcl_WrongNumArgs(interp, 1, objv, "address (in hex)");
        return TCL_ERROR;
    }

    /*
     * Convert the string-valued argument to a number and initialize
     * variables that will be used/returned in the lookup.
     */
    addr = Tcl_GetString(objv[1]);
    sscanf(addr, "%lx", (long unsigned int *) &(bfdinfo->addr));

    bfdinfo->found    = 0;
    bfdinfo->file     = NULL;
    bfdinfo->function = NULL;
    bfdinfo->line     = 0;

    /* Do the lookup */
    /* 
     * @@ bfdinfo->exe is the first member in the structure, this
     * seems a little fragile!
     */
    bfd_map_over_sections(bfdinfo->abfd, lookup_address, 
                          (PTR) (bfdinfo + offsetof(bfdinfo_t, exe)));

    /* Report error if the address wasn't found */
    if ( ! bfdinfo->found ) {
        Tcl_SetObjResult(interp, Tcl_NewStringObj(PSBFD_ENOTFOUND,
                                                  strlen(PSBFD_ENOTFOUND)));
        return TCL_ERROR;
    }

    /* 
     * The address was found - create the list that will be returned.
     * The list is a 3-element list, consisting of:
     *    {function file line}
     * If either of function or file aren't available, then their
     * elements will be have been set to "?" (so the check against
     * NULL isn't really necessary).
     */
    nullstring   = Tcl_NewStringObj("", 0);

    listv = (Tcl_Obj **) Tcl_Alloc(sizeof(Tcl_Obj *) * 3);
    if ( bfdinfo->function != NULL ) {
        listv[0] = Tcl_NewStringObj(bfdinfo->function,
                                    strlen(bfdinfo->function));
    }
    else {
        listv[0] = nullstring;
    }
    if ( bfdinfo->file != NULL ) {
        listv[1] = Tcl_NewStringObj(bfdinfo->file, strlen(bfdinfo->file));
    }
    else {
        listv[1] = nullstring;
    }
    listv[2] = Tcl_NewLongObj(bfdinfo->line);

    list = Tcl_NewListObj(3, listv);

    Tcl_SetObjResult(interp, list);

    return TCL_OK;
}


/* lookup_address : internal function invoked repeatedly by
 *                  bfd_map_over_sections().
 */
static void
lookup_address(bfd *abfd, asection *section, PTR bfdinfo)
{
    bfdinfo_t *bfd = (bfdinfo_t *) bfdinfo;
    bfd_vma vm_address;
    bfd_size_type section_size;
    const char *file = NULL, *function = NULL;
    unsigned int line = 0, len;
    int found;
    
    /* Was the address already found? */
    if ( bfd->found ) {
        return;
    }

    /* Only deal with specific BFD section types */
    if ( (bfd_get_section_flags(abfd, section) & SEC_ALLOC) == 0 ) {
        return;
    }

    /* 
     * Get the starting address and size of this section and check to
     * see if the address we are processing falls in that range.
     */
    vm_address = bfd_get_section_vma(abfd, section);
    if ( bfd->addr < vm_address ) {
        return;
    }

#if defined(bfd_get_section_size_before_reloc)
    section_size = bfd_get_section_size_before_reloc(section);
#else
    /* Could also use bfd_get_section_size(section) */
    section_size = bfd_section_size(abfd, section);
#endif
    if ( bfd->addr >= (vm_address + section_size) ) {
        return;
    }

    /* Grab the info we need */
    found = bfd_find_nearest_line(abfd, section, bfd->symtab,
                                  bfd->addr - vm_address,
                                  &file, &function, &line);
    
    /*
     * @@ Not sure where the storage for the character strings
     * is obtained
     */

    if ( ! bfd->found ) {
        bfd->found = found;
        if ( file == NULL ) {
            file = UNKNOWN_SYMBOL;
        }
        len = strlen(file);
        bfd->file = Tcl_Alloc(len+1);
        if ( bfd->file == NULL ) {
            bfd->found = 0;
            return;
        }
        strcpy(bfd->file, file);
        /* fprintf(stderr, "function = %s\n", function); */

        if ( function == NULL ) {
            function = UNKNOWN_SYMBOL;
        }
        len = strlen(function);
        bfd->function = Tcl_Alloc(len+1);
        if ( bfd->function == NULL) {
            bfd->found = 0;
            Tcl_Free(bfd->file);
            return;
        }
        strcpy(bfd->function, function);

        bfd->line = line;
        /*
        printf("returning:\n\tfile=\"%s\"\n\tfunction=\"%s\" (%s)\n\tline=%d\n",
               bfd->file, bfd->function, function, bfd->line);
        */
    }

    return;
}

#else /* ! BFD_ENABLED */

int
BFD_lookupObjCmd(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[])
{
    Tcl_SetObjResult(interp, Tcl_NewStringObj(PSBFD_ENOTINSTALLED,                                               strlen(PSBFD_ENOTINSTALLED)));
    return TCL_ERROR;
}

#endif
