/* $Id: psinv.c,v 1.94 2013/04/16 15:56:41 ruiliu Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *   Copyright(C) 2002-2010, 2012-2013, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *           National Center for Supercomputing Applications 
 *             University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * psinv: display machine inventory and information regarding the PerfSuite build
 *
 * Author: Rick Kufrin
 * Date:   12/2002
 *
 * This file is part of PerfSuite.
 */

#define PROGRAM_VERSION "1.1"

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if HAVE_SYS_UTSNAME_H
# include <sys/utsname.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#if HAVE_TIME_H
# include <time.h>
#endif

#if HAVE_GETOPT_H
# include <getopt.h>
#endif

#if HAVE_GNU_LIBC_VERSION_H
# include <gnu/libc-version.h>
#endif

#include <perfsuite.h>
#include <psinternal.h>
#include <cpuid.h>

#if HAVE_BFD_H
# include <bfd.h>
#endif

#if HAVE_PAPI
# include <papi.h>
# if PAPI3
#  include <papicompat.h>
# endif
#endif

#if HAVE_PERFMON
# include <perfmon/pfmlib.h>
#endif

static void
version()
{
    printf("\t--------------------------------------------\n");
    printf("\t               %s\n", PACKAGE_STRING);
    printf("\t                  psinv %s\n", PROGRAM_VERSION);
    printf("\n");
    printf("\t          University of Illinois/NCSA\n");
    printf("\t              Open Source License\n");
    printf("\n");
    /*printf("\tEmail: %s\n", PACKAGE_BUGREPORT);*/
    printf("\t        http://perfsuite.ncsa.uiuc.edu/\n");
    printf("\t       http://perfsuite.sourceforge.net/\n");
    printf("\t--------------------------------------------\n");
  
    return;
}

static void
print_cacheinfo(ps_cpu_cachearray_t caches)
{
    unsigned int size;
    int level, maxlevel, skip;
    ps_cpu_cachelist_t *cachelist;
    ps_cpu_cache_t *cache;

    puts("\nCache and TLB Information -");
    printf("Cache levels:\t\t%d\n", caches.levels);
    /* @@ ADD printf("Caches/TLBs:\t\t%d\n", caches.ncaches); */

    puts("\nCache Details -");
    for ( level = 0; level < caches.levels; level++ ) {
        printf("Level %d:\n", level+1);
        cachelist = caches.cachelist[level];
        while ( cachelist != NULL ) {
            skip = 0;
            cache = cachelist->cache;
            switch ( cache->type ) {
            case ICACHE:
            case ITRACECACHE:
            case DCACHE:
            case UCACHE:      break;
            default:          skip = 1; break;
            }
            if ( ! skip ) {
	        fputs("\tType:\t\t", stdout);
                switch ( cache->type ) {
                case ICACHE:      puts("Instruction"); break;
                case ITRACECACHE: puts("Instruction Trace"); break;
                case DCACHE:      puts("Data"); break;
                case UCACHE:      puts("Unified"); break;
                default:          skip = 1; break;
                }
                if ( cache->type == ITRACECACHE ) {
                    printf("\tSize:\t\t%dK uOps\n", cache->specs.itcache.uops);
                }
                else {
                    size = cache->specs.cache.size;
                    fputs("\tSize:\t\t", stdout);
                    if ( size < 1024 ) {
                        printf("%d KB\n", size);
                    }
                    else {
                        printf("%.2f MB\n", size/1024.0);
                    }
                    printf("\tLine size:\t%d bytes\n",
                           cache->specs.cache.linesize);
                }
                fputs("\tAssociativity:\t", stdout);
                switch ( cache->nway ) {
                case FULLY_ASSOCIATIVE:
                    puts("Fully associative");
                    break;
                case DIRECT_MAPPED:
                    puts("Direct mapped");
                    break;
                default:
                    printf("%d-way set associative\n", cache->nway);
                    break;
                }
                puts("");
            }
        
            cachelist = cachelist->next;
        }
    }

    /* Note: PSConfig triggers off this text. */
    puts("TLB Details -");
    maxlevel = 0;
    for (level = 0; level < caches.levels; level++) {
        cachelist = caches.cachelist[level];
        while ( cachelist != NULL ) {
            switch ( cachelist->cache->type ) {
            case ITLB:
            case DTLB:
            case UTLB:
	        if ( level > maxlevel ) {
	            maxlevel = level;
	        }
            default:
	        break;
            }
            cachelist = cachelist->next;
        }
    }
    for ( level = 0; level <= maxlevel; level++ ) {
        printf("Level %d:\n", level+1);
        cachelist = caches.cachelist[level];
        while ( cachelist != NULL ) {
            cache = cachelist->cache;
            switch ( cache->type ) {
            case ITLB:
            case DTLB:
            case UTLB:
                fputs("\tType:\t\t", stdout);
                switch ( cache->type ) {
                case ITLB: puts("Instruction"); break;
                case DTLB: puts("Data"); break;
                case UTLB: puts("Unified"); break;
                default: break;
                }
	        printf("\tEntries:\t%d\n", cache->specs.tlb.entries);
	        fputs("\tPagesize (KB):\t", stdout);
	        for (size = 0; size < cache->specs.tlb.npagesize; size++) {
	            printf("%d ", cache->specs.tlb.pagesize[size]);
	        }
	        puts("");
	        fputs("\tAssociativity:\t", stdout);
	        if ( cache->nway == FULLY_ASSOCIATIVE ) {
	            puts("Fully associative");
	        }
	        else {
	            printf("%d-way set associative\n",
                           cache->nway);
	        }
	        puts("");
                break;
            default:
                break;
            }
            cachelist = cachelist->next;
        }
    }

    return;
}

static void
print_sysinfo()
{
#if HAVE_UNAME
    int ret;
    struct utsname utsbuf;
#endif

    /* @@ Put in libperfsuite */
#undef MEGABYTE
#define MEGABYTE 1048576.0
#undef KILOBYTE
#define KILOBYTE 1024.0
    puts("System Information -");
#if HAVE_UNAME
    ret = uname(&utsbuf);
    if ( ret == 0 ) {
        printf("Node Name:\t\t%s\n", utsbuf.nodename);
        printf("OS Name:\t\t%s\n",  utsbuf.sysname);
        printf("OS Release:\t\t%s\n",  utsbuf.release);
        printf("OS Build/Version:\t%s\n",  utsbuf.version);
        printf("OS Machine:\t\t%s\n",  utsbuf.machine);
    }
#endif
    printf("Processors:\t\t%ld\n",sysconf(_SC_NPROCESSORS_CONF));
/*
    printf("CPUs Online:\t\t%d\n",sysconf(_SC_NPROCESSORS_ONLN));
*/
    printf("Total Memory (MB):\t%.2f\n",
           ((double)sysconf(_SC_PHYS_PAGES)*(double)getpagesize())/MEGABYTE);
    printf("System Page Size (KB):\t%.2f\n", getpagesize()/KILOBYTE);
/*
    printf("Available:\t\t%.2f MB\n",
           (sysconf(_SC_AVPHYS_PAGES)*getpagesize())/MEGABYTE); 
*/

    puts("");

    return;
}

int
ps_cpuid_dump(ps_cpuid_t *cpuinfo)
{
    double cpuspeed;
    const unsigned int POWER_REV_STRLEN = 16;
    char IBM_power_rev[POWER_REV_STRLEN];

    puts("Processor Information -");
    fputs("Vendor:\t\t\t", stdout);
    switch ( cpuinfo->vendor ) {
        case INTEL: puts("Intel"); break;
        case AMD:   puts("AMD"); break;
        case IBM:   puts("IBM"); break;
        default:    puts("(unknown)"); break;
    }

    /* Old code to print "Process family: ..." removed. */

    if ( cpuinfo->brand_string != NULL ) {
        printf("Brand:\t\t\t%s\n",cpuinfo->brand_string);
    }

    /* Old code to print "Model: ..." removed. */

    printf("CPUID info:\t\tfamily: %u, model: %u, stepping: %u\n",
           cpuinfo->cpuid_family, cpuinfo->cpuid_model,
           cpuinfo->cpuid_stepping);

    if (cpuinfo->vendor == IBM) {
        snprintf (IBM_power_rev, POWER_REV_STRLEN, "%d.%d",
                  (cpuinfo->revision & 0xFFFF) >> 8,
                  (cpuinfo->revision & 0xFF));
        printf("Revision:\t\t%s (%04X)\n", IBM_power_rev, cpuinfo->revision);
    } else {
        printf("Revision:\t\t%d\n", cpuinfo->revision);
    }

    if ( ps_cpuspeed(&cpuspeed) == PS_SUCCESS ) {
        printf("Clock Speed:\t\t%.2f MHz\n", cpuspeed);
    }

    print_cacheinfo(cpuinfo->cachearray);

    return 0;
}

static inline void
xml_eventheader(FILE *fd, char *class)
{
    char timebuf[64], hostbuf[64];
    time_t tp;

    time(&tp);
    strcpy(timebuf, ctime(&tp));
    timebuf[strlen(timebuf)-1] = '\0';
    gethostname(hostbuf, 64);
    fputs("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n", fd);
    fprintf(fd, "<ps_hwpc_eventlist class=\"%s\">\n", class);
    fprintf(fd, "<!-- Generated by PerfSuite psinv at %s on %s -->\n",
            timebuf, hostbuf);
    fputs("\n", fd);
}

static inline void
xml_event(FILE *fd, char *type, char *name, char *desc)
{
    if ( desc != NULL ) {
        fprintf(fd, "\t<!-- %s -->\n", desc);
    }
    fprintf(fd, "\t<ps_hwpc_event type=\"%s\" name=\"%s\" />\n", type, name);
}

static inline void
xml_eventfooter(FILE *fd)
{
    fputs("</ps_hwpc_eventlist>\n", fd);
}

#if HAVE_PAPI
int
PAPI_preset_info_compar(const void *av, const void *bv)
{
    PAPI_preset_info_t *a = (PAPI_preset_info_t *) av;
    PAPI_preset_info_t *b = (PAPI_preset_info_t *) bv;

    return strcmp(a->event_name, b->event_name);
}
#endif

int
print_papi_events(int xmlonly)
#if HAVE_PAPI
{
    const PAPI_preset_info_t *info;
    int papi_ret, event_count, avail_count=0, derived=0, non_derived=0, i;
#if ! PAPI3
    int flopscode, ipscode;
#endif

    papi_ret = PAPI_library_init(PAPI_VER_CURRENT);

    if (papi_ret != PAPI_VER_CURRENT) {
        fputs("Error: PAPI library version mismatch.\n", stderr);
        return -1;
    }

    if ( ! xmlonly ) {
        puts("PAPI Version Information -");
#if PAPI3
# if defined(PAPI_LIB_VERSION)
        papi_ret = PAPI_get_opt(PAPI_LIB_VERSION, NULL);
# endif
        printf("Major version:\t\t%d\n", PAPI_VERSION_MAJOR(papi_ret));
        printf("Minor version:\t\t%d\n", PAPI_VERSION_MINOR(papi_ret));
        printf("Revision:\t\t%d\n",      PAPI_VERSION_REVISION(papi_ret));
        printf("Patch level:\t\t%d\n",   PAPI_VERSION_INCREMENT(papi_ret));
#else
        printf("Major version:\t\t%d\n", papi_ret);
#endif
        puts("");
    }

    info = PAPI_query_all_events_verbose();
    if ( info == NULL) {
        fputs("Error: could not retrieve PAPI event information.\n", stderr);
        return -1;
    }

    /*
     * @@ This is awful gamey, but it seems to be the criteria
     * Plus, we're not really "using" PAPI, otherwise we might
     * be better off copying over things to a new array and sorting
     * that.
     */
    for (event_count=0; info[event_count].event_name; event_count++ );

    qsort((PAPI_preset_info_t *)info, event_count,
          sizeof(PAPI_preset_info_t), PAPI_preset_info_compar);

    for (i=0; i<event_count; i++) {
       if (info[i].avail) {
           avail_count++;
           if (info[i].flags & PAPI_DERIVED) {
               derived++;
           }
           else {
               non_derived++;
           }
        }
    }

    if ( xmlonly ) {
        xml_eventheader(stdout, "PAPI");
        puts("\t<!-- ========================================= -->");
        puts("\t<!-- Each event is preceded by its description -->");
        puts("\t<!-- ========================================= -->");
        puts("");
#if ! PAPI3
        papi_ret = PAPI_event_name_to_code("PAPI_FLOPS", &flopscode);
        if ( papi_ret != PAPI_OK ) {
            fprintf(stderr, "Error retrieving code for PAPI_FLOPS: %s\n",
                    PAPI_strerror(papi_ret));
            exit(1);
        }
        papi_ret = PAPI_event_name_to_code("PAPI_IPS", &ipscode);
        if ( papi_ret != PAPI_OK ) {
            fprintf(stderr, "Error retrieving code for PAPI_IPS: %s\n",
                    PAPI_strerror(papi_ret));
            exit(1);
        }
#endif
        for ( i = 0; i < event_count; i++ ) {
            if ( info[i].avail ) {
#if ! PAPI3
                /* Forget these guys */
                if ( (info[i].event_code == flopscode) ||
                     (info[i].event_code == ipscode) ) {
                    continue;
                }
#endif
                xml_event(stdout, "preset", info[i].event_name, 
                          info[i].event_descr);
            }
        }
        xml_eventfooter(stdout);

        return 0;
    }

    puts("PAPI Standard Event Information -");
    printf("Standard events:\t%d\n", avail_count);
    printf("Non-derived events:\t%d\n", non_derived);
    printf("Derived events:\t\t%d\n\n", derived);

    puts("PAPI Standard Event Details -");
    puts("Non-derived:");
    for (i=0; i<event_count; i++) {
        if ( info[i].avail && (! (info[i].flags & PAPI_DERIVED) ) ) {
            printf("\t%s:\t%s\n", info[i].event_name, info[i].event_descr);
        }
    }

    puts("Derived:");
    for (i=0; i<event_count; i++) {
        if ( info[i].avail && (info[i].flags & PAPI_DERIVED )) {
            printf("\t%s:\t%s\n", info[i].event_name, info[i].event_descr);
        }
    }

    puts("");

    return 0;
}
#else
{
    fputs("Sorry, not configured with PAPI support ", stderr);
    fputs(" (-p option not functional).\n", stderr);
    return 0;
}
#endif /* HAVE_PAPI */


int
print_native_events(int xmlonly)
{
#if HAVE_PERFMON
    char *name;
    unsigned int version;
    int ret, i, count;

    ret = pfm_initialize();
    if ( ret != PFMLIB_SUCCESS ) {
        fprintf(stderr, "Error initializing libpfm: %s\n", pfm_strerror(ret));
        return -1;
    }

    if ( xmlonly ) {
        xml_eventheader(stdout, "perfmon");
    }
    else {
        puts("Perfmon Version Information -");
        ret = pfm_get_version(&version);
        if ( ret == PFMLIB_SUCCESS ) {
            printf("Major version:\t\t%u\n", PFMLIB_MAJ_VERSION(version));
            printf("Minor version:\t\t%u\n", PFMLIB_MIN_VERSION(version));
            puts("");
        }
        else {
            fprintf(stderr, "Error retrieving Perfmon version: %s\n",
                    pfm_strerror(ret));
            return -1;
        }
    }

    i = pfm_get_first_event();
    count = 0;

    for( ; i != -1; i = pfm_get_next_event(i) ) {
        pfm_get_event_name(i, &name);
        count++;
    }

    if ( ! xmlonly ) {
        printf("Perfmon Event Information -\n");
        printf("Number of events:\t%d\n\n", count);

        printf("Perfmon Event Details -\n");
    }

    for( i = pfm_get_first_event(); i != -1; i = pfm_get_next_event(i) ) {
        pfm_get_event_name(i, &name);
        if ( xmlonly ) {
            xml_event(stdout, "preset", name, NULL);
        }
        else {
            printf("\t%s\n", name);
        }
    }

    if ( xmlonly ) {
        xml_eventfooter(stdout);
    }

    return 0;
#else
    return -1;
#endif /* HAVE_PERFMON */
}

void
print_external_software(void)
{
    int ret;
#if defined(TCLSH)
    char cmd[256];
#endif

    puts("External Software Version Information -");
    fputs("Tcl shell:\t\t", stdout); fflush(stdout);
#if defined(TCLSH)
    sprintf(cmd, "echo 'puts $tcl_patchLevel' | %s", TCLSH);
    ret = system(cmd);
#else
    puts("n/a");
#endif
    fputs("tDOM versions:\t\t", stdout); fflush(stdout);
#if defined(TCLSH) && defined(TDOMLIBDIR)
    sprintf(cmd, "echo \"lappend auto_path %s; package require tdom; puts [package versions tdom]\" | %s", TDOMLIBDIR, TCLSH);
#if defined(TDOMLIBDIR)
    ret = system(cmd);
#endif
#else
    puts("n/a");
#endif

    return;
}

void
print_build_info(void)
{
    int cplus_demangling_available;

    puts("Details of the PerfSuite Build -");
    puts("--------------------------------");
    printf("PerfSuite:\t\t%s\n", PACKAGE_VERSION);
#if defined(__TIMESTAMP__)
    printf("psinv.c Last Modified:\t%s\n", __TIMESTAMP__);
#endif
#if defined(__DATE__) && defined(__TIME__)
    printf("Build date:\t\t%s %s\n", __DATE__, __TIME__);
#endif
    printf("Built on:\t\t%s\n", PS_BUILD_HOST);
    fputs("Built for:\t\t", stdout);
#if defined(__LP64__)
    puts("64-bit systems");
#else
    puts("32-bit systems");
#endif

    fputs("Architecture:\t\t", stdout);
#if IA32
    fputs("IA32 (including x86 and x86-64)", stdout);
#elif IA64
    fputs("IA64 (Itanium and Itanium 2)", stdout);
#elif PPC32
    fputs("PPC32", stdout);
#elif PPC64
    fputs("PPC64", stdout);
#else
    fputs("(not detected)", stdout);
#endif
    puts("");

    fputs("Real time clock:\t", stdout);
#if RTC_GETTIMEOFDAY
    puts("gettimeofday");
#elif RTC_CLOCK_GETTIME
    puts("clock_gettime");
#elif RTC_MMTIMER
# if defined(MMTIMER_NAME)
    printf("%s", MMTIMER_NAME);
# else
    fputs("mmtimer", stdout);
# endif
# if defined(MMTIMER_DESC)
    printf(" %s", MMTIMER_DESC);
# endif
# if defined(MMTIMER_VERSION)
    printf(" (version %s)\n", MMTIMER_VERSION);
# endif
#elif defined(IA32) || defined(IA64)
    puts("CPU Time Stamp Counter (\"rdtscll\" instruction)");
#elif defined(PPC32) || defined(PPC64)
    puts("PowerPC CPU Timebase Register (\"mftb\" family instructions)");
#endif

    fputs("GCC Version:\t\t", stdout);
#if defined(__VERSION__)
    printf("%s\n", __VERSION__);
#else
    puts("n/a");
#endif
    fputs("GNU C Library:\t\t", stdout);
#if HAVE_GNU_LIBC_VERSION_H
    printf("%s\n", gnu_get_libc_version());
#else
    puts("n/a");
#endif

    cplus_demangling_available = 0;
    fputs("Source Code Mapping:\t", stdout);
#if BFD_ENABLED
#if HAVE_BFD_H && defined(BFD_LIB_SONAME)
    printf("using %s\n", BFD_LIB_SONAME);
#if CPLUS_DEMANGLE_PIC
    cplus_demangling_available = 1;
#endif
#else
    puts("not available (libbfd configured, but bfd.h or libbfd not found)");
#endif
#else
    puts("using \"addr2line\"");
    cplus_demangling_available = 1;
#endif

    fputs("C++ name demangling:\t", stdout);
    if (cplus_demangling_available) {
        puts("available");
    } else {
        puts("unavailable");
    }

    fputs("MD5 checksums:\t\t", stdout);
#if HAVE_MD5_RFC
    puts("available");
#else
    puts("not available");
#endif

    printf("Configuration command line:\n\t\t\t%s\n", PS_CONFIG_CMD_LINE);
    printf("The working directory where the above ran:\n\t\t\t%s\n", PS_BUILD_PWD);
    return;
}

#define PSINV_OPTION_FLAGS "behpvVx"
static int do_cpuinfo = 1, do_help = 0, do_events = 0,
           do_external_software = 0, do_papi = 0, do_sysinfo = 1,
           do_version = 0, do_xmlevents = 0, do_build_info = 0;

static void
usage()
{
    printf("%s (version %s) system information utility\n\n",
           PACKAGE_NAME, PACKAGE_VERSION);

    puts("Usage: psinv [option]"); 
    puts("");
    puts("Options:");
    puts("\t-b, --build\tPrint the details regarding this build");
    puts("\t-e, --events\tList available native hardware performance events");
    puts("\t\t\t(\"--events=xml\" lists native events in XML format)");
    puts("\t-h, --help\tHelp (this message)");
    puts("\t-p, --papi\tList available PAPI standard events");
    puts("\t\t\t(\"--papi=xml\" lists PAPI standard events in XML format)");
#if ! HAVE_PAPI
    puts("\t            \t  (option not available on this system)");
#endif
    puts("\t-v, --verbose\tVerbose mode (show everything)");
    puts("\t-V, --version\tReport psinv version");

    return;
}

int
main(int argc, char **argv)
{
    ps_cpuid_t p;
    int c, option_index;
    extern char *optarg;

    static struct option options[] = {
        {"build",     no_argument,       &do_build_info       ,'b'},
        {"events",    optional_argument, NULL,                 'e'},
        {"help",      no_argument,       &do_help,             'h'},
        {"papi",      optional_argument, NULL,                 'p'},
        {"verbose",   no_argument,       NULL,                 'v'},
        {"version",   no_argument,       &do_version,          'V'},
        {"xternal",  no_argument,       &do_external_software,   1},
        {0, 0, 0, 0}
    };

    while ( 1 ) {
        option_index = 0;
        c = getopt_long(argc, argv, PSINV_OPTION_FLAGS, options, &option_index);
        if ( c == -1 ) {
            break;
        }
        
        switch ( c ) {
        case 0:
            /* Long option found if flag is non-NULL */
            break;
        case 'b':
            do_build_info = 1;
            break;
        case 'e':
        case 'p':
            if ( c == 'p' ) {
                do_papi = 1;
            }
            else {
                do_events = 1;
            }
            if ( optarg != 0 ) {
                if ( optarg[0] == '=' ) {
                    optarg++;
                }
                if ( strcmp(optarg, "xml") == 0 ) {
                    do_xmlevents = 1;
                }
            }
            break;
        case 'h':
            usage();
            exit(0);
        case 'v':
#if HAVE_PERFMON
            do_events = 1;
#endif
#if HAVE_PAPI
            do_papi = 1;
#endif
            break;
        case 'V':
            do_version = 1;
            break;
        case 'x':
            do_external_software = 1;
            break;
        default:
            usage();
            exit(1);
        }
    }

    if ( do_build_info ) {
        print_build_info();
        exit(0);
    }

    if ( do_external_software ) {
        print_external_software();
        exit(0);
    }

    if ( do_help ) {
        usage();
        exit(0);
    }

    if ( do_xmlevents ) {
        if ( do_papi ) {
            print_papi_events(do_xmlevents);
        }
        else {
            if ( print_native_events(do_xmlevents) != 0 ) {
                fputs("The \"-e\" (or \"--events\") option ", stdout);
                puts("is not supported on this system");
                exit(0);
            }
        }
        exit(0);
    }

    if ( do_version ) {
        version();
        exit(0);
    }

    if ( do_sysinfo ) {
        print_sysinfo();
    }

    if ( do_cpuinfo && (_ps_cpuid(&p) == PS_SUCCESS) ) {
        ps_cpuid_dump(&p);
    }  
    else if ( do_cpuinfo ) {
        puts("Sorry, no information is known about this CPU.");
    }

    if ( do_papi ) {
        print_papi_events(do_xmlevents);
    }

    if ( do_events ) {
        if ( print_native_events(do_xmlevents) != 0 ) {
            fputs("The \"-e\" (or \"--events\") option ", stdout);
            puts("is not supported on this system");
            exit(0);
        }
    }

    exit(0);
}
