/* $Id: environment.c,v 1.20 2006/10/02 14:10:00 rkufrin Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *        Copyright(C) 2002-2006, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * environment.c: routines related to environment manipulation for psrun
 *
 * Author: Rick Kufrin
 * Date:   12/14/2002
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#if HAVE_GETOPT_H
# include <getopt.h>
#endif

#include <hwpc.h>

#include "psrun.h"
#include "resource.h"
#include "options.h"
#include "environment.h"

static inline void
secserror(char *val)
{
    fprintf(stderr, "psrun warning: %s must be a ", PSRUN_DORESOURCE_ENV);
    fputs("positive non-zero integer\n", stderr);
    fprintf(stderr, "               %s current value: \"%s\"\n",
            PSRUN_DORESOURCE_ENV, val);
    return;
}

void
setup_environment(int argc, char **argv)
{
    int i;
    char *envval, *exebase, *preloadenv, *preloadps, *preloadorig;

    /*
     * Safety check.
     */
    envval = getenv(PSRUN_PID_ENV);
    if ( envval != NULL ) {
        fprintf(stderr, "psrun warning: the environment variable %s is set.\n",
                PSRUN_PID_ENV);
    }

    if ( psrun_options.debug_on ) {
        fprintf(stderr, "Pthread support %senabled\n",
               psrun_options.do_pthreads ? "" : "not ");
    
        fputs("Executing ", stderr);
        for ( i = optind; i < argc; i++ ) {
            fprintf(stderr, "%s ", argv[i]);
        }
        fputs("\n", stderr);
    }

    /*
     * Adjust the output file if specified, default is 
     * base name of executable.  Highest priority is
     * "-o" option, then environment, and finally default.
     *
     * @@ This does an unnecessary reset of the environment
     * variable if already set and not overridden, but it
     * seems easier to follow the logic.
     */
    if ( psrun_options.outputfile == NULL ) {
        psrun_options.outputfile = getenv(FNAMEFROMENV);
        if ( psrun_options.outputfile == NULL ) {
            exebase = strrchr(argv[optind], '/');
            if ( exebase != NULL ) {
                exebase++;
            }
            else {
                exebase = argv[optind];
            }
            psrun_options.outputfile = exebase;
        }
    }
    if ( setenv(FNAMEFROMENV, psrun_options.outputfile, OVERWRITE) != 0 ) {
        fputs("psrun: internal error (filename setenv)\n", stderr);
        exit(1);
    }

    /*
     * Adjust the output format if specified.  If not specified,
     * set to whatever the current environment contains (which
     * may be NULL)
     */
    if ( psrun_options.format != NULL ) {
        if ( setenv(FORMATFROMENV, psrun_options.format, OVERWRITE) != 0 ) {
            fputs("psrun: internal error (format setenv)\n", stderr);
            exit(1);
        }
    }
    else {
        psrun_options.format = getenv(FORMATFROMENV);
    }

    /*
     * Check to see if we should return psrun or command's
     * exit status.
     */
    if ( ! psrun_options.user_exitstatus ) {
        if ( getenv(PSRUN_EXITSTATUS_ENV) != NULL ) {
            psrun_options.user_exitstatus = 1;
        }
    }

    /*
     * Arrange for resource monitoring if not specified on
     * the command line but present in the environment.
     */
    if ( ! psrun_options.do_resource ) {
        envval = getenv(PSRUN_DORESOURCE_ENV);
        if ( envval != NULL ) {
            psrun_options.do_resource = 1;
            if ( strlen(envval) > 0 ) {
               if ( isint(envval) ) {
                   i = atoi(envval);
                   if ( i > 0 ) {
                       psrun_options.sleepsecs = atoi(envval);
                   }
                   else {
                       secserror(envval);
                   }
               }
               else {
                   secserror(envval);
               }
            }
        }
    }
        
    /*
     * Assign function pointer to the proper output routine if
     * doing resource monitoring.
     */
    if ( psrun_options.do_resource ) {
        if ( (psrun_options.format != NULL) &&
             (strcasecmp(psrun_options.format, PS_FORMATSTRING_TEXT) == 0) ) {
            psrun_resource_write = psrun_resource_text_write;
        }
        else {
            psrun_resource_write = psrun_resource_xml_write;
        }
    }

    /*
     * Check to see if the user requested resource monitoring only.
     * If so, no need to set up the environment further.
     */
    if ( psrun_options.do_hwpc == 0 ) {
#if DEBUG
        printf("%s:%d : Requested psrun resource monitoring only.\n",
               __FILE__, __LINE__);
#endif
        return;
    }

    /*
     * Arrange for MD5 checksums if requested
     */
    if ( psrun_options.do_md5 == 1 ) {
        if ( setenv(MD5FROMENV, "", OVERWRITE) != 0 ) {
            fputs("psrun: internal error (md5 checksum setenv)\n", stderr);
            exit(1);
        }
    }

    /*
     * Preload the proper library. First check to see if
     * any setting currently exists for preloading.  If so,
     * add the psrun library to it.
     */
    preloadps = psrun_options.do_pthreads ?
        PSRUN_PRELOAD_PTHREAD_OBJ : PSRUN_PRELOAD_OBJ;
    preloadorig = getenv(PSRUN_PRELOAD_ENV);
    if ( preloadorig != NULL ) {
        if ( setenv(PSRUN_PRELOAD_ORIG_ENV, preloadorig, OVERWRITE) != 0 ) {
            fprintf(stderr, "psrun: error caching original preload env\n");
            exit(1);
        }
        preloadenv = malloc(strlen(preloadorig) + strlen(preloadps) + 2);
        if ( preloadenv == NULL ) {
            fprintf(stderr, "psrun: out of memory for preload environment\n");
            exit(1);
        }
        sprintf(preloadenv, "%s%c%s",
                preloadps, PSRUN_PRELOAD_SEP, preloadorig);
    }
    else {
        preloadenv = preloadps;
    }
    if ( setenv(PSRUN_PRELOAD_ENV, preloadenv, OVERWRITE) != 0 ) {
        fprintf(stderr, "psrun: internal error (preload)\n");
        exit(1);
    }

    /*
     * Arrange for an annotation if specified
     */
    if ( psrun_options.annotation != NULL ) {
        if ( setenv(ANNOTATIONFROMENV, psrun_options.annotation,
                    OVERWRITE) != 0 ) {
            fputs("psrun: internal error (annotation setenv)\n", stderr);
            exit(1);
        }
    }

    /*
     * Adjust the event configuration file if specified
     */
    if ( psrun_options.configfile != NULL ) {
        if ( setenv(CONFIGFROMENV, psrun_options.configfile,
                    OVERWRITE) != 0 ) {
            fputs("psrun: internal error (config filename setenv)\n", stderr);
            exit(1);
        }
        if ( psrun_options.sysconfig ) {
            if ( psrun_options.sysconfigdir == NULL ) {
                psrun_options.sysconfigdir = "";
            }
            if ( setenv(SYSCONFIGFROMENV, psrun_options.sysconfigdir,
                        OVERWRITE) != 0 ) {
                fputs("psrun: internal error (sysconfig setenv)\n", stderr);
                exit(1);
            }
        }
    }

    /*
     * Adjust the counting domain if specified
     */
    if ( psrun_options.domain != NULL ) {
        if ( setenv(DOMAINFROMENV, psrun_options.domain, OVERWRITE) != 0 ) {
            fputs("psrun: internal error (counting domain setenv)\n", stderr);
            exit(1);
        }
    }

    /*
     * Adjust the profiling threshold if specified
     */
    if ( psrun_options.threshold != NULL ) {
        if ( setenv(THRESHOLDFROMENV, psrun_options.threshold,
                    OVERWRITE) != 0 ) {
            fputs("psrun: internal error (threshold setenv)\n", stderr);
            exit(1);
        }
    }

    /*
     * Arrange to monitor child processes if requested
     */
    if ( psrun_options.do_fork ) {
        if ( setenv(PSRUN_DOFORK_ENV, "", OVERWRITE) != 0 ) {
            fputs("psrun: internal error (do_fork putenv)\n", stderr);
            exit(1);
        }
    }

    /*
     * Arrange for signal handling if requested.
     */
    if ( psrun_options.siglist != NULL ) {
        if ( setenv(SIGNALFROMENV, psrun_options.siglist,
                    OVERWRITE) != 0 ) {
            fputs("psrun: internal error (signal setenv)\n", stderr);
            exit(1);
        }
    }

    return;
}
