/* $Id: libpsrun.c,v 1.84 2013/01/29 22:12:34 ruiliu Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *   Copyright(C) 2002-2010, 2012-2013, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * libpsrun.c: support routines for the PerfSuite psrun utility
 *
 * This file is pre-loaded on startup and causes hardware
 * performance counting routines in the PerfSuite libpshwpc
 * library to be executed before the program's main() function
 * is executed and also before the program exits.
 *
 * It's important to keep in mind that these routines are executed
 * as part of the process being measured by psrun.
 *
 * Author: Rick Kufrin
 * Date: 12/14/2002
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#define _GNU_SOURCE  /* For definition of RTLD_NEXT */

#include <stdio.h>

#if _REENTRANT
# if HAVE_PTHREAD
#  include <pthread.h>
# endif
#endif

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_DLFCN_H
# include <dlfcn.h>
#endif

#include <perfsuite.h>
#include <psdebug.h>

#include "psrun.h"

#define LIBPSRUN_NON_EXTERN 1
#include "libpsrun.h"

#include "libpsrun_misc.h"
#include "environment.h"

#if PSRUN_SIGPROF_IGN
# if HAVE_SIGNAL_H
#  include <signal.h>
# endif
#endif

/*
 * Function pointers we'll need.
 */
int (*_psrun_hwpc_init)     PARAMS((void));
int (*_psrun_hwpc_start)    PARAMS((void));
int (*_psrun_hwpc_stop)     PARAMS((char *));
int (*_psrun_hwpc_shutdown) PARAMS((void));
int (*_cancel_children_threads) PARAMS((void));

pid_t (*_psrun_real_fork)   PARAMS((void)) = NULL;

#if _REENTRANT
int (*_psrun_real_pthread_create) (pthread_t * restrict thread,
                                   const pthread_attr_t * restrict attr,
                                   void *(*start_routine) (void *),
                                   void * restrict arg);
#endif

/*
 * These must be the same error processing routines that are
 * used in the PerfSuite core library.  We recompile them into
 * this shared object because the core library can't be created
 * as a shared library on IA-32 (if it could, we could just dlopen
 * it, which would be much better...)
 */
#include <error.c>

static void psrun_init PARAMS((void)) __attribute__((constructor));

#if HAVE_ON_EXIT
static void psrun_fini PARAMS((int status, void *arg));
#elif HAVE_ATEXIT
static void psrun_fini PARAMS((void));
#else
static void psrun_fini PARAMS((void)) __attribute__((destructor));
#endif

#if DEBUG
# if HAVE_ON_EXIT
#  define EXITLABEL "on_exit()"
# elif HAVE_ATEXIT
#  define EXITLABEL "atexit()"
# else
#  define EXITLABEL "gcc destructor"
# endif
#endif

static void *ps_hwpc_dlhandle;

int _psrun_initialized  = 0;

int _psrun_ignoreerrors = 0;

#ifdef _REENTRANT
# ifdef LIBPSRUN_REENTRANT
static void *pthread_dlhandle;
# endif
#endif /* _REENTRANT */

static void
psrun_init(void)
{
    char *preloadorig, *error, *envval;
    int ret;
#if PSRUN_SIGPROF_IGN
    struct sigaction sigact, origsigact, currentsigact;
#endif

#if DEBUG
    pid_t pid;
    extern char **environ;
    int count;
#endif

#if PSRUN_SIGPROF_IGN
    /*
     * Ignore SIGPROF signals till startup is complete.
     * @@ This is experimental, have observed signals
     * terminating things early.  This seems to lessen
     * the problem but does not eliminate it.  This has
     * to occur as early as possible, so we do it before
     * anything else.
     */
    memset(&sigact, 0, sizeof(sigact));
    sigact.sa_handler = SIG_IGN;
    sigact.sa_flags = SA_RESTART;
    ret = sigaction(SIGPROF, &sigact, &origsigact);
    ret = sigaction(SIGPROF, NULL, &sigact);
# if DEBUG
    /* Conform to PS_DEBUG output format */
    printf("%s:%d : SIGPROF ignored on startup. Handler=%p, flags=%x\n",
           __FILE__, __LINE__, sigact.sa_handler, sigact.sa_flags);
# endif
#endif

#if DEBUG
    ps_setdebuglevel();
    pid = getpid();
    _psrun_showenv("entry of psrun_init");
#endif

    PS_DEBUG(PS_DEBUG_INFO, printf("[PID %d] entering psrun_init\n", pid);)
    PS_DEBUG(PS_DEBUG_VERBOSE, \
             puts("Environment at process start:");\
             for (count = 0; \
                  (envval = environ[count]) != NULL; \
                  count++) { \
                 printf("\t%s\n", envval); \
             } \
             printf("Defined environment variables at process start: %d\n", \
                    count); \
    )

    /*
     * Open the base PerfSuite hardware performance shared library
     * and get the addresses of functions that we'll need to call.
     */ 

    ps_hwpc_dlhandle = dlopen(PS_HWPC_SHLIB, DLRESOLVE_MODE);
    if ( ! ps_hwpc_dlhandle  ) {
        fprintf(stderr, "%s\n", dlerror());
        exit(1);
    }

    _psrun_hwpc_init     = dlsym(ps_hwpc_dlhandle, PS_HWPC_INIT);
    CHECK_DLSYM

    _psrun_hwpc_start    = dlsym(ps_hwpc_dlhandle, PS_HWPC_START);
    CHECK_DLSYM

    _psrun_hwpc_stop     = dlsym(ps_hwpc_dlhandle, PS_HWPC_STOP);
    CHECK_DLSYM

    _psrun_hwpc_shutdown = dlsym(ps_hwpc_dlhandle, PS_HWPC_SHUTDOWN);
    CHECK_DLSYM

    _cancel_children_threads = dlsym(ps_hwpc_dlhandle, "cancel_children_threads");
    CHECK_DLSYM

    _psrun_real_fork     = dlsym(RTLD_NEXT, FORK);
    CHECK_DLSYM

#ifdef _REENTRANT
# ifdef LIBPSRUN_REENTRANT
    /*
     * Open the POSIX thread library and get the address of
     * pthread_create(), which we will trap before the thread
     * is actually created so that we can also start measurement
     * of the new thread.
     */
    envval = getenv(PSRUN_PTHREADLIB_ENV);
    if ( envval == NULL ) {
        envval = PTHREAD_SHLIB;
    }
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("[PID %d] using \"%s\" for pthread shared library.\n", \
                    pid, envval);)
    pthread_dlhandle = dlopen(envval, DLRESOLVE_MODE);
    if ( ! pthread_dlhandle ) {
        PS_DEBUG(PS_DEBUG_FATAL, fprintf(stderr, "%s\n", dlerror());)
        exit(1);
    }
    _psrun_real_pthread_create = dlsym(pthread_dlhandle, PTHREAD_CREATE);
    CHECK_DLSYM
# endif
#endif

    /*
     * Register a function to be called at process exit.  This is
     * what will eventually write out the performance data.
     */
#if HAVE_ON_EXIT
    ret = on_exit(psrun_fini, NULL);
#elif HAVE_ATEXIT
    ret = atexit(psrun_fini);
# if HAVE_ON_EXIT || HAVE_ATEXIT
    if ( ret != 0 ) {
        LIBPSRUN_FATAL("registering exit function", ret);
    }
# endif
#endif

    /*
     * Remove the preloaded library from the environment if the
     * user has not requested monitoring of child processes.
     * Be careful to restore any settings that existed when
     * psrun was first invoked.
     */
    if ( (getenv(PSRUN_DOFORK_ENV)  == NULL) && 
         (getenv(PSRUN_PRELOAD_ENV) != NULL) ) {
        preloadorig = getenv(PSRUN_PRELOAD_ORIG_ENV);
        if ( preloadorig != NULL ) {
            PS_DEBUG(PS_DEBUG_INFO, \
                     printf("[PID %d]: ", pid); \
                     printf("restoring %s to \"%s\"\n", \
                            PSRUN_PRELOAD_ENV, preloadorig);)
            ret = setenv(PSRUN_PRELOAD_ENV, preloadorig, OVERWRITE);
            if ( ret != 0 ) {
                LIBPSRUN_FATAL("restoring original preload to environment",ret);
            }
        }
        else {
            PS_DEBUG(PS_DEBUG_INFO, \
                     printf("[PID %d]: ", pid); \
                     printf("unsetting %s\n", PSRUN_PRELOAD_ENV);)
            ret = unsetenv(PSRUN_PRELOAD_ENV);
            if ( ret != 0 ) {
                LIBPSRUN_FATAL("removing preload from environment", ret);
            }
        }
    }
    else {
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("[PID %d] monitoring of child processes has ", pid); \
                        puts("been selected.");)
    }

    /*
     * This is an experimental safety-check - not used much in practice.
     */

    envval = getenv(PSRUN_ONERROR_ENV);
    if ( envval != NULL ) {
        if ( strcmp(envval, "ignore") == 0 ) {
            _psrun_ignoreerrors = 1;
            PS_DEBUG(PS_DEBUG_INFO, \
                     puts("Ignoring runtime measurement errors."); \
            )
        }
    }

    /*
     * Initialize and start performance measurement and then
     * turn control over to the user program.
     */

    ret = (*_psrun_hwpc_init)();
    if ( ret != PS_SUCCESS ) {
        if ( ! _psrun_ignoreerrors ) {
            LIBPSRUN_FATAL(ps_strerror(ret), ret);
        }
        PS_DEBUG(PS_DEBUG_WARNING, \
                 printf("Error during initialization (ignored): %s\n", \
                         ps_strerror(ret)); \
        )
        return;
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("[PID %d] psrun_init: hwpc_init returned %d\n", \
             pid, ret); \
    )

    ret = (*_psrun_hwpc_start)();
    if ( ret != PS_SUCCESS ) {
        if ( ! _psrun_ignoreerrors ) {
            LIBPSRUN_FATAL(ps_strerror(ret), ret);
        }
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("Error starting measurement (ignored): %s\n", \
                 ps_strerror(ret)); \
        )
        return;
    }

    PS_DEBUG(PS_DEBUG_INFO, \
             printf("[PID %d] psrun_init: hwpc_start returned %d\n", \
                    pid, ret); \
            _psrun_showenv("exit of psrun_init"); \
    )

#if PSRUN_SIGPROF_IGN
    /*
     * SIGPROF handling may or may not have changed during
     * execution.  If it didn't, restore it to what it was
     * when we entered this function.
     */
    ret = sigaction(SIGPROF, NULL, &currentsigact);
    PS_DEBUG(PS_DEBUG_INFO, \
             printf("psrun signal handler = %p, flags = %x\n", \
             sigact.sa_handler, sigact.sa_flags); \
             printf("current signal handler = %p, flags = %x\n", \
             currentsigact.sa_handler, currentsigact.sa_flags);)
    if ( (currentsigact.sa_handler == sigact.sa_handler) &&
         (currentsigact.sa_flags   == sigact.sa_flags) ) {
        PS_DEBUG(PS_DEBUG_INFO, \
                 puts("Restoring original SIGPROF handling.");)
        ret = sigaction(SIGPROF, &origsigact, NULL);
    } else {
        PS_DEBUG(PS_DEBUG_INFO, \
                 puts("SIGPROF handling has changed. Not restoring.");)
    }
#endif

    _psrun_initialized = 1;

    return;
}

static void
#if HAVE_ON_EXIT
psrun_fini(int status, void *arg)
#else
psrun_fini(void)
#endif
{
    pid_t pid, masterpid;
    char *env;
    int ret;

#if BLOCKERRORSONEXIT
    _psrun_block_errors();
#endif

    pid = getpid();
    PS_DEBUG(PS_DEBUG_INFO, printf("[PID %d] entering psrun_fini.\n", pid); \
                            _psrun_showenv("entry of psrun_fini"); \
    )
#if HAVE_ON_EXIT
    PS_DEBUG(PS_DEBUG_INFO, printf("[PID %d] exit status=%d.\n", pid, status);)
#endif

    /*
     * If the user has not selected monitoring of child processes,
     * return without writing out performance data unless this
     * process is the original process exec'd by psrun.
     *
     * If this process is instead a descendant of the original 
     * process, performance monitoring will not have been started 
     * after the fork() occurred, so even trying to access the data
     * would be an error.
     *
     * If PSRUN_PID_ENV has not been set in the environment, then 
     * we assume that this library has been preloaded directly and
     * not through psrun, so we'll force this process to be
     * recognized as the "original process".
     */

    env = getenv(PSRUN_PID_ENV);
    if ( env == NULL ) {
        masterpid = pid;
    }
    else {
        sscanf(env, "%d", &masterpid);
    }

    if ( (getenv(PSRUN_DOFORK_ENV) == NULL) && (pid != masterpid) ) {
        PS_DEBUG(PS_DEBUG_INFO, \
                 printf("[PID %d] psrun_fini: exiting without hwpc_stop\n", \
                        pid); \
        )
        return;
    }

    /* Give the children threads a chance to terminate normally,
     * in case exit() is called instead of pthread_exit() or pthread_cancel().
     * Rui added 2012-11-14.
     */
    PS_DEBUG(PS_DEBUG_INFO, puts("about to call cancel_children_threads in psrun_fini()...");)
    _cancel_children_threads();

    /*
     * Output the final performance data.
     */

    ret = (*_psrun_hwpc_stop)(NULL);
    if ( ret != PS_SUCCESS ) {
        LIBPSRUN_FATAL(ps_strerror(ret), ret);
    }

    /*
     * Shutdown the counters.
     */
    ret = (*_psrun_hwpc_shutdown)();
    if ( ret != PS_SUCCESS ) {
        LIBPSRUN_FATAL(ps_strerror(ret), ret);
    }

/*
 * The following call to dlclose() has been removed from compilation due
 * to errors encountered that cause the underlying libraries (possibly libdl?)
 * to print the following error message:
 *
 * Inconsistency detected by ld.so: dl-close.c: 719: _dl_close: Assertion
 * `map->l_init_called' failed!
 *
 * This behavior is unexpected, and did not occur in prior versions of
 * PerfSuite and system software.  Removing the call to dlclose() seems to
 * improve the situation.
 */

#if PSRUN_DLCLOSE
    dlclose(ps_hwpc_dlhandle);

# ifdef _REENTRANT
#  ifdef LIBPSRUN_REENTRANT
    dlclose(pthread_dlhandle);
#  endif /* _REENTRANT */
# endif  /* LIBPSRUN_REENTRANT */
#endif   /* PSRUN_DLCLOSE */

    PS_DEBUG(PS_DEBUG_INFO, printf("[PID %d] exiting psrun_fini using %s\n",
                                   pid, EXITLABEL);)

    return;
    /* It seems that an exit handler should not call exit() in itself,
     * see "man atexit".  So Rui changed this to return. 2012-11-14.
     */

#if HAVE_ON_EXIT
    exit(status);
#else
    exit(ret);
#endif
}
