/* $Id: libpsrun_misc.c,v 1.21 2010/08/26 20:15:27 rkufrin Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *        Copyright(C) 2002-2010, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * libpsrun_misc.c: support routines for the PerfSuite psrun utility
 *
 * Author: Rick Kufrin
 * Date:   2/7/2004
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#if HAVE_DLFCN_H
# define _GNU_SOURCE
# include <dlfcn.h>
#endif

#include <stdio.h>

#if _REENTRANT
# if HAVE_PTHREAD_H
#  include <pthread.h>
# endif
#endif

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if HAVE_SIGNAL_H
# include <signal.h>
#endif

#if HAVE_SYS_TIME_H
# include <sys/time.h>
#endif

#if HAVE_SYS_RESOURCE_H
# include <sys/resource.h>
#endif

#include "psrun.h"
#include "libpsrun.h"
#include "libpsrun_misc.h"
#include "environment.h"

#if HAVE_FORK
pid_t
fork(void) {
    pid_t pid;
    pid_t (*tmp_fork)(void);
    int ret;

    /*
     * This check was introduced after SGI reported that MPT MPI, version
     * 1.21 or higher, could invoke this fork() wrapper before psrun had
     * been initialized.  Therefore, we guard against that and simply
     * return the forked process ID if so.
     */

    if ( ! _psrun_initialized ) {
        tmp_fork = dlsym(RTLD_NEXT, FORK);
        pid = (*tmp_fork)();
        return pid;
    }

    pid = (*_psrun_real_fork)();

    if ( pid == 0 ) {
        /*
         * The child process still has performance measurement
         * running.  Shut it off in this process and do not
         * report any results.
         */
#if DEBUG
        fprintf(stderr, "[psrun: Child process forked, PID=%d]\n", getpid());
#endif
        ret = (*_psrun_hwpc_shutdown)();
        if ( ret != PS_SUCCESS ) {
            LIBPSRUN_FATAL(ps_strerror(ret), ret);
        }

#if DEBUG
        printf("fork(): _psrun_hwpc_shutdown returned %d\n", ret);
#endif

        /*
         * If the user has requested monitoring of child processes,
         * then start it here explicitly.  This is in effect the
         * same thing done through library preloading for the
         * original process spawned.
         */
        if ( getenv(PSRUN_DOFORK_ENV) != NULL  ) {
            if ( getenv(FNAMEFROMENV) != NULL ) {
                unsetenv(FNAMEFROMENV);
            }
            ret = (*_psrun_hwpc_init)();
            if ( ret != PS_SUCCESS ) {
                LIBPSRUN_FATAL(ps_strerror(ret), ret);
            }

#if DEBUG
            printf("fork(): _psrun_hwpc_init returned %d\n", ret);
#endif

            ret = (*_psrun_hwpc_start)();
            if ( ret != PS_SUCCESS ) {
                LIBPSRUN_FATAL(ps_strerror(ret), ret);
            }

#if DEBUG
            printf("fork(): _psrun_hwpc_start returned %d\n", ret);
#endif
        }
    }
#if DEBUG
    else {
        fprintf(stderr, "[psrun: Parent process forked, PID=%d]\n", getpid());
    }
#endif

    return pid;
}
#endif /* HAVE_FORK */

#if DEBUG
#if defined(FMT)
#undef FMT
#endif
#define FMT "[PID %d] %-14s = %s\n"
void
_psrun_showenv(char *msg)
{
    int pid;

    pid = getpid();

    printf("[PID %d] Environment (%s)\n", pid, msg);
    printf(FMT, pid, PSRUN_DOFORK_ENV,  getenv(PSRUN_DOFORK_ENV));
    printf(FMT, pid, PSRUN_PRELOAD_ENV, getenv(PSRUN_PRELOAD_ENV));
    printf(FMT, pid, PSRUN_PID_ENV,     getenv(PSRUN_PID_ENV));
    printf(FMT, pid, FNAMEFROMENV,      getenv(FNAMEFROMENV));

    return;
}
#endif

#if BLOCKERRORSONEXIT

/*
 * Have seen problems with (at least) RedHat 9
 * where a SEGV and coredump occurs in the wrapup 
 * after the user application finishes and the 
 * atexit() routine has executed.  This is an 
 * attempt to sidestep these errors that would 
 * prefer not to do, but haven't yet found another 
 * solution.
 */

static void
segv_handler(int arg)
{
    struct sigaction act, oldact;
    int ret;

#if DEBUG
    pid_t pid;

    pid = getpid();
    fprintf(stderr, "[PID %d] Notice: caught SIGSEGV in wrap-up.\n", pid);
#endif

    act.sa_handler = SIG_IGN;
    ret = sigaction(SIGSEGV, &act, &oldact);

    return;
}

void
_psrun_block_errors(void)
{
    struct sigaction act, oldact;
    struct rlimit rlimit;
    int ret;

    act.sa_handler = segv_handler;
    ret = sigaction(SIGSEGV, &act, &oldact);
    if ( ret != 0 ) {
#if DEBUG
        fputs("Warning: error setting SEGV signal action in psrun epilog.\n", \
              stderr);
#endif
    }
    if ( getrlimit(RLIMIT_CORE, &rlimit) != 0 ) {
#if DEBUG
        fputs("Warning: error getting resource limits in psrun epilog.\n", \
              stderr);
#endif
    }
    rlimit.rlim_cur = rlimit.rlim_max = 0;
    if ( setrlimit(RLIMIT_CORE, &rlimit) != 0 ) {
#if DEBUG
        fputs("Warning: error setting resource limits in psrun epilog.\n", \
              stderr);
#endif
    }

    return;
}
#endif /* BLOCKERRORSONEXIT */
