/* $Id: libpsrun_r.c,v 1.37 2012/11/15 22:54:18 ruiliu Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *     Copyright(C) 2002-2006, 2012, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *           National Center for Supercomputing Applications
 *             University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * libpsrun_r.c: POSIX thread-specific support for the
 *               PerfSuite "psrun" utility.  These routines
 *               are pre-loaded before the user's main()
 *               function is called and before the program exits.
 *
 * This source code is based on ideas from the "trapper" tool
 * by Phil Mucci.
 *
 * Author: Rick Kufrin
 * Date: 12/14/2002
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>

#include <pthread.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#include <perfsuite.h>
#include "psrun.h"
#include "libpsrun.h"
#include "libpsrun_r.h"

BEGIN_C_DECLS
static void psrun_pthread_fini PARAMS((void *));
END_C_DECLS

/*
 * We replace the users' original start routine for
 * thread creation with our own, which first sets up
 * performance counting, and then calls the original
 * start routine with the supplied argument.
 *
 * The important thing is that it must be done after
 * the thread has been created.
 */

static void *
psrun_thread_init(void *arg)
{
    void *(*user_start_routine)(void *) =
          ((user_arg_t *) arg)->user_start_routine;
    void *user_arg = ((user_arg_t *) arg)->user_arg;
    void *user_ret;
    int ret, errorseen;

#if DEBUG
    pthread_t thread_id = pthread_self();
    fprintf(stderr, "Thread %lu created...\n", thread_id);
    if ( _psrun_ignoreerrors ) {
        fputs("Thread ignoring runtime measurement errors.\n", stderr);
    }
#endif

    free(arg);

    errorseen = 0;

    ret = (*_psrun_hwpc_start)();
#if DEBUG
    fprintf(stderr, "Thread %lu returned from hwpc_start with status %d\n", thread_id, ret);
#endif
    if ( ret != PS_SUCCESS ) {
        if ( ! _psrun_ignoreerrors ) {
            LIBPSRUN_FATAL(ps_strerror(ret), ret);
        }
        else {
            fputs("Error in thread performance measurement start ", stderr);
            fprintf(stderr, "(ignored): %s\n", ps_strerror(ret));
        }
        errorseen = 1;
    }

    if ( ! errorseen ) {
        pthread_cleanup_push(psrun_pthread_fini, NULL);

        user_ret = user_start_routine(user_arg);

        pthread_cleanup_pop(1);
    }
    else {
        user_ret = user_start_routine(user_arg);
    }

    return user_ret;
}

/*
 * Override the original pthread creation
 * routine with our own, which arranges for libpshwpc
 * functions to be called by individual threads on
 * creation and at pthread_exit or cancellation time
 * via cleanup handlers.
 */
int
pthread_create(pthread_t * restrict thread,
               const pthread_attr_t * restrict attr,
               void *(*start_routine) (void *),
               void * restrict arg)
{
    user_arg_t *user_arg;
    int ret;

    if ( start_routine == NULL ) {
        return -1;
    }

    user_arg = (user_arg_t *) malloc(sizeof(user_arg_t));
    if ( user_arg == NULL ) {
        return -1;
    }
    user_arg->user_start_routine = start_routine;
    user_arg->user_arg = arg;

    ret = _psrun_real_pthread_create(thread, attr, psrun_thread_init,
                                     user_arg);
   
    return ret;
}

static void
psrun_pthread_fini(void *arg)
{
    int ret;

#if DEBUG
    pthread_t thread_id = pthread_self();
    fprintf(stderr, "Thread %lu executing cleanup handler\n", thread_id);
#endif

    ret = (*_psrun_hwpc_stop)(NULL);
    if ( ret != PS_SUCCESS ) {
        LIBPSRUN_FATAL(ps_strerror(ret), ret);
    }

#if DEBUG
    fprintf(stderr, "Thread %lu returned from hwpc_stop with status %d\n", thread_id, ret);
#endif

    return;
}
