/* $Id: options.c,v 1.21 2006/10/02 14:10:00 rkufrin Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *        Copyright(C) 2002-2006, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * options.c: routines related to psrun option processing
 *
 * Author: Rick Kufrin
 * Date:   12/14/2002
 *
 * This file is part of PerfSuite.
 *
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>

#if HAVE_CTYPE_H
# include <ctype.h>
#endif

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_GETOPT_H
# include <getopt.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#include <hwpc.h>
#if HAVE_PERFMON
# include <hwpc-perfmon.h>
#endif

#include "psrun.h"
#include "resource.h"
#include "environment.h"
#include "options.h"
#include "info.h"

/*
 * Option defaults
 */

psrun_options_t psrun_options = {
    .do_pthreads     = 0, 
    .do_fork         = 0,
    .do_hwpc         = 1,
    .do_resource     = 0,
    .do_md5          = 0,
    .debug_on        = 0,
    .sysconfig       = 0,
    .user_exitstatus = 0,
    .sleepsecs       = PSRUN_DEFAULT_SLEEPSECS,
    .outputfile      = NULL,
    .configfile      = NULL,
    .sysconfigdir    = NULL,
    .domain          = NULL,
    .annotation      = NULL,
    .threshold       = NULL,
    .siglist         = NULL,
    .format          = NULL
};

/*
 * "+" is necessary to preserve options meant for the
 * user's command (not this program) to parse
 */
#define PSRUN_OPTION_FLAGS "+a:c:C::d:fF:hio:pr::R::S:t:VX"

void
parse_options(int argc, char **argv)
{
    int c, option_index;

    /*
     * Long options accepted by this program
     */
    static struct option options[] = {
        {"annotate",     required_argument, NULL, 'a'},
        {"config",       required_argument, NULL, 'c'},
        {"domain",       required_argument, NULL, 'd'},
        {"exituser",     no_argument,       &(psrun_options.user_exitstatus),1},
        {"fork",         no_argument,       &(psrun_options.do_fork), 1},
        {"format",       required_argument, NULL, 'F'},
        {"info",         no_argument,       NULL, 'i'},
        {"md5",          no_argument,       &(psrun_options.do_md5), 1},
        {"output",       required_argument, NULL, 'o'},
        {"pthreads",     no_argument,       &(psrun_options.do_pthreads), 1},
        {"resource",     optional_argument, NULL, 'r'},
        {"resourceonly", optional_argument, NULL, 'R'},
        {"signal",       required_argument, NULL, 'S'},
        {"sysconfig",    optional_argument, NULL, 'C'},
        {"threshold",    required_argument, NULL, 't'},
        {"version",      no_argument,       NULL, 'V'},
        {"help",         no_argument,       NULL, 'h'},
        {"Debug",        no_argument,       &(psrun_options.debug_on), 1},
        {0, 0, 0, 0}
    };

    /* Parse the command line */
    while ( 1 ) {
        /*
         * Check for long option
         */
        option_index = 0;
        c = getopt_long(argc, argv, PSRUN_OPTION_FLAGS, options, &option_index);
        if ( c == -1 ) {
            break;
        }

        /*
         * Long option supplied if 0 was returned, otherwise
         * it's a single-character option
         */
        switch ( c ) {
        case 0:
            break;
        case 'a':
            psrun_options.annotation = optarg;
            break;
        case 'c':
            psrun_options.configfile = optarg;
            break;
        case 'C':
            psrun_options.sysconfig = 1;
            if ( optarg != 0 ) {
                psrun_options.sysconfigdir = optarg;
            }
            break;
        case 'd':
            psrun_options.domain = optarg;
            break;
        case 'f':
            psrun_options.do_fork = 1;
            break;
        case 'F':
            psrun_options.format = optarg;
            break;
        case 'h':
            usage();
            exit(0);
        case 'i':
            info();
            exit(0);
        case 'o':
            psrun_options.outputfile = optarg;
            break;
        case 'p':
            psrun_options.do_pthreads = 1;
            break;
        case 'r':
        case 'R':
            psrun_options.do_resource = 1;
            if ( c == 'R' ) {
                psrun_options.do_hwpc = 0;
            }
            if ( optarg != 0 ) {
                if ( ! isint(optarg) ) {
                   fprintf(stderr, "Illegal value for sampling interval: %s\n",
                           optarg);
                   exit(1);
                }
                psrun_options.sleepsecs = atoi(optarg);
            }
            if ( psrun_options.sleepsecs < 1 ) {
                fputs("\nError: resource sampling interval must be ", stderr);
                fputs("1 or greater.\n", stderr);
                exit(1);
            }
            break;
        case 'S':
            psrun_options.siglist = optarg;
            break;
        case 't':
            if ( ! isint(optarg) ) {
               fprintf(stderr, "Illegal value for threshold: %s\n", optarg);
               exit(1);
            }
            psrun_options.threshold = optarg;
            break;
        case 'V':
            version();
            exit(0);
        case 'X':
            psrun_options.user_exitstatus = 1;
            break;
        default:
            usage();
            exit(1);
        }
    }
  
    /*
     * Was a command supplied?
     */
    if ( optind == argc ) {
        usage();
        puts("\nError: no command was supplied.");
        exit(1);
    }

    return;
}

/*
 * Helper function that checks a string to see if it's
 * an integer (contains only decimal digits).  Used to
 * check user-supplied arguments.
 */

int
isint(char *string)
{
    int c;

    for ( c = 0; string[c] != '\0'; c++ ) {
        if ( ! isdigit((int)string[c]) ) {
            return 0;
        }
    }

    return 1;
}

void
usage(void)
{
    char *fmt = "\t%-21s\t%s\n";

    banner();

    puts("\nUsage: psrun [option] command [args]");

    puts("\nOptions:");
    printf(fmt, "-a, --annotate TEXT", "Include annotation TEXT in output");
    printf(fmt, "-c, --config FILE", "Specify configuration file");
    printf(fmt, "-C[DIR], --sysconfig[=DIR]", "");
    puts("\t                     \tCheck DIR for config file (default standard)");
    printf(fmt, "-d, --domain DOMAIN", "Select counting domain (default user)");
    printf(fmt, "-f, --fork", "Monitor child processes (default no)");
    printf(fmt, "-F, --format FORMAT", "Select output format (text|xml)");
    printf(fmt, "-i, --info", "Show capabilities of this install of psrun");
    printf(fmt, "--md5", "Include MD5 checksums in output");
    printf(fmt, "-o, --output PREFIX",
                "Specify output file prefix (default <command>,");
    puts("\t                     \t\"stdout\" and \"stderr\" are recognized)");
    printf(fmt, "-p, --pthreads", "Enable POSIX thread support");
    printf(fmt, "-r[N], --resource[=N]",
                "Collect resource usage information, N gives");
    printf("\t                     \tthe sampling interval in seconds (default %d)\n", PSRUN_DEFAULT_SLEEPSECS);
    printf(fmt, "-R[N], --resourceonly[=N]", "");
    puts("\t                     \tDo resource usage monitoring only");
    printf(fmt, "-S, --signal SIGLIST",
                "Specify list of signals to catch (numeric)");
    printf(fmt, "-t, --threshold INT",
                "Specify overflow threshold (for profiling)");
    printf(fmt, "-V, --version", "Report psrun version");
    printf(fmt, "-X, --exituser",
                "Exit with exit status of <command>, not psrun");
    printf(fmt, "-h, --help", "Help (this message)");

    puts("\nEnvironment:");
    printf(fmt,
           CONTROLFROMENV, "Control hardware performance counting (on|off)");
    printf(fmt,
           ANNOTATIONFROMENV, "Specify annotation text for output");
    printf(fmt, CONFIGFROMENV, "Specify configuration file");
    printf(fmt, DOMAINFROMENV, "Select domain (user|kernel|all)");
    printf(fmt, EXCLUDEDBFROMENV, "Specify file name of exclusion database");
    printf(fmt, FNAMEFROMENV, "Specify output file prefix");
    printf(fmt, FORMATFROMENV, "Select output format (text|xml)");
    printf(fmt, MD5FROMENV, "Include MD5 checksums in output (if set)");
    printf(fmt, MPXUSECFROMENV, "Specify multiplexing interval in microseconds");
    printf(fmt, NOHOSTNAMEFROMENV, "Do not use hostname in output file name");
    printf(fmt, NOPIDFROMENV, "Do not use process ID in output file name");
#if HAVE_PERFMON
    printf(fmt,
           SAMPLE_ENTRIES_ENV, "Specify perfmon sample buffer size (bytes)");
#endif
    printf(fmt, SIGNALFROMENV, "Specify list of signal numbers to catch");
    puts("\t                     \tMultiple signals should be comma-separated");
    printf(fmt, SYSCONFIGFROMENV, "Check alternate directory for configs");
    puts("\t                     \tIf not set to a value, use standard directory");
    printf(fmt, THRESHOLDFROMENV, "Specify overflow threshold (for profiling)");
    printf(fmt, TIMELIMITFROMENV, "Specify minimum CPU seconds for output");
    printf(fmt, PSRUN_DOFORK_ENV, "Monitor child processes (if set)");
    printf(fmt, PSRUN_DORESOURCE_ENV, "Monitor resource usage (if set)");
    printf(fmt, 
           PSRUN_EXITSTATUS_ENV, "Exit with exit status of <command> (if set)");
    printf(fmt, 
           PSRUN_PTHREADLIB_ENV, "Specify alternate POSIX threads library");

    showconfig();

    return;
}
