/* $Id: psrun.c,v 1.99 2006/06/04 16:12:10 rkufrin Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *        Copyright(C) 2002-2006, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * psrun: run a command, activating performance measurement through runtime
 *        library preloading. Optionally also monitor resource usage.
 *
 * Author: Rick Kufrin
 * Date:   12/14/2002
 *
 * This file is part of PerfSuite.
 *
 * @@ With resource monitoring, this becomes useful even
 * without libpshwpc... need to arrange for that.
 *
 * @@ If psrun itself dies, the child continues to run.
 * Is this the desired behavior?
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if HAVE_PTHREAD
# include <pthread.h>
#endif

#if HAVE_STRING_H
# include <string.h>
#endif

#define _USE_BSD

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#if HAVE_SYS_RESOURCE_H
# include <sys/resource.h>
#endif

#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif

#include <hwpc.h>
#include <pshwpc.h>

#include "psrun.h"
#include "resource.h"
#include "options.h"
#include "info.h"
#include "environment.h"
#include "signals.h"

extern char **environ;

/*
 * This structure is global and not file-private
 * so that it can be accessed from within the signal 
 * handler.
 */
psrun_resinfo_t res;

static char pidenv[32];

/* Make sure this survives high optimization levels */
static pthread_t monitor_thread;

int
main(int argc, char **argv)
{
    int ret;
    pid_t pid_ret;

    parse_options(argc, argv);
    setup_environment(argc, argv);
    setup_signals();

    memset(&res, 0, sizeof(psrun_resinfo_t));

    /*
     * For resource monitoring, create a monitor thread
     * and pass in the options necessary as well as variables
     * containing information to be written out after execution.
     *
     * The resource thread is blocked waiting on a condition
     * until the command to be measured starts.
     */
    if ( psrun_options.do_resource ) {
        ret = resource_init(&monitor_thread, &res);
        if ( ret != 0 ) {
            fputs("Error initializing resource thread.\n", stderr);
            exit(1);
        }
    }

    /*
     * Fork and exec the program to run.
     * Problems will be reported here.
     */
    res.pid = fork();
    if ( res.pid == 0 ) {
        sprintf(pidenv, "%d", getpid());
        setenv(PSRUN_PID_ENV, pidenv, OVERWRITE);
        execvp(argv[optind], &argv[optind]);
        perror(argv[optind]);
        _exit(1);
    }
    else if ( res.pid < 0 ) {
        perror("fork");
        exit(1);
    }

    fclose(stdin);

    /* Allow the resource thread to proceed */
    if ( psrun_options.do_resource ) {
        ret = resource_start();
        if ( ret != 0 ) {
            psrun_options.do_resource = 0;
            fputs("Warning: error starting resource monitor.\n", stderr);
        }
    }

    /* Hang around till the program's exited.  */
    pid_ret = wait4(res.pid, &(res.status), 0, &(res.rusage));
    if ( pid_ret != res.pid ) {
        perror("wait4");
        exit(1);
    }

    /* Deal with resource reporting if it was requested.  */
    if ( psrun_options.do_resource ) {
        ret = resource_stop(&monitor_thread, &res);
        if ( ret != 0 ) {
            fputs("Warning: error writing resource data.\n", stderr);
        }
    }

    /*
     * Made it... return exit status according to the 
     * user's selection.  Default is the status of psrun, 
     * not the command that was monitored.
     */
    ret = 0;
    if ( psrun_options.user_exitstatus ) {
        if ( WIFEXITED(res.status) != 0 ) {
            ret = WEXITSTATUS(res.status);
        }
        else if ( WIFSIGNALED(res.status) ) {
            ret = WTERMSIG(res.status) + 128;
        }
    }

    exit(ret);
}
