/* $Id: resource.c,v 1.66 2009/02/10 17:22:22 rkufrin Exp $ */

/****************************************************************************
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *        Copyright(C) 2002-2006, The Board of Trustees of the
 *            University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *           National Center for Supercomputing Applications
 *             University of Illinois at Urbana-Champaign
 * 
 *                  http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * resource.c : support for optional resource monitoring output
 *              in XML or text format from psrun
 *
 * Author: Rick Kufrin
 * Date:   2/2003
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>

#if HAVE_ERRNO_H
# include <errno.h>
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if HAVE_PTHREAD
# include <pthread.h>
#endif

#if HAVE_TIME_H
# include <time.h>
#endif

#if HAVE_STRING_H
# if HAVE_STRSIGNAL
#  define __USE_GNU
# endif
# include <string.h>
#endif

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#if HAVE_SYS_TIME_H
# include <sys/time.h>
#endif

#if HAVE_SYS_RESOURCE_H
# include <sys/resource.h>
#endif

#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif

#include <hwpc.h>

#include "psrun.h"
#include "resource.h"
#include "environment.h"
#include "options.h"
#include "signals.h"

BEGIN_C_DECLS
static int get_memoryuse         PARAMS((psrun_resinfo_t *res,
                                         float *vmsizemb, float *rssizemb));
static int get_starttime         PARAMS((char *buf));
static FILE *select_stream       PARAMS((char *suffix, pid_t pid));
static int set_exitstatus        PARAMS((psrun_resinfo_t *res, char *buf));
END_C_DECLS

int (*psrun_resource_write)(psrun_resinfo_t *);

/* Shared between the main and monitor threads */
static int running;
static pthread_cond_t  res_condvar = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t res_mutex   = PTHREAD_MUTEX_INITIALIZER;
static time_t starttime;

/***************************************************************
 * Inlined functions
 ***************************************************************/

static inline void
indent(FILE *fd, int amt)
{
    int i;

    for ( i = 0; i < amt; i++ ) {
        fprintf(fd, "%c", INDENT_CHAR);
    }

    return;
}

static inline void
output_banner(FILE *fd, char *banner)
{
    int i;

    fprintf(fd, "%s\n", banner);
    for (i = 0; i < MAJORSEPLENGTH; i++) {
        fprintf(fd, "%c", MAJORSEPCHAR);
    }
    fputs("\n", fd);

    return;
}

static inline void
output_value(FILE *fd, char *label, char *string)
{
    int len = strlen(label);
    int i;

    fprintf(fd, "%s", label);

    for (i = len; i < LABELSEPPOS; i++) {
        fputs(" ", fd);
    }

    fprintf(fd, "%c ", LABELSEPCHAR);

    fprintf(fd, "%s\n", string);

    return;
}

/***************************************************************
 * Resource control functions called by the main program
 ***************************************************************/

/*
 * resource_init: initialize variables that are input to the
 * monitor thread, lock a mutex that will be used with
 * condition variable, and create the monitor thread
 */

int
resource_init(pthread_t *monitor_thread, psrun_resinfo_t *res)
{
    int ret;

    res->sleepsecs = psrun_options.sleepsecs;
    res->count = 0;
    res->loadhist = NULL;
    ret = pthread_mutex_lock(&res_mutex);
    if ( ret != 0 ) {
#if DEBUG
        fputs("error locking resource mutex.\n", stderr);
#endif
        return ret;
    }
    running = 0;
    ret = pthread_create(monitor_thread, NULL, monitor_resource, res);
    if ( ret != 0 ) {
        psrun_options.do_resource = 0;
#if DEBUG
        fputs("error creating resource monitor thread.\n", stderr);
#endif
    }

    /* This will be 0 (return value of pthread_create) on success */
    return ret;
}

/*
 * resource_start: signal the monitor thread that it's OK to
 * proceed through a condition variable
 */

int
resource_start(void)
{
    int ret;

    time(&starttime);

    running = 1;
    ret = pthread_mutex_unlock(&res_mutex);
    if ( ret != 0 ) {
#if DEBUG
        fputs("error unlocking resource mutex.\n", stderr);
#endif
        return ret;
    }
    ret = pthread_cond_signal(&res_condvar);
    if ( ret != 0 ) {
#if DEBUG
        fputs("error signaling resource mutex to proceed.\n", stderr);
#endif
    }

    /* This will be zero on success (return value of pthread_cond_signal) */
    return ret;
}

/*
 * resource_stop: terminate the monitor thread
 */

int
resource_stop(pthread_t *monitor_thread, psrun_resinfo_t *res)
{
    int ret;

    /* Cancel the monitor thread */
    ret = pthread_cancel(*monitor_thread);
    if ( ret != 0 ) {
        fputs("Error cancelling monitor thread.\n", stderr);
    }

    ret = pthread_join(*monitor_thread, NULL);
    if ( ret == 0 ) {
        /* Only write the data if it was actually sampled. */
        if ( (res->loadhist != NULL) && (res->count > 0) ) {
            psrun_resource_write(res);
        }
    }
#if DEBUG
    else {
        fputs("error joining with resource monitor thread\n", stderr);
    }
#endif

    /*
     * This will be zero on success (return value of pthread_join).
     * Note that this doesn't imply that data was actually written.
     */
    return ret;
}

/***************************************************************
 * Resource monitoring function called by the monitor thread
 ***************************************************************/

/*
 * monitor_resource: thread that continually monitors resource
 * usage and load average of the system.  This is all the monitor
 * thread does. It exits through cancellation by the main thread.
 *
 * The monitor thread does no output.
 */

void *
monitor_resource(void *data)
{
    psrun_resinfo_t *res = (psrun_resinfo_t *) data;
    int load, ret, oldtype;
    double loadavg[3];
    struct timespec req, rem;

    /* Have the main thread do the signal handling */
    block_signals();

    pthread_cleanup_push((void *) pthread_mutex_unlock, &res_mutex);

    /*
     * Wait until the main thread says let's rock 'n roll...
     */
    ret = pthread_mutex_lock(&res_mutex);
    if ( ret != 0 ) {
        fputs("psrun monitor thread: error locking mutex.\n", stderr);
    }
    while ( ! running ) {
        ret = pthread_cond_wait(&res_condvar, &res_mutex);
        if ( ret != 0 ) {
            fputs("Error waiting on condition to proceed in monitor thread\n",
                  stderr);
        }
    }
    pthread_cleanup_pop(1);

    /*
     * Introduce a little delay just in case.  This helps in the event
     * that the command failed to exec() - but is still not foolproof.
     * The reason why we do this is to guard against output that
     * is meaningless for our purposes, for example, the user is trying
     * to monitor an executable that doesn't exist.  For some slice of
     * time, the process/PID will exist as a forked copy of psrun, 
     * which we don't care about.  So let's give it a chance to exit
     * before we start trying to poke at it.
     */
    req.tv_sec  = 0;
    req.tv_nsec = 50000000;      /* Wait for 1/2 second               */
    ret = nanosleep(&req, &rem); /* We don't really care about return */

    /* Allocate/initialize data structures */
    req.tv_sec = res->sleepsecs;
    req.tv_nsec = 0;
    res->maxrss = res->maxvsize = 0;
    memset(res->loadavg, 0, sizeof(double) * 3);
    memset(res->loadmax, 0, sizeof(double) * 3);
    res->loadbins = PSRUN_DEFAULTLOADBINS;
    res->loadhist = (int *) malloc(res->loadbins * sizeof(int));
    if ( res->loadhist == NULL ) {
        fputs("Error allocating memory for load information\n", stderr);
    }
    else {
        memset(res->loadhist, 0, res->loadbins * sizeof(int));
    }

    /*
     * Respond to cancellation requests immediately, otherwise
     * the main thread can be left waiting until the next
     * cancellation point is reached (on Linux, that means
     * we'd have to wait until returned from the sleep(), which
     * is where this thread spends most of its time).
     *
     * We don't worry about resetting cancellation back to deferred
     * type because if in cancellation mode, we hold no more mutexes
     * after this point.
     */
    ret = pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype); 
    if ( ret != 0 ) {
        fputs("Error setting cancellation type to asynchronous.\n", stderr);
    }

    while ( 1 ) {
        /* Get the current resource usage information */
        pthread_testcancel();

        ret = ps_procstat(res->pid, &res->procstat);

        pthread_testcancel();

        if ( ret == PS_SUCCESS ) {
            /* Has memory usage exceeded high-water marks? */
            if ( res->procstat.rss > res->maxrss ) {
                res->maxrss = res->procstat.rss;
            }
            if ( res->procstat.vsize > res->maxvsize ) {
                res->maxvsize = res->procstat.vsize;
            }

            /*
             * Get the current system load.  This is also available
             * in sysinfo(2), but not portable 
             */
            ret = getloadavg(loadavg, 3);
            if ( ret == 3 ) {
                /*
                 * Note the maximum 1-min load average, and store all
                 * three associated with it
                 */
                if ( res->loadavg[LOADAVG1MIN] < loadavg[LOADAVG1MIN] ) {
                    memcpy(res->loadavg, loadavg, sizeof(double) * 3);
                    res->loadmax[LOADAVG1MIN] = loadavg[LOADAVG1MIN];
#if DEBUG
                    printf("New 1-min load avg: %.3f %.3f %.3f\n",
                           loadavg[0], loadavg[1], loadavg[2]);
#endif
                }

                /*
                 * Store individual maximum load averages.
                 * Already did 1-min max above.  Do the other two 
                 */
                for ( load = 1; load < 3; load++) {
                    if ( loadavg[load] > res->loadmax[load] ) {
                        res->loadmax[load] = loadavg[load];
                    }
                }

                /*
                 * Increment the histogram bin corresponding to the
                 * current load average
                 */
                if ( res->loadhist != NULL ) {
                    load = (int) (loadavg[0] * 10.0 + 0.5);
                    if ( load >= res->loadbins ) {
                        load = res->loadbins - 1;
                    }
                    res->loadhist[load]++;
                    /*
                      printf("Load = %.2f, incrementing bin %d (%d)\n",
                             loadavg[0], load, res->loadhist[load]);
                    */
                }
            }
            res->count++;
        }
        else {
#if DEBUG
            fprintf(stderr, "resource thread ps_procstat returned %s\n",
                    ps_strerror(ret));
            if ( ret == PS_ESYS ) {
                perror("System error details");
            }
#endif
            pthread_cleanup_push((void *) pthread_mutex_unlock, &res_mutex);
            ret = pthread_mutex_lock(&res_mutex);
            if ( ret != 0 ) {
                fputs("Error acquiring mutex in monitor thread.\n", stderr);
            }
            if ( ! running ) {
# if DEBUG
                puts("resource thread exiting after ps_procstat");
# endif
                pthread_exit(NULL);
            }
            pthread_cleanup_pop(1);
        }

        pthread_testcancel();

        ret = nanosleep(&req, &rem);
        /*
         * Don't want to exit always on non-zero return here, since
         * things like control-Z will then cause this thread to
         * exit when the process is resumed.  Instead, we'll
         * check for interrupted system call return before we
         * decide that it's time to call it a day.
         */
        if ( ret != 0 ) {
            if ( errno != EINTR ) {
                perror("psrun: resource monitor exiting");
                pthread_exit(NULL);
            }
# if DEBUG
            else {
                puts("resource thread continuing after interrupted sleep");
            }
# endif
        }
    }

    return data;
}

/***************************************************************
         Output routines, XML and plain text

   These are used by the main program, not the monitor thread.
 ***************************************************************/

int
psrun_resource_xml_write(psrun_resinfo_t *res)
{
    char *envval, buf[256];
    FILE *fd = NULL;
    int i, exited, ret;
    float vmsizemb, rssizemb;
    long clk_tck;

    fd = select_stream("xml", res->pid);
    if ( fd == NULL ) {
        return PS_ESYS;
    }

    fputs("<?xml version=\"1.0\" encoding=\"UTF-8\" ", fd);
#if defined(PSRUN_DTDDIR)
    fputs("standalone=\"no\"", fd);
#endif
    fputs("?>\n", fd);
#if defined(PSRUN_DTDDIR)
    fprintf(fd,
            "<!DOCTYPE resourcereport\n\tSYSTEM \"%s/resourcereport.dtd\"\n>\n",
            PSRUN_DTDDIR);
#endif
    fprintf(fd, "<%s version=\"%s\" generator=\"psrun\"\n",
            RESOURCEREPORTTAG, PSRUN_RESOURCE_VERSION);
    fputs("    xmlns:perfsuite=\"http://perfsuite.ncsa.uiuc.edu/\">\n", fd);

    /*  @@ This can be inconsistent */
    /*
      indent(fd, INDENT);
      fprintf(fd, "<%s>%s</%s>\n", COMMANDTAG, res->procstat.comm, COMMANDTAG);
    */

    envval = getenv(ANNOTATIONFROMENV);
    if ( envval != NULL ) {
        indent(fd, INDENT);
        fprintf(fd, "<%s>%s</%s>\n", ANNOTATIONTAG, envval, ANNOTATIONTAG);
    }

    indent(fd, INDENT);
    fprintf(fd, "<%s>%d</%s>\n", PIDTAG, res->pid, PIDTAG);

    ret = get_starttime(buf);
    if ( ret == 0 ) {
        indent(fd, INDENT);
        fprintf(fd, "<%s>%s</%s>\n", STARTTIMETAG, buf, STARTTIMETAG);
    }
   
    exited = set_exitstatus(res, buf);
    indent(fd, INDENT);
    fprintf(fd, "<%s type=\"", EXITCODETAG);
    fprintf(fd, "%s", buf);  /* "exit", "signal", or "other" */
    fprintf(fd,"\" base=\"10\">%d</%s>\n", res->code, EXITCODETAG);

    indent(fd, INDENT);
    fprintf(fd, "<%s frequency=\"%d\" units=\"seconds\">%d</%s>\n",
            SAMPLESTAG, res->sleepsecs, res->count, SAMPLESTAG);

    indent(fd, INDENT);
    fprintf(fd, "<%s min=\"%.1f\" max=\"%.1f\" resolution=\"%.1f\" ",
            LOADAVGSTAG, 0.0, PSRUN_DEFAULTLOADBINS / 10.0, 0.1);
    fprintf(fd, "outofrange=\"%d\">\n",
            res->loadhist[PSRUN_DEFAULTLOADBINS-1]);
    for ( i = 0; i < PSRUN_DEFAULTLOADBINS - 1; i++ ) {
        if ( res->loadhist[i] > 0 ) {
            indent(fd, INDENT*2);
            fprintf(fd, "<%s load=\"%.1f\">%d</%s>\n",
                    LOADAVGTAG, i / 10.0, res->loadhist[i], LOADAVGTAG);
        }
    }
    indent(fd, INDENT);
    fprintf(fd, "</%s>\n", LOADAVGSTAG);

    indent(fd, INDENT);
    fprintf(fd, "<%s>\n", MAXLOADSTAG);
    indent(fd, INDENT*2);
    fprintf(fd, "<%s interval=\"1\" unit=\"minutes\">%.2f</%s>\n",
            MAXLOADTAG, res->loadmax[LOADAVG1MIN], MAXLOADTAG);
    indent(fd, INDENT*2);
    fprintf(fd, "<%s interval=\"5\" unit=\"minutes\">%.2f</%s>\n",
            MAXLOADTAG, res->loadmax[LOADAVG5MIN], MAXLOADTAG);
    indent(fd, INDENT*2);
    fprintf(fd, "<%s interval=\"15\" unit=\"minutes\">%.2f</%s>\n",
            MAXLOADTAG, res->loadmax[LOADAVG15MIN], MAXLOADTAG);
    indent(fd, INDENT);
    fprintf(fd, "</%s>\n", MAXLOADSTAG);

    ret = get_memoryuse(res, &vmsizemb, &rssizemb);
    indent(fd, INDENT);
    fprintf(fd, "<%s>\n", MEMORYUSETAG);
    indent(fd, INDENT*2);
    fprintf(fd, "<%s units=\"MB\">%.4f</%s>\n", MAXRSSTAG, rssizemb, MAXRSSTAG);
    indent(fd, INDENT*2);
    fprintf(fd, "<%s units=\"MB\">%.4f</%s>\n", MAXVMTAG, vmsizemb, MAXVMTAG);
    indent(fd, INDENT);
    fprintf(fd, "</%s>\n", MEMORYUSETAG);

    clk_tck = sysconf(_SC_CLK_TCK);
    indent(fd, INDENT);
    fprintf(fd, "<%s src=\"self\" units=\"seconds\">\n", CPUTIMETAG);
    indent(fd, INDENT*2);
    fprintf(fd, "<%s>%.2f</%s>\n",
            USERTIMETAG, ((double)res->procstat.utime)/clk_tck, USERTIMETAG);
    indent(fd, INDENT*2);
    fprintf(fd, "<%s>%.2f</%s>\n",
            SYSTEMTIMETAG, ((double)res->procstat.stime)/clk_tck, SYSTEMTIMETAG);
    indent(fd, INDENT);
    fprintf(fd, "</%s>\n", CPUTIMETAG);

    indent(fd, INDENT);
    fprintf(fd, "<%s src=\"self\">\n", FAULTSTAG);
    indent(fd, INDENT*2);
    fprintf(fd, "<%s>%lu</%s>\n", MAJORTAG, res->procstat.majflt, MAJORTAG);
    indent(fd, INDENT*2);
    fprintf(fd, "<%s>%lu</%s>\n", MINORTAG, res->procstat.minflt, MINORTAG);
    indent(fd, INDENT);
    fprintf(fd, "</%s>\n", FAULTSTAG);

    indent(fd, INDENT);
    fprintf(fd, "<%s>%ld</%s>\n", SWAPSTAG, res->rusage.ru_nswap, SWAPSTAG);

    fprintf(fd, "</%s>\n", RESOURCEREPORTTAG);
    fclose(fd);

    return PS_SUCCESS;
}

int
psrun_resource_text_write(psrun_resinfo_t *res)
{
    char buf[256]; /* @@ MAXPATHLEN? */
    FILE *fd;
    int exited, ret;
    float vmsizemb, rssizemb;
    long clk_tck;

    fd = select_stream("txt", res->pid);
    if ( fd == NULL ) {
        return PS_ESYS;
    }

    fputs("\n", fd);
    output_banner(fd, "Resource/Exit Information");

    ret = get_starttime(buf);
    if ( ret == 0 ) {
        output_value(fd, "Start time", buf);
    }
   
    exited = set_exitstatus(res, buf);
    output_value(fd, "Exit cause", buf);
    if ( exited >= 0 ) {
        sprintf(buf, "%d", res->code);
        if ( exited ) {
            output_value(fd, "Exit code (base 10)", buf);
        }
        else {
#if HAVE_STRSIGNAL
            strcat(buf," (");
            strcat(buf, strsignal(res->code));
            strcat(buf,")");
#endif
            output_value(fd, "Signal (base 10)", buf);
        }
    }

    get_memoryuse(res, &vmsizemb, &rssizemb);
    sprintf(buf, "%.4f", rssizemb);
    output_value(fd, "Max. resident set size (MB)", buf);
    sprintf(buf, "%.4f", vmsizemb);
    output_value(fd, "Max. virtual memory (MB)", buf);

    clk_tck = sysconf(_SC_CLK_TCK);
    sprintf(buf, "%.2f", ((double)res->procstat.utime)/clk_tck);
    output_value(fd, "CPU seconds (user)", buf);
    sprintf(buf, "%.2f", ((double)res->procstat.stime)/clk_tck);
    output_value(fd, "CPU seconds (system)", buf);

    sprintf(buf, "%lu", res->procstat.majflt);
    output_value(fd, "Major faults", buf);
    sprintf(buf, "%lu", res->procstat.minflt);
    output_value(fd, "Minor faults", buf);

    sprintf(buf, "%ld", res->rusage.ru_nswap);
    output_value(fd, "Swaps", buf);

    /* @@ Need mean, min, max load avg, stddev load average,
     * and max 1/5/15 min load avgs
     */

    return PS_SUCCESS;
}

/****************************************************************
                    Internal functions
 ****************************************************************/

static FILE *
select_stream(char *suffix, pid_t pid)
{
    FILE *fd = NULL;
    char buf[256], hostname[256]; /* @@ MAXPATHLEN? */
    char *envval, *firstdot;
    int nohostname, nopid;

    envval = getenv(NOHOSTNAMEFROMENV);
    nohostname = ( envval != NULL );

    envval = getenv(NOPIDFROMENV);
    nopid = ( envval != NULL );

    envval = getenv(FNAMEFROMENV);
    if ( envval == NULL ) {
        fputs("psrun internal error: cannot determine output prefix. ", stderr);
        fputs("Please report as a bug.\n", stderr);
        return NULL;
    }
    else if ( strcasecmp(envval, "stdout") == 0 ) {
        fd = stdout;
    }
    else if ( strcasecmp(envval, "stderr") == 0 ) {
        fd = stderr;
    }
    if ( fd == NULL ) {
        if ( nohostname ) {
            if ( nopid ) {
                sprintf(buf, "%s.%s.%s",
                        envval, PSRUN_RESOURCE_SUFFIX, suffix);
            }
            else {
                sprintf(buf, "%s.%d.%s.%s",
                        envval, pid, PSRUN_RESOURCE_SUFFIX, suffix);
            }
        }
        else {
            gethostname(hostname, 256);
            firstdot = strchr(hostname, '.');
            if ( firstdot != NULL ) {
                *firstdot = '\0';
            }
            if ( nopid ) {
                sprintf(buf, "%s.%s.%s.%s",
                        envval, hostname, PSRUN_RESOURCE_SUFFIX, suffix);
            }
            else {
                sprintf(buf, "%s.%d.%s.%s.%s",
                        envval, pid, hostname, PSRUN_RESOURCE_SUFFIX, suffix);
            }
        }

        fd = fopen(buf, "w");
    }

    return fd;
}

static int
get_starttime(char *buf)
{
    ctime_r(&starttime,buf);
    buf[strlen(buf)-1] = '\0';

    return 0;
}

static int
set_exitstatus(psrun_resinfo_t *res, char *buf)
{
    int exited;

    if ( WIFEXITED(res->status) != 0 ) {
        res->code = WEXITSTATUS(res->status);
        strcpy(buf, "exit");
        exited = 1;
    }
    else if ( WIFSIGNALED(res->status) ) {
        res->code = WTERMSIG(res->status);
        strcpy(buf, "signal");
        exited = 0;
    }
    else {
        strcpy(buf, "other");
        exited = -1;
    }

    return exited;
}

static int
get_memoryuse(psrun_resinfo_t *res, float *vmsizemb, float *rssizemb)
{
    int pagesize;

    pagesize  = getpagesize();
    *vmsizemb  = res->maxvsize/MEGABYTE;
    *rssizemb = (pagesize*res->maxrss)/MEGABYTE;

    return 0;
}


/*
 * block_signals: executed by the psrun resource monitor thread, which is
 *                only created if the user requests resource monitoring.
 */

int
block_signals(void)
{
    sigset_t sigset, osigset;
    int ret;

    ret = sigfillset(&sigset);
    if ( ret != 0 ) {
        return PS_ESYS;
    }

    ret = pthread_sigmask(SIG_BLOCK, &sigset, &osigset);
    if ( ret != 0 ) {
        return PS_ESYS;
    }

    return PS_SUCCESS;
}

