/* $Id: signals.c,v 1.11 2006/08/30 23:27:09 rkufrin Exp $ */

/****************************************************************************
 *
 *                      University of Illinois/NCSA
 *                          Open Source License
 * 
 *          Copyright(C) 2004-2006, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                             Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * signals.c: signal handling for psrun
 *
 * Author: Rick Kufrin
 * Date:   11/7/2004
 *
 * This file is part of PerfSuite.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#endif

#if HAVE_PTHREAD
# include <pthread.h>
#endif

#if HAVE_SIGNAL_H
# include <signal.h>
#endif

#include <perfsuite.h>
#include <psinternal.h>
#include <hwpc.h>

#include "psrun.h"
#include "environment.h"
#include "options.h"
#include "signals.h"

BEGIN_C_DECLS
static RETSIGTYPE forward_handler PARAMS((int signo, siginfo_t *siginfo,
                                          void *ucontext));
static RETSIGTYPE null_handler    PARAMS((int, siginfo_t *, void *));
END_C_DECLS

static int sigs[] = {
    SIGSEGV 
    /*, SIGUSR1 */
};
static int NUMSIGS = sizeof(sigs)/sizeof(int);

/*
 * setup_signals: initialization routine executed by the main psrun
 *                thread.
 */

void
setup_signals(void)
{
    struct sigaction act, oldact;
    int signo, i, ret;
    char *envval, **siglist;

    /*
     * Install default handler for signals we want to catch,
     * as listed in the sigs array.
     *
     * Include SA_RESTART in the flags as it is likely
     * we'll be in the wait4() system call in the main program
     * when a signal arrives.
     */
    act.sa_sigaction = null_handler;
    act.sa_flags     = (SA_SIGINFO | SA_RESTART | SA_ONESHOT);
    for ( i = 0; i < NUMSIGS; i++ ) {
        ret = sigaction(sigs[i], &act, &oldact);
        if ( ret != 0 ) {
            fprintf(stderr,
                    "psrun: error setting signal handler for signal %d.\n",
                    sigs[i]);
        }
    }

    /*
     * Install catch-and-forward handlers for signals specified
     * by the user.
     */
    envval = getenv(SIGNALFROMENV);
    if ( envval != NULL ) {
        ret = _ps_tokenize(envval, SIGNALSEP, &siglist);
        if ( ret != PS_SUCCESS ) {
            fprintf(stderr, "psrun: error parsing signal list: \"%s\"\n",
                    psrun_options.siglist);
        }
        else {
            act.sa_sigaction = forward_handler;
            for ( i = 0; siglist[i] != NULL; i++ ) {
                if ( sscanf(siglist[i], "%d", &signo) == 1 ) {
                    ret = sigaction(signo, &act, &oldact);
                    if ( ret != 0 ) {
                        fputs("psrun: error setting signal handler", stderr);
                        fprintf(stderr," for signal %d.\n", signo);
                    }
#if DEBUG
                printf("Installed forwarding handler for signal %d\n", signo);
#endif
                }
                free(siglist[i]);
            }
            free(siglist);
        }
    }

    return;
}

/*
 * forward_handler: signal handler that forwards on the received signal
 *                  to the process that psrun has forked.
 */

static RETSIGTYPE
forward_handler(int signo, siginfo_t *siginfo, void *ucontext)
{
#if DEBUG
    fprintf(stderr, "psrun: caught signal %d (forwarding to PID %d)\n",
            signo, res.pid);
#endif
    kill(res.pid, signo);
    return;
}


/*
 * null_handler: signal-handler that notes signal receipt and returns
 */

static RETSIGTYPE
null_handler(int signo, siginfo_t *siginfo, void *ucontext)
{
    pid_t pid;

    pid = getpid();
    fprintf(stderr, "psrun [PID %d]: caught signal %d.\n",
                    pid, siginfo->si_signo);

    return;
}
